// --------------------------------------------------------------------------------------------------------------------
// <copyright file="CodepageMapData.cs" company="Microsoft Corporation">
//   Copyright (c) 2008, 2009, 2010 All Rights Reserved, Microsoft Corporation
//
//   This source is subject to the Microsoft Permissive License.
//   Please see the License.txt file for more information.
//   All other rights reserved.
//
//   THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY 
//   KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
//   IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
//   PARTICULAR PURPOSE.
// </copyright>
// <summary>
//   Data for detecting and mapping code pagess.
// </summary>
// --------------------------------------------------------------------------------------------------------------------

namespace Microsoft.Exchange.Data.Globalization
{
    using System;

    /// <summary>
    /// The character set type of the code page.
    /// </summary>
    internal enum CodePageKind : byte
    {
        /// <summary>
        /// Unknown character set
        /// </summary>
        Unknown = 0,  
          
        /// <summary>
        /// Single Byte character set
        /// </summary>
        Sbcs,               
        
        /// <summary>
        /// Double Byte character set
        /// </summary>
        Dbcs,               
        
        /// <summary>
        /// Multi-byte character set
        /// </summary>
        Mbcs,   
        
        /// <summary>
        /// Unicode character set
        /// </summary>
        Unicode
    }

    /// <summary>
    /// An indication of the code page ASCII support.
    /// </summary>
    internal enum CodePageAsciiSupport : byte
    {
        /// <summary>
        /// Unknown ASCII support
        /// </summary>
        Unknown = 0,            
        
        /// <summary>
        /// No ASCII support
        /// </summary>
        None,               

        /// <summary>
        /// Incomplete ASCII support
        /// </summary>
        Incomplete,         
        
        /// <summary>
        /// Some remapping is required
        /// </summary>
        Remap,              
        
        /// <summary>
        /// ASCII support is fine for most purposes
        /// </summary>
        Fine,               
        
        /// <summary>
        /// Complete ASCII support
        /// </summary>
        Complete
    }

    /// <summary>
    /// An indication of the code page Unicode support.
    /// </summary>
    internal enum CodePageUnicodeCoverage : byte
    {
        /// <summary>
        /// Unknown Unicode coverage
        /// </summary>
        Unknown = 0,            
        
        /// <summary>
        /// Partial Unicode coverage
        /// </summary>
        Partial,  
        
        /// <summary>
        /// Complete Unicode coverage
        /// </summary>
        Complete 
    }

    /// <summary>
    /// Flags detailing if the codepage is detectable and/or is a 7bit code page.
    /// </summary>
    [Flags]
    internal enum CodePageFlags : byte
    {
        /// <summary>
        /// The codepage is not detectable or 7 bit
        /// </summary>
        None,

        /// <summary>
        /// The code page is detectable
        /// </summary>
        Detectable, 
        
        /// <summary>
        /// The code page is a 7bit codepage
        /// </summary>
        SevenBit           
    }

    /// <summary>
    /// Contains the data and functions necessary to map a code page.
    /// </summary>
    internal class CodePageMapData
    {
        /// <summary>
        /// An array holding all the known codepages.
        /// </summary>
        private static readonly CodePage[] AllKnownCodePages =
        {            
            new CodePage(
                37,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None,
                28591,
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                437,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C6, 0x9048, 0x10), new CodePageRange(0x207F, 0x25A0, 0x71F6, 0x08) }),            
            new CodePage(
                500,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                708,
                1256, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FB, 0xA43D, 0x20), new CodePageRange(0x060C, 0x0652, 0x5EDD, 0x01), new CodePageRange(0x2500, 0x25A0, 0x76C2, 0x40), new CodePageRange(0xF8C1, 0xF8C7, 0xFFFF, 0x00) }),           
            new CodePage(
                720,
                1256, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FB, 0xA43C, 0x10), new CodePageRange(0x0621, 0x0652, 0x5EF2, 0x01), new CodePageRange(0x207F, 0x25A0, 0x756F, 0x02) }),
            new CodePage(
                737,
                1253, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03CE, 0x8C58, 0x01), new CodePageRange(0x207F, 0x25A0, 0x767A, 0x20) }),
            new CodePage(
                775,
                1257, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x017E, 0x9F52, 0x80), new CodePageRange(0x2019, 0x25A0, 0x71ED, 0x01) }),
            new CodePage(
                850,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0192, 0x9F17, 0x40), new CodePageRange(0x2017, 0x2017, 0xFFFF, 0x00), new CodePageRange(0x2500, 0x25A0, 0x7B5F, 0x01) }),
            new CodePage(
                852,
                1250, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02DD, 0x9797, 0x10), new CodePageRange(0x2500, 0x25A0, 0x7B5F, 0x01) }),
            new CodePage(
                855,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x045F, 0x7A91, 0x02), new CodePageRange(0x2116, 0x25A0, 0x7775, 0x01) }),
            new CodePage(
                857,
                1254, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x015F, 0xA0BA, 0x20), new CodePageRange(0x2500, 0x25A0, 0x7B5F, 0x01), new CodePageRange(0xF8BB, 0xF8BD, 0xFFFF, 0x00) }),
            new CodePage(
                858,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0192, 0x9F19, 0x10), new CodePageRange(0x2017, 0x20AC, 0xA618, 0x10), new CodePageRange(0x2500, 0x25A0, 0x7B5F, 0x01) }),
            new CodePage(
                860,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C6, 0x9027, 0x01), new CodePageRange(0x207F, 0x25A0, 0x73D3, 0x04) }),
            new CodePage(
                861,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C6, 0x9045, 0x20), new CodePageRange(0x207F, 0x25A0, 0x71F6, 0x08) }),
            new CodePage(
                862,
                1255, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x05EA, 0x6C0B, 0x08), new CodePageRange(0x207F, 0x25A0, 0x71F6, 0x08) }),
            new CodePage(
                863,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C6, 0x9045, 0x80), new CodePageRange(0x2017, 0x25A0, 0x71D9, 0x40) }),
            new CodePage(
                864,
                1256, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0669, 0x5E67, 0x80), new CodePageRange(0x2219, 0x25A0, 0x940F, 0x10), new CodePageRange(0xF8BE, 0xF8C0, 0xFFFF, 0x00), new CodePageRange(0xFE7D, 0xFEFC, 0xA61C, 0x40) }),
            new CodePage(
                865,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C6, 0x904B, 0x08), new CodePageRange(0x207F, 0x25A0, 0x71F6, 0x08) }),
            new CodePage(
                866,
                1251, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x045E, 0x7BEE, 0x40), new CodePageRange(0x2116, 0x25A0, 0x7763, 0x40) }),
            new CodePage(
                869,
                1253, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03CE, 0x89E6, 0x04), new CodePageRange(0x2015, 0x2019, 0x5F7D, 0x08), new CodePageRange(0x2500, 0x25A0, 0x7B5F, 0x01) }),
            new CodePage(
                870,
                1250, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28592, 
                new[] { new CodePageRange(0x0000, 0x02DD, 0x9794, 0x80) }),
            new CodePage(
                874,
                874, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00A0, 0xA5F3, 0x02), new CodePageRange(0x0E01, 0x0E5B, 0xA699, 0x80), new CodePageRange(0x2013, 0x20AC, 0xA5F5, 0x20), new CodePageRange(0xF8C1, 0xF8C8, 0xFFFF, 0x00) }),
            new CodePage(
                875,
                1253, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03CE, 0x88F1, 0x40), new CodePageRange(0x2015, 0x2019, 0x5F7D, 0x08) }),
            new CodePage(
                932,
                932, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x804D, 0x40), new CodePageRange(0x2010, 0x266F, 0x5F78, 0x08), new CodePageRange(0x3000, 0x33CD, 0x8C7A, 0x10), new CodePageRange(0x4E00, 0x9FA0, 0x0000, 0x20), new CodePageRange(0xE000, 0xE757, 0xFFFF, 0x00), new CodePageRange(0xF8F0, 0xFA2D, 0xA0CA, 0x04), new CodePageRange(0xFF01, 0xFFE5, 0xA533, 0x80) }),
            new CodePage(
                936,
                936, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x800A, 0x08), new CodePageRange(0x2010, 0x2642, 0x65D8, 0x08), new CodePageRange(0x3000, 0x33D5, 0x8795, 0x02), new CodePageRange(0x4E00, 0x9FA5, 0xFFFF, 0x00), new CodePageRange(0xE000, 0xE864, 0xFFFF, 0x00), new CodePageRange(0xF8F5, 0xFA29, 0xA208, 0x04), new CodePageRange(0xFE30, 0xFFE5, 0x9D55, 0x01) }),
            new CodePage(
                949,
                949, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x804E, 0x10), new CodePageRange(0x2015, 0x266D, 0x64D1, 0x80), new CodePageRange(0x3000, 0x33DD, 0x84A4, 0x01), new CodePageRange(0x4E00, 0x9F9C, 0x0000, 0x40), new CodePageRange(0xAC00, 0xD7A3, 0xFFFF, 0x00), new CodePageRange(0xE000, 0xE0BB, 0xFFFF, 0x00), new CodePageRange(0xF8F7, 0xFA0B, 0xA229, 0x40), new CodePageRange(0xFF01, 0xFFE6, 0xA44D, 0x80) }),
            new CodePage(
                950,
                950, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C9, 0x8DB5, 0x04), new CodePageRange(0x2013, 0x2642, 0x67EB, 0x04), new CodePageRange(0x3000, 0x33D5, 0x88A4, 0x10), new CodePageRange(0x4E00, 0x9FA4, 0x0000, 0x04), new CodePageRange(0xE000, 0xFA0D, 0x51A1, 0x10), new CodePageRange(0xFE30, 0xFFE5, 0x9D95, 0x04) }),
            new CodePage(
                1026,
                1254, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28599, 
                new[] { new CodePageRange(0x0000, 0x015F, 0xA0BC, 0x08) }),
            new CodePage(
                1047,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                1140,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1141,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1142,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1143,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1144,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1145,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1146,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1147,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1148,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1149,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xA33E, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                1200,
                1200, 
                CodePageKind.Unicode,
                CodePageAsciiSupport.None,
                CodePageUnicodeCoverage.Complete,
                CodePageFlags.None, 
                0, 
                null),
            new CodePage(
                1201,
                1200, 
                CodePageKind.Unicode,
                CodePageAsciiSupport.None,
                CodePageUnicodeCoverage.Complete,
                CodePageFlags.None, 
                0, 
                null),
            new CodePage(
                1250,
                1250, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02DD, 0x979A, 0x08), new CodePageRange(0x2013, 0x2122, 0xA22F, 0x01) }),
            new CodePage(
                1251,
                1251, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0491, 0x7718, 0x08), new CodePageRange(0x2013, 0x2122, 0xA33D, 0x04) }),
            new CodePage(
                1252,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02DC, 0x9865, 0x04), new CodePageRange(0x2013, 0x2122, 0xA22F, 0x01) }),
            new CodePage(
                1253,
                1253, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03CE, 0x8B6B, 0x02), new CodePageRange(0x2013, 0x2122, 0xA33C, 0x08), new CodePageRange(0xF8F9, 0xF8FB, 0xFFFF, 0x00) }),
            new CodePage(
                1254,
                1254, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02DC, 0x97CD, 0x20), new CodePageRange(0x2013, 0x2122, 0xA22F, 0x01) }),
            new CodePage(
                1255,
                1255, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x05F4, 0x6BAF, 0x10), new CodePageRange(0x200E, 0x2122, 0xA338, 0x80), new CodePageRange(0xF88D, 0xF896, 0xFFFF, 0x00) }),
            new CodePage(
                1256,
                1256, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x06D2, 0x58D1, 0x01), new CodePageRange(0x200C, 0x2122, 0xA228, 0x01) }),
            new CodePage(
                1257,
                1257, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02DB, 0x9A72, 0x80), new CodePageRange(0x2013, 0x2122, 0xA22F, 0x01), new CodePageRange(0xF8FC, 0xF8FD, 0xFFFF, 0x00) }),
            new CodePage(
                1258,
                1258, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0323, 0x96C5, 0x02), new CodePageRange(0x2013, 0x2122, 0xA33C, 0x02) }),
            new CodePage(
                1361,
                949, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x849F, 0x40), new CodePageRange(0x1100, 0x11DF, 0xA538, 0x10), new CodePageRange(0x2015, 0x266D, 0x64D1, 0x80), new CodePageRange(0x3000, 0x33DD, 0x8608, 0x04), new CodePageRange(0x4E00, 0x9F9C, 0x0000, 0x40), new CodePageRange(0xAC00, 0xD7A3, 0xFFFF, 0x00), new CodePageRange(0xE000, 0xE0BB, 0xFFFF, 0x00), new CodePageRange(0xF8EC, 0xFA0B, 0xA21C, 0x08), new CodePageRange(0xFF01, 0xFFE6, 0xA44D, 0x80) }),
            new CodePage(
                10000,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C0, 0x9304, 0x02), new CodePageRange(0x2013, 0x25CA, 0x6C35, 0x01), new CodePageRange(0xF8FF, 0xFB02, 0x9D4E, 0x80) }),
            new CodePage(
                10001,
                932, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x8455, 0x80), new CodePageRange(0x2010, 0x266F, 0x5FA4, 0x01), new CodePageRange(0x3000, 0x33CD, 0x8C7A, 0x10), new CodePageRange(0x4E00, 0x9FA0, 0x0000, 0x20), new CodePageRange(0xE000, 0xE757, 0xFFFF, 0x00), new CodePageRange(0xF929, 0xFA2D, 0xA103, 0x04), new CodePageRange(0xFF01, 0xFFE5, 0xA533, 0x80) }),
            new CodePage(
                10002,
                950, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C9, 0x8F3A, 0x02), new CodePageRange(0x2013, 0x2642, 0x6A92, 0x20), new CodePageRange(0x3000, 0x33D5, 0x88A4, 0x10), new CodePageRange(0x4E00, 0x9FA4, 0x0000, 0x04), new CodePageRange(0xE000, 0xFA0D, 0x519D, 0x40), new CodePageRange(0xFE30, 0xFFE5, 0x9D95, 0x04) }),
            new CodePage(
                10003,
                949, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                51949, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x7EF1, 0x02), new CodePageRange(0x2015, 0x266D, 0x64D1, 0x80), new CodePageRange(0x3000, 0x33DD, 0x84A4, 0x01), new CodePageRange(0x4E00, 0x9F9C, 0x0000, 0x40), new CodePageRange(0xAC00, 0xD79D, 0x0000, 0x80), new CodePageRange(0xF8E6, 0xFA0B, 0xA21A, 0x20), new CodePageRange(0xFF01, 0xFFE6, 0xA44D, 0x80) }),
            new CodePage(
                10004,
                1256, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FC, 0xA33F, 0x01), new CodePageRange(0x060C, 0x06D5, 0xA539, 0x01), new CodePageRange(0x2026, 0x2026, 0xFFFF, 0x00), new CodePageRange(0xF827, 0xF840, 0xFFFF, 0x00) }),
            new CodePage(
                10005,
                1255, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FC, 0xA340, 0x20), new CodePageRange(0x05B0, 0x05EA, 0xA69E, 0x08), new CodePageRange(0x2013, 0x20AA, 0xA603, 0x01), new CodePageRange(0xF7FC, 0xF826, 0xFFFF, 0x00) }),
            new CodePage(
                10006,
                1253, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03CE, 0x8C76, 0x20), new CodePageRange(0x2013, 0x2265, 0x9A9A, 0x40), new CodePageRange(0xF8A0, 0xF8A0, 0xFFFF, 0x00) }),
            new CodePage(
                10007,
                1251, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x045F, 0x7BEE, 0x10), new CodePageRange(0x2013, 0x2265, 0x9AAA, 0x20) }),
            new CodePage( 
                10008,
                936, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x7FFC, 0x20),  new CodePageRange(0x2015, 0x2642, 0x6BAB, 0x40), new CodePageRange(0x3000, 0x3229, 0x9CC5, 0x02), new CodePageRange(0x4E00, 0x9FA0, 0x0000, 0x10), new CodePageRange(0xF8D8, 0xF8FF, 0xFFFF, 0x00), new CodePageRange(0xFF01, 0xFFE5, 0x9FD5, 0x20) }),
            new CodePage(
                10010,
                1250, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C0, 0x93EE, 0x01), new CodePageRange(0x2013, 0x25CA, 0x6E1B, 0x04), new CodePageRange(0xF8FF, 0xF8FF, 0xFFFF, 0x00) }),
            new CodePage(
                10017,
                1251, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0491, 0x7711, 0x80), new CodePageRange(0x2013, 0x2265, 0x9B42, 0x04) }),
            new CodePage(
                10021,
                874, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Fine,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00BB, 0xA539, 0x20), new CodePageRange(0x0E01, 0x0E59, 0xA69B, 0x01), new CodePageRange(0x200B, 0x2122, 0xA224, 0x02), new CodePageRange(0xFEFF, 0xFEFF, 0xFFFF, 0x00) }),
            new CodePage(
                10029,
                1250, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02C7, 0x9A8D, 0x01), new CodePageRange(0x2013, 0x25CA, 0x71A4, 0x10) }),
            new CodePage(
                10079,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C0, 0x940C, 0x20), new CodePageRange(0x2013, 0x25CA, 0x6FB7, 0x02), new CodePageRange(0xF8FF, 0xF8FF, 0xFFFF, 0x00) }),
            new CodePage(
                10081,
                1254, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C0, 0x908A, 0x40), new CodePageRange(0x2013, 0x25CA, 0x70C2, 0x20), new CodePageRange(0xF8A0, 0xF8FF, 0xA694, 0x02) }),
            new CodePage(
                10082,
                1250, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C0, 0x917F, 0x04), new CodePageRange(0x2013, 0x25CA, 0x7159, 0x80),  new CodePageRange(0xF8FF, 0xF8FF, 0xFFFF, 0x00) }),
            new CodePage(
                12000,
                1200, 
                CodePageKind.Unicode,
                CodePageAsciiSupport.None,
                CodePageUnicodeCoverage.Complete,
                CodePageFlags.None, 
                0, 
                null),
            new CodePage(
                12001,
                1200, 
                CodePageKind.Unicode,
                CodePageAsciiSupport.None,
                CodePageUnicodeCoverage.Complete,
                CodePageFlags.None, 
                0, 
                null),
            new CodePage(
                20000,
                950, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x8003, 0x80), new CodePageRange(0x2013, 0x273D, 0x51A6, 0x01), new CodePageRange(0x3000, 0x33D5, 0x8882, 0x01), new CodePageRange(0x415A, 0x415A, 0xFFFF, 0x00), new CodePageRange(0x4E00, 0x9FA4, 0x0000, 0x02), new CodePageRange(0xE000, 0xEDD2, 0x51A5, 0x08), new CodePageRange(0xF8F0, 0xF8F1, 0xFFFF, 0x00), new CodePageRange(0xFE30, 0xFFE5, 0x9D52, 0x08) }),
            new CodePage(
                20001,
                950, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x8343, 0x02), new CodePageRange(0x2013, 0x273D, 0x51A6, 0x01), new CodePageRange(0x3000, 0x33D5, 0x8882, 0x01), new CodePageRange(0x415A, 0x415A, 0xFFFF, 0x00), new CodePageRange(0x4E00, 0x9FA4, 0x0000, 0x02), new CodePageRange(0xE000, 0xF8F0, 0x51A1, 0x20), new CodePageRange(0xFE30, 0xFFE5, 0x9D52, 0x08) }),
            new CodePage(
                20002,
                950, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x7C00, 0x01), new CodePageRange(0x2013, 0x273D, 0x51A6, 0x01), new CodePageRange(0x3000, 0x33D5, 0x8882, 0x01), new CodePageRange(0x415A, 0x415A, 0xFFFF, 0x00), new CodePageRange(0x4E00, 0x9FA4, 0x0000, 0x02), new CodePageRange(0xE000, 0xE401, 0xFFFF, 0x00), new CodePageRange(0xF8BA, 0xF8E7, 0xFFFF, 0x00), new CodePageRange(0xFE30, 0xFFE5, 0x9D52, 0x08) }),
            new CodePage(
                20003,
                950, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x7C00, 0x01), new CodePageRange(0x2013, 0x273D, 0x573C, 0x80), new CodePageRange(0x3000, 0x33D5, 0x8882, 0x01), new CodePageRange(0x4E00, 0x5203, 0x84A0, 0x10), new CodePageRange(0x5206, 0x9FA4, 0x0406, 0x02), new CodePageRange(0xE000, 0xF317, 0xFFFF, 0x00), new CodePageRange(0xF8A3, 0xFA0D, 0xA0AC, 0x10), new CodePageRange(0xFE32, 0xFFE5, 0x9F08, 0x08) }),
            new CodePage(
                20004,
                950, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02D9, 0x9A75, 0x10), new CodePageRange(0x2035, 0x223C, 0x9CFD, 0x20), new CodePageRange(0x3002, 0x3129, 0xA217, 0x10), new CodePageRange(0x4E00, 0x9FA4, 0x0000, 0x08), new CodePageRange(0xE000, 0xE31B, 0x977E, 0x40), new CodePageRange(0xF882, 0xF8A2, 0xFFFF, 0x00), new CodePageRange(0xFE35, 0xFF5D, 0xA20F, 0x80) }),
            new CodePage(
                20005,
                950, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03C9, 0x8CC0, 0x40), new CodePageRange(0x2013, 0x2642, 0x6987, 0x02), new CodePageRange(0x3000, 0x33D5, 0x88A0, 0x20), new CodePageRange(0x4E00, 0x65E1, 0x51A5, 0x02), new CodePageRange(0x65E2, 0x7591, 0x17E2, 0x02), new CodePageRange(0x7594, 0x7675, 0x2794, 0x02), new CodePageRange(0x7678, 0x8FB4, 0x2878, 0x02), new CodePageRange(0x8FBF, 0x9FA4, 0x51A5, 0x04), new CodePageRange(0xE000, 0xF36D, 0xFFFF, 0x00), new CodePageRange(0xF870, 0xFA0D, 0x9F0B, 0x01), new CodePageRange(0xFE30, 0xFFE5, 0x9EEF, 0x02) }),
            new CodePage(
                20105,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Incomplete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00A4, 0xA5F2, 0x04), new CodePageRange(0x203E, 0x2207, 0x9D4F, 0x10) }),
            new CodePage(
                20106,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Incomplete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FC, 0xA33F, 0x10) }),
            new CodePage(
                20107,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Incomplete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FC, 0xA43C, 0x01) }),
            new CodePage(
                20108,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Incomplete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00F8, 0xA43E, 0x40) }),
            new CodePage(
                20127,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable | CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x007F, 0xFFFF, 0x00) }),
            new CodePage(
                20261,
                1252, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Incomplete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02DD, 0x97AF, 0x01), new CodePageRange(0x1E02, 0x2126, 0x9540, 0x04), new CodePageRange(0xF8DD, 0xF8FF, 0xFFFF, 0x00) }),
            new CodePage(
                20269,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Incomplete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0332, 0x944B, 0x40), new CodePageRange(0x2015, 0x2193, 0x9F4B, 0x04), new CodePageRange(0x2669, 0x2669, 0xFFFF, 0x00), new CodePageRange(0xF8F6, 0xF8FF, 0xFFFF, 0x00) }),
            new CodePage(
                20273,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                20277,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                20278,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                20280,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                20284,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                20285,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                20290,
                932, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00AC, 0xA5F1, 0x08), new CodePageRange(0x203E, 0x203E, 0xFFFF, 0x00), new CodePageRange(0xFF61, 0xFF9F, 0xFFFF, 0x00) }),
            new CodePage(
                20297,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                20420,
                1256, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Incomplete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00F7, 0xA44C, 0x02), new CodePageRange(0x060C, 0x0669, 0xA697, 0x04), new CodePageRange(0x2007, 0x2007, 0xFFFF, 0x00), new CodePageRange(0xF8F4, 0xF8FC, 0xA6D9, 0x08), new CodePageRange(0xFE7C, 0xFEFC, 0xA618, 0x80) }),
            new CodePage(
                20423,
                1253, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Incomplete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03CE, 0x88A7, 0x80) }),
            new CodePage(
                20424,
                1255, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00F7, 0xA44C, 0x08), new CodePageRange(0x05D0, 0x05EA, 0xFFFF, 0x00), new CodePageRange(0x2017, 0x203E, 0xA6AE, 0x10) }),
            new CodePage(
                20833,
                949, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00AC, 0xA544, 0x08), new CodePageRange(0x203E, 0x20A9, 0xA68F, 0x20), new CodePageRange(0xFFA0, 0xFFDC, 0xA69C, 0x40) }),
            new CodePage(
                20838,
                874, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00AC, 0xA545, 0x04), new CodePageRange(0x0E01, 0x0E5B, 0xA699, 0x80) }),
            new CodePage(
                20866,
                1251, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x8052, 0x01), new CodePageRange(0x2219, 0x25A0, 0x940C, 0x80) }),
            new CodePage(
                20871,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28591, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                20880,
                1251, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x045F, 0x7BA3, 0x80), new CodePageRange(0x2116, 0x2116, 0xFFFF, 0x00) }),
            new CodePage(
                20905,
                1254, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02D9, 0x9A78, 0x08) }),
            new CodePage(
                20924,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x017E, 0xA0A5, 0x02), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
            new CodePage(
                20932,
                932, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x045F, 0x7BAA, 0x08), new CodePageRange(0x2010, 0x266F, 0x618B, 0x04), new CodePageRange(0x3000, 0x33CD, 0x8C7D, 0x08), new CodePageRange(0x4E00, 0x9FA5, 0x0000, 0x01), new CodePageRange(0xE000, 0xE757, 0xFFFF, 0x00), new CodePageRange(0xF8F0, 0xFA2D, 0xA0D1, 0x80), new CodePageRange(0xFF01, 0xFFE5, 0xA537, 0x40) }),
            new CodePage(
                20936,
                936, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x7FFC, 0x20), new CodePageRange(0x2015, 0x2642, 0x6BAB, 0x40), new CodePageRange(0x3000, 0x3229, 0x9CC5, 0x02), new CodePageRange(0x4E00, 0x9FA0, 0x0000, 0x10), new CodePageRange(0xF8D8, 0xF8FF, 0xFFFF, 0x00), new CodePageRange(0xFF01, 0xFFE5, 0x9FD5, 0x20) }),
            new CodePage(
                20949,
                949, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                51949, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x7EF1, 0x02), new CodePageRange(0x2015, 0x266D, 0x64D1, 0x80), new CodePageRange(0x3000, 0x33DD, 0x84A4, 0x01), new CodePageRange(0x4E00, 0x9F9C, 0x0000, 0x40), new CodePageRange(0xAC00, 0xD79D, 0x0000, 0x80), new CodePageRange(0xF8E6, 0xFA0B, 0xA21A, 0x20), new CodePageRange(0xFF01, 0xFFE6, 0xA44D, 0x80) }),
            new CodePage(
                21025,
                1251, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Remap,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                28595, 
                new[] { new CodePageRange(0x0000, 0x045F, 0x7B9C, 0x20), new CodePageRange(0x2116, 0x2116, 0xFFFF, 0x00) }),
            new CodePage(
                21866,
                1251, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0491, 0x775C, 0x10), new CodePageRange(0x2219, 0x25A0, 0x9412, 0x08) }),
            new CodePage(
                28591,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00FF, 0xFFFF, 0x00) }),
            new CodePage(
                28592,
                1250, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02DD, 0x9794, 0x80) }),
            new CodePage(
                28593,
                1254, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02D9, 0x9A78, 0x08), new CodePageRange(0xF7F5, 0xF7FB, 0xFFFF, 0x00) }),
            new CodePage(
                28594,
                1257, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x02DB, 0x99E9, 0x02) }),
            new CodePage(
                28595,
                1251, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x045F, 0x7B9C, 0x20), new CodePageRange(0x2116, 0x2116, 0xFFFF, 0x00) }),
            new CodePage(
                28596,
                1256, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00AD, 0x7BA3, 0x80), new CodePageRange(0x060C, 0x0652, 0x5EDD, 0x01),  new CodePageRange(0xF7C8, 0xF7F4, 0xFFFF, 0x00) }),
            new CodePage(
                28597,
                1253, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x03CE, 0x88AE, 0x08), new CodePageRange(0x2015, 0x2015, 0xFFFF, 0x00), new CodePageRange(0xF7C2, 0xF7C7, 0xFFFF, 0x00) }),
            new CodePage(
                28598,
                1255, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                38598, 
                new[] { new CodePageRange(0x0000, 0x00F7, 0xA44D, 0x04), new CodePageRange(0x05D0, 0x05EA, 0xFFFF, 0x00), new CodePageRange(0x2017, 0x203E, 0xA6D6, 0x10), new CodePageRange(0xF79C, 0xF7C1, 0xFFFF, 0x00) }),
            new CodePage(
                28599,
                1254, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x015F, 0xA0BC, 0x08) }),
            new CodePage(
                28603,
                1257, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x017E, 0xA0A9, 0x01), new CodePageRange(0x2019, 0x201E, 0x71ED, 0x01) }),
            new CodePage(
                28605,
                1252, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x017E, 0xA0AA, 0x40), new CodePageRange(0x20AC, 0x20AC, 0xFFFF, 0x00) }),
           new CodePage(
               29001,
               1252, 
               CodePageKind.Sbcs,
               CodePageAsciiSupport.Incomplete,
               CodePageUnicodeCoverage.Partial,
               CodePageFlags.None, 
               0, 
               new[] { new CodePageRange(0x0000, 0x03CE, 0x8C76, 0x80), new CodePageRange(0x203C, 0x266C, 0x6604, 0x01), new CodePageRange(0xF8FF, 0xF8FF, 0xFFFF, 0x00) }),
            new CodePage(
                38598,
                1255, 
                CodePageKind.Sbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x00F7, 0xA44D, 0x04), new CodePageRange(0x05D0, 0x05EA, 0xFFFF, 0x00), new CodePageRange(0x2017, 0x203E, 0xA6D6, 0x10), new CodePageRange(0xF79C, 0xF7C1, 0xFFFF, 0x00) }),
            new CodePage(
                50220,
                932, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Fine,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable | CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x845C, 0x08), new CodePageRange(0x2010, 0x266F, 0x5F78, 0x08), new CodePageRange(0x3000, 0x33CD, 0x8C7A, 0x10), new CodePageRange(0x4E00, 0x9FA0, 0x0000, 0x20), new CodePageRange(0xE000, 0xE757, 0xFFFF, 0x00), new CodePageRange(0xF8F0, 0xFA2D, 0xA0CA, 0x04), new CodePageRange(0xFF01, 0xFFE5, 0x9E26, 0x01) }),
            new CodePage(
                50221,
                932, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Fine,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x7D64, 0x04), new CodePageRange(0x2010, 0x266F, 0x5F78, 0x08), new CodePageRange(0x3000, 0x33CD, 0x8C7A, 0x10), new CodePageRange(0x4E00, 0x9FA0, 0x0000, 0x20), new CodePageRange(0xE000, 0xE757, 0xFFFF, 0x00), new CodePageRange(0xF8F0, 0xFA2D, 0xA0CA, 0x04), new CodePageRange(0xFF01, 0xFFE5, 0xA533, 0x80) }),
            new CodePage(
                50222,
                932, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Fine,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x7D64, 0x04), new CodePageRange(0x2010, 0x266F, 0x5F78, 0x08), new CodePageRange(0x3000, 0x33CD, 0x8C7A, 0x10), new CodePageRange(0x4E00, 0x9FA0, 0x0000, 0x20), new CodePageRange(0xE000, 0xE757, 0xFFFF, 0x00), new CodePageRange(0xF8F0, 0xFA2D, 0xA0CA, 0x04), new CodePageRange(0xFF01, 0xFFE5, 0xA533, 0x80) }),
            new CodePage(
                50225,
                949, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Fine,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable | CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x81B6, 0x04), new CodePageRange(0x2015, 0x266D, 0x64D1, 0x80), new CodePageRange(0x3000, 0x33DD, 0x84A4, 0x01), new CodePageRange(0x4E00, 0x9F9C, 0x0000, 0x40), new CodePageRange(0xAC00, 0xD79D, 0x2B9E, 0x80), new CodePageRange(0xE000, 0xE0BB, 0xFFFF, 0x00), new CodePageRange(0xF900, 0xFA0B, 0xFFFF, 0x00), new CodePageRange(0xFF01, 0xFFE6, 0xA44D, 0x80) }),
            new CodePage(
                50227,
                936, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                936, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x800A, 0x08), new CodePageRange(0x2010, 0x2642, 0x65D8, 0x08), new CodePageRange(0x3000, 0x33D5, 0x8795, 0x02), new CodePageRange(0x4E00, 0x9FA5, 0xFFFF, 0x00), new CodePageRange(0xE000, 0xE864, 0xFFFF, 0x00), new CodePageRange(0xF8F5, 0xFA29, 0xA208, 0x04), new CodePageRange(0xFE30, 0xFFE5, 0x9D55, 0x01) }),
            new CodePage(
                51932,
                932, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x804D, 0x40), new CodePageRange(0x2010, 0x266F, 0x5F78, 0x08), new CodePageRange(0x3000, 0x33CD, 0x8C7A, 0x10), new CodePageRange(0x4E00, 0x9FA0, 0x0000, 0x20), new CodePageRange(0xF8F0, 0xFA2D, 0xA0D1, 0x80), new CodePageRange(0xFF01, 0xFFE5, 0xA533, 0x80) }),
            new CodePage(
                51936,
                936, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                936, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x800A, 0x08), new CodePageRange(0x2010, 0x2642, 0x65D8, 0x08), new CodePageRange(0x3000, 0x33D5, 0x8795, 0x02), new CodePageRange(0x4E00, 0x9FA5, 0xFFFF, 0x00), new CodePageRange(0xE000, 0xE864, 0xFFFF, 0x00), new CodePageRange(0xF8F5, 0xFA29, 0xA208, 0x04), new CodePageRange(0xFE30, 0xFFE5, 0x9D55, 0x01) }),
            new CodePage(
                51949,
                949, 
                CodePageKind.Dbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x7EF1, 0x02), new CodePageRange(0x2015, 0x266D, 0x64D1, 0x80), new CodePageRange(0x3000, 0x33DD, 0x84A4, 0x01), new CodePageRange(0x4E00, 0x9F9C, 0x0000, 0x40), new CodePageRange(0xAC00, 0xD79D, 0x0000, 0x80), new CodePageRange(0xF8E6, 0xFA0B, 0xA21A, 0x20), new CodePageRange(0xFF01, 0xFFE6, 0xA44D, 0x80) }),
            new CodePage(
                52936,
                936, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.Detectable | CodePageFlags.SevenBit, 
                0, 
                new[] { new CodePageRange(0x0000, 0x0451, 0x844E, 0x20), new CodePageRange(0x2014, 0x2642, 0x6B2A, 0x80), new CodePageRange(0x3000, 0x3229, 0x9CED, 0x40), new CodePageRange(0x4E00, 0x9FA0, 0x0000, 0x10), new CodePageRange(0xE000, 0xE233, 0xFFFF, 0x00), new CodePageRange(0xE766, 0xE814, 0xA544, 0x02), new CodePageRange(0xF8F5, 0xF8F5, 0xFFFF, 0x00), new CodePageRange(0xFE31, 0xFFE5, 0x9F05, 0x20) }),
            new CodePage(
                54936,
                936, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Complete,
                CodePageFlags.None, 
                0, 
                null),
            new CodePage(
                57002,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                57003,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                57004,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                57005,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                57006,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                57007,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                57008,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                57009,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                57010,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                57011,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Partial,
                CodePageFlags.None, 
                0, 
                new[] { new CodePageRange(0x0000, 0x009F, 0xFFFF, 0x00), new CodePageRange(0x0901, 0x0D6F, 0x78F5, 0x04) }),
            new CodePage(
                65000,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Incomplete,
                CodePageUnicodeCoverage.Complete,
                CodePageFlags.SevenBit, 
                0, 
                null),
            new CodePage(
                65001,
                1200, 
                CodePageKind.Mbcs,
                CodePageAsciiSupport.Complete,
                CodePageUnicodeCoverage.Complete,
                CodePageFlags.None, 
                0, 
                null),
        };

        /// <summary>
        /// Character bitmaps.
        /// </summary>
        private static readonly byte[] CharacterBitmap =
        {
            0xFF, 0xFF, 0x01, 0x7F, 0x81, 0x01, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x7F, 0x1F, 0x7F, 0x3F, 0x0E, 
            0xBF, 0xFF, 0x81, 0x90, 0xFF, 0xFF, 0xFF, 0xA1, 0x7F, 0xFF, 0x90, 0x90, 0x90, 0x90, 0x7F, 0x0F, 
            0x80, 0x21, 0x10, 0x01, 0x91, 0x10, 0x2F, 0x10, 0x31, 0x00, 0x31, 0x1F, 0x90, 0xFF, 0x0F, 0x81, 
            0x9F, 0xAF, 0x7F, 0x0E, 0x10, 0x01, 0x3B, 0x00, 0xFF, 0xFF, 0x10, 0x7F, 0xA7, 0x10, 0x10, 0x3B, 
            0x81, 0x01, 0x6F, 0x7F, 0x01, 0x7F, 0x00, 0x17, 0x1E, 0x10, 0x00, 0xFF, 0x10, 0xFF, 0x7F, 0x7F, 
            0x10, 0x01, 0x1E, 0x1E, 0x90, 0x21, 0x7F, 0x21, 0xFF, 0x7F, 0x01, 0x00, 0x9F, 0x7F, 0x7F, 0x7F, 
            0x10, 0x10, 0x21, 0x01, 0x00, 0x00, 0x10, 0x00, 0x01, 0x1F, 0x00, 0x40, 0x00, 0x40, 0x00, 0x00, 
            0x90, 0xB1, 0x00, 0x7F, 0x81, 0x01, 0x40, 0xC0, 0x80, 0x01, 0x80, 0x00, 0x00, 0x00, 0x7F, 0x07, 
            0xA1, 0x80, 0x6F, 0x86, 0x8E, 0xAB, 0xFF, 0x00, 0x7F, 0xB1, 0xA1, 0xFF, 0xFF, 0x1F, 0x7F, 0x10, 
            0xC0, 0x7F, 0x7F, 0x1E, 0xFF, 0x7F, 0x01, 0x01, 0x71, 0x2F, 0x10, 0x7F, 0xA1, 0x81, 0x6F, 0xBF, 
            0xBB, 0xFF, 0x7F, 0x00, 0x7F, 0x7F, 0x7F, 0x10, 0xFF, 0x90, 0x80, 0x7F, 0xFF, 0x7F, 0x7F, 0x81, 
            0xA1, 0x00, 0x10, 0x3F, 0x00, 0x10, 0x6F, 0x00, 0x80, 0x8F, 0x7F, 0x90, 0x80, 0x80, 0x00, 0x10, 
            0x7F, 0xFF, 0x3F, 0x1F, 0xFF, 0x10, 0x3F, 0x7F, 0x8E, 0x1E, 0x7F, 0x7F, 0x80, 0x7F, 0x31, 0x21, 
            0x01, 0x10, 0x00, 0x10, 0x7F, 0xFF, 0x7F, 0xFF, 0x7F, 0x7F, 0x0F, 0x01, 0x0E, 0x7F, 0x3F, 0x7F, 
            0x81, 0xBF, 0x01, 0x7F, 0xFF, 0x7F, 0x00, 0x80, 0x9F, 0x0E, 0x90, 0x10, 0x90, 0x21, 0x21, 0x81, 
            0xFF, 0x8F, 0x7F, 0x9F, 0x0E, 0x9F, 0xFF, 0x7F, 0x00, 0x00, 0x00, 0x7F, 0xA1, 0x9F, 0x01, 0x1F, 
            0xAF, 0x7F, 0x0F, 0x21, 0x86, 0x06, 0x80, 0x00, 0x0F, 0x7F, 0x7F, 0x4F, 0x81, 0xFF, 0x7F, 0xFF, 
            0x7F, 0xFF, 0x07, 0x0E, 0x0E, 0x0F, 0x01, 0x11, 0x9E, 0x1F, 0x31, 0x10, 0xA1, 0x2F, 0x10, 0x10, 
            0x90, 0x00, 0x16, 0x00, 0x10, 0x10, 0x10, 0x10, 0x00, 0x80, 0x10, 0x10, 0x8E, 0x8E, 0x01, 0x7F, 
            0x3F, 0x01, 0x10, 0x0F, 0xFF, 0x81, 0x7F, 0x01, 0xFF, 0x21, 0x7F, 0x0F, 0xFF, 0x7F, 0x0F, 0x0E, 
            0x01, 0x0E, 0x01, 0x7F, 0x80, 0x80, 0x7F, 0xEF, 0x4F, 0x8F, 0x00, 0x01, 0x0F, 0x7F, 0x7F, 0x7F, 
            0xFF, 0x7F, 0x0F, 0x3F, 0x8F, 0x7F, 0x2F, 0x3F, 0x9F, 0x7F, 0x7F, 0x7F, 0x7F, 0x3F, 0x3F, 0x1F, 
            0x1F, 0x8E, 0x0E, 0x9F, 0x1E, 0x10, 0x00, 0x1E, 0x00, 0x7F, 0x0F, 0x0E, 0x9F, 0x80, 0x07, 0x7F, 
            0xFF, 0x01, 0x00, 0xFF, 0x9E, 0xAF, 0xFF, 0x0F, 0x0F, 0x0F, 0x4F, 0xBF, 0xBF, 0x8F, 0x5F, 0xFF, 
            0x0E, 0xCF, 0x87, 0x7F, 0x5F, 0x0F, 0x6F, 0x0E, 0xFF, 0x9F, 0x61, 0x7F, 0x81, 0x7F, 0x01, 0x7F, 
            0x8F, 0x7F, 0x2F, 0x01, 0x3F, 0x0E, 0x6F, 0x1F, 0x6F, 0x01, 0x2F, 0x7F, 0x8E, 0xFF, 0x0F, 0x01, 
            0x31, 0x21, 0x00, 0x10, 0x80, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x21, 0x10, 0x21, 0x7F, 0x7F, 
            0x00, 0x00, 0x0F, 0x0E, 0x00, 0x7F, 0x6F, 0x8F, 0x00, 0x0F, 0x0E, 0x0F, 0x01, 0x01, 0x01, 0x7F, 
            0x87, 0x8F, 0x6F, 0x7F, 0xFF, 0x1F, 0x01, 0x0E, 0x81, 0x6F, 0x7F, 0x0F, 0x01, 0x2F, 0x7F, 0x1F, 
            0xBF, 0xFF, 0x01, 0xCE, 0x6F, 0x00, 0x0E, 0x7F, 0x3F, 0x0E, 0x7F, 0x2F, 0x9F, 0x7F, 0x1E, 0x7F, 
            0xCF, 0x7F, 0x01, 0x31, 0xA1, 0x21, 0x10, 0x00, 0x10, 0x10, 0x10, 0x00, 0x0E, 0x10, 0x7F, 0x7F, 
            0x01, 0x5E, 0x01, 0x7F, 0x0E, 0x6F, 0x2F, 0x0E, 0xFF, 0x80, 0x7F, 0x00, 0x81, 0x01, 0x3F, 0xA1, 
            0x8F, 0x81, 0x40, 0x00, 0x01, 0x2F, 0x6F, 0x0F, 0x80, 0xEF, 0x01, 0xEF, 0x1F, 0xFF, 0x0F, 0x3F, 
            0x01, 0x6F, 0x7F, 0x0F, 0xBF, 0x0E, 0x6F, 0x0F, 0x1F, 0x7F, 0x7F, 0x0F, 0x5F, 0x01, 0x6F, 0x7F, 
            0x0E, 0x7F, 0x2F, 0x6F, 0x61, 0x3F, 0x7F, 0x4F, 0x7F, 0x3F, 0x7F, 0x6F, 0x7F, 0x7F, 0x11, 0x0E, 
            0x8F, 0x8E, 0x01, 0x0F, 0x80, 0x0F, 0x21, 0x8E, 0x80, 0x21, 0x90, 0x41, 0x1E, 0x00, 0x10, 0x00, 
            0xAF, 0x8F, 0x21, 0xFF, 0x00, 0x8F, 0xAF, 0x7F, 0x7F, 0x6F, 0x8F, 0x0E, 0x9F, 0x8E, 0x9F, 0x7F, 
            0xA1, 0x07, 0x01, 0x0F, 0x80, 0x7F, 0xA1, 0x07, 0x00, 0x01, 0x7F, 0x0E, 0xFF, 0x8E, 0x0E, 0x8F, 
            0x8F, 0x8E, 0x0F, 0x0F, 0x0E, 0xFF, 0x01, 0x01, 0x8E, 0x8E, 0x0F, 0x0E, 0xB1, 0x0F, 0x0E, 0x0E, 
            0xAF, 0x01, 0x2F, 0x0E, 0x6F, 0x6F, 0x7F, 0x1E, 0xE1, 0x80, 0x0E, 0x90, 0x80, 0xAF, 0x10, 0x10, 
            0x7F, 0x01, 0x0E, 0x0F, 0x81, 0xFF, 0x01, 0x0E, 0x10, 0x00, 0x01, 0x0E, 0x8E, 0x7F, 0x0F, 0x01, 
            0x01, 0x6F, 0x0F, 0x01, 0x2F, 0x0E, 0x0F, 0x00, 0x6F, 0x6F, 0x2F, 0x0F, 0x0F, 0x0E, 0x0F, 0x01, 
            0x01, 0x01, 0x0F, 0x1E, 0x00, 0x10, 0x00, 0x10, 0x10, 0x10, 0x01, 0x00, 0x7F, 0x6F, 0x0E, 0x0F, 
            0x0F, 0x0E, 0x7F, 0x6F, 0x2F, 0x6F, 0x06, 0x6F, 0x0E, 0x01, 0x1F, 0x1E, 0x80, 0x8F, 0xEF, 0x0E, 
            0x81, 0x0E, 0x2F, 0x01, 0x8F, 0x6F, 0x0E, 0x0F, 0x0E, 0x6F, 0x6F, 0x0E, 0x81, 0xA1, 0x0F, 0xFF, 
            0x81, 0xEF, 0x00, 0x0F, 0x0F, 0x6F, 0x7F, 0x06, 0xA1, 0x80, 0x7F, 0x0E, 0x81, 0x0F, 0x61, 0x01, 
            0x06, 0x00, 0x01, 0x2F, 0x0F, 0x6F, 0x1F, 0x7F, 0x8F, 0x0F, 0x0E, 0x80, 0x1E, 0xFF, 0x3F, 0x07, 
            0x0E, 0x0F, 0x01, 0x16, 0xA1, 0x3F, 0x0F, 0x00, 0x8E, 0x6F, 0x01, 0x7F, 0x80, 0x0E, 0x0F, 0x4E, 
            0x6F, 0x61, 0x2F, 0x0F, 0x6F, 0x0E, 0x5F, 0x9F, 0x8F, 0x6F, 0x0E, 0x1F, 0x0F, 0x81, 0x01, 0x00, 
            0x8F, 0x0E, 0x7F, 0x0E, 0x2F, 0x2F, 0x21, 0x0F, 0x2F, 0x01, 0x2F, 0x01, 0x0F, 0x01, 0x01, 0x6F, 
            0x0E, 0x7F, 0x0E, 0x01, 0x0E, 0x0E, 0x0E, 0x01, 0x01, 0x0E, 0x6F, 0x00, 0x81, 0x8F, 0x06, 0x01, 
            0x8E, 0x0F, 0xEF, 0x0F, 0x8F, 0x0F, 0x00, 0x6F, 0x0F, 0x0F, 0x6F, 0x2F, 0xEF, 0x8E, 0x00, 0xBF, 
            0x7F, 0xFF, 0x81, 0xFF, 0x7F, 0x7F, 0x7F, 0x6F, 0xFF, 0xFF, 0x21, 0x7F, 0x6F, 0x7F, 0x61, 0x01, 
            0xA1, 0x10, 0x6F, 0x01, 0x3F, 0x1F, 0x10, 0x0F, 0x01, 0x0E, 0x3F, 0x0E, 0xFF, 0x8E, 0x0E, 0x0F, 
            0x00, 0x0E, 0x7F, 0x0E, 0xA1, 0xFF, 0x01, 0x4E, 0x7F, 0x6F, 0x61, 0x7F, 0x7F, 0x7F, 0x7F, 0x00, 
            0x10, 0x7F, 0x00, 0x10, 0x10, 0x7F, 0x7F, 0x7F, 0x7F, 0x90, 0x00, 0x10, 0x7F, 0x10, 0x01, 0x00, 
            0xFF, 0x10, 0x3B, 0x01, 0x81, 0x31, 0x61, 0x0E, 0x92, 0x3F, 0x6F, 0x01, 0x31, 0x7F, 0x01, 0x2F, 
            0xA1, 0xAF, 0x7F, 0x2F, 0x0E, 0xFF, 0x3B, 0x7F, 0x0F, 0x31, 0x00, 0x10, 0x90, 0x21, 0x0E, 0x00, 
            0x7F, 0x01, 0x3F, 0x01, 0x3F, 0x7F, 0x21, 0x00, 0x21, 0x21, 0x6F, 0x3B, 0x7F, 0x01, 0x00, 0x10, 
            0x3F, 0x3F, 0x31, 0x31, 0x21, 0x31, 0x7F, 0x7F, 0x81, 0x86, 0x01, 0x10, 0x9F, 0x7F, 0x2F, 0x01, 
            0x90, 0x00, 0x01, 0x00, 0x7F, 0x2F, 0x7F, 0x90, 0x8F, 0xF1, 0x0E, 0xFF, 0x7F, 0xEF, 0x86, 0x11, 
            0x0E, 0x11, 0x01, 0x01, 0x8E, 0x01, 0x21, 0x06, 0x0F, 0x00, 0x00, 0x31, 0xEF, 0x7F, 0x47, 0x00, 
            0x3F, 0x7F, 0x01, 0x00, 0x10, 0x01, 0x21, 0x21, 0x80, 0xA1, 0x21, 0x10, 0x21, 0x31, 0x01, 0x10, 
            0xFF, 0xEF, 0x01, 0x1F, 0x81, 0x3F, 0x7F, 0x01, 0xFF, 0x7F, 0x7F, 0x10, 0x10, 0x7F, 0x21, 0x10, 
            0xFF, 0x9F, 0x13, 0x7F, 0xA1, 0x01, 0x7F, 0x7F, 0x7F, 0x0E, 0x7F, 0x21, 0x8E, 0x10, 0x7F, 0x00, 
            0x8E, 0x7F, 0x1F, 0x0F, 0xA1, 0x21, 0x1F, 0x7F, 0x11, 0x10, 0x10, 0x10, 0x8E, 0xFF, 0x0E, 0x00, 
            0x10, 0x0E, 0x01, 0x00, 0x7F, 0x6F, 0x00, 0x21, 0x9F, 0x7F, 0x6F, 0x10, 0x00, 0x10, 0x7F, 0x00, 
            0x7F, 0x0F, 0x0F, 0x3F, 0x00, 0x0F, 0x7F, 0x7F, 0x7F, 0x71, 0x7F, 0x7F, 0x01, 0x10, 0x00, 0x10, 
            0x10, 0x1E, 0x10, 0x7F, 0xAF, 0x81, 0x0E, 0x6F, 0x80, 0x0F, 0xFF, 0x6F, 0xFF, 0x7F, 0x8E, 0x21, 
            0x10, 0x10, 0x0E, 0x80, 0xFF, 0x8F, 0x7F, 0x81, 0x01, 0x80, 0x0F, 0x6F, 0x1F, 0xCE, 0x3F, 0x0F, 
            0x01, 0x5F, 0x06, 0x21, 0x21, 0x31, 0x01, 0x10, 0x00, 0x7F, 0x7F, 0x0E, 0x0E, 0x0E, 0x0F, 0x7F, 
            0x21, 0x21, 0x7F, 0x21, 0x2F, 0x6F, 0x00, 0x0F, 0x0F, 0x01, 0x0E, 0x0E, 0x8E, 0xFF, 0x00, 0x7F, 
            0x8F, 0x1E, 0x1F, 0x6F, 0x8E, 0x01, 0x00, 0x6F, 0x7F, 0x6F, 0x0F, 0x0E, 0x8F, 0xEF, 0x00, 0x80, 
            0x10, 0xEF, 0x61, 0x1F, 0x21, 0x01, 0x0F, 0x0F, 0x8F, 0x8E, 0x81, 0x7F, 0xA1, 0x10, 0x10, 0xFF, 
            0xFF, 0x90, 0x90, 0x7F, 0x41, 0x01, 0x2F, 0x01, 0x90, 0xFF, 0x7F, 0xFF, 0xBF, 0xBF, 0x86, 0xA1, 
            0x01, 0xB1, 0x10, 0x90, 0xA1, 0xA1, 0x01, 0x01, 0x81, 0x21, 0x01, 0x0F, 0xAF, 0x01, 0x7F, 0x10, 
            0x2F, 0x6F, 0x06, 0x7F, 0x81, 0xE1, 0x01, 0xFF, 0x81, 0xFF, 0x00, 0x10, 0x01, 0x6F, 0x00, 0x01, 
            0x90, 0x81, 0x7F, 0x06, 0x81, 0x6F, 0x5F, 0x2F, 0xFF, 0x6F, 0x00, 0x6F, 0x01, 0x6F, 0x6F, 0x7F, 
            0xA1, 0x01, 0x6F, 0x6F, 0x7F, 0x81, 0x2F, 0x21, 0x01, 0x0F, 0x01, 0x06, 0x01, 0x00, 0x00, 0x06, 
            0x1F, 0x0F, 0x21, 0x6F, 0x0F, 0x6F, 0x01, 0x0F, 0x6F, 0x3B, 0x7F, 0x4E, 0x06, 0x00, 0x7F, 0x7F, 
            0x31, 0x21, 0x21, 0x01, 0x00, 0x7F, 0x3F, 0x21, 0xFF, 0x86, 0x0F, 0x8F, 0x81, 0x7F, 0x06, 0x7F, 
            0xFF, 0x0F, 0x86, 0x81, 0x00, 0x7F, 0x7F, 0x7F, 0x81, 0xFF, 0x3F, 0x81, 0x8F, 0xBF, 0x01, 0x0F, 
            0x7F, 0x7F, 0x0E, 0xFF, 0xA1, 0x81, 0x10, 0x01, 0x81, 0x01, 0x7F, 0x01, 0x81, 0x0F, 0x10, 0x6F, 
            0x0F, 0x2F, 0x01, 0x21, 0x8E, 0x81, 0x00, 0x8E, 0xAB, 0xFF, 0x31, 0x31, 0x0F, 0x0F, 0x1F, 0x7F, 
            0xEF, 0xFF, 0x01, 0x7F, 0xC6, 0x3F, 0x21, 0x7F, 0xFF, 0x7F, 0x7F, 0x01, 0x0F, 0x6F, 0x10, 0x10, 
            0x80, 0xFF, 0x7F, 0x7F, 0xEF, 0x90, 0x10, 0x7F, 0xA1, 0x01, 0x7F, 0x01, 0xFF, 0x00, 0x7F, 0x10, 
            0xFF, 0x1F, 0x10, 0x1F, 0x10, 0x01, 0x7F, 0x10, 0x40, 0x3B, 0x00, 0x10, 0x0F, 0x01, 0x3F, 0x7F, 
            0x7F, 0x7F, 0x2F, 0x3F, 0x71, 0x7F, 0x00, 0x7F, 0xBF, 0x8F, 0x10, 0x2F, 0x8E, 0x40, 0x01, 0x7F, 
            0x80, 0x00, 0xBF, 0x01, 0x7F, 0x10, 0x10, 0x01, 0x81, 0x90, 0x06, 0x90, 0x10, 0x90, 0x0F, 0x06, 
            0x00, 0x00, 0x8E, 0xE1, 0x8F, 0x10, 0x2F, 0x0E, 0xFF, 0x0F, 0x7F, 0x00, 0x8E, 0x1F, 0x0E, 0x7F, 
            0x61, 0x01, 0x10, 0x10, 0x0F, 0x7F, 0x71, 0x06, 0xB1, 0x31, 0x01, 0x01, 0x0E, 0x6F, 0x31, 0x01, 
            0x21, 0x00, 0x2F, 0x21, 0x07, 0x01, 0x3B, 0x01, 0x01, 0x0E, 0x01, 0x7F, 0x00, 0x01, 0x00, 0x10, 
            0x01, 0x10, 0x31, 0x6F, 0x00, 0x01, 0x00, 0x00, 0x7F, 0x7F, 0x7F, 0x7F, 0xB1, 0x7F, 0x21, 0x01, 
            0x80, 0x10, 0x01, 0x01, 0xFF, 0x01, 0x7F, 0x7F, 0x10, 0x31, 0x01, 0x7F, 0x80, 0xA1, 0x01, 0xBF, 
            0x11, 0x6F, 0x6F, 0x7F, 0x7F, 0x7F, 0x1F, 0x01, 0xBF, 0xFF, 0x7F, 0x7F, 0xFF, 0x7F, 0x3F, 0x7F, 
            0xFF, 0x7F, 0x7F, 0x7F, 0x00, 0x1F, 0x31, 0x31, 0x7F, 0x90, 0x21, 0x9E, 0x1E, 0x90, 0x00, 0x00, 
            0x00, 0x3F, 0x01, 0x7F, 0xFF, 0x80, 0x1E, 0x0E, 0xFF, 0x7F, 0x7F, 0xAF, 0xFF, 0x7F, 0x7F, 0x7F, 
            0x7F, 0x7F, 0x1E, 0x11, 0x80, 0x90, 0x10, 0x90, 0x0E, 0x86, 0x01, 0x7F, 0x0E, 0x7F, 0x1E, 0x7F, 
            0xFF, 0x11, 0x00, 0x10, 0x06, 0x0E, 0x7F, 0x1F, 0x1F, 0x7F, 0x0F, 0x7F, 0x3F, 0x3F, 0x3F, 0x11, 
            0x0E, 0x1F, 0x10, 0x4E, 0x91, 0x1F, 0x2F, 0x0E, 0x7F, 0x7F, 0x00, 0x7F, 0xFF, 0x2F, 0x7F, 0x00, 
            0x3F, 0x06, 0x6F, 0x1F, 0x01, 0x0E, 0x3F, 0x07, 0x7F, 0x21, 0x7F, 0x10, 0x00, 0x01, 0x2F, 0x07, 
            0x10, 0x61, 0x10, 0x10, 0x1E, 0x10, 0x10, 0x10, 0x90, 0x10, 0x00, 0x10, 0x90, 0x00, 0x01, 0x21, 
            0x8E, 0x0E, 0x1F, 0x06, 0x1F, 0x0E, 0x1F, 0x0F, 0xFF, 0x81, 0x61, 0x0F, 0x0E, 0x01, 0x01, 0x0E, 
            0x2F, 0x7F, 0x16, 0x7F, 0x8F, 0xFF, 0x3F, 0x3F, 0x1E, 0x00, 0x0E, 0x7F, 0xFF, 0x7F, 0x0E, 0x0F, 
            0x7F, 0x0F, 0x1E, 0x01, 0xBF, 0x81, 0x7F, 0x0E, 0x0F, 0x81, 0x21, 0x3F, 0x7F, 0x0F, 0x7F, 0x31, 
            0xFF, 0x8F, 0x3F, 0x00, 0x90, 0x1F, 0x1F, 0x00, 0x86, 0x90, 0x9E, 0x10, 0x21, 0x10, 0x00, 0x01, 
            0x8E, 0x8F, 0x2F, 0x90, 0x71, 0xAF, 0x9F, 0x1F, 0x7F, 0x3F, 0x9F, 0x7F, 0xFF, 0x1F, 0x0F, 0x3F, 
            0x86, 0x1F, 0x21, 0x7F, 0x90, 0x00, 0x0E, 0x0F, 0x7F, 0x01, 0x0F, 0x1F, 0x2F, 0x7F, 0x2F, 0x1F, 
            0x7F, 0x7F, 0x3F, 0x0E, 0x7F, 0x0E, 0x1F, 0x3F, 0xFF, 0xFF, 0x01, 0x00, 0x90, 0x11, 0x1F, 0x1E, 
            0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x8E, 0x10, 0xA1, 0x90, 0x10, 0x80, 0x10, 0x90, 0x1E, 0x10, 
            0x07, 0x6F, 0xAF, 0x00, 0x8E, 0xFF, 0xBF, 0x1E, 0xFF, 0x3F, 0x1F, 0x8E, 0x81, 0xFF, 0xFF, 0x81, 
            0x00, 0x06, 0x7F, 0x9E, 0x80, 0x80, 0x01, 0x86, 0x8E, 0x80, 0xFF, 0x86, 0x1F, 0x3F, 0x01, 0xBF, 
            0x81, 0x9F, 0x00, 0x0E, 0xEF, 0x0F, 0x7F, 0x7F, 0x8F, 0x1F, 0x0E, 0x0E, 0x0F, 0x01, 0x4F, 0x3F, 
            0xFF, 0x9E, 0x0E, 0x80, 0xBF, 0x81, 0x21, 0x0E, 0x00, 0x00, 0x0E, 0x10, 0xC0, 0x00, 0x00, 0x00, 
            0x10, 0x00, 0x10, 0x10, 0x10, 0x00, 0x0E, 0x1E, 0x00, 0x00, 0x1F, 0x01, 0x1E, 0x0E, 0x3F, 0x7F, 
            0x1E, 0x7F, 0x0F, 0x3F, 0x0E, 0x4F, 0x0F, 0x1E, 0x2F, 0x2F, 0x00, 0x0F, 0x1F, 0x01, 0x7F, 0x10, 
            0x27, 0x1F, 0x00, 0x1E, 0x7F, 0x2F, 0x7F, 0x01, 0x06, 0x11, 0x1F, 0x06, 0x21, 0x0F, 0x0E, 0x6F, 
            0x1F, 0x0F, 0x0E, 0x61, 0x80, 0x9E, 0xFF, 0x2F, 0x81, 0x00, 0x01, 0x81, 0xBF, 0x21, 0xCF, 0x8E, 
            0x01, 0x1F, 0x0E, 0x6F, 0x9F, 0x9E, 0x1F, 0x90, 0x00, 0x80, 0x1E, 0x80, 0x10, 0x10, 0x90, 0x00, 
            0x80, 0x80, 0x00, 0x00, 0x80, 0x1E, 0x1E, 0x1E, 0x90, 0x00, 0x00, 0x3F, 0x7F, 0x0E, 0x3F, 0x0F, 
            0xFF, 0x9F, 0x1F, 0xBF, 0xFF, 0x80, 0x61, 0x7F, 0x1F, 0x7F, 0x7F, 0x3F, 0x8E, 0x0E, 0x07, 0x1F, 
            0x80, 0x1F, 0x0F, 0x0F, 0x9F, 0x06, 0x00, 0x01, 0x7F, 0x7F, 0x6F, 0x00, 0x7F, 0x7F, 0x21, 0x3F, 
            0x80, 0x8E, 0x0E, 0x0F, 0x0F, 0x0E, 0x0E, 0x7F, 0xAF, 0x61, 0x6F, 0x6F, 0xEF, 0x0F, 0x6F, 0x00, 
            0x21, 0x1E, 0x0F, 0x1E, 0x00, 0x1E, 0x21, 0x10, 0x00, 0x10, 0x00, 0x1E, 0x00, 0x10, 0x10, 0x0F, 
            0x0E, 0x01, 0x06, 0x0F, 0xBF, 0xFF, 0x0F, 0x6F, 0x8E, 0x1F, 0x0E, 0x8F, 0x9F, 0x16, 0x8F, 0x0E, 
            0x8E, 0x1F, 0x1F, 0x1F, 0xFF, 0x8E, 0x1E, 0x81, 0x01, 0x86, 0x6F, 0x8F, 0x7F, 0x9E, 0x01, 0x7F, 
            0x80, 0x1E, 0x0F, 0x7F, 0xFF, 0x1E, 0x1E, 0x0E, 0x9E, 0x0F, 0x10, 0x10, 0x10, 0x00, 0x00, 0x1E, 
            0x00, 0x00, 0x1E, 0x10, 0x00, 0x10, 0x0F, 0x3F, 0x00, 0x2F, 0x0E, 0x00, 0x8E, 0x7F, 0x7F, 0x0F, 
            0x1E, 0x1E, 0x0E, 0x00, 0x0E, 0x01, 0x6F, 0x00, 0x1F, 0x7F, 0x01, 0x00, 0x1E, 0x0F, 0x1F, 0x1F, 
            0x8F, 0x01, 0x0F, 0x0E, 0x6F, 0x0E, 0x2F, 0x6F, 0xB1, 0x01, 0x00, 0x3F, 0x8E, 0x0E, 0x10, 0x1E, 
            0x80, 0x00, 0x00, 0x10, 0x10, 0x00, 0x00, 0x1E, 0x80, 0xAF, 0x0E, 0x80, 0x1F, 0x10, 0x0E, 0x6F, 
            0x0F, 0x31, 0x7F, 0x0F, 0xFF, 0x0F, 0x7F, 0x01, 0x2F, 0x1F, 0x0E, 0x1F, 0x01, 0x0F, 0x0E, 0x1F, 
            0x0F, 0x0F, 0x2F, 0x01, 0x01, 0x0E, 0x0F, 0x00, 0x0E, 0x0F, 0x0E, 0x01, 0x3F, 0x11, 0x3F, 0x01, 
            0xA1, 0x80, 0x00, 0x4E, 0x91, 0x00, 0x00, 0x1E, 0x9E, 0x1E, 0x0E, 0x21, 0x10, 0x00, 0x0F, 0x00, 
            0x8F, 0x81, 0x1F, 0x8F, 0x3F, 0x8E, 0x0F, 0x00, 0x7F, 0x1F, 0x3F, 0x7F, 0xBF, 0x0F, 0x0E, 0x0F, 
            0x0E, 0x1F, 0x0F, 0x0E, 0x6F, 0x01, 0x1E, 0x06, 0x2F, 0x0E, 0x21, 0x10, 0x10, 0x00, 0x0E, 0x06, 
            0xAF, 0x0E, 0x0E, 0x0E, 0x0F, 0x1F, 0x7F, 0x2F, 0x81, 0x00, 0x21, 0x01, 0x0F, 0x0E, 0x1E, 0x3F, 
            0x0E, 0x00, 0x00, 0x1E, 0x21, 0x0F, 0x00, 0x0E, 0x0E, 0x0F, 0x0F, 0x00, 0x0E, 0x87, 0x01, 0x01, 
            0x21, 0x00, 0x21, 0x10, 0x80, 0x6F, 0x0F, 0x0F, 0x8F, 0x01, 0x0E, 0x0F, 0xCF, 0x0F, 0x6F, 0x10, 
            0x00, 0x01, 0x0E, 0x0F, 0x6F, 0x8E, 0x2F, 0x9F, 0x00, 0x80, 0x00, 0x00, 0x7F, 0x0E, 0x0F, 0x00, 
            0xAF, 0x2F, 0x2F, 0x2F, 0x80, 0x0F, 0x0E, 0x00, 0xAF, 0x0F, 0x5F, 0x0F, 0x0F, 0x47, 0x21, 0x01, 
            0x81, 0x6F, 0x00, 0x2F, 0x1E, 0x80, 0x00, 0x3F, 0x21, 0x01, 0x7F, 0x7F, 0x81, 0x9F, 0x7F, 0x1F, 
            0xFF, 0x1F, 0x10, 0xA1, 0x9F, 0x0F, 0x81, 0x0F, 0x01, 0x00, 0x0E, 0x1F, 0x80, 0x91, 0x2F, 0x80, 
            0x7F, 0x91, 0x21, 0x21, 0x10, 0x10, 0x01, 0x0F, 0x80, 0x7F, 0x7F, 0x00, 0x00, 0x31, 0x10, 0x3F, 
            0x21, 0x0F, 0x0F, 0x7F, 0x7F, 0x00, 0x10, 0x0F, 0x7F, 0x3F, 0x1F, 0x6F, 0x0F, 0x6F, 0x00, 0x21, 
            0x00, 0x01, 0x6F, 0x6F, 0x8E, 0x81, 0x6F, 0x00, 0xEF, 0x00, 0x0F, 0x8F, 0xBF, 0x01, 0x0E, 0x7F, 
            0x0F, 0x00, 0x07, 0x1F, 0x81, 0x81, 0x21, 0xA1, 0xFF, 0x9F, 0x9F, 0x00, 0x1F, 0x7F, 0x1F, 0x1F, 
            0xFF, 0x80, 0x00, 0x1F, 0x8F, 0x00, 0x00, 0x21, 0xA1, 0x10, 0x10, 0x7F, 0x00, 0x01, 0x1F, 0x01, 
            0xFF, 0x8E, 0x71, 0x80, 0x80, 0x8F, 0x01, 0x7F, 0x00, 0x06, 0x7F, 0x8E, 0x9F, 0xDF, 0x7F, 0x3F, 
            0xFF, 0x7F, 0x0F, 0x00, 0x80, 0x00, 0x00, 0x10, 0x00, 0x21, 0x10, 0x10, 0x90, 0x90, 0x10, 0x90, 
            0x90, 0xFF, 0x0F, 0x00, 0x7F, 0x21, 0x7F, 0x01, 0x1F, 0x3F, 0x7F, 0x1F, 0x00, 0x1F, 0x41, 0x1F, 
            0x4F, 0x0F, 0x0E, 0x1F, 0x0F, 0x4F, 0x1E, 0x1F, 0x00, 0x01, 0x01, 0x1F, 0x5F, 0x0E, 0x01, 0x21, 
            0x0E, 0x01, 0x7F, 0x1F, 0x10, 0x10, 0x10, 0x00, 0x61, 0x21, 0x00, 0x7F, 0x1F, 0x00, 0x00, 0x0E, 
            0x00, 0x00, 0x10, 0x7F, 0x0F, 0x0F, 0x00, 0x0F, 0x0E, 0x0F, 0x0F, 0x1E, 0x01, 0x0F, 0x0F, 0x0F, 
            0xFF, 0x91, 0x7F, 0x7F, 0xBF, 0x0E, 0x10, 0x11, 0x81, 0x11, 0xA1, 0x90, 0x21, 0x10, 0x1E, 0x00, 
            0xA1, 0x80, 0x10, 0xA1, 0x90, 0x8E, 0x0E, 0x00, 0x1F, 0x0E, 0x0E, 0x80, 0x8E, 0x8F, 0x00, 0x0E, 
            0xA1, 0x06, 0x1E, 0x7F, 0x80, 0x00, 0x2F, 0x6F, 0x61, 0x00, 0x00, 0x7F, 0x8F, 0x80, 0x7F, 0x9F, 
            0x86, 0x80, 0x3F, 0x21, 0x3F, 0x1F, 0x21, 0x00, 0x90, 0x10, 0x10, 0x00, 0x2F, 0x11, 0x01, 0x7F, 
            0x7F, 0x07, 0x0E, 0x2F, 0x1F, 0x0E, 0x01, 0x0F, 0x00, 0x0F, 0x00, 0x00, 0x8E, 0x1F, 0x0E, 0x10, 
            0x4F, 0x0E, 0x0E, 0x0E, 0x7F, 0x0F, 0x0F, 0x6F, 0x1F, 0x7F, 0x7F, 0x0E, 0x6F, 0x1F, 0x01, 0x01, 
            0x7F, 0x06, 0x7F, 0x01, 0x0F, 0x6F, 0x7F, 0x1E, 0x4F, 0x4F, 0x6F, 0x3F, 0x0F, 0x1F, 0x0E, 0x00, 
            0x8E, 0x90, 0x00, 0x00, 0x8E, 0x31, 0x00, 0x00, 0x80, 0x3F, 0x00, 0x0F, 0x06, 0x2F, 0x5F, 0x01, 
            0x1F, 0x7F, 0x00, 0x0E, 0x7F, 0x8E, 0x01, 0x4F, 0x0E, 0x0E, 0x7F, 0x00, 0x8E, 0x0F, 0x0E, 0x6F, 
            0x7F, 0x6F, 0x0F, 0x0E, 0xEF, 0x7F, 0x06, 0x0E, 0x0E, 0x07, 0x61, 0x06, 0x00, 0x2F, 0x00, 0x0F, 
            0x21, 0x21, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x8E, 0x0F, 0x6F, 0x6F, 0x1F, 0x1F, 0x0E, 0x4F, 
            0x01, 0x7F, 0x2F, 0x0E, 0x7F, 0x0F, 0x00, 0x6F, 0x7F, 0x2F, 0x6F, 0x06, 0x00, 0x0E, 0x7F, 0x01, 
            0x00, 0x41, 0x6F, 0x0E, 0xCF, 0x1E, 0x00, 0x01, 0x8F, 0x21, 0x00, 0x3F, 0x10, 0x0E, 0x00, 0x0E, 
            0x21, 0x06, 0x21, 0x00, 0x0E, 0x6F, 0x0E, 0x00, 0x01, 0x6F, 0x0E, 0x0E, 0x4F, 0x0E, 0x7F, 0x0F, 
            0x1F, 0x1F, 0x0E, 0x7F, 0x00, 0x7F, 0x0E, 0x0F, 0x0F, 0x5F, 0x0F, 0x0E, 0x01, 0x01, 0x0E, 0x0F, 
            0x0F, 0x0E, 0x10, 0x7F, 0x0F, 0x00, 0x01, 0x21, 0x0E, 0x10, 0x10, 0x00, 0xEF, 0x8F, 0x7F, 0x7F, 
            0x8F, 0x0F, 0x01, 0x0E, 0x80, 0x0E, 0x0F, 0x00, 0x7F, 0x5F, 0x00, 0xAF, 0x8E, 0x00, 0x6F, 0x0E, 
            0x00, 0x8F, 0x21, 0x6F, 0x00, 0x00, 0x00, 0x00, 0x21, 0x21, 0x6F, 0x65, 0x17, 0x0E, 0x6F, 0x06, 
            0x00, 0x7F, 0x0F, 0x00, 0x01, 0x7F, 0x0E, 0x6F, 0x0F, 0x06, 0x21, 0x00, 0x21, 0x01, 0x4F, 0x06, 
            0x01, 0x7F, 0x0F, 0x6F, 0x8F, 0x7F, 0x0F, 0x21, 0x6F, 0x6F, 0x0F, 0x0E, 0x21, 0x0F, 0x6F, 0x6F, 
            0x00, 0x01, 0x0F, 0x06, 0x7F, 0x21, 0x00, 0x0E, 0x0E, 0x0F, 0x00, 0x7F, 0x7F, 0x00, 0x31, 0x6F, 
            0xB1, 0x21, 0x21, 0x11, 0x8E, 0x00, 0x10, 0x21, 0x80, 0x7F, 0x6F, 0x61, 0x2F, 0x6F, 0x0E, 0x06, 
            0x80, 0x80, 0x3B, 0x0E, 0x10, 0x81, 0x0F, 0x10, 0x00, 0x21, 0x21, 0x21, 0x0F, 0x1E, 0x0E, 0x7F, 
            0x21, 0x00, 0x0F, 0x01, 0x5F, 0x7F, 0x7F, 0x0E, 0x21, 0x7F, 0x7F, 0x21, 0x7F, 0x01, 0x00, 0x10, 
            0x0E, 0x01, 0x6F, 0x01, 0x1F, 0x3F, 0x00, 0x7F, 0x81, 0xFF, 0x7F, 0x7F, 0xAF, 0x7F, 0x7F, 0x9F, 
            0x81, 0x7F, 0x21, 0x01, 0x10, 0x01, 0x01, 0x7F, 0xBF, 0x90, 0x10, 0x80, 0x1E, 0x00, 0x6F, 0x01, 
            0x0E, 0x10, 0x10, 0x01, 0xFF, 0x0E, 0x01, 0x7F, 0xFF, 0x7F, 0x0E, 0x10, 0x80, 0x00, 0x7F, 0x7F, 
            0x6F, 0x7F, 0x01, 0x2F, 0xFF, 0xFF, 0x10, 0x7F, 0x3F, 0x01, 0x7F, 0x21, 0x0E, 0x21, 0x01, 0x01, 
            0xFF, 0x0F, 0x7F, 0x21, 0x80, 0x31, 0x00, 0x6F, 0xA1, 0x2F, 0x6F, 0x4F, 0x61, 0x4F, 0x6F, 0x01, 
            0x8E, 0x86, 0x0F, 0xFF, 0x7F, 0x81, 0x1F, 0x06, 0x7F, 0x1F, 0x00, 0x0F, 0x8F, 0xFF, 0x0E, 0x0E, 
            0x8E, 0x3F, 0x7F, 0x7F, 0xFF, 0x0E, 0x10, 0x10, 0x10, 0x00, 0x7F, 0x01, 0x81, 0xBF, 0x0F, 0x8E, 
            0x8E, 0x80, 0x9F, 0x7F, 0x00, 0x01, 0xFF, 0xDF, 0x8E, 0xFF, 0x00, 0x21, 0x80, 0x2F, 0x1E, 0x01, 
            0x8E, 0x0E, 0x0E, 0x3F, 0x3F, 0x7F, 0x0E, 0x0F, 0xFF, 0x90, 0x10, 0x90, 0xE1, 0x81, 0x1F, 0x1F, 
            0x01, 0x0E, 0x3F, 0x0F, 0x8E, 0x8E, 0x0E, 0x01, 0x80, 0x7F, 0x2F, 0x7F, 0x0F, 0x0E, 0x7F, 0x00, 
            0x0E, 0x0F, 0x00, 0x4F, 0x01, 0x0E, 0x7F, 0x86, 0x0F, 0xE1, 0x1F, 0x7F, 0x0E, 0x0F, 0x0E, 0x0E, 
            0x7F, 0x7F, 0x1F, 0x7F, 0x7F, 0x00, 0x06, 0x10, 0x1E, 0x61, 0x7F, 0x0E, 0x7F, 0x5F, 0x0F, 0x01, 
            0x0E, 0x0E, 0x00, 0x1F, 0x0F, 0x3F, 0x6F, 0x01, 0x7F, 0x0E, 0x6F, 0x21, 0xDE, 0x86, 0x4F, 0x01, 
            0x80, 0x0F, 0x0F, 0x0E, 0x8F, 0x0E, 0x2F, 0x07, 0x40, 0x10, 0x0E, 0x7F, 0x86, 0x8E, 0x06, 0xFF, 
            0x8F, 0xFF, 0x00, 0x7F, 0x11, 0x10, 0x10, 0x90, 0x90, 0xBF, 0x0E, 0x00, 0x9F, 0x01, 0x01, 0x0E, 
            0x80, 0x7F, 0x01, 0x1F, 0x00, 0x0E, 0x0E, 0x0E, 0xFF, 0x86, 0x21, 0xCE, 0x7F, 0x80, 0x07, 0x7F, 
            0x7F, 0x00, 0x00, 0x1F, 0x81, 0xFF, 0x00, 0x01, 0x81, 0x7F, 0x01, 0x00, 0x80, 0x0F, 0x06, 0x21, 
            0x01, 0x10, 0x10, 0x0E, 0x90, 0xAF, 0x7F, 0x8E, 0x8E, 0x8E, 0x00, 0x00, 0x7F, 0x00, 0x0E, 0x00, 
            0xBF, 0x6F, 0x0E, 0x0E, 0x87, 0x01, 0x7F, 0x0F, 0x0F, 0x7F, 0x1E, 0x00, 0x0F, 0x0E, 0x00, 0x00, 
            0x0F, 0x8E, 0x0E, 0x8E, 0x00, 0x1F, 0x06, 0x0E, 0x0E, 0x00, 0x7F, 0x0E, 0x86, 0x8E, 0x0F, 0x0E, 
            0x8E, 0x00, 0x7F, 0x01, 0x8E, 0x0F, 0x6F, 0x1F, 0x00, 0x0E, 0x3F, 0x00, 0xAF, 0x8F, 0x00, 0x80, 
            0x0E, 0x80, 0x00, 0x00, 0x10, 0x10, 0x10, 0x1F, 0x8E, 0x00, 0x1F, 0x0F, 0x8E, 0x06, 0x01, 0x3F, 
            0x00, 0x00, 0x00, 0x0E, 0x0E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0E, 0x0F, 0x0E, 0x80, 0x0E, 0x0E, 
            0x0F, 0x00, 0x7F, 0x0F, 0x0E, 0x0E, 0x01, 0x0E, 0x00, 0x01, 0x7F, 0x7F, 0x0F, 0x0E, 0x0F, 0x0F, 
            0x01, 0x00, 0x0F, 0x00, 0x40, 0x0E, 0x0E, 0x0F, 0x80, 0x06, 0x10, 0x00, 0x0F, 0x00, 0x0E, 0x0E, 
            0x8E, 0x0F, 0x1F, 0x1F, 0x8E, 0x1F, 0x0E, 0x0E, 0x9F, 0x0E, 0x0F, 0x0F, 0x2F, 0x2F, 0x3F, 0x0F, 
            0x8E, 0xFF, 0x7F, 0x80, 0x4F, 0x80, 0x0F, 0x0E, 0x0F, 0x7F, 0x0E, 0x2F, 0xFF, 0x06, 0x00, 0x01, 
            0xA1, 0x00, 0x10, 0x00, 0x90, 0x0E, 0x3F, 0x2F, 0x1E, 0x06, 0x0F, 0x0E, 0x1F, 0x86, 0x06, 0x86, 
            0x1F, 0xFF, 0x0E, 0x3F, 0x00, 0x0F, 0x7F, 0x00, 0x8E, 0xFF, 0x0F, 0x1E, 0x8E, 0x0E, 0x0F, 0x00, 
            0x80, 0x10, 0x00, 0x0E, 0x0E, 0x0F, 0x0F, 0x0E, 0x8E, 0x86, 0x25, 0xAF, 0x00, 0x8F, 0x00, 0x0E, 
            0x01, 0x07, 0x06, 0x0E, 0x80, 0x4E, 0x00, 0x06, 0x0F, 0x7F, 0x00, 0x6F, 0x6F, 0x00, 0x00, 0x06, 
            0x0E, 0x00, 0x00, 0x0E, 0x0E, 0x00, 0x7F, 0x1F, 0x80, 0x0F, 0x0E, 0x0F, 0x00, 0x0F, 0x0E, 0x00, 
            0x8E, 0x0F, 0x21, 0x06, 0x0E, 0x0F, 0x0E, 0x0E, 0x0E, 0x00, 0x6F, 0x00, 0x2F, 0x0F, 0x0E, 0x0E, 
            0x2F, 0x00, 0x3F, 0x00, 0x1F, 0x00, 0x21, 0x10, 0x0F, 0x00, 0x00, 0x00, 0x86, 0x8E, 0x2F, 0x0E, 
            0xFF, 0x01, 0x00, 0x6F, 0x80, 0x2F, 0x00, 0x0E, 0x0E, 0x00, 0x00, 0x0F, 0x8F, 0x0E, 0x0E, 0x80, 
            0x7F, 0xFF, 0x01, 0x1E, 0x7F, 0x7F, 0x2F, 0x7F, 0xFF, 0x90, 0x7F, 0x3F, 0xFF, 0x7F, 0x01, 0x7F, 
            0x80, 0x00, 0x1E, 0x7F, 0x7F, 0x3F, 0x31, 0x00, 0x81, 0xFF, 0x10, 0xEF, 0x1E, 0x80, 0x0F, 0x01, 
            0x7F, 0x7F, 0x0E, 0x3F, 0x80, 0xFF, 0x00, 0x0E, 0xEF, 0x00, 0x7F, 0x0E, 0xC1, 0x1F, 0x01, 0x0F, 
            0x3B, 0x1F, 0x00, 0x3F, 0x9F, 0xFF, 0x01, 0xFF, 0x7F, 0xFF, 0x81, 0x7F, 0x7F, 0xA1, 0x8F, 0x7F, 
            0x01, 0x01, 0x0E, 0x5F, 0x01, 0x7F, 0x41, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x31, 0x10, 0x21, 
            0x10, 0x10, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x0E, 0x0F, 0x01, 0x10, 0x10, 0xCF, 0x8F, 0x6F, 0x01, 
            0xFF, 0x01, 0x01, 0x7F, 0xFF, 0x7F, 0x7F, 0x01, 0x7F, 0x7F, 0x01, 0x00, 0x81, 0x90, 0x10, 0xFF, 
            0x6F, 0x8F, 0x7F, 0x61, 0x7F, 0x7F, 0x7F, 0x7F, 0x80, 0xA1, 0x0E, 0x0E, 0xFF, 0x0F, 0x8E, 0x81, 
            0xFF, 0x8E, 0xFF, 0x7F, 0x6F, 0x00, 0x0F, 0x41, 0xAF, 0x8F, 0x01, 0xA1, 0x00, 0xB1, 0xFF, 0x7F, 
            0x0F, 0xFF, 0x6F, 0x86, 0xFF, 0xFF, 0x6F, 0x6F, 0xFF, 0x6F, 0x0E, 0x6F, 0xEF, 0x00, 0x7F, 0x4F, 
            0x3F, 0x0F, 0x0E, 0x21, 0x81, 0xEF, 0x6F, 0x80, 0xFF, 0x90, 0xFF, 0x10, 0x10, 0x01, 0x21, 0x31, 
            0x80, 0xFF, 0x21, 0x0E, 0x7F, 0x21, 0x31, 0x6F, 0xEF, 0x7F, 0x7F, 0x6F, 0x0F, 0x6F, 0x6F, 0x7F, 
            0x0E, 0x7F, 0x0E, 0x21, 0x10, 0x9E, 0x7F, 0x01, 0x10, 0x40, 0x3F, 0x00, 0x10, 0x10, 0x21, 0x0F, 
            0x21, 0x00, 0x3F, 0x01, 0x7F, 0x16, 0x01, 0x10, 0x6F, 0x01, 0x0E, 0x01, 0x1F, 0x21, 0x01, 0x00, 
            0x0F, 0x7F, 0x01, 0x0E, 0x10, 0x01, 0x01, 0x0E, 0xFF, 0xFF, 0x7F, 0x7F, 0xFF, 0x31, 0x7F, 0x7F, 
            0xFF, 0x3F, 0x90, 0x00, 0x0E, 0x7F, 0x6F, 0x0E, 0xFF, 0x90, 0x31, 0xFF, 0x0E, 0xEF, 0xFF, 0x3F, 
            0x3F, 0x7F, 0x00, 0xA1, 0x8E, 0xFF, 0x0E, 0x00, 0x8E, 0x1F, 0x01, 0x40, 0x8F, 0x0E, 0x31, 0x00, 
            0x7F, 0x31, 0x4F, 0x1F, 0xEF, 0xFF, 0x10, 0x8F, 0x8F, 0x8F, 0x0E, 0x00, 0x6F, 0x0F, 0x3F, 0x7F, 
            0x81, 0xFF, 0x00, 0x06, 0x8F, 0x01, 0x21, 0x00, 0x80, 0x7F, 0x1F, 0x07, 0x0F, 0x01, 0x0E, 0x10, 
            0x00, 0x10, 0x10, 0x80, 0x80, 0x80, 0x06, 0x81, 0x1F, 0x0E, 0x0F, 0x0E, 0xBF, 0x1E, 0x00, 0x0F, 
            0x7F, 0x7F, 0x0F, 0x0E, 0x3F, 0x06, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0F, 0x00, 0x0F, 
            0x0F, 0x6F, 0x1F, 0x1F, 0x0E, 0x0E, 0x2F, 0x0E, 0xAF, 0xFF, 0x0F, 0xFF, 0xFF, 0x1E, 0x0E, 0x0E, 
            0x86, 0x7F, 0x81, 0x7F, 0x01, 0x5F, 0x2F, 0x7F, 0xFF, 0x80, 0x61, 0xA1, 0x21, 0x90, 0x61, 0x10, 
            0x40, 0x10, 0x00, 0x00, 0x90, 0x21, 0x06, 0x2F, 0x86, 0x0F, 0x0E, 0x1F, 0x0E, 0x00, 0x0E, 0x0E, 
            0x0E, 0x00, 0x1F, 0x0E, 0x06, 0x00, 0x0E, 0x0F, 0x8E, 0xFF, 0x0E, 0x0E, 0x00, 0x01, 0x0E, 0x0E, 
            0x61, 0x31, 0x00, 0x00, 0x10, 0x10, 0x10, 0x00, 0x7F, 0x21, 0x3F, 0x00, 0x0E, 0x3F, 0x0F, 0x61, 
            0x7F, 0x0F, 0x01, 0x00, 0x4F, 0x21, 0x6F, 0x0E, 0x06, 0x0E, 0x21, 0x7F, 0x80, 0x6F, 0x00, 0x0E, 
            0x8E, 0x0F, 0x10, 0x10, 0x80, 0x00, 0x1F, 0x7F, 0x00, 0x00, 0x00, 0x2F, 0x0E, 0xCF, 0x7F, 0x8E, 
            0x00, 0xEF, 0x0F, 0x00, 0x7F, 0x21, 0x7F, 0x6F, 0xA1, 0x6F, 0x2F, 0x3F, 0x80, 0x0E, 0x1E, 0x2F, 
            0x8E, 0x00, 0x2F, 0x0F, 0x1F, 0x0E, 0x1F, 0x7F, 0x0E, 0xFF, 0x00, 0x81, 0x00, 0x10, 0x1E, 0x00, 
            0x0E, 0x0F, 0x06, 0x0E, 0x9F, 0x8E, 0x8E, 0x0E, 0x86, 0x07, 0x0E, 0x80, 0x86, 0x9F, 0x90, 0x0F, 
            0x06, 0x0E, 0x2F, 0x0F, 0x80, 0x86, 0x01, 0x9E, 0x01, 0x86, 0x1F, 0x7F, 0x7F, 0x80, 0x2F, 0x80, 
            0xEF, 0x0F, 0x27, 0x21, 0x80, 0x0F, 0x00, 0x00, 0x90, 0x0F, 0x00, 0x10, 0x21, 0x10, 0x0E, 0x01, 
            0x01, 0x81, 0x0F, 0x8E, 0x01, 0x0E, 0x00, 0x0E, 0x0E, 0x7F, 0x01, 0x1E, 0xFF, 0x21, 0x00, 0x7F, 
            0x01, 0x0E, 0x0E, 0x21, 0x10, 0x00, 0x21, 0x0E, 0x00, 0x0F, 0x0F, 0x00, 0x0E, 0x0E, 0x0F, 0x0F, 
            0x0E, 0x0F, 0x3F, 0x01, 0x2F, 0x00, 0x0E, 0x6F, 0x8F, 0x0E, 0x0F, 0x61, 0xA1, 0x06, 0x00, 0x00, 
            0xA1, 0x00, 0x0F, 0x0F, 0x0F, 0x0F, 0x00, 0x0E, 0x00, 0x1F, 0x0E, 0x01, 0x0E, 0x7F, 0x0E, 0x0F, 
            0x4F, 0x0E, 0x6F, 0x00, 0x80, 0x00, 0x00, 0x0F, 0x8E, 0x0E, 0x4E, 0x01, 0xAF, 0x0E, 0x2F, 0x06, 
            0x0F, 0x0E, 0x0F, 0x00, 0x8F, 0x0E, 0x00, 0xBF, 0x6F, 0x21, 0x6F, 0x00, 0x6F, 0x6F, 0x00, 0x00, 
            0x8E, 0x00, 0x0E, 0x0F, 0x80, 0x10, 0x0E, 0x0F, 0x80, 0x2F, 0x00, 0x0F, 0x21, 0x7F, 0x01, 0x0E, 
            0xA1, 0x0E, 0x6F, 0xA1, 0x0E, 0x0E, 0x6F, 0x00, 0x0F, 0x01, 0x00, 0x3B, 0x00, 0x7F, 0x7F, 0x0E, 
            0x0F, 0x7F, 0x5E, 0x21, 0x01, 0x7F, 0x7F, 0x7F, 0x7F, 0x11, 0x00, 0x7F, 0x00, 0x00, 0x7F, 0x10, 
            0x0E, 0x7F, 0x7F, 0x7F, 0x7F, 0x21, 0x00, 0x7F, 0x81, 0x8F, 0x00, 0x21, 0x80, 0x7F, 0x7F, 0x86, 
            0x81, 0x10, 0xFF, 0x7F, 0x06, 0x10, 0x7F, 0x01, 0x90, 0x80, 0x0E, 0xA1, 0xFF, 0x81, 0x06, 0x90, 
            0x10, 0xFF, 0x01, 0x00, 0x9F, 0x9F, 0xFF, 0x8E, 0x9F, 0x7F, 0x3F, 0xFF, 0x90, 0xFF, 0x80, 0x8F, 
            0x0F, 0x0E, 0x0E, 0x0E, 0x8E, 0xEF, 0x10, 0x90, 0x0F, 0x8E, 0x00, 0x6F, 0x00, 0xFF, 0x11, 0x21, 
            0xA1, 0x90, 0x01, 0x6F, 0x8E, 0x01, 0x6F, 0x3F, 0xFF, 0x00, 0x00, 0x10, 0x10, 0x7F, 0x0F, 0x40, 
            0xEF, 0x86, 0x10, 0xAF, 0xFF, 0xFF, 0x00, 0x61, 0x00, 0x81, 0x06, 0x0F, 0xFF, 0x86, 0x2F, 0x0E, 
            0x81, 0x01, 0x00, 0x0E, 0x3F, 0x7F, 0x01, 0x2F, 0x0F, 0x0E, 0x00, 0x1F, 0x07, 0x8E, 0x11, 0x6F, 
            0x0E, 0x7F, 0x7F, 0x6F, 0x21, 0x00, 0x0E, 0x0E, 0x0F, 0x0E, 0x0F, 0x0F, 0x0F, 0x0F, 0x07, 0x0E, 
            0x0F, 0x00, 0x7F, 0x7F, 0x7F, 0x2F, 0x3F, 0x40, 0x7F, 0x6F, 0x3B, 0x7F, 0x7F, 0x7F, 0x6F, 0x3B, 
            0x1F, 0x21, 0x0E, 0x21, 0xFF, 0x80, 0x10, 0x7F, 0x8E, 0x0E, 0x7F, 0x9F, 0x8E, 0x06, 0x81, 0x7F, 
            0x10, 0x10, 0x00, 0x10, 0x90, 0xFF, 0x3F, 0xFF, 0x00, 0x90, 0xFF, 0x0E, 0x7F, 0x00, 0x10, 0x10, 
            0xFF, 0x80, 0x0F, 0x0E, 0x8F, 0x1F, 0x7F, 0xFF, 0x8F, 0x00, 0x0F, 0x6F, 0x0F, 0x7F, 0x0E, 0x00, 
            0x8E, 0x8F, 0x0E, 0x9F, 0x8E, 0xFF, 0x7F, 0x7F, 0x7F, 0x1E, 0x00, 0x00, 0x80, 0x81, 0x5F, 0x01, 
            0x80, 0x2F, 0x6F, 0x21, 0x8E, 0x0E, 0x0F, 0x0E, 0x6F, 0x7F, 0x7F, 0x0F, 0x0F, 0x80, 0x0F, 0xA1, 
            0xE1, 0x9F, 0x1F, 0x7F, 0x0F, 0x8F, 0x7F, 0x0E, 0x8E, 0x0E, 0x6F, 0x7F, 0x8F, 0x2F, 0x0F, 0x6F, 
            0x6F, 0x2F, 0x6F, 0x6F, 0x00, 0x0F, 0x0E, 0x0E, 0x3F, 0x2F, 0x10, 0x01, 0x6F, 0x00, 0x0E, 0x0E, 
            0x21, 0x2F, 0x0E, 0x6F, 0xBB, 0x80, 0xFF, 0x7F, 0xA1, 0x00, 0x7F, 0x61, 0xA1, 0x00, 0x3F, 0x3F, 
            0x10, 0x7F, 0x1F, 0x31, 0xFF, 0x8E, 0x01, 0x8F, 0x1F, 0xA1, 0x7F, 0x3F, 0x21, 0x21, 0x01, 0x7F, 
            0xA1, 0x71, 0x0E, 0x7F, 0xEF, 0x7F, 0x21, 0x7F, 0x7F, 0x01, 0x0E, 0x7F, 0x01, 0x0F, 0x00, 0x7F, 
            0x10, 0x21, 0x0F, 0x0F, 0x8F, 0x31, 0x7F, 0x7F, 0x0F, 0x7F, 0x10, 0x01, 0x81, 0x3F, 0x0F, 0x31, 
            0x0F, 0x7F, 0x00, 0x0E, 0x21, 0x6F, 0x0F, 0x2F, 0x2F, 0x10, 0x50, 0x01, 0x7F, 0x01, 0x21, 0x01, 
            0x9F, 0x21, 0x00, 0x0E, 0x0F, 0x21, 0x0E, 0x01, 0xEF, 0x8E, 0x6F, 0x0E, 0xEF, 0x01, 0x6F, 0x0E, 
            0x9B, 0x21, 0x10, 0x31, 0x0F, 0x10, 0x7F, 0x7F, 0x9F, 0xEF, 0x00, 0x41, 0x21, 0x3F, 0x00, 0x00, 
            0x01, 0x3B, 0x7F, 0x01, 0x9F, 0x0E, 0x71, 0x6F, 0x80, 0x7F, 0x7F, 0x6F, 0x7F, 0x7F, 0x00, 0x0F, 
            0x7F, 0x7F, 0x01, 0x3F, 0x0F, 0x01, 0x0E, 0x7F, 0x0F, 0x7F, 0x01, 0x10, 0x7F, 0x0F, 0x0F, 0x6F, 
            0xFF, 0xFF, 0x3F, 0x21, 0xB1, 0x7F, 0x06, 0xFF, 0xBF, 0x9F, 0xFF, 0x7F, 0x7F, 0x01, 0x00, 0x01, 
            0xFF, 0xEF, 0x7F, 0xA1, 0x00, 0x90, 0x0F, 0x7F, 0x7F, 0xFF, 0x80, 0x0E, 0x9F, 0x01, 0x6F, 0x0E, 
            0x6F, 0x7F, 0x01, 0x00, 0x81, 0x0E, 0x0E, 0x01, 0x7F, 0x6F, 0x7F, 0x0F, 0x0F, 0x3F, 0x7F, 0x0F, 
            0x81, 0x01, 0x0E, 0x21, 0x21, 0x5E, 0x06, 0x7F, 0x81, 0x6F, 0x00, 0x0E, 0x3F, 0x7F, 0x0E, 0x0E, 
            0x0E, 0x0E, 0x00, 0x7F, 0x11, 0x7F, 0x10, 0x01, 0x01, 0x1F, 0x00, 0x01, 0x7F, 0x7F, 0x00, 0x16, 
            0x1F, 0x1F, 0x1F, 0x01, 0x8F, 0x8E, 0x7F, 0x7F, 0xFF, 0x7F, 0x00, 0x00, 0xA1, 0x3F, 0x2F, 0x00, 
            0x7F, 0x1F, 0x01, 0x0E, 0x3F, 0x0E, 0x00, 0x10, 0x0F, 0x81, 0x1F, 0x7F, 0x01, 0x1F, 0x1F, 0x0F, 
            0xA1, 0x3F, 0x01, 0x0F, 0x86, 0x7F, 0x01, 0x0E, 0xBF, 0x00, 0x0F, 0x3F, 0x01, 0x7F, 0x10, 0x7F, 
            0x9E, 0x10, 0x10, 0x90, 0x10, 0x90, 0x10, 0x01, 0x00, 0x0E, 0x1F, 0x0E, 0x8E, 0x9F, 0x3F, 0x7F, 
            0xAF, 0x06, 0x7F, 0x0F, 0x9F, 0x3F, 0x7F, 0x0F, 0x01, 0x3F, 0x0F, 0x3F, 0xFF, 0xFF, 0x0E, 0x81, 
            0x7F, 0x7F, 0x0E, 0x00, 0x07, 0x7F, 0x3F, 0x7F, 0x7F, 0x3F, 0x7F, 0x3F, 0x0E, 0x0F, 0x0E, 0x7F, 
            0x00, 0x21, 0x0E, 0x0F, 0x0E, 0x1F, 0x00, 0x0E, 0x00, 0x0E, 0x21, 0x10, 0x10, 0x00, 0x00, 0x10, 
            0x0F, 0x7F, 0x7F, 0x7F, 0x86, 0x8E, 0x2F, 0x0F, 0x81, 0x0F, 0x21, 0x31, 0x8F, 0x7F, 0x80, 0x00, 
            0x7F, 0x01, 0x75, 0x0E, 0x8F, 0xFF, 0x01, 0x81, 0x06, 0xFF, 0xBF, 0x8E, 0x00, 0x7F, 0x0E, 0x2F, 
            0xA1, 0x81, 0x7F, 0x7F, 0xFF, 0x6F, 0x06, 0x1F, 0xFF, 0x7F, 0xFF, 0x3F, 0x7F, 0x7F, 0x06, 0x7F, 
            0xFF, 0x81, 0x0E, 0x90, 0x80, 0xA1, 0x10, 0x21, 0x10, 0x10, 0x10, 0x90, 0x90, 0x90, 0x01, 0x1F, 
            0x8E, 0x2F, 0x01, 0x3F, 0xBF, 0x6F, 0x0F, 0x0E, 0x0F, 0x7F, 0x2F, 0x21, 0xFF, 0xFF, 0x0F, 0x80, 
            0x8E, 0x81, 0x3F, 0x01, 0x7F, 0x0F, 0x7F, 0x2F, 0x81, 0x80, 0x7F, 0x7F, 0x06, 0x1F, 0x01, 0x7F, 
            0x7F, 0x00, 0x0F, 0x7F, 0x41, 0x01, 0x31, 0x61, 0x9F, 0x21, 0x21, 0x10, 0x10, 0x10, 0x00, 0x10, 
            0x4F, 0x1F, 0x7F, 0x61, 0x7F, 0x2F, 0x6F, 0x0F, 0x7F, 0x0E, 0x0E, 0x1F, 0x7F, 0x6F, 0x0F, 0x0E, 
            0x06, 0x0E, 0x01, 0x06, 0x0F, 0x7F, 0x3F, 0x6F, 0x0F, 0x0F, 0x1F, 0x1F, 0x0E, 0x0E, 0x0F, 0x0F, 
            0x80, 0xFF, 0x00, 0x2F, 0x8F, 0x3F, 0x00, 0x80, 0xBF, 0x0F, 0x7F, 0x0F, 0x7F, 0x1F, 0x21, 0x7F, 
            0xFF, 0x6F, 0x0F, 0xA1, 0x8E, 0x81, 0x1E, 0x31, 0x31, 0x10, 0x00, 0x10, 0x90, 0x10, 0x10, 0x10, 
            0x7F, 0x6F, 0x2F, 0x7F, 0x3F, 0x0F, 0x7F, 0x2F, 0x0F, 0x7F, 0x7F, 0x6F, 0x0F, 0x01, 0x0E, 0x0E, 
            0x3F, 0xFF, 0x01, 0x2F, 0x0E, 0x0E, 0x7F, 0x01, 0xFF, 0xFF, 0x0F, 0x0E, 0x01, 0x7F, 0x7F, 0x7F, 
            0x0F, 0x21, 0x0F, 0x0F, 0x0F, 0x7F, 0x0F, 0x01, 0x8E, 0x81, 0x7F, 0xEF, 0x8F, 0x8E, 0x01, 0x7F, 
            0x31, 0x21, 0x01, 0x1E, 0x10, 0x00, 0x10, 0x7F, 0x2F, 0x0E, 0x01, 0x1F, 0x2F, 0x00, 0x0E, 0x0E, 
            0x61, 0x01, 0x0E, 0x00, 0x6F, 0x01, 0x0F, 0x6F, 0x00, 0x01, 0x00, 0x06, 0x61, 0x21, 0x6F, 0x7F, 
            0x86, 0x8F, 0x61, 0x00, 0xEF, 0x0F, 0x0F, 0x6F, 0xFF, 0x0F, 0x7F, 0x6F, 0x7F, 0x2F, 0x3F, 0x06, 
            0x80, 0x90, 0x0E, 0x6F, 0x0E, 0xFF, 0x0E, 0x00, 0x6F, 0x61, 0x2F, 0x0E, 0x00, 0x7F, 0x0F, 0x6F, 
            0x01, 0x06, 0x7F, 0x6F, 0x40, 0x2F, 0x0E, 0x7F, 0x7F, 0x00, 0x06, 0x6F, 0x0F, 0x00, 0x6F, 0x21, 
            0x7F, 0x2F, 0x0F, 0x2F, 0x2F, 0x3F, 0x6F, 0x7F, 0x01, 0x0E, 0x0E, 0x01, 0x0F, 0x40, 0x6F, 0x01, 
            0x0F, 0x41, 0x6F, 0x0F, 0x01, 0x00, 0x00, 0x21, 0x00, 0x0E, 0x6F, 0x1F, 0x0E, 0x0F, 0x7F, 0x00, 
            0x6F, 0x6F, 0x0F, 0x0F, 0xFF, 0x00, 0x2F, 0x01, 0x61, 0x61, 0x6F, 0x0E, 0x01, 0x1F, 0x00, 0x0F, 
            0x01, 0x0E, 0x0E, 0x00, 0x6F, 0x01, 0x00, 0x7F, 0x1F, 0x7F, 0x0F, 0x6F, 0x7F, 0x0F, 0x6F, 0x0E, 
            0x8E, 0x0E, 0x6F, 0x0E, 0x0E, 0x0E, 0x6F, 0x10, 0x0F, 0x01, 0x6F, 0x00, 0x0F, 0x00, 0x7F, 0x0E, 
            0x01, 0x0F, 0x1F, 0x6F, 0x00, 0x8E, 0x2F, 0x6F, 0x7F, 0x6F, 0x7F, 0x7F, 0xAF, 0xAF, 0x01, 0x01, 
            0xA1, 0x10, 0x10, 0x00, 0x90, 0x01, 0x0E, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x81, 0x01, 0x0F, 0x8F, 
            0x0E, 0x81, 0x01, 0x2F, 0x0E, 0x0F, 0x7F, 0x0F, 0x0E, 0x0F, 0x0E, 0x0E, 0x01, 0x0F, 0x0E, 0x01, 
            0x0E, 0x0E, 0x6F, 0x00, 0x21, 0x1E, 0x6F, 0x6F, 0x6F, 0x0E, 0x6F, 0x0E, 0x6F, 0x2F, 0x2F, 0x7F, 
            0x6F, 0x0F, 0x00, 0x0F, 0x0F, 0x00, 0x10, 0x4F, 0x7F, 0x27, 0x7F, 0x10, 0x7F, 0x7F, 0x7F, 0x10, 
            0x7F, 0x7F, 0x7F, 0x21, 0x2F, 0x1F, 0x7F, 0x10, 0x10, 0x06, 0x7F, 0x3F, 0x01, 0x21, 0x21, 0x7F, 
            0x0F, 0x7F, 0x1F, 0x0F, 0x1E, 0x1E, 0x21, 0x0F, 0x00, 0x0F, 0x7F, 0x0F, 0x10, 0x0E, 0x7F, 0x21, 
            0x6F, 0x40, 0x2F, 0x3F, 0x7F, 0x00, 0x4E, 0x10, 0x21, 0x01, 0x0E, 0x21, 0x80, 0x1F, 0x5E, 0x7F, 
            0x7F, 0x7F, 0x0F, 0x1F, 0x01, 0x00, 0x0F, 0x7F, 0x7F, 0x7F, 0x0E, 0x7F, 0x11, 0x7F, 0x3F, 0x00, 
            0x0F, 0x9F, 0x1F, 0x7F, 0x1F, 0x21, 0x01, 0x00, 0xFF, 0x06, 0x0F, 0x3F, 0x8F, 0x00, 0x2F, 0x00, 
            0xAF, 0x0E, 0x0E, 0x3F, 0x0F, 0x06, 0x1E, 0x10, 0xA1, 0x90, 0x10, 0x10, 0x10, 0x1F, 0x7F, 0x1F, 
            0x16, 0x6F, 0x0E, 0x1F, 0x0E, 0x00, 0x7F, 0x06, 0x00, 0x7F, 0x0F, 0x0E, 0x7F, 0x0F, 0x3F, 0x7F, 
            0x7F, 0x0E, 0x21, 0x2F, 0x7F, 0x00, 0x0E, 0x0E, 0x0E, 0x7F, 0x7F, 0x00, 0x0E, 0x01, 0x0F, 0x0F, 
            0x81, 0xFF, 0x7F, 0x3F, 0xAF, 0x7F, 0x3F, 0x7F, 0xFF, 0x00, 0x10, 0x71, 0x21, 0x00, 0x21, 0x10, 
            0x90, 0x90, 0x10, 0x80, 0x10, 0x90, 0x21, 0x00, 0x1F, 0x0E, 0x0E, 0x7F, 0xBF, 0x8E, 0x0E, 0x86, 
            0x8E, 0x7F, 0x00, 0x87, 0x8E, 0xFF, 0x8F, 0x01, 0x0E, 0x7F, 0x01, 0x37, 0xFF, 0xFF, 0x0F, 0x9F, 
            0x80, 0x80, 0x7F, 0x00, 0x1F, 0xB1, 0x3F, 0x7F, 0xFF, 0xFF, 0x3F, 0x0E, 0xFF, 0x7F, 0x1F, 0x6F, 
            0xFF, 0x2F, 0x7F, 0x7F, 0x7F, 0x01, 0x5F, 0x7F, 0xFF, 0xFF, 0x1F, 0xFF, 0xFF, 0xA1, 0x00, 0x10, 
            0x21, 0x21, 0x10, 0x10, 0x80, 0x90, 0x10, 0x10, 0x90, 0x10, 0x01, 0x0E, 0xFF, 0x7F, 0x7F, 0x7F, 
            0x06, 0x7F, 0x0F, 0x7F, 0x9F, 0xAF, 0x3F, 0xFF, 0x0E, 0x8E, 0x06, 0x06, 0x1F, 0x1F, 0x7F, 0x7F, 
            0x8E, 0x7F, 0x3F, 0x0F, 0x81, 0x00, 0x00, 0x7F, 0xBF, 0x7F, 0x01, 0x0F, 0x2F, 0x0F, 0x1E, 0x0E, 
            0x8F, 0x7F, 0x00, 0x0F, 0x0E, 0x8E, 0x1F, 0x00, 0x01, 0x21, 0x10, 0x10, 0x80, 0x90, 0x90, 0xB1, 
            0x90, 0x10, 0x10, 0x90, 0x90, 0x10, 0x80, 0xA1, 0x3F, 0x0F, 0x1F, 0x7F, 0x8E, 0x8F, 0x00, 0xFF, 
            0x80, 0x80, 0x1E, 0x06, 0x0E, 0x01, 0x8F, 0x00, 0x8E, 0x9F, 0x7F, 0x4E, 0x8F, 0x7F, 0x6F, 0x21, 
            0x8E, 0x0F, 0x1F, 0x1F, 0x0F, 0x1E, 0x1F, 0x0E, 0x86, 0xFF, 0x07, 0x9F, 0xFF, 0xBF, 0x1F, 0x7F, 
            0x7F, 0x06, 0x01, 0x01, 0x8F, 0xFF, 0x0E, 0x2F, 0x8F, 0x06, 0x0E, 0x01, 0xA1, 0x00, 0x10, 0x10, 
            0x00, 0x10, 0x10, 0x10, 0x80, 0x8F, 0x01, 0xFF, 0xEF, 0xBF, 0x00, 0x2F, 0x01, 0x1F, 0x7F, 0x0E, 
            0x86, 0x1F, 0x6F, 0x00, 0x81, 0x0F, 0x3F, 0x7F, 0x87, 0x00, 0x7F, 0x7F, 0x0F, 0x0F, 0x00, 0x41, 
            0x3F, 0x0E, 0x1F, 0x6F, 0x0F, 0x8E, 0x00, 0x1F, 0x7F, 0x7F, 0x1F, 0x00, 0xFF, 0x86, 0xBF, 0x3F, 
            0x9F, 0x06, 0x7F, 0x00, 0x8F, 0x01, 0xFF, 0x0E, 0x7F, 0x0F, 0x01, 0x6F, 0x8E, 0x8E, 0x0F, 0xAF, 
            0x7F, 0xEF, 0x7F, 0x3F, 0x0F, 0xFF, 0x01, 0x7F, 0xFF, 0xFF, 0x7F, 0x2F, 0xBF, 0x1F, 0x1F, 0x0F, 
            0x9E, 0x0E, 0x21, 0x10, 0x31, 0x21, 0x00, 0x10, 0x10, 0x00, 0x10, 0x21, 0x10, 0x8F, 0x3F, 0x00, 
            0x6F, 0x01, 0x06, 0x2F, 0xFF, 0x0F, 0x7F, 0x0E, 0x87, 0x3F, 0x0E, 0x06, 0x86, 0x1E, 0x1F, 0x7F, 
            0x7F, 0x01, 0x3F, 0x07, 0x81, 0x0F, 0x7F, 0x8E, 0x8E, 0x0E, 0x6F, 0x6F, 0x0F, 0x0E, 0x10, 0x0E, 
            0x9F, 0x7F, 0x00, 0x3F, 0x8E, 0x0F, 0x00, 0x06, 0x8E, 0x3F, 0x1F, 0x0E, 0x01, 0x5E, 0x6F, 0x0E, 
            0x8E, 0x8E, 0x1F, 0x8F, 0x7F, 0x2F, 0x7F, 0x40, 0x11, 0x0F, 0x21, 0x00, 0x80, 0x90, 0x00, 0x10, 
            0x90, 0x10, 0x10, 0x00, 0x80, 0x10, 0x2F, 0x00, 0x00, 0x0F, 0x0F, 0x1E, 0x9E, 0xEF, 0x0E, 0xFF, 
            0x1F, 0x80, 0x0F, 0x3F, 0x5F, 0x0E, 0x6F, 0x6F, 0x8F, 0x80, 0x0E, 0x1E, 0xFF, 0x00, 0x1F, 0x8E, 
            0x9F, 0x1E, 0x4F, 0x0E, 0x0F, 0x0F, 0x3F, 0x0E, 0xAF, 0x81, 0x1F, 0x8E, 0x7F, 0xFF, 0x06, 0x8F, 
            0x0F, 0x80, 0x80, 0x0E, 0xBF, 0x0F, 0x2F, 0x0E, 0x00, 0x0E, 0x71, 0x00, 0x00, 0x1F, 0x6F, 0x1F, 
            0x0E, 0x1E, 0x21, 0x0E, 0x10, 0x10, 0x10, 0x10, 0x90, 0x00, 0x10, 0x0F, 0x00, 0x0E, 0x2F, 0x01, 
            0x8E, 0x8F, 0x1F, 0x0F, 0x9F, 0x00, 0x00, 0x00, 0xFF, 0x06, 0x01, 0x0F, 0x0F, 0x0F, 0x1E, 0x0F, 
            0xEF, 0x0F, 0x00, 0x01, 0x00, 0x86, 0x0E, 0x3F, 0x0E, 0x7F, 0x00, 0x0E, 0x8E, 0x9F, 0x0E, 0x6F, 
            0x8E, 0x00, 0x0E, 0x0F, 0x8F, 0x0E, 0x2F, 0x0E, 0x7F, 0x5E, 0x7F, 0x0F, 0x80, 0x8F, 0x00, 0x8E, 
            0x00, 0x80, 0x1E, 0x21, 0x10, 0x1F, 0x00, 0x1F, 0xEF, 0x86, 0x0E, 0x0E, 0x0E, 0x00, 0x00, 0x0F, 
            0x8F, 0x51, 0x7F, 0x6F, 0x00, 0x3F, 0x1E, 0x0E, 0x8F, 0x1F, 0x6F, 0x81, 0x0E, 0xAF, 0x7F, 0x0F, 
            0x0E, 0x01, 0x0E, 0x0F, 0xFF, 0xEF, 0x0F, 0x00, 0x81, 0x3F, 0x00, 0x6F, 0x9F, 0xFF, 0x7F, 0x00, 
            0x7F, 0x0E, 0x6F, 0x0F, 0x80, 0x90, 0x00, 0x90, 0x10, 0xA1, 0x10, 0x6F, 0x3F, 0x0F, 0x0F, 0x0F, 
            0x10, 0x6F, 0x3F, 0x0E, 0x4F, 0x7F, 0x00, 0x6F, 0x00, 0x0F, 0x4F, 0x0F, 0x01, 0x7F, 0x7F, 0x0E, 
            0x1F, 0x01, 0x7F, 0x00, 0x6F, 0x01, 0x0E, 0x1F, 0x7F, 0x06, 0x6F, 0x06, 0x80, 0x80, 0x10, 0x00, 
            0xAF, 0x61, 0x7F, 0xAF, 0x9F, 0x41, 0x7F, 0x61, 0x0E, 0x0F, 0x01, 0x8E, 0xEF, 0x8F, 0x00, 0xAF, 
            0x0F, 0xAF, 0x6F, 0x06, 0x6F, 0x01, 0xA1, 0x0F, 0x8E, 0x80, 0x6F, 0x8F, 0x86, 0x2F, 0x6F, 0x0F, 
            0xFF, 0x0F, 0x00, 0x06, 0x4F, 0x21, 0x0E, 0x0E, 0x81, 0x9F, 0x01, 0x00, 0x8E, 0x8E, 0x0E, 0x0F, 
            0x0E, 0x00, 0x10, 0x8F, 0x8F, 0x8E, 0x0F, 0x0E, 0xFF, 0x0F, 0x00, 0x0F, 0xAF, 0x6F, 0x01, 0x01, 
            0x0E, 0x0E, 0x0F, 0x6F, 0xAF, 0x96, 0x0F, 0x00, 0x80, 0x8F, 0x6F, 0x7F, 0x6F, 0x0E, 0x1F, 0x7F, 
            0x00, 0x01, 0x0E, 0x0E, 0x3B, 0x31, 0x7F, 0x6F, 0x7F, 0x7F, 0x01, 0x7F, 0x01, 0x0F, 0x7F, 0x7F, 
            0x00, 0x0E, 0x00, 0x0F, 0x00, 0xFF, 0x0E, 0x01, 0x7F, 0x1F, 0x0E, 0x00, 0x10, 0x61, 0x61, 0x7F, 
            0x01, 0x7F, 0x01, 0x0E, 0x0F, 0x3F, 0x7F, 0x6F, 0x2F, 0x3F, 0x00, 0x10, 0x0E, 0x3F, 0x7F, 0x01, 
            0x01, 0x00, 0x7F, 0x7F, 0x0E, 0x06, 0x7F, 0x0F, 0x86, 0x80, 0x0E, 0x11, 0xFF, 0x40, 0x00, 0x0E, 
            0xB1, 0x00, 0xFF, 0x0E, 0x7F, 0x6F, 0x06, 0x7F, 0xEF, 0x8E, 0x0F, 0x0E, 0x86, 0x81, 0x40, 0x0E, 
            0x0E, 0x0F, 0x6F, 0x6F, 0x8E, 0x41, 0x00, 0x7F, 0xA1, 0x21, 0x01, 0x10, 0xFF, 0x00, 0x21, 0x00, 
            0x7F, 0x7F, 0x0F, 0x10, 0x0E, 0x0F, 0x0E, 0x7F, 0x01, 0x7F, 0x00, 0x7F, 0x7F, 0x0F, 0x0E, 0x7F, 
            0x0F, 0x7F, 0x0E, 0x01, 0x7F, 0x7F, 0x01, 0x7F, 0x0E, 0x10, 0x0E, 0x7F, 0x6F, 0x31, 0x0F, 0x7F, 
            0x7F, 0x00, 0x0F, 0x0F, 0x01, 0x00, 0x0E, 0x6F, 0x0E, 0x7F, 0x00, 0x0E, 0x7F, 0x7F, 0x00, 0x0F, 
            0x80, 0x7F, 0x0F, 0x3F, 0x3F, 0x7F, 0x3F, 0x00, 0x01, 0x01, 0x00, 0x7F, 0x7F, 0x0E, 0x1F, 0x7F, 
            0x0F, 0x00, 0x7F, 0x0E, 0x01, 0x00, 0x1F, 0x7F, 0x81, 0xA1, 0x0E, 0x2F, 0x80, 0x0E, 0x0E, 0x8F, 
            0xFB, 0x2F, 0xBF, 0x40, 0x00, 0x7F, 0x7F, 0x31, 0xFF, 0xFF, 0x00, 0x00, 0x7F, 0xFF, 0x1E, 0x1E, 
            0x1F, 0x7F, 0x0F, 0x0E, 0xCF, 0x87, 0x10, 0x10, 0x80, 0x01, 0x7F, 0x6F, 0x4E, 0x46, 0x01, 0x41, 
            0x3F, 0x00, 0x3F, 0x3F, 0x0F, 0x06, 0x7F, 0x6F, 0x8F, 0x6F, 0x7F, 0x0E, 0x7F, 0x0F, 0x7F, 0x7F, 
            0xCE, 0x4F, 0x0F, 0x7F, 0x7F, 0x7F, 0x01, 0x00, 0x00, 0x10, 0x00, 0x00, 0x2F, 0x1F, 0x61, 0x7F, 
            0x7F, 0x0F, 0x0F, 0x01, 0xAF, 0x7F, 0x0F, 0x7F, 0x7F, 0x01, 0x01, 0x0F, 0x81, 0x7F, 0x27, 0x7F, 
            0xC1, 0x7F, 0x0E, 0x0F, 0xFF, 0x3F, 0x7F, 0x01, 0x00, 0x0F, 0x4E, 0x60, 0xB1, 0x80, 0x10, 0x21, 
            0x01, 0x7F, 0x6F, 0x7F, 0x61, 0x0F, 0x01, 0x0E, 0x80, 0x6F, 0x0F, 0x71, 0x1F, 0x00, 0x01, 0x7F, 
            0x00, 0x07, 0x3F, 0x10, 0x10, 0x10, 0x10, 0x31, 0x01, 0x6F, 0x1E, 0x4F, 0x0F, 0x6F, 0x6F, 0x7F, 
            0x01, 0x1F, 0x2F, 0x00, 0xFF, 0xEF, 0x7F, 0x61, 0xFF, 0x61, 0x0F, 0x41, 0x8F, 0x00, 0x7F, 0x7F, 
            0x3F, 0x0E, 0x06, 0x61, 0xFF, 0x81, 0x7F, 0x5F, 0x4F, 0x87, 0x7F, 0x0F, 0x0F, 0x00, 0x1F, 0x01, 
            0x8F, 0x21, 0x10, 0x21, 0x7F, 0x00, 0x0E, 0x7F, 0x6F, 0x6F, 0x0E, 0x0F, 0x1F, 0x0F, 0x61, 0x00, 
            0x4F, 0x7F, 0x01, 0x00, 0x8E, 0x0E, 0x7F, 0x7F, 0x6F, 0x27, 0x01, 0x01, 0x81, 0x7F, 0x00, 0x0F, 
            0x6F, 0x0E, 0x6F, 0x00, 0x01, 0x00, 0x21, 0x10, 0x1E, 0x0E, 0x06, 0x6F, 0x00, 0x01, 0x7F, 0x0E, 
            0x0E, 0x0F, 0x6F, 0x40, 0x7F, 0x0F, 0x00, 0x06, 0xAF, 0x7F, 0x0E, 0x4F, 0xA1, 0x0E, 0x7F, 0x21, 
            0x0F, 0x21, 0x01, 0x01, 0x6F, 0x00, 0x4E, 0x6F, 0x0F, 0x6F, 0x0E, 0x06, 0x0F, 0x00, 0x01, 0x0F, 
            0x00, 0x00, 0x0E, 0x00, 0x01, 0x00, 0x6F, 0x00, 0x0E, 0x7F, 0x2F, 0x1F, 0x7F, 0x7F, 0x06, 0x01, 
            0x6F, 0x00, 0x00, 0x0E, 0x0E, 0x00, 0x7F, 0x00, 0x0F, 0xBF, 0x00, 0x0F, 0x0F, 0x0E, 0x0F, 0x00, 
            0xFF, 0x80, 0x7F, 0x7F, 0xFF, 0x21, 0x0E, 0x7F, 0xEF, 0x7F, 0xE1, 0x21, 0x7F, 0x21, 0x7F, 0xFF, 
            0xFF, 0x8E, 0x00, 0x6F, 0x0E, 0x0F, 0x00, 0x01, 0x7F, 0x7F, 0x16, 0x7F, 0x80, 0x7F, 0x21, 0x2F, 
            0x9E, 0x00, 0x0E, 0x0F, 0xFF, 0x7F, 0x21, 0x7F, 0x0E, 0x01, 0x00, 0x7F, 0x81, 0x7F, 0x61, 0x7F, 
            0x0F, 0x0E, 0x0F, 0x0E, 0x00, 0x00, 0x7F, 0x6F, 0xFF, 0x80, 0x7F, 0x7F, 0xFF, 0x7F, 0x6F, 0x10, 
            0x80, 0x7F, 0x00, 0x0F, 0x7F, 0x1E, 0x61, 0x21, 0xAF, 0x8E, 0x7F, 0x8E, 0x0E, 0xFF, 0x0F, 0x2F, 
            0x10, 0x21, 0x10, 0x10, 0x80, 0x8F, 0x7F, 0x0F, 0x9F, 0xFF, 0x80, 0x0E, 0x9F, 0x8E, 0xFF, 0x7F, 
            0x7F, 0x7F, 0x00, 0xFF, 0x81, 0x8F, 0x7F, 0x8E, 0x80, 0xAF, 0x0E, 0x00, 0x7F, 0x8F, 0xFF, 0x7F, 
            0xBF, 0xB1, 0x21, 0x21, 0xA1, 0x31, 0x21, 0x00, 0x90, 0x10, 0x3F, 0x00, 0x0F, 0x7F, 0x01, 0x7F, 
            0xFF, 0xEF, 0x3F, 0xFF, 0x8F, 0xFF, 0x0F, 0x7F, 0x0E, 0x0E, 0x0E, 0x4F, 0xBF, 0x86, 0x7F, 0x7F, 
            0x80, 0x1F, 0x00, 0x06, 0x9F, 0x2F, 0x0E, 0xFF, 0x00, 0x7F, 0x00, 0x7F, 0xAF, 0x8E, 0x0F, 0xC1, 
            0x7F, 0x87, 0x0E, 0x4F, 0x0E, 0xFF, 0x01, 0xFF, 0x9F, 0x0F, 0x7F, 0x01, 0xFF, 0x7F, 0x10, 0x06, 
            0xA1, 0x21, 0x31, 0x10, 0x00, 0x10, 0x21, 0x10, 0x10, 0xA1, 0x10, 0x10, 0x00, 0x10, 0x0E, 0x7F, 
            0x5F, 0x01, 0x0F, 0x7F, 0xA7, 0x9F, 0x7F, 0x7F, 0xFF, 0xAF, 0x0E, 0xAF, 0x81, 0x81, 0xC0, 0x00, 
            0x2F, 0x3F, 0x0F, 0x1F, 0xFF, 0x8F, 0xAF, 0x80, 0x8F, 0x8F, 0xAF, 0x0E, 0x86, 0x0E, 0xAF, 0x7F, 
            0xFF, 0xFF, 0x5F, 0x7F, 0xFF, 0x00, 0x00, 0x01, 0xBF, 0x1F, 0x7F, 0x0E, 0x1F, 0x7F, 0x3F, 0x0E, 
            0x90, 0x81, 0x3F, 0x8E, 0x2F, 0x9E, 0x0F, 0x2F, 0x00, 0x7F, 0x06, 0x0E, 0xFF, 0x8E, 0xAF, 0x7F, 
            0xDF, 0x7F, 0x0F, 0x7F, 0xFF, 0x6F, 0xCF, 0x8F, 0x0E, 0x01, 0x0E, 0x61, 0x86, 0x90, 0xE1, 0xBF, 
            0x90, 0xA1, 0x21, 0x21, 0x21, 0x90, 0x80, 0x90, 0x90, 0x10, 0x10, 0x10, 0x90, 0x00, 0x10, 0x10, 
            0x80, 0x10, 0x4E, 0x7F, 0x0F, 0x00, 0x7F, 0x7F, 0x8E, 0x81, 0x0E, 0x80, 0x8E, 0x1F, 0x21, 0x0F, 
            0x06, 0x7F, 0x61, 0x00, 0x80, 0x86, 0x06, 0x01, 0x87, 0x3F, 0x7F, 0x2F, 0x81, 0x8F, 0x8E, 0x4F, 
            0x81, 0x0F, 0x3F, 0x9F, 0xAF, 0x8E, 0x00, 0x90, 0x7F, 0xFF, 0x06, 0x8F, 0x7F, 0x7F, 0x10, 0x01, 
            0xFF, 0xFF, 0x7F, 0x7F, 0xBF, 0x1F, 0x3F, 0x00, 0xFF, 0x1E, 0x11, 0x0E, 0x1F, 0x27, 0x7F, 0x0E, 
            0xFF, 0xFF, 0x21, 0xFF, 0xFF, 0x91, 0x00, 0x01, 0x01, 0x21, 0x00, 0x01, 0xA1, 0xA1, 0x00, 0x21, 
            0x90, 0x10, 0x10, 0x31, 0x10, 0x10, 0x10, 0x31, 0x10, 0x10, 0x00, 0x1F, 0x80, 0x4E, 0x0F, 0x8F, 
            0x01, 0x8F, 0x07, 0x00, 0x3F, 0x0F, 0x7F, 0x3F, 0x8E, 0x8F, 0x01, 0x0F, 0x8F, 0x06, 0x2F, 0x6F, 
            0x8E, 0x7F, 0x0F, 0x3F, 0x01, 0x7F, 0x1F, 0x0E, 0x81, 0x8E, 0x0E, 0x86, 0x0E, 0xA1, 0x00, 0x7F, 
            0x0E, 0x0E, 0x0E, 0x7F, 0xEF, 0x80, 0x0F, 0x7F, 0x81, 0x00, 0x01, 0x2F, 0x8F, 0x6F, 0x00, 0x6F, 
            0x2F, 0x4F, 0x7F, 0x0F, 0x8E, 0x81, 0x31, 0xFF, 0x7F, 0x8F, 0x0F, 0x0E, 0x0E, 0x7F, 0x0F, 0x7F, 
            0xFF, 0xEF, 0x0F, 0x7F, 0x8E, 0x7F, 0x61, 0x00, 0x80, 0xA1, 0xA1, 0x01, 0x21, 0x00, 0x00, 0x00, 
            0x90, 0x80, 0x10, 0x80, 0x6F, 0xC1, 0x2F, 0x0E, 0x27, 0x7F, 0x21, 0x7F, 0x8F, 0xFF, 0x0E, 0x21, 
            0x8F, 0x0F, 0x7F, 0x0F, 0xAF, 0x7F, 0x0F, 0x6F, 0x7F, 0x01, 0x7F, 0x00, 0x8F, 0x87, 0x06, 0x6F, 
            0xFF, 0xAF, 0x00, 0x3F, 0x0E, 0x01, 0x06, 0x6F, 0xCF, 0x8E, 0x0F, 0x0F, 0x8F, 0x01, 0x7F, 0x2F, 
            0x9F, 0x1F, 0x6F, 0x0E, 0x0E, 0x1F, 0x0F, 0x0E, 0x8E, 0xFF, 0x7F, 0x0F, 0x1F, 0x8F, 0x00, 0x00, 
            0x61, 0x31, 0x00, 0x00, 0xAF, 0xFF, 0x0F, 0x0E, 0xAF, 0x01, 0x0F, 0x3F, 0xAF, 0x7F, 0x7F, 0x2F, 
            0x1F, 0x0F, 0x7F, 0x0F, 0x8E, 0x8E, 0x01, 0x8F, 0x00, 0xA1, 0x21, 0x21, 0x21, 0x00, 0x00, 0x10, 
            0x90, 0xA1, 0x21, 0x21, 0x90, 0x27, 0x21, 0x40, 0xA1, 0x00, 0x2F, 0x00, 0x00, 0x10, 0x00, 0x0E, 
            0xFF, 0x81, 0x0E, 0x8F, 0x3F, 0x8F, 0x21, 0x8E, 0x0F, 0x3F, 0x00, 0x07, 0x8E, 0xFF, 0x00, 0x7F, 
            0x8E, 0x06, 0x1F, 0x00, 0x8E, 0x0F, 0x00, 0x00, 0x0E, 0x0F, 0xEF, 0x0E, 0x8E, 0x80, 0x0F, 0x86, 
            0x00, 0x8E, 0x8E, 0xEF, 0xFF, 0xE1, 0x8E, 0xDF, 0x8E, 0x6F, 0x7F, 0x0F, 0xA7, 0x3F, 0x7F, 0x0E, 
            0xFF, 0x61, 0x6F, 0x5F, 0x01, 0x0F, 0x1F, 0x00, 0xEF, 0x0F, 0x2F, 0xFF, 0x0F, 0x6F, 0x7F, 0x6F, 
            0x0F, 0x1F, 0x01, 0x21, 0xAF, 0xEF, 0x0E, 0x7F, 0xFF, 0x7F, 0x0F, 0x0F, 0xB1, 0xA1, 0xA1, 0x01, 
            0x11, 0x21, 0x3F, 0x86, 0x90, 0x80, 0x1E, 0x90, 0x90, 0x90, 0xA1, 0x00, 0x00, 0x1F, 0xAF, 0x00, 
            0x8E, 0xAF, 0x01, 0x06, 0xBF, 0x7F, 0x8F, 0x8E, 0xB1, 0x0E, 0x86, 0x7F, 0x7F, 0x00, 0x0E, 0x00, 
            0xAF, 0x87, 0x00, 0x8E, 0x8E, 0xCF, 0x8F, 0x7F, 0x1F, 0x0E, 0x06, 0x1F, 0x8E, 0x9F, 0x6F, 0x0F, 
            0x8E, 0x3F, 0x21, 0x0E, 0xFF, 0x0E, 0x0E, 0x1F, 0x01, 0x0E, 0x01, 0x7F, 0x8F, 0x8E, 0x2F, 0xAF, 
            0x80, 0xFF, 0x0E, 0x6F, 0x0E, 0x01, 0x06, 0x21, 0x81, 0x8E, 0x3F, 0x6F, 0xFF, 0x6F, 0xBF, 0x0E, 
            0xFF, 0x01, 0x00, 0x2F, 0x1E, 0x00, 0x01, 0x01, 0xA1, 0x2F, 0x00, 0x10, 0x00, 0xA1, 0x21, 0x10, 
            0x10, 0x00, 0x2F, 0x00, 0x8E, 0x0F, 0x0E, 0x2F, 0xEF, 0x00, 0x40, 0x27, 0x8E, 0x3F, 0x0F, 0x01, 
            0x00, 0x0F, 0x3F, 0x0F, 0x0E, 0x01, 0x0E, 0x0E, 0x0E, 0x21, 0x00, 0x6F, 0x0E, 0x7F, 0x0F, 0x7F, 
            0x8F, 0x8F, 0x6F, 0x01, 0x8E, 0x2F, 0x0E, 0x0E, 0x86, 0x06, 0xFF, 0x21, 0x21, 0x0E, 0x00, 0x0F, 
            0x80, 0xCF, 0x21, 0xEF, 0x2F, 0x8F, 0x8E, 0x7F, 0x1E, 0x6F, 0x01, 0x2F, 0x80, 0x8F, 0x6F, 0x7F, 
            0x8F, 0x7F, 0x21, 0x6F, 0x81, 0x06, 0x06, 0x0E, 0x1F, 0x21, 0x31, 0x21, 0x80, 0x80, 0x21, 0x90, 
            0x21, 0x90, 0x0F, 0x00, 0x0F, 0x7F, 0x21, 0x01, 0xEF, 0x6F, 0x6F, 0x0F, 0x86, 0x7F, 0x1F, 0x0F, 
            0x8E, 0x0E, 0x00, 0x00, 0x7F, 0x01, 0x27, 0x3F, 0xEF, 0x8F, 0x01, 0x6F, 0x80, 0x86, 0x07, 0x06, 
            0x1F, 0x0F, 0x41, 0x40, 0x86, 0x07, 0x0F, 0x00, 0x7F, 0x7F, 0x0E, 0x1F, 0x00, 0x0E, 0x0E, 0x6F, 
            0x06, 0x7F, 0x2F, 0x00, 0x0F, 0x10, 0x2F, 0x0F, 0x0E, 0x0E, 0x0F, 0x6F, 0x00, 0x0E, 0x00, 0x0E, 
            0x80, 0x91, 0x21, 0x21, 0x80, 0x00, 0x0E, 0x00, 0xA1, 0x10, 0x80, 0x00, 0x10, 0x00, 0x2F, 0x2F, 
            0xFF, 0x8F, 0x00, 0x8F, 0x7F, 0x86, 0x81, 0x8F, 0x00, 0x4F, 0x00, 0x81, 0x86, 0xEF, 0x7F, 0x00, 
            0x3F, 0x1F, 0x06, 0x0E, 0x8E, 0x0E, 0x0E, 0x7F, 0x00, 0x00, 0x0E, 0x0F, 0x6F, 0x81, 0x0F, 0x8F, 
            0x3F, 0x8E, 0x6F, 0x6F, 0x06, 0x0F, 0x06, 0x00, 0x8E, 0x10, 0x21, 0x11, 0x3F, 0x0E, 0x2F, 0x0F, 
            0x01, 0x01, 0x00, 0x6F, 0x0F, 0x00, 0x06, 0x06, 0x2F, 0x0E, 0x06, 0x6F, 0x80, 0x8F, 0x01, 0x01, 
            0x00, 0x21, 0x6F, 0x6F, 0x80, 0x0E, 0x0F, 0x06, 0x81, 0x01, 0x00, 0x0E, 0x8F, 0x06, 0x00, 0x06, 
            0x0F, 0x2F, 0x00, 0x6F, 0x81, 0x81, 0x01, 0x80, 0x00, 0x8E, 0x6F, 0x6F, 0x0F, 0x0F, 0x2F, 0x2F, 
            0x8E, 0x8E, 0x21, 0x00, 0xA1, 0x0E, 0x00, 0x0F, 0x2F, 0x00, 0x2F, 0x0E, 0x0F, 0x00, 0x06, 0x06, 
            0x0F, 0x0F, 0x01, 0x0E, 0x00, 0x80, 0x40, 0x00, 0x0E, 0x0E, 0x2F, 0x6F, 0x8F, 0x81, 0x00, 0x00, 
            0x86, 0x00, 0x0F, 0x0F, 0xEF, 0xA1, 0x81, 0x01, 0x0E, 0x0F, 0x6F, 0x0E, 0xC0, 0x80, 0x00, 0x87, 
            0x0F, 0x8F, 0x6F, 0x06, 0x00, 0x00, 0x6F, 0x0F, 0x8E, 0x0E, 0x0E, 0x01, 0x00, 0x21, 0x07, 0x21, 
            0x7F, 0x7F, 0x10, 0x7F, 0x11, 0x0E, 0x00, 0x31, 0x0F, 0x00, 0x00, 0x01, 0x8F, 0x0E, 0x00, 0x0F, 
            0x00, 0x0E, 0x7F, 0x0E, 0x8E, 0x01, 0x0F, 0x3F, 0x2F, 0x3F, 0x7F, 0x0F, 0x0E, 0x6F, 0x7F, 0x0F, 
            0x00, 0x0E, 0x0E, 0x3F, 0x00, 0x06, 0x5F, 0x7F, 0x0E, 0x3F, 0x0F, 0x0E, 0x7F, 0x0F, 0x6F, 0x00, 
            0xEF, 0x86, 0x01, 0x21, 0xAF, 0x0E, 0x07, 0x00, 0x81, 0x3F, 0x00, 0x2F, 0x0E, 0x01, 0x0E, 0x6F, 
            0x8F, 0xEF, 0x7F, 0xFF, 0x7F, 0x5E, 0x7F, 0x1F, 0x00, 0x21, 0x7F, 0x01, 0x80, 0x0E, 0x01, 0x21, 
            0x81, 0x00, 0x4E, 0x21, 0xA1, 0x01, 0x0E, 0x4E, 0x6F, 0x3F, 0x00, 0x7F, 0x90, 0x01, 0x0F, 0x6F, 
            0x2F, 0x10, 0x1F, 0x7F, 0x7F, 0x01, 0x7F, 0x10, 0x8E, 0xFF, 0x7F, 0x31, 0x86, 0x3F, 0x06, 0x0E, 
            0x80, 0x0E, 0x10, 0x10, 0x0E, 0x2F, 0x7F, 0x0F, 0xEF, 0x8E, 0x10, 0x9F, 0x00, 0x80, 0x6F, 0x0F, 
            0x0F, 0x10, 0x0F, 0x0F, 0x2F, 0x0E, 0x0E, 0x0E, 0x01, 0x01, 0x3F, 0x2F, 0x01, 0x0F, 0x4F, 0x6F, 
            0x0F, 0x21, 0x6F, 0x3F, 0x31, 0x7F, 0x0E, 0x7F, 0x01, 0x01, 0x6F, 0x21, 0x6F, 0x0F, 0x01, 0x7F, 
            0xAF, 0xD0, 0x10, 0x0F, 0x8F, 0x7F, 0x6F, 0x8E, 0x8E, 0x0F, 0x8E, 0x7F, 0x0F, 0x7F, 0x21, 0x5E, 
            0x8E, 0x80, 0x7F, 0xFF, 0x7F, 0x90, 0xFF, 0x5E, 0x6F, 0x90, 0x8F, 0x7F, 0x80, 0x80, 0x0E, 0x21, 
            0x86, 0x11, 0x86, 0x0F, 0x8E, 0x00, 0x0F, 0x8F, 0x0E, 0x00, 0x10, 0x7F, 0xEF, 0x80, 0x01, 0xBF, 
            0x8E, 0x81, 0x0E, 0x3F, 0x00, 0x10, 0x00, 0x07, 0x8E, 0x9F, 0x00, 0x0E, 0x8E, 0x1E, 0x0E, 0x0F, 
            0x8E, 0x0E, 0x0F, 0x06, 0x0F, 0x1F, 0x1E, 0x10, 0xEF, 0x87, 0x00, 0x86, 0x86, 0x9F, 0x01, 0x7F, 
            0x1F, 0x7F, 0x01, 0x7F, 0xBF, 0x1E, 0x1E, 0x21, 0x9E, 0x1F, 0x1E, 0x3F, 0x00, 0x0E, 0x00, 0x1F, 
            0x0E, 0x16, 0x10, 0x6F, 0x3F, 0x0E, 0x1F, 0x1F, 0x1F, 0x90, 0x1E, 0x0E, 0x0F, 0x00, 0x1F, 0x1E, 
            0x1E, 0x00, 0x10, 0x0F, 0x7F, 0x11, 0x0F, 0x61, 0x7F, 0x00, 0x01, 0x0F, 0x00, 0x10, 0x6F, 0x2F, 
            0x7F, 0x7F, 0x7F, 0x0E, 0x00, 0x00, 0x1E, 0x10, 0x00, 0x10, 0x1F, 0x0F, 0x8E, 0x8F, 0x6F, 0x0F, 
            0xFF, 0x00, 0x0F, 0x00, 0x9F, 0x7F, 0x00, 0x7F, 0x00, 0x0F, 0x61, 0x1F, 0xBF, 0xFF, 0x7F, 0xFF, 
            0xFF, 0x9E, 0x21, 0x00, 0x10, 0x0E, 0x0E, 0x0F, 0xFF, 0x1E, 0x7F, 0x0F, 0x00, 0x4F, 0x00, 0x2F, 
            0x7F, 0x0E, 0x7F, 0x2F, 0x1F, 0x00, 0x5F, 0x00, 0x0F, 0x11, 0x6F, 0x07, 0x00, 0xFF, 0x7F, 0x00, 
            0x06, 0x7F, 0x7F, 0x7F, 0x8E, 0x5F, 0x7F, 0x4F, 0x7F, 0x1F, 0x06, 0x0E, 0x7F, 0x2F, 0x0E, 0x1E, 
            0x7F, 0x00, 0x6F, 0x7F, 0x5F, 0x4F, 0x6F, 0x01, 0x8F, 0x7F, 0x6F, 0x7F, 0x0F, 0x0E, 0x00, 0x11, 
            0x00, 0x31, 0x21, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x7F, 0x0E, 0x1E, 0x7F, 0x00, 
            0x0F, 0x3F, 0x11, 0x7F, 0x0F, 0x00, 0x0E, 0x0E, 0x7F, 0x7F, 0x2F, 0x7F, 0xFF, 0xFF, 0x7F, 0x7F, 
            0x8E, 0x6F, 0x4F, 0x0E, 0xFF, 0x3F, 0x0F, 0x0E, 0x00, 0x7F, 0x7F, 0x00, 0xFF, 0x8F, 0x00, 0x8F, 
            0x1F, 0x87, 0x0F, 0xFF, 0x1F, 0x7F, 0x1F, 0x7F, 0x80, 0xAF, 0x2F, 0x7F, 0x0F, 0x2F, 0x16, 0x00, 
            0x9F, 0x7F, 0x7F, 0x7F, 0x00, 0x7F, 0x00, 0x0F, 0x7F, 0x0F, 0x31, 0x5F, 0x8F, 0x0E, 0x5F, 0x7F, 
            0x7F, 0x3F, 0x0F, 0x7F, 0x81, 0x1E, 0x10, 0x10, 0x10, 0x06, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 
            0x0E, 0x1E, 0x00, 0x0E, 0x3F, 0x00, 0x00, 0x1F, 0x86, 0x0E, 0x0F, 0x7F, 0x7F, 0x0E, 0x1F, 0x0F, 
            0x8E, 0x4F, 0x3F, 0x01, 0x80, 0x00, 0x0E, 0x7F, 0x86, 0x7F, 0x1F, 0x7F, 0x00, 0x0E, 0x7F, 0x2F, 
            0x86, 0x80, 0x0E, 0x00, 0x00, 0x7F, 0x01, 0x5F, 0x0F, 0x6F, 0x7F, 0x3F, 0x0E, 0x0E, 0x1F, 0x07, 
            0x06, 0x1F, 0x7F, 0x3F, 0x0E, 0x7F, 0x6F, 0x06, 0x6F, 0x5F, 0x0E, 0x7F, 0x1F, 0x7F, 0x7F, 0x0F, 
            0x0E, 0xFF, 0x0E, 0x10, 0x21, 0x31, 0x10, 0x10, 0x90, 0x80, 0x10, 0x10, 0x80, 0x10, 0x10, 0x10, 
            0x80, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00, 0x01, 0x8E, 0xFF, 0x7F, 0x00, 0x71, 0x80, 0x1F, 0x0F, 
            0x1E, 0x0F, 0x0E, 0x7F, 0xAF, 0x0F, 0x7F, 0x0F, 0x06, 0x7F, 0x7F, 0x00, 0x6F, 0x0E, 0x7F, 0x3F, 
            0x0F, 0x00, 0x00, 0x00, 0x7F, 0x0E, 0x0E, 0x7F, 0xFF, 0xEF, 0x0E, 0x0E, 0x1F, 0x0E, 0x06, 0x4E, 
            0x0E, 0x00, 0x1F, 0x0E, 0x0E, 0x7F, 0x0E, 0x6F, 0x7F, 0x5E, 0x0E, 0x06, 0x7F, 0x4E, 0x7F, 0x00, 
            0x0E, 0x5F, 0x0F, 0x7F, 0x1F, 0x7F, 0x21, 0x0F, 0x0F, 0x21, 0x00, 0x31, 0xA1, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x00, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x1F, 0x1E, 0x2F, 0x00, 0x1E, 0x7F, 
            0x00, 0x00, 0x7F, 0x0E, 0x0F, 0x7F, 0x00, 0x07, 0xBF, 0x01, 0x0E, 0x06, 0x2F, 0x0F, 0x0E, 0x1F, 
            0x7F, 0x00, 0x0E, 0x40, 0x5F, 0x7F, 0x7F, 0x7F, 0x0F, 0x0E, 0x0F, 0x7F, 0x3F, 0x0E, 0x01, 0x6F, 
            0x0E, 0x7F, 0x2F, 0x0E, 0x8E, 0x2F, 0x1F, 0x0E, 0x7F, 0x7F, 0x4E, 0x0F, 0x0E, 0x1F, 0x7F, 0x0F, 
            0x1F, 0x7F, 0x0F, 0x0E, 0xBF, 0x0F, 0x3F, 0x80, 0x6F, 0x8F, 0x6F, 0x7F, 0x3F, 0x0E, 0x7F, 0x0F, 
            0x0F, 0x7F, 0x2F, 0x7F, 0x0F, 0x6F, 0x0F, 0x7F, 0x61, 0x7F, 0x6F, 0x7F, 0x3F, 0x0E, 0x00, 0x00, 
            0x0F, 0x00, 0x00, 0x0E, 0x01, 0x3F, 0x00, 0x21, 0x21, 0x21, 0x31, 0x21, 0x90, 0x90, 0x10, 0x00, 
            0x90, 0x10, 0x00, 0x21, 0x90, 0x21, 0x10, 0x50, 0x00, 0x6F, 0x7F, 0x6F, 0x8E, 0xBF, 0x01, 0x6F, 
            0x7F, 0xFF, 0x8F, 0x7F, 0x7F, 0x7F, 0x6F, 0xAF, 0x8E, 0xB1, 0x00, 0x7F, 0xEF, 0x7F, 0xAF, 0x7F, 
            0x8E, 0x0E, 0x5F, 0x0E, 0x5E, 0x06, 0x4F, 0x00, 0xFF, 0xAF, 0x7F, 0x8F, 0xEF, 0xCE, 0xEF, 0x0E, 
            0x0E, 0x0E, 0x00, 0x7F, 0xDF, 0x8F, 0x0E, 0x0E, 0x81, 0x0F, 0x6F, 0x0F, 0x80, 0x7F, 0x3F, 0x01, 
            0x00, 0x0F, 0x0F, 0x1F, 0x9F, 0x8E, 0x7F, 0x81, 0xFF, 0x80, 0x0E, 0x7F, 0x6F, 0x0F, 0x4F, 0x7F, 
            0x8E, 0x0E, 0x0F, 0x0F, 0x8E, 0x0E, 0x8E, 0x6F, 0x8F, 0x0E, 0x00, 0x7F, 0x00, 0x00, 0x7F, 0x6F, 
            0x80, 0x0E, 0x6F, 0x4F, 0x0E, 0x80, 0x21, 0x0E, 0x0E, 0x0E, 0x40, 0x01, 0x00, 0x01, 0x31, 0x31, 
            0x21, 0x00, 0x21, 0x10, 0x00, 0x10, 0x10, 0x00, 0x0E, 0x1E, 0x00, 0x00, 0x21, 0x0F, 0x0E, 0x3F, 
            0x7F, 0x00, 0x0E, 0x0F, 0x0E, 0x00, 0x6F, 0x0E, 0xBF, 0x8F, 0x00, 0x0E, 0xFF, 0x6F, 0x0E, 0x7F, 
            0x8F, 0x06, 0x7F, 0x0E, 0x0E, 0x7F, 0x0E, 0x1F, 0x80, 0x80, 0x7F, 0xCE, 0x80, 0x81, 0x07, 0x7F, 
            0x0E, 0x5F, 0x3F, 0x0F, 0x9E, 0x80, 0x7F, 0x3F, 0x80, 0x0E, 0x7F, 0x11, 0x8E, 0x3F, 0x0E, 0x2F, 
            0x0F, 0x1F, 0x7F, 0x0F, 0xEF, 0xEF, 0x0E, 0x9F, 0x0F, 0xEF, 0x0F, 0x7F, 0x6F, 0x0F, 0x4F, 0x1F, 
            0x8E, 0x7F, 0x0E, 0x7F, 0xFF, 0x3F, 0x0E, 0x10, 0x0E, 0x00, 0x10, 0x00, 0x0E, 0x21, 0x31, 0x10, 
            0x10, 0x10, 0x10, 0x00, 0x90, 0x10, 0x10, 0x00, 0x10, 0x10, 0x00, 0x0F, 0xAF, 0x0F, 0x0F, 0x6F, 
            0x80, 0x0E, 0x2F, 0x00, 0x7F, 0x0E, 0x06, 0x2F, 0x6F, 0x1F, 0x00, 0x0F, 0x0E, 0x0F, 0x6F, 0x6F, 
            0x0E, 0xEF, 0x7F, 0x0E, 0x01, 0x0E, 0x7F, 0x0E, 0x8F, 0xBF, 0x0F, 0x00, 0x81, 0x0F, 0x0E, 0x7F, 
            0x80, 0x61, 0x0E, 0x3F, 0x7F, 0x7F, 0x01, 0x00, 0x8F, 0x8E, 0x0F, 0x81, 0x0E, 0x80, 0x0E, 0x0E, 
            0x7F, 0x06, 0x6F, 0x6F, 0x90, 0x0E, 0x0F, 0x0E, 0x80, 0x1F, 0x1F, 0x7F, 0xEF, 0x1F, 0x0E, 0x1F, 
            0x0F, 0x7F, 0x6F, 0x1F, 0x00, 0x0E, 0x1F, 0x0E, 0x6F, 0x8E, 0x0E, 0x0F, 0x0F, 0x00, 0x3F, 0x6F, 
            0x0E, 0x6F, 0x00, 0x0E, 0x00, 0x21, 0x10, 0x10, 0x00, 0x00, 0x00, 0x10, 0x00, 0x10, 0x0E, 0x0F, 
            0x0E, 0x4F, 0x0F, 0x0F, 0x6F, 0x0E, 0x00, 0x4F, 0x7F, 0x01, 0x4F, 0x6F, 0xB1, 0x0F, 0x5F, 0x6F, 
            0x8E, 0x0F, 0x5F, 0x0E, 0xEF, 0x00, 0x3F, 0x0E, 0x01, 0x0E, 0x0E, 0x06, 0x8F, 0xFF, 0x7F, 0xAF, 
            0x6F, 0x80, 0x1E, 0x0E, 0x31, 0x00, 0x0E, 0x06, 0xBF, 0x00, 0x7F, 0x0E, 0x7F, 0x4F, 0x4F, 0x06, 
            0x2F, 0x61, 0x2F, 0x01, 0x7F, 0x0E, 0x2F, 0x0F, 0xFF, 0x1E, 0x00, 0x0F, 0x1F, 0x5F, 0x7F, 0x00, 
            0x4F, 0x21, 0x0F, 0x0F, 0xCF, 0x8E, 0x0F, 0x6F, 0x80, 0x00, 0x01, 0x00, 0x90, 0x00, 0x0E, 0x01, 
            0x0F, 0x3F, 0x0E, 0x6F, 0xEF, 0x8F, 0x0F, 0xDF, 0x0F, 0x8E, 0x2F, 0x0E, 0x0E, 0x0E, 0x4F, 0x4F, 
            0x8F, 0x6F, 0x0E, 0x7F, 0x0E, 0x21, 0x1F, 0x00, 0x06, 0x7F, 0x0E, 0x00, 0x0F, 0x0E, 0x40, 0x0E, 
            0x7F, 0x6F, 0x7F, 0x6F, 0x8E, 0x01, 0x2F, 0x0F, 0x0F, 0x16, 0x4F, 0x0E, 0x8E, 0x0E, 0x0E, 0x0E, 
            0x80, 0x10, 0x10, 0x00, 0x2F, 0x6F, 0x00, 0x00, 0x2F, 0x00, 0x41, 0x2F, 0x0E, 0x06, 0x1F, 0x6F, 
            0x7F, 0x7F, 0x0E, 0x0E, 0x6F, 0x00, 0x0E, 0x0E, 0x8F, 0x4F, 0x00, 0x6F, 0xEF, 0x0E, 0x3F, 0x7F, 
            0x8F, 0x6F, 0x0E, 0x21, 0x0E, 0x21, 0x21, 0x0E, 0x80, 0x81, 0x2F, 0x8E, 0x0F, 0x81, 0x6F, 0x0E, 
            0x0F, 0x6F, 0x00, 0x00, 0x8E, 0x6F, 0x4F, 0x2F, 0x80, 0x6F, 0x0E, 0x6F, 0x8E, 0x0F, 0x0E, 0x6F, 
            0x00, 0x7F, 0x00, 0x00, 0x8E, 0x6F, 0x0E, 0x0F, 0x6F, 0x8E, 0x7F, 0x7F, 0x46, 0x6F, 0x61, 0x6F, 
            0x8F, 0x8E, 0x0E, 0x5F, 0x8E, 0x00, 0x2F, 0x6F, 0xEF, 0x0E, 0x06, 0x0E, 0x21, 0x00, 0x00, 0x4F, 
            0xAF, 0x8E, 0x2F, 0x8E, 0x0F, 0x9E, 0x00, 0x4F, 0x0E, 0x1F, 0x0E, 0x0E, 0x0F, 0x00, 0x6F, 0x0E, 
            0x0E, 0x06, 0x0E, 0x0F, 0x0F, 0x0E, 0x0E, 0x00, 0x0F, 0x0F, 0x0F, 0x01, 0x7F, 0x00, 0x01, 0x10, 
            0x40, 0x6F, 0x0E, 0x00, 0x01, 0x07, 0x06, 0x0E, 0xEF, 0x80, 0x0E, 0x0E, 0x80, 0x4F, 0x1F, 0x8E, 
            0x8E, 0x8E, 0x06, 0x6F, 0x0F, 0x0F, 0x0E, 0x00, 0x8E, 0x8E, 0x0E, 0xFF, 0x11, 0x90, 0x01, 0x31, 
            0x7F, 0x0E, 0x00, 0x00, 0x8E, 0x91, 0x1F, 0x00, 0xFF, 0x00, 0x0E, 0x00, 0xFF, 0x6F, 0x11, 0x10, 
            0x10, 0x01, 0x06, 0x0E, 0x8E, 0xFF, 0x0F, 0x80, 0x00, 0xB1, 0x7F, 0x00, 0x00, 0x00, 0x7F, 0x00, 
            0x80, 0x8E, 0x7F, 0x0E, 0x9F, 0x1F, 0x1F, 0x01, 0xCF, 0x7F, 0x4E, 0x01, 0x10, 0x10, 0x00, 0x0E, 
            0x80, 0xCE, 0x00, 0x80, 0xCF, 0x80, 0x00, 0x00, 0x00, 0x0E, 0x00, 0x7F, 0xFF, 0xFF, 0x3F, 0x7F, 
            0x0F, 0x1F, 0x00, 0x7F, 0x0F, 0x0E, 0x00, 0x5F, 0x7F, 0x71, 0x2F, 0x31, 0x10, 0x10, 0x0E, 0x00, 
            0x10, 0x90, 0x10, 0x10, 0x00, 0x0E, 0x06, 0x0E, 0x7F, 0x00, 0x1F, 0x6F, 0x00, 0x0E, 0x0E, 0x6F, 
            0x00, 0x0F, 0x0E, 0x0F, 0x07, 0x01, 0x01, 0x0E, 0x5F, 0x7F, 0x0E, 0x10, 0x0F, 0x6F, 0x0E, 0x71, 
            0x0E, 0x0E, 0x0E, 0x00, 0x9F, 0x80, 0x10, 0x10, 0x90, 0x10, 0x00, 0x90, 0x90, 0x10, 0x00, 0x1E, 
            0x0E, 0x61, 0x00, 0x0E, 0x8E, 0x80, 0x0E, 0x9E, 0x8E, 0xFF, 0x0F, 0x06, 0x0E, 0x7F, 0x00, 0x0E, 
            0x8E, 0x80, 0x0E, 0x01, 0xEF, 0x01, 0x0F, 0x01, 0x80, 0x7F, 0x1E, 0x0F, 0x4F, 0x0E, 0x06, 0x21, 
            0x96, 0x80, 0x00, 0x96, 0x21, 0x90, 0x10, 0x0E, 0x10, 0x7F, 0x7F, 0x06, 0xAF, 0x00, 0x4F, 0x0E, 
            0x8F, 0x6F, 0x0E, 0x0E, 0x80, 0x06, 0x7F, 0x00, 0x06, 0x00, 0x00, 0x01, 0x80, 0x81, 0x0E, 0x9F, 
            0xDF, 0x9F, 0x06, 0x00, 0x00, 0x00, 0x7F, 0x00, 0x81, 0x00, 0x06, 0x00, 0x21, 0x00, 0x00, 0x00, 
            0x00, 0x0F, 0x06, 0x0E, 0x0E, 0x11, 0x2F, 0x6F, 0x00, 0x6F, 0x51, 0x0F, 0x7F, 0x06, 0x7F, 0x00, 
            0x4F, 0x00, 0x0F, 0x0E, 0x8E, 0xA1, 0x6F, 0x01, 0x8E, 0x6F, 0x0F, 0x00, 0xFF, 0x0E, 0x5F, 0x0E, 
            0x0F, 0x06, 0x2F, 0x0E, 0xFF, 0xEF, 0x7F, 0xFF, 0x1F, 0xEF, 0x0E, 0x00, 0x6F, 0x00, 0x7F, 0x00, 
            0x86, 0x80, 0x1E, 0x10, 0x80, 0x00, 0x00, 0x00, 0x9E, 0x01, 0x10, 0x0E, 0x00, 0x7F, 0x00, 0x00, 
            0x8F, 0x06, 0x0E, 0x00, 0x7F, 0x8F, 0x0E, 0x0F, 0x21, 0x4E, 0x7F, 0x00, 0x81, 0x00, 0x00, 0x7F, 
            0x0E, 0x00, 0x4F, 0x00, 0x7F, 0x21, 0x00, 0x06, 0x10, 0x7F, 0x0F, 0x0F, 0x0E, 0x0E, 0x0E, 0x7F, 
            0x1F, 0x8E, 0x41, 0x00, 0x0E, 0x0E, 0x00, 0x0E, 0xBF, 0x06, 0x0E, 0x00, 0x7F, 0x00, 0x00, 0x0F, 
            0x8F, 0x6F, 0x0F, 0x1F, 0x00, 0x16, 0x00, 0x00, 0x0E, 0x7F, 0x41, 0x00, 0x0E, 0x0E, 0x6F, 0x0F, 
            0x0F, 0x6F, 0x0E, 0x7F, 0x8F, 0x0E, 0x06, 0x0E, 0x6F, 0x6F, 0x0E, 0x0F, 0x01, 0x00, 0x7F, 0x0E, 
            0x6F, 0x00, 0x6F, 0x01, 0x7F, 0x7F, 0x0F, 0x21, 0x0E, 0x0F, 0x0F, 0x06, 0x0F, 0x00, 0x00, 0x6F, 
            0xBF, 0x8E, 0x0E, 0x00, 0x8E, 0x7F, 0x6F, 0x7F, 0x8E, 0x00, 0x00, 0x00, 0x2F, 0x6F, 0x7F, 0x00, 
            0x86, 0x8E, 0x06, 0x80, 0x0E, 0xA1, 0x00, 0x00, 0x0F, 0x3F, 0x00, 0x6F, 0xEF, 0x0E, 0x6F, 0x6F, 
            0xC1, 0x0E, 0x0E, 0x0E, 0x80, 0x06, 0x7F, 0x0F, 0x01, 0x01, 0x0E, 0x00, 0x0E, 0x2F, 0x00, 0x00, 
            0x6F, 0x80, 0x00, 0x0F, 0x0E, 0x00, 0x00, 0x01, 0x80, 0x0E, 0x0F, 0x6F, 0x00, 0x1F, 0x0E, 0x0F, 
            0x00, 0x00, 0x0E, 0x0E, 0x01, 0x00, 0x06, 0x06, 0x3F, 0x0E, 0x7F, 0x01, 0x7F, 0xEF, 0x00, 0x01, 
            0x7F, 0x10, 0x61, 0x00, 0x81, 0xFF, 0x7F, 0x10, 0x9F, 0x1F, 0x6F, 0x7F, 0xA1, 0x7F, 0x6F, 0x3F, 
            0x61, 0x0F, 0x0F, 0x01, 0x8E, 0x81, 0x6F, 0xFF, 0x7F, 0x8E, 0x06, 0x2F, 0x7F, 0x10, 0x01, 0x0F, 
            0x81, 0x00, 0x7F, 0x0F, 0x80, 0x01, 0x1F, 0x00, 0xEF, 0x7F, 0x0F, 0x7F, 0x01, 0x7F, 0x0F, 0x7F, 
            0x8F, 0x7F, 0x7F, 0x0F, 0x00, 0x00, 0x10, 0x7F, 0x01, 0x7F, 0x0E, 0x01, 0x8E, 0x80, 0x1F, 0x1F, 
            0x8E, 0x01, 0x7F, 0x0E, 0xAF, 0x10, 0x0E, 0x0F, 0x0F, 0x7F, 0x10, 0x0F, 0x8F, 0xEF, 0x3F, 0x1F, 
            0x7F, 0xFF, 0x21, 0x00, 0x1F, 0x0E, 0x0E, 0x21, 0x0E, 0x0F, 0x10, 0x16, 0x0E, 0x1F, 0x0F, 0x10, 
            0x0E, 0x0E, 0x3F, 0x0F, 0x00, 0x0E, 0x2F, 0x0E, 0x0E, 0x00, 0x0E, 0x0F, 0x00, 0x0E, 0x0E, 0x10, 
            0x21, 0x0E, 0x6F, 0x0E, 0x8E, 0x8E, 0x0E, 0x6F, 0x8F, 0x0E, 0x0E, 0x00, 0xFF, 0x11, 0x0F, 0x7F, 
            0x1E, 0x21, 0x21, 0x00, 0x9F, 0x8E, 0x31, 0x90, 0x10, 0xB1, 0x0E, 0x00, 0x00, 0x0E, 0x21, 0x0E, 
            0xCE, 0x9F, 0x7F, 0x3F, 0xFF, 0x0E, 0x2F, 0x01, 0x90, 0x8F, 0x0E, 0x0E, 0x0F, 0x10, 0x7F, 0x00, 
            0xFF, 0x8E, 0x3F, 0x80, 0x8E, 0x81, 0x0F, 0x7F, 0x0F, 0x7F, 0x0E, 0x21, 0x8E, 0x80, 0x10, 0x0F, 
            0xBF, 0x7F, 0x21, 0x0E, 0x8E, 0x01, 0x0E, 0x00, 0x1E, 0x7F, 0x0E, 0x0E, 0xB1, 0xB1, 0x10, 0x90, 
            0x90, 0x90, 0x10, 0x1F, 0x1F, 0x00, 0x0E, 0x3F, 0xFF, 0x6F, 0x1F, 0x1F, 0x7F, 0x6F, 0x0F, 0x0E, 
            0x0E, 0x1E, 0x01, 0x10, 0x01, 0x01, 0x00, 0x0F, 0x0E, 0x00, 0x7F, 0x0F, 0x0E, 0x81, 0x10, 0x0E, 
            0x00, 0x0E, 0x0F, 0x1F, 0x00, 0x10, 0x7F, 0x3F, 0x0F, 0x0F, 0x00, 0x7F, 0x7F, 0x7F, 0x1F, 0x21, 
            0x00, 0x10, 0x1F, 0x0E, 0x21, 0x7F, 0x0E, 0x0F, 0x01, 0x7F, 0x71, 0x71, 0x11, 0x0E, 0x31, 0x21, 
            0x8E, 0x9F, 0x0F, 0x0E, 0xBF, 0x0F, 0x6F, 0x7F, 0x90, 0x10, 0x8F, 0x0F, 0x0E, 0x01, 0x7F, 0x7F, 
            0x8E, 0x80, 0x0E, 0x8F, 0xEF, 0xEF, 0x00, 0x00, 0x00, 0x06, 0x0E, 0x00, 0x8E, 0x9F, 0x2F, 0x21, 
            0xDF, 0x0E, 0x5F, 0x00, 0x80, 0x00, 0x01, 0x7F, 0x0F, 0x06, 0x06, 0x0E, 0x80, 0x8F, 0x0F, 0x8F, 
            0x1F, 0x8E, 0x0E, 0x21, 0x00, 0x0E, 0x0F, 0x0F, 0xEF, 0x0F, 0x6F, 0x0F, 0x9F, 0x10, 0x0F, 0x1F, 
            0x6F, 0x01, 0x6F, 0x0E, 0x00, 0x6F, 0x0E, 0x2F, 0x6F, 0x01, 0x6F, 0x6F, 0x8F, 0x0E, 0x1E, 0x0E, 
            0x0F, 0x0F, 0x0E, 0x0F, 0xFF, 0x0F, 0x4F, 0x7F, 0x8E, 0x7F, 0x06, 0x7F, 0x00, 0x00, 0x5F, 0x00, 
            0x01, 0x10, 0x0E, 0x0F, 0x4E, 0x0F, 0x7F, 0x4F, 0x41, 0x00, 0x00, 0x10, 0x01, 0x06, 0x01, 0x5F, 
            0x8F, 0x8E, 0x1F, 0x00, 0x8E, 0x0F, 0x0F, 0x40, 0x8E, 0x7F, 0x01, 0x1F, 0x0E, 0x4F, 0x10, 0x10, 
            0x90, 0x80, 0x7F, 0xFF, 0x0E, 0x87, 0x0E, 0x1F, 0x0E, 0x4F, 0x90, 0x3F, 0x8F, 0x21, 0x0E, 0x0F, 
            0xFF, 0x00, 0x7F, 0x0E, 0x00, 0x0F, 0x0F, 0x0E, 0x3F, 0x71, 0x7F, 0x0F, 0x4F, 0x7F, 0x21, 0x51, 
            0x10, 0x10, 0x2F, 0x0F, 0x0E, 0x00, 0x6F, 0x06, 0x8E, 0x5F, 0x0E, 0x0E, 0x0E, 0x4F, 0x7F, 0x00, 
            0x7F, 0x01, 0x00, 0x6F, 0x40, 0x7F, 0x41, 0x1F, 0x0E, 0x5F, 0x6F, 0x0E, 0x00, 0x7F, 0x2F, 0x00, 
            0x00, 0x21, 0x10, 0x00, 0x8F, 0x2F, 0x0E, 0x41, 0xAF, 0x41, 0x0F, 0x01, 0x0E, 0x4F, 0x6F, 0x0F, 
            0x0F, 0x4F, 0x00, 0x7F, 0x0F, 0x7F, 0x7F, 0xEF, 0x0E, 0xFF, 0x1F, 0x0E, 0x06, 0x0E, 0x00, 0x10, 
            0x90, 0x01, 0x00, 0x40, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x1F, 0x5F, 0x00, 0x06, 0x00, 0x00, 
            0x4E, 0x4E, 0x7F, 0x06, 0x0F, 0x7F, 0x7F, 0x00, 0x5F, 0x2F, 0x7F, 0x4F, 0xDF, 0x8F, 0x7F, 0x4F, 
            0xDF, 0x0F, 0x2F, 0x7F, 0xFF, 0x7F, 0x7F, 0x00, 0x40, 0x01, 0x6F, 0x00, 0x90, 0x80, 0x00, 0xEF, 
            0x4F, 0xFF, 0x0E, 0x41, 0x4F, 0x00, 0x0F, 0x01, 0x80, 0x00, 0x0E, 0x4F, 0x00, 0x0F, 0x0E, 0x0E, 
            0x06, 0x0F, 0x0F, 0x00, 0x0E, 0x7F, 0x00, 0x5F, 0x00, 0x7F, 0x7F, 0x7F, 0xFF, 0x01, 0x7F, 0x7F, 
            0x61, 0x00, 0x6F, 0x2F, 0xEF, 0x40, 0x01, 0x0F, 0x41, 0x6F, 0x6F, 0x01, 0x00, 0x1F, 0x0F, 0x6F, 
            0x3F, 0x0F, 0x0F, 0x2F, 0x00, 0x0E, 0x31, 0x10, 0x00, 0x0E, 0x00, 0x00, 0x0E, 0x0E, 0x7F, 0x0E, 
            0x9F, 0x9F, 0x40, 0x7F, 0x80, 0x0F, 0x0F, 0x5F, 0x8E, 0x6F, 0x0E, 0x7F, 0x00, 0x00, 0x10, 0x01, 
            0x9F, 0x81, 0x07, 0x00, 0x06, 0x8E, 0x00, 0x0E, 0x4F, 0x01, 0x0F, 0x00, 0xDF, 0x00, 0x7F, 0x6F, 
            0x8F, 0x4F, 0x21, 0x4F, 0x80, 0x4E, 0x0F, 0x7F, 0x5F, 0x1F, 0x4F, 0x0F, 0x80, 0x0E, 0x01, 0x0F, 
            0x6F, 0x8F, 0x4F, 0x00, 0x00, 0x4F, 0x06, 0x06, 0x8E, 0xC1, 0x1E, 0x0F, 0x80, 0x6F, 0x06, 0xCF, 
            0x8E, 0x0E, 0x86, 0x06, 0x00, 0x0E, 0x40, 0x00, 0x81, 0x81, 0x6F, 0x8E, 0x01, 0x00, 0x00, 0x6F, 
            0x01, 0x00, 0x10, 0x01, 0xEF, 0x06, 0x0E, 0x0E, 0xCF, 0x06, 0x4F, 0x0F, 0xFF, 0x0E, 0x1F, 0x0F, 
            0x7F, 0x0E, 0x7F, 0x7F, 0x9F, 0x8E, 0x7F, 0x21, 0x0F, 0x2F, 0x01, 0x01, 0x0E, 0x00, 0x7F, 0x11, 
            0xA1, 0xA1, 0x21, 0x00, 0x9F, 0x0E, 0x3F, 0x7F, 0xA1, 0x00, 0x01, 0x0F, 0x01, 0x0E, 0x0E, 0x1F, 
            0x8E, 0xE1, 0x0E, 0xAF, 0x7F, 0xA1, 0x01, 0x0E, 0x0E, 0x00, 0x00, 0x0E, 0xAF, 0xBF, 0xA1, 0x1E, 
            0x8E, 0x7F, 0x0F, 0x3F, 0x8E, 0x6F, 0x8F, 0x0F, 0x7F, 0x10, 0x7F, 0x40, 0xBF, 0x8E, 0x21, 0xFF, 
            0x81, 0x8F, 0x0E, 0x61, 0x01, 0xFF, 0x6F, 0x01, 0xFF, 0x9F, 0x0F, 0x7F, 0xFF, 0x1E, 0x0E, 0x3F, 
            0xFF, 0x7F, 0x7F, 0x7F, 0x00, 0x10, 0x01, 0x7F, 0xFF, 0x8F, 0x7F, 0xB1, 0xA1, 0x8F, 0x1F, 0x0F, 
            0x1F, 0x00, 0x00, 0x01, 0xA1, 0x90, 0x21, 0x4F, 0x1F, 0x21, 0x21, 0x3F, 0x7F, 0x21, 0x1F, 0x7F, 
            0x01, 0x61, 0x01, 0x40, 0x7F, 0x00, 0x00, 0x01, 0x00, 0x7F, 0x3F, 0x7F, 0x7F, 0x6F, 0x01, 0x0F, 
            0x21, 0x01, 0x6F, 0x0E, 0x2F, 0x7F, 0x7F, 0x21, 0x00, 0x21, 0x7F, 0x2F, 0x0E, 0x21, 0x00, 0x6F, 
            0x6F, 0x01, 0x10, 0x21, 0x31, 0x40, 0x6F, 0x2F, 0x7F, 0x1F, 0x41, 0x01, 0x81, 0x8F, 0x0F, 0x7F, 
            0x8E, 0x01, 0x21, 0x10, 0x86, 0x01, 0x7F, 0x6F, 0x00, 0x21, 0x6F, 0x7F, 0x8E, 0x80, 0x61, 0xFF, 
            0x8F, 0xFF, 0x1B, 0x01, 0x3F, 0x0F, 0x10, 0x10, 0x8E, 0x9F, 0x3F, 0x00, 0x81, 0x7F, 0x00, 0x10, 
            0x90, 0x10, 0x0F, 0x3F, 0x1F, 0x7F, 0x00, 0x0E, 0x80, 0x80, 0x06, 0xFF, 0x10, 0x80, 0x10, 0x10, 
            0x1E, 0x71, 0x7F, 0x7F, 0x91, 0x7F, 0x0E, 0x00, 0xFF, 0x7F, 0x07, 0x0E, 0xFF, 0x7F, 0x7F, 0x0F, 
            0x0F, 0x0F, 0x7F, 0x31, 0x9F, 0xFF, 0x01, 0x7F, 0x90, 0x90, 0x3F, 0x0E, 0x0F, 0x7F, 0x0F, 0x0F, 
            0x8E, 0x0E, 0x7F, 0x00, 0x7F, 0x7F, 0x10, 0x0F, 0xFF, 0x6F, 0x0E, 0x7F, 0x01, 0x0E, 0x3F, 0x0F, 
            0x0F, 0x8F, 0x7F, 0xBF, 0x1F, 0x00, 0x1E, 0x1F, 0x10, 0x21, 0x10, 0x10, 0x81, 0x86, 0x01, 0x0F, 
            0xFF, 0x1F, 0x6F, 0x27, 0xFF, 0x0E, 0x06, 0x0E, 0x0E, 0x1F, 0x6F, 0x0E, 0xFF, 0x8E, 0x2F, 0xBF, 
            0x5F, 0xBF, 0x01, 0x1F, 0x01, 0x10, 0x00, 0x01, 0x8F, 0x2F, 0x1F, 0x2F, 0x1F, 0x6F, 0x00, 0x0F, 
            0x1E, 0x0E, 0x01, 0x0F, 0x06, 0x1F, 0x0F, 0x10, 0x10, 0x5F, 0x06, 0x17, 0x0F, 0x8F, 0x0F, 0x7F, 
            0x7F, 0x6F, 0x7F, 0x0F, 0x7F, 0x1F, 0x5F, 0x2F, 0x0E, 0x1F, 0x10, 0x10, 0x00, 0x1F, 0x00, 0x0E, 
            0x3F, 0x06, 0x0F, 0x1F, 0x3F, 0x1F, 0x00, 0x00, 0x1F, 0x01, 0x0F, 0x61, 0x1F, 0x0E, 0x10, 0x10, 
            0x91, 0x01, 0x6F, 0x1F, 0x81, 0x01, 0x2F, 0x2F, 0x2F, 0x0F, 0x01, 0x01, 0x7F, 0x10, 0x40, 0x00, 
            0x0E, 0x00, 0x6F, 0x0E, 0x10, 0x01, 0x7F, 0x06, 0x2F, 0x0F, 0x06, 0x00, 0xBF, 0x00, 0x10, 0x0F, 
            0x8E, 0x6F, 0x2F, 0x10, 0xCF, 0x0F, 0x00, 0x21, 0x21, 0x6F, 0x2F, 0x10, 0x6F, 0x8F, 0x0F, 0x91, 
            0x6F, 0x0F, 0x6F, 0x00, 0x01, 0x0E, 0x2B, 0x00, 0xFF, 0x8E, 0x21, 0x7F, 0xEF, 0x7F, 0x7F, 0x8E, 
            0xA1, 0x0F, 0xBF, 0x21, 0x7F, 0x01, 0x7F, 0x7F, 0xBF, 0x86, 0x0E, 0xBF, 0x01, 0x81, 0x7F, 0x0E, 
            0x61, 0x10, 0x0E, 0x7F, 0x80, 0x07, 0x3F, 0x00, 0x00, 0x3F, 0x2F, 0x2F, 0x2F, 0x0F, 0x2F, 0x01, 
            0x01, 0x00, 0x01, 0x01, 0x1F, 0x01, 0x2F, 0x01, 0x01, 0x80, 0x0F, 0x0E, 0x00, 0x0F, 0x7F, 0x0E, 
            0x2F, 0x10, 0x10, 0x00, 0x3F, 0x0E, 0x00, 0x21, 0x2F, 0x21, 0x6F, 0x0E, 0x00, 0x0F, 0x0E, 0x7F, 
            0x00, 0x01, 0x7F, 0x6F, 0x0E, 0x1F, 0x7F, 0x00, 0x7F, 0x0F, 0x7F, 0x01, 0x81, 0x3F, 0x1F, 0x10, 
            0x90, 0x10, 0x7F, 0x0E, 0x9F, 0x00, 0x71, 0x31, 0x10, 0x01, 0x0E, 0x7F, 0xEF, 0x0E, 0x6F, 0xFF, 
            0x01, 0x6F, 0x00, 0x6F, 0x6F, 0x3F, 0x0F, 0x6F, 0x81, 0x0E, 0x2F, 0x00, 0x8F, 0x0E, 0x7F, 0x1E, 
            0x8F, 0x1F, 0x7F, 0x0E, 0x7F, 0x0E, 0x01, 0x06, 0xFF, 0x1F, 0x06, 0xAF, 0x5F, 0x80, 0x7F, 0x00, 
            0x01, 0x7F, 0x00, 0x0E, 0xFF, 0x0E, 0x01, 0x3F, 0xFF, 0x7F, 0x0F, 0x7F, 0xA1, 0x10, 0x01, 0x00, 
            0x0E, 0x0E, 0x0F, 0x0E, 0x81, 0x8F, 0x00, 0x81, 0x00, 0x1F, 0x1F, 0x2F, 0x01, 0x0E, 0x61, 0x3F, 
            0x7F, 0x00, 0x1F, 0x0E, 0x21, 0x2F, 0x31, 0x0E, 0x1F, 0x7F, 0x00, 0x00, 0x00, 0x1F, 0x01, 0x1F, 
            0x00, 0x0E, 0x0E, 0x0E, 0x0F, 0x1F, 0x1F, 0x7F, 0x7F, 0x0F, 0x7F, 0x06, 0xFF, 0x8F, 0x0F, 0x00, 
            0xF1, 0x10, 0x01, 0x10, 0x8E, 0x0F, 0x2F, 0x3F, 0x00, 0x00, 0x0F, 0x0E, 0x8E, 0xCF, 0x0F, 0x8F, 
            0x10, 0x90, 0x0F, 0x00, 0x0E, 0x0E, 0x0F, 0x01, 0x80, 0x8E, 0x3F, 0x7F, 0x8F, 0x00, 0x0F, 0x0F, 
            0x8F, 0x7F, 0x1F, 0x7F, 0x01, 0xBF, 0x7F, 0x0F, 0xBF, 0x8E, 0x0F, 0xFF, 0x1F, 0x86, 0x0E, 0x0E, 
            0x01, 0x00, 0x01, 0x01, 0x81, 0x80, 0x00, 0x00, 0x80, 0x7F, 0x01, 0x00, 0x8E, 0x1F, 0x77, 0x7F, 
            0x1F, 0x0E, 0x0E, 0x06, 0x9F, 0x9E, 0x00, 0x8E, 0x86, 0x8E, 0x00, 0x6F, 0x1F, 0x1F, 0x3F, 0x0E, 
            0x80, 0x7F, 0x10, 0x00, 0x01, 0x0F, 0x01, 0x0E, 0x00, 0x0E, 0x0F, 0x0E, 0x0E, 0x0E, 0x6F, 0x1F, 
            0x3F, 0x0E, 0x1F, 0x06, 0x00, 0xFF, 0x00, 0x1F, 0x06, 0x10, 0x1F, 0x0E, 0x7F, 0x6F, 0x01, 0x01, 
            0x7F, 0x07, 0x06, 0x7F, 0x0E, 0x1F, 0x2F, 0x06, 0x00, 0x21, 0x0E, 0x7F, 0x6F, 0x3F, 0x01, 0x7F, 
            0x00, 0x00, 0x0E, 0x01, 0x06, 0x00, 0x00, 0x2F, 0x80, 0x8F, 0x0E, 0x00, 0x8E, 0x3F, 0x0E, 0x0E, 
            0x8E, 0x01, 0x81, 0x0E, 0x0E, 0x0F, 0x00, 0x7F, 0x8E, 0x8F, 0x2F, 0xFF, 0x7F, 0x80, 0x0F, 0x01, 
            0x0F, 0x00, 0x7F, 0x7F, 0x81, 0xFF, 0x01, 0x3F, 0x86, 0x7F, 0x01, 0x10, 0x9F, 0x7F, 0x7F, 0x6F, 
            0x0F, 0x0F, 0x0E, 0x7F, 0x81, 0x80, 0x10, 0x8E, 0x1F, 0x8E, 0x0E, 0x0F, 0x2F, 0x1E, 0x10, 0x10, 
            0x90, 0x90, 0x7F, 0x0E, 0x80, 0x0F, 0x0F, 0x00, 0x80, 0x1F, 0x00, 0x00, 0x3F, 0x1F, 0x0F, 0x0E, 
            0x8E, 0x9F, 0x7F, 0x8E, 0xBF, 0xA1, 0x10, 0x10, 0x10, 0x00, 0x10, 0x00, 0x90, 0x1F, 0x00, 0x1E, 
            0xAF, 0x2F, 0x0F, 0x1F, 0x80, 0x7F, 0x7F, 0x7F, 0x0E, 0x16, 0x06, 0x0E, 0x5E, 0x9F, 0x0F, 0x0E, 
            0x1F, 0x0E, 0x6F, 0x0E, 0x7F, 0x0F, 0x00, 0x1F, 0x9E, 0x90, 0x31, 0x10, 0x90, 0x00, 0x10, 0x21, 
            0x90, 0x00, 0x00, 0x0F, 0x01, 0xFF, 0x00, 0x11, 0x8F, 0x8E, 0x0E, 0x80, 0x1F, 0x8E, 0x3F, 0x40, 
            0x1E, 0x80, 0x17, 0x00, 0x80, 0x90, 0x10, 0x10, 0x80, 0x00, 0x00, 0x00, 0x8F, 0x7F, 0x07, 0x00, 
            0x0F, 0x01, 0x0E, 0x01, 0x2F, 0x0E, 0x00, 0x80, 0x0F, 0x8E, 0x1F, 0x7F, 0x7F, 0x1E, 0x1F, 0x6F, 
            0x8E, 0x8E, 0x21, 0x00, 0xA1, 0x00, 0x00, 0x10, 0x00, 0x0E, 0x21, 0x0E, 0x7F, 0x00, 0x0F, 0x0E, 
            0x06, 0x65, 0x00, 0x86, 0x0E, 0x8E, 0x2F, 0x7F, 0x00, 0x1E, 0x01, 0x00, 0xFF, 0xF1, 0x7F, 0x0F, 
            0x80, 0x7F, 0x00, 0x3F, 0x8F, 0x2F, 0x0E, 0x7F, 0x1F, 0x0E, 0x3F, 0x10, 0x10, 0x81, 0x0F, 0x9F, 
            0x06, 0x9F, 0x0E, 0x7F, 0x0F, 0x1E, 0x00, 0x7F, 0x8F, 0x6F, 0x2F, 0x0E, 0x8F, 0x0F, 0x00, 0x21, 
            0x9F, 0x11, 0x1F, 0x1F, 0x1E, 0x21, 0x00, 0x00, 0x00, 0x90, 0x6F, 0xCF, 0x6F, 0x01, 0x3F, 0x01, 
            0x00, 0x7F, 0x00, 0x0E, 0x86, 0xFF, 0x21, 0x01, 0x8F, 0x1F, 0x7F, 0x7F, 0x8F, 0x0E, 0x4F, 0x0E, 
            0x7F, 0xAF, 0x01, 0x01, 0xBF, 0x1F, 0x01, 0x80, 0x00, 0x90, 0x2F, 0x0F, 0x00, 0x0E, 0x0E, 0x0F, 
            0x8F, 0x0F, 0x06, 0x0E, 0x81, 0x0E, 0x01, 0x2F, 0xFF, 0x0E, 0x0F, 0x00, 0x7F, 0x0E, 0x00, 0x6F, 
            0x80, 0x00, 0x1F, 0x0F, 0x3F, 0xC0, 0x01, 0x1F, 0x00, 0x0E, 0x1F, 0x4F, 0x8E, 0xAF, 0x06, 0x07, 
            0x81, 0x7F, 0x0E, 0x00, 0x8E, 0x1E, 0x01, 0x21, 0x00, 0x0E, 0x00, 0x00, 0x0F, 0x00, 0x6F, 0x00, 
            0x07, 0xA7, 0x27, 0x1E, 0x0E, 0x00, 0x40, 0x06, 0x80, 0xAF, 0x01, 0x0E, 0x8F, 0x0E, 0x1F, 0x01, 
            0x81, 0x06, 0x00, 0x0E, 0x16, 0x0F, 0x2F, 0x0F, 0x8E, 0x8F, 0x6F, 0xEF, 0x6F, 0x8F, 0x00, 0x0E, 
            0x21, 0x0F, 0x00, 0x00, 0x91, 0x8F, 0x00, 0x00, 0x8E, 0x0E, 0x7F, 0x11, 0xB1, 0x0F, 0x7F, 0x0F, 
            0x7F, 0x7F, 0x0E, 0x00, 0x8F, 0x0F, 0x17, 0xEF, 0x7F, 0xFF, 0x0F, 0x0F, 0x0E, 0x00, 0x00, 0x0F, 
            0xEF, 0x8F, 0x0E, 0x3F, 0x8F, 0x2F, 0xFF, 0xFF, 0x81, 0x80, 0x7F, 0x5F, 0x5F, 0x7F, 0x7F, 0x3F, 
            0xFF, 0x8E, 0x31, 0x8E, 0x0E, 0xFF, 0x00, 0x1F, 0x7F, 0x8F, 0x0E, 0x0F, 0x80, 0x7F, 0x00, 0x10, 
            0x86, 0x00, 0x0F, 0x0E, 0x8E, 0x00, 0x00, 0x31, 0x10, 0x0F, 0x7F, 0x01, 0x8F, 0x86, 0x01, 0x6F, 
            0x31, 0xFF, 0x0E, 0x00, 0x31, 0x31, 0x00, 0x00, 0x8E, 0x80, 0x3F, 0x0F, 0x81, 0x6F, 0x6F, 0x7F, 
            0x8E, 0x41, 0x0E, 0x0F, 0x2F, 0x0F, 0x0F, 0x0E, 0x8F, 0x80, 0x1E, 0xAF, 0x0F, 0xA1, 0x80, 0x80, 
            0x0E, 0x0F, 0x86, 0x00, 0x0E, 0x80, 0xEF, 0xFF, 0x8F, 0x01, 0x6F, 0x0F, 0x0E, 0x0E, 0xEF, 0x81, 
            0xAF, 0xCF, 0x8E, 0xFF, 0x8F, 0x80, 0x0E, 0x8E, 0x8F, 0xFF, 0xBF, 0x9F, 0x00, 0x7F, 0xFF, 0x6F, 
            0x7F, 0x7F, 0x81, 0x10, 0x01, 0x0E, 0x10, 0x01, 0x0F, 0x7F, 0xC1, 0xFF, 0x00, 0x9E, 0x8E, 0x8F, 
            0x00, 0x7F, 0x7F, 0x00, 0x01, 0x7F, 0x8F, 0x80, 0x7F, 0x00, 0x81, 0x00, 0x0E, 0x0F, 0x87, 0x7F, 
            0x0F, 0x21, 0x01, 0x3F, 0x7F, 0x01, 0x7F, 0x7F, 0x00, 0xFF, 0x0F, 0x9F, 0x2F, 0x1F, 0x0E, 0x10, 
            0x31, 0x01, 0x80, 0x00, 0x00, 0x00, 0x8E, 0x0E, 0x1F, 0x00, 0x8E, 0x7F, 0x01, 0x10, 0x00, 0x00, 
            0x7F, 0x00, 0x1F, 0x1F, 0x0E, 0x4E, 0x10, 0x01, 0x6F, 0x01, 0x0F, 0x7F, 0x0F, 0x7F, 0xB1, 0x80, 
            0x0E, 0x0F, 0x8E, 0x0E, 0x7F, 0x87, 0x80, 0x7F, 0xAF, 0x6F, 0x7F, 0x0F, 0x6F, 0x00, 0x9F, 0xEF, 
            0x7F, 0x81, 0x86, 0x90, 0x80, 0x00, 0x0E, 0x81, 0x86, 0x80, 0x80, 0x0F, 0x00, 0x01, 0xEF, 0x0F, 
            0x0F, 0x6F, 0xA1, 0x10, 0x01, 0x01, 0x40, 0x7F, 0x01, 0x1F, 0x81, 0xFF, 0x7F, 0xFF, 0xA1, 0xFF, 
            0x6F, 0x00, 0x21, 0x21, 0x0F, 0x01, 0xFF, 0x8F, 0x8F, 0x61, 0x8E, 0x0E, 0x0F, 0xEF, 0xEF, 0x21, 
            0x21, 0x10, 0x00, 0x00, 0x0E, 0x01, 0x87, 0xFF, 0x0E, 0x81, 0x8E, 0x8E, 0x0F, 0x01, 0x00, 0x8F, 
            0x0F, 0x6F, 0xEF, 0x21, 0x00, 0x01, 0x80, 0x0F, 0x0E, 0x6F, 0x81, 0x6F, 0x0E, 0x0F, 0x0E, 0x00, 
            0x7F, 0x0E, 0x00, 0x80, 0x7F, 0x8F, 0x7F, 0x10, 0x1F, 0x7F, 0x7F, 0x06, 0x00, 0x61, 0x8F, 0xFF, 
            0x1F, 0x7F, 0x81, 0x31, 0x7F, 0x8F, 0x9F, 0x06, 0xFF, 0x0E, 0x8F, 0x0F, 0x0E, 0x90, 0x80, 0x8E, 
            0x0F, 0x91, 0x7F, 0xE1, 0x8F, 0x7F, 0x3F, 0x3F, 0x7F, 0x0E, 0x80, 0x80, 0x10, 0x10, 0x8F, 0x7F, 
            0x1F, 0x00, 0x8E, 0x0F, 0x80, 0x10, 0x10, 0x00, 0x1E, 0x6F, 0xEF, 0x8E, 0x1F, 0x90, 0x6F, 0xEF, 
            0x21, 0x0E, 0x0E, 0x1F, 0x0E, 0x0F, 0xAF, 0x0E, 0x0E, 0x01, 0xEF, 0x01, 0x01, 0x00, 0x8E, 0x3F, 
            0x0E, 0x0E, 0x00, 0x21, 0x6F, 0x6F, 0x01, 0xEF, 0x21, 0x01, 0xEF, 0xFF, 0x01, 0x21, 0x01, 0x21, 
            0x00, 0x2F, 0xA1, 0xA1, 0x00, 0x21, 0x90, 0x40, 0x0E, 0x7F, 0xA1, 0x01, 0x21, 0x61, 0x10, 0x7F, 
            0x7F, 0x21, 0xA1, 0xFF, 0x0E, 0xFF, 0x3F, 0xA1, 0x01, 0x01, 0x61, 0x2F, 0x01, 0x7F, 0x8E, 0x7F, 
            0x21, 0x01, 0x80, 0x00, 0x01, 0x00, 0xEF, 0x0E, 0x21, 0x7F, 0x7F, 0x07, 0x00, 0x1F, 0x01, 0xFF, 
            0x06, 0x86, 0x21, 0x10, 0x3F, 0x0E, 0x3F, 0x01, 0x3F, 0x0E, 0xBF, 0xB1, 0x00, 0x00, 0x8E, 0x7F, 
            0x0E, 0xFF, 0x86, 0x8E, 0x91, 0x10, 0x00, 0x00, 0x2F, 0xFF, 0x8E, 0xFF, 0x00, 0x8E, 0x7F, 0x81, 
            0xFF, 0x00, 0x0E, 0x0F, 0x1E, 0x3F, 0xFF, 0x81, 0x3F, 0x01, 0x9F, 0x1F, 0x7F, 0x4F, 0x9F, 0x0F, 
            0x0F, 0x1F, 0x0E, 0x3F, 0x0F, 0x2F, 0x21, 0x00, 0x16, 0x61, 0x90, 0x8E, 0x10, 0x01, 0x10, 0x01, 
            0x0F, 0x01, 0x80, 0x00, 0x0E, 0x3F, 0x6F, 0x1F, 0x2F, 0x7F, 0x0E, 0x7F, 0x7F, 0x6F, 0x0F, 0x7F, 
            0x7F, 0x7F, 0x7F, 0x00, 0x7F, 0x81, 0x7F, 0x00, 0x1E, 0x00, 0x10, 0x10, 0x00, 0x31, 0x80, 0x80, 
            0x5F, 0x0E, 0x90, 0x0E, 0x0F, 0x2F, 0x8F, 0x2F, 0x00, 0x0F, 0x01, 0x00, 0x61, 0x0F, 0xFF, 0x81, 
            0x2F, 0xBF, 0x1F, 0x0F, 0x2F, 0x7F, 0x0E, 0x1F, 0x06, 0x10, 0xA1, 0x10, 0x00, 0x40, 0x90, 0x00, 
            0x10, 0x00, 0x8E, 0x00, 0x0F, 0x1E, 0x21, 0x6F, 0x0E, 0x01, 0x0E, 0x6F, 0x0E, 0x3F, 0x0F, 0x6F, 
            0x1F, 0x0F, 0x21, 0x00, 0x7F, 0x7F, 0x8F, 0xFF, 0x2F, 0x2F, 0xE1, 0x0F, 0x3F, 0xFF, 0xA1, 0xA1, 
            0x0F, 0x7F, 0x10, 0x00, 0x0E, 0x01, 0x90, 0x90, 0x10, 0x90, 0x31, 0x90, 0x1F, 0x7F, 0x00, 0x0F, 
            0x01, 0x7F, 0x81, 0x00, 0x7F, 0x07, 0x01, 0x0E, 0x7F, 0x0E, 0x01, 0x01, 0x01, 0x01, 0x0E, 0x00, 
            0x6F, 0x7F, 0x01, 0x00, 0x6F, 0x01, 0x7F, 0x7F, 0x01, 0x6F, 0x07, 0x2F, 0x3F, 0x00, 0x8E, 0xA1, 
            0x00, 0x10, 0xC0, 0x10, 0x0F, 0x8E, 0x81, 0x01, 0x8E, 0x1F, 0x1F, 0x0F, 0x0E, 0x3F, 0x8E, 0x8E, 
            0x6F, 0x9E, 0x0E, 0x8E, 0xEF, 0x3F, 0x3F, 0x0E, 0x8F, 0x6F, 0x9F, 0x86, 0x80, 0x21, 0x90, 0x00, 
            0x0E, 0x1E, 0x8F, 0x2F, 0x8F, 0x01, 0x21, 0x3F, 0x0F, 0x0F, 0x81, 0x8E, 0x11, 0x8E, 0x9F, 0x8E, 
            0x2F, 0x0F, 0x0F, 0x8F, 0x01, 0x0E, 0x8F, 0xFF, 0x00, 0x0F, 0x8E, 0x1F, 0x1F, 0x2F, 0x8F, 0x1F, 
            0x06, 0x2F, 0x61, 0x21, 0x21, 0x00, 0x90, 0xA1, 0x00, 0x8F, 0x80, 0x81, 0x0E, 0x0E, 0x4E, 0x3F, 
            0x0F, 0x6F, 0x8E, 0xAF, 0x00, 0x0F, 0x1F, 0x7F, 0x0F, 0x0E, 0xFF, 0x6F, 0x0F, 0x0E, 0x00, 0x00, 
            0x0E, 0x01, 0x00, 0x0F, 0x01, 0x00, 0x01, 0x2F, 0x3F, 0x0F, 0x01, 0x0E, 0x0E, 0x6F, 0x6F, 0x7F, 
            0x3F, 0x10, 0x00, 0x6F, 0x00, 0x0E, 0x01, 0x0E, 0x0E, 0x0E, 0x0F, 0x00, 0x6F, 0x7F, 0x00, 0x21, 
            0x2F, 0x01, 0x01, 0x0F, 0x2F, 0x01, 0x21, 0x0E, 0x21, 0x0F, 0x8F, 0x8F, 0x8F, 0x01, 0x81, 0x2F, 
            0x6F, 0x01, 0x80, 0x0F, 0x6F, 0x2F, 0x0E, 0x0F, 0x00, 0x0F, 0x86, 0x8E, 0x6F, 0x81, 0x8F, 0x8E, 
            0x01, 0x00, 0x0F, 0x7F, 0x10, 0x2F, 0x80, 0x80, 0x0E, 0x0F, 0x8E, 0x10, 0x11, 0x1F, 0x21, 0x06, 
            0x06, 0x2F, 0x21, 0x61, 0x0E, 0x0E, 0x81, 0x01, 0x06, 0xFF, 0x0E, 0xA1, 0x06, 0x2F, 0x00, 0x01, 
            0x21, 0x1E, 0xFF, 0x00, 0x0F, 0x7F, 0x86, 0x7F, 0x7F, 0x00, 0x80, 0x21, 0x10, 0x10, 0x1F, 0x7F, 
            0x01, 0x27, 0x3F, 0x0E, 0x31, 0xFF, 0x81, 0x65, 0x2F, 0x00, 0x10, 0xA1, 0x00, 0x21, 0xF1, 0x0E, 
            0x01, 0x7F, 0x7F, 0x7F, 0x06, 0x0E, 0x01, 0x01, 0x00, 0x7F, 0x0F, 0x0F, 0x1F, 0x3F, 0x7F, 0x0F, 
            0x21, 0x10, 0x21, 0x00, 0x01, 0x3F, 0x00, 0x11, 0x1F, 0x01, 0xFF, 0x86, 0x1E, 0x11, 0xAF, 0x01, 
            0x86, 0x0E, 0xAF, 0x01, 0x0F, 0x1F, 0x7F, 0x1F, 0x21, 0x1F, 0x81, 0x80, 0x3F, 0x8F, 0x6F, 0xFF, 
            0xFF, 0x80, 0x2F, 0x00, 0x00, 0x00, 0x81, 0xEF, 0x16, 0x01, 0x0E, 0x01, 0x0E, 0x00, 0x86, 0x3F, 
            0x0E, 0x0E, 0x2F, 0x00, 0x2F, 0x01, 0x2F, 0x0E, 0x3F, 0x00, 0xA1, 0xFF, 0x00, 0x06, 0x6F, 0x00, 
            0x6F, 0x0E, 0xEF, 0x8F, 0x6F, 0x6F, 0x6F, 0x4F, 0x4F, 0x0F, 0x7F, 0x6F, 0x06, 0x6F, 0x0E, 0x0F, 
            0x6F, 0x0F, 0x0F, 0x0F, 0x6F, 0x2F, 0x0F, 0xEF, 0x6F, 0x6F, 0x6F, 0x6F, 0x2F, 0x0F, 0x8F, 0x0E, 
            0x7F, 0x6F, 0xFF, 0x01, 0x00, 0x00, 0x81, 0x10, 0x0E, 0x0E, 0x01, 0x7F, 0x6F, 0x01, 0xEF, 0xFF, 
            0x6F, 0x0F, 0x2F, 0xEF, 0x00, 0x6F, 0x0E, 0x00, 0x0E, 0x6F, 0xEF, 0x06, 0x0F, 0x0F, 0x0F, 0x2F, 
            0x0F, 0x0F, 0x6F, 0x6F, 0x6F, 0x6F, 0x6F, 0x0F, 0x21, 0x00, 0x0E, 0x21, 0x21, 0x01, 0x27, 0x2F, 
            0x6F, 0x0F, 0x0E, 0x0F, 0x0E, 0x0E, 0xAF, 0x81, 0x0E, 0x01, 0x81, 0x2F, 0x2F, 0x01, 0xEF, 0x06, 
            0x00, 0x6F, 0x6F, 0x2F, 0x00, 0x81, 0xEF, 0x8F, 0x6F, 0x8E, 0x4F, 0x8E, 0x80, 0x0E, 0x7F, 0x0E, 
            0x0F, 0x6F, 0xEF, 0x6F, 0x00, 0x21, 0x61, 0x10, 0x01, 0x6F, 0x8F, 0x0F, 0x0E, 0x2F, 0x00, 0x4F, 
            0x0E, 0x0F, 0x01, 0x0F, 0x0E, 0x0F, 0x8F, 0x80, 0x0F, 0x21, 0x00, 0x01, 0x0F, 0x0F, 0xCE, 0xEF, 
            0x00, 0x0F, 0x8E, 0x6F, 0x0E, 0x00, 0x8F, 0x01, 0x00, 0x21, 0x21, 0x21, 0x6F, 0x0F, 0x8F, 0xAF, 
            0x6F, 0x0E, 0xEF, 0x2F, 0x00, 0x00, 0x1F, 0x0F, 0x00, 0x0E, 0x8F, 0x21, 0x6F, 0x6F, 0xBF, 0x2F, 
            0x2F, 0x6F, 0xEF, 0x01, 0x6F, 0x6F, 0x01, 0x2F, 0x6F, 0x0F, 0xEF, 0xEF, 0x0E, 0x6F, 0x6F, 0x6F, 
            0x0E, 0x0E, 0x0F, 0x01, 0x0F, 0x0F, 0x8F, 0x6F, 0x00, 0x06, 0x6F, 0x6F, 0x0F, 0x01, 0x0F, 0x21, 
            0x00, 0x21, 0x2F, 0x00, 0x00, 0x21, 0x61, 0x0F, 0x6F, 0x0F, 0x6F, 0x0E, 0x21, 0x6F, 0x6F, 0x0E, 
            0x6F, 0x6F, 0x81, 0xCE, 0x21, 0x01, 0x8F, 0x0E, 0xEF, 0x6F, 0x8F, 0x01, 0xEF, 0x01, 0x0E, 0x6F, 
            0x0E, 0x8F, 0xAF, 0x8E, 0x6F, 0x81, 0x0F, 0x8E, 0x00, 0x0F, 0x0F, 0xEF, 0x00, 0x00, 0x00, 0x00, 
            0x01, 0x21, 0x00, 0x0E, 0x21, 0x21, 0x00, 0x00, 0x0F, 0x6F, 0x6F, 0x2F, 0x0E, 0x0E, 0x0E, 0x0E, 
            0x0F, 0x0F, 0x2F, 0x0E, 0x06, 0x4F, 0x0E, 0x0F, 0x00, 0x00, 0x8E, 0xEF, 0x0F, 0x4F, 0xEF, 0x6F, 
            0x0F, 0x6F, 0xAF, 0x6F, 0x0E, 0x06, 0x21, 0x01, 0x01, 0x0E, 0x8E, 0xEF, 0x01, 0x8F, 0x6F, 0xCF, 
            0x0E, 0x6F, 0x2F, 0x0F, 0x0E, 0x2F, 0x8F, 0xEF, 0x8E, 0x2F, 0xAF, 0x3F, 0x0E, 0xEF, 0xEF, 0x8F, 
            0x8E, 0x7F, 0x0E, 0x6F, 0x0F, 0x0F, 0xEF, 0xDF, 0x0E, 0x8E, 0xA1, 0xA1, 0x86, 0xA1, 0x01, 0x80, 
            0x0F, 0x8E, 0xEF, 0x8E, 0x6F, 0x6F, 0xAF, 0x0E, 0x0F, 0x2F, 0xAF, 0x00, 0x0E, 0x21, 0x6F, 0x0F, 
            0x0E, 0x4F, 0x8F, 0x8E, 0x00, 0x81, 0xA1, 0xA1, 0x06, 0x6F, 0x6F, 0x4F, 0x00, 0x6F, 0x81, 0x8F, 
            0x6F, 0x00, 0x8E, 0x0F, 0x0E, 0x0E, 0x8E, 0x0E, 0x0F, 0x6F, 0x0E, 0x2F, 0x6F, 0x2F, 0x8F, 0x21, 
            0x0E, 0x0F, 0x7F, 0xA1, 0x00, 0x00, 0x0F, 0x0F, 0x2F, 0x21, 0x2F, 0x0E, 0x6F, 0x0F, 0x21, 0x6F, 
            0x21, 0x0F, 0x21, 0x6F, 0x2F, 0x0F, 0x6F, 0x0E, 0x4F, 0x0E, 0x0F, 0x7F, 0x6F, 0x01, 0x01, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x10, 0x90, 0x90, 0x10, 0x10, 0x10, 0x80, 0x10, 
            0x90, 0x90, 0x10, 0x10, 0x00, 0x10, 0x90, 0x90, 0x10, 0x90, 0x90, 0x80, 0x00, 0x10, 0x10, 0x10, 
            0x10, 0x90, 0x90, 0x90, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x90, 0x10, 0x90, 0x80, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 
            0x10, 0x10, 0x90, 0x10, 0x00, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x90, 0x90, 
            0x10, 0x10, 0x90, 0x90, 0x10, 0x10, 0x90, 0x10, 0x10, 0x00, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x80, 0x10, 
            0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x10, 0x90, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 
            0x90, 0x90, 0x90, 0x10, 0x10, 0x10, 0xFF, 0x00, 0x7F, 0x0E, 0xFF, 0x01, 0x01, 0x0F, 0x0F, 0x0F, 
            0x00, 0x00, 0x10, 0x0F, 0x1F, 0x3F, 0x80, 0x21, 0x0E, 0x00, 0x0E, 0x0E, 0x6F, 0x2F, 0x21, 0x0F, 
            0x7F, 0x3F, 0x01, 0x01, 0x7F, 0x7F, 0x00, 0x10, 0x3F, 0x00, 0x10, 0x0F, 0x0F, 0x0F, 0x0E, 0x3F, 
            0x2F, 0x1F, 0x10, 0x0F, 0x01, 0x07, 0x0F, 0x2F, 0x3F, 0x3F, 0xFF, 0x6F, 0x0E, 0x0F, 0xFF, 0x00, 
            0x6F, 0x11, 0xFF, 0x0E, 0x10, 0x6F, 0x0E, 0x6F, 0x21, 0x7F, 0x8E, 0x8E, 0x0E, 0x8F, 0x1F, 0x0F, 
            0x01, 0x10, 0x21, 0x2F, 0x00, 0x6F, 0xAF, 0xAF, 0x6F, 0x0E, 0xFF, 0x0F, 0x7F, 0x0F, 0xFF, 0x01, 
            0x01, 0x0F, 0x0E, 0x00, 0x7F, 0x0E, 0x0F, 0x81, 0x00, 0x80, 0x7F, 0x8E, 0x0F, 0x3F, 0x7F, 0x10, 
            0x0E, 0x3F, 0x8F, 0xA1, 0x7F, 0x0E, 0x8F, 0x1E, 0x6F, 0xEF, 0x81, 0x00, 0x0E, 0x0F, 0x21, 0x3F, 
            0x1E, 0x0E, 0xFF, 0x8E, 0x01, 0x8E, 0x2F, 0x8E, 0x7F, 0x7F, 0x0E, 0x0E, 0x1F, 0x7F, 0x8E, 0x1F, 
            0x0F, 0x7F, 0x0E, 0x0F, 0x00, 0x7F, 0x21, 0x0E, 0x01, 0x06, 0x7F, 0x0E, 0x7F, 0x0E, 0x5F, 0x0F, 
            0x0E, 0x0E, 0xEF, 0x00, 0x7F, 0x7F, 0x00, 0x0E, 0x10, 0x00, 0x80, 0x0F, 0x0E, 0x00, 0x06, 0x5F, 
            0x7F, 0x7F, 0x0E, 0x0F, 0x00, 0x1F, 0x3F, 0x00, 0x0F, 0x7F, 0x0E, 0x6F, 0x0F, 0x0E, 0x1F, 0x0F, 
            0x7F, 0x1E, 0x0F, 0x3F, 0x0E, 0x06, 0x80, 0x06, 0x06, 0x6F, 0x81, 0x3F, 0x7F, 0x0F, 0x8F, 0x0F, 
            0x7F, 0x7F, 0x00, 0x7F, 0x3F, 0x7F, 0xFF, 0x8F, 0x01, 0x40, 0x01, 0x3F, 0x7F, 0x1F, 0x0F, 0x0F, 
            0x7F, 0x0F, 0xBF, 0x81, 0x1F, 0x7F, 0x1F, 0x7F, 0x7F, 0x3F, 0x80, 0x06, 0x3F, 0x01, 0x0F, 0x0F, 
            0x11, 0x2F, 0x90, 0x80, 0x0F, 0x10, 0x1F, 0x90, 0x3F, 0x16, 0x1E, 0x00, 0x0F, 0x40, 0x81, 0x80, 
            0x0F, 0x10, 0x80, 0x7F, 0x0F, 0x1F, 0xFF, 0x9F, 0x90, 0x0F, 0x01, 0x31, 0x01, 0x7F, 0x8F, 0xFF, 
            0x01, 0x80, 0x10, 0xDE, 0x81, 0x00, 0x7F, 0x0E, 0x8E, 0x00, 0xFF, 0x10, 0x10, 0x10, 0x80, 0x0E, 
            0x06, 0x0E, 0xBF, 0x00, 0x10, 0x00, 0x6F, 0x00, 0x7F, 0x00, 0x7F, 0x00, 0x06, 0x0E, 0x6F, 0x21, 
            0x01, 0x21, 0x21, 0x00, 0x07, 0x00, 0x81, 0x86, 0x21, 0x10, 0x90, 0x00, 0x0E, 0x01, 0x80, 0x6F, 
            0xEF, 0x1F, 0x6F, 0x6F, 0x21, 0x0F, 0xAF, 0xEF, 0x0E, 0xAF, 0x00, 0x80, 0x00, 0x6F, 0x6F, 0x7F, 
            0x90, 0x01, 0x8E, 0x90, 0xFF, 0x6F, 0xFF, 0x7F, 0x01, 0xFF, 0x8E, 0xFF, 0xFF, 0x80, 0x01, 0x0E, 
            0x0E, 0x80, 0x8E, 0xBF, 0x00, 0x8E, 0xFF, 0x80, 0xBF, 0x8E, 0x3F, 0x7F, 0x1F, 0xFF, 0x81, 0x90, 
            0x10, 0x7F, 0xFF, 0x0E, 0x10, 0x7F, 0x81, 0x01, 0x00, 0x7F, 0x7F, 0x1F, 0x21, 0x3F, 0x9E, 0xFF, 
            0x00, 0xFF, 0xFF, 0x80, 0x7F, 0x0E, 0x00, 0x10, 0x0F, 0x01, 0xFF, 0x00, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x1E, 0x7F, 0x7F, 0x0E, 0x31, 0x0E, 0x0F, 0x0E, 0x0E, 0x00, 0x7F, 0x1F, 0x7F, 0x0F, 
            0x0E, 0x0E, 0x01, 0x00, 0x0F, 0x0F, 0x3F, 0x1F, 0x0F, 0x3F, 0x7F, 0x7F, 0x1E, 0x3F, 0x7F, 0x00, 
            0x0F, 0x7F, 0x00, 0x0E, 0x7F, 0x7F, 0x06, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x1F, 0x01, 0x3F, 
            0x1F, 0x7F, 0x9F, 0x9F, 0xFF, 0x0E, 0x91, 0x00, 0x7F, 0x8F, 0x9F, 0x0E, 0xBF, 0xFF, 0x0F, 0x00, 
            0x0E, 0x0E, 0xFF, 0x81, 0x00, 0xEF, 0x7F, 0xE1, 0x6F, 0xB1, 0x7F, 0x01, 0x80, 0x10, 0x90, 0x90, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x0E, 0x9F, 0x01, 0x5F, 0x0E, 0x71, 0x6F, 0x01, 0x0E, 0x9F, 0x8E, 
            0x01, 0x8E, 0x8E, 0xEF, 0x0F, 0x0E, 0x00, 0x07, 0x00, 0x6F, 0x80, 0x4E, 0x1F, 0x0E, 0x80, 0x7F, 
            0x0F, 0x31, 0x90, 0x21, 0x00, 0x01, 0x10, 0x00, 0x10, 0x6F, 0x0F, 0x00, 0x01, 0x0E, 0xFF, 0x80, 
            0x00, 0x00, 0x00, 0x0E, 0x0E, 0x01, 0xBF, 0x0F, 0x10, 0x00, 0x9F, 0x7F, 0x1F, 0x0E, 0x6F, 0x0E, 
            0x7F, 0x7F, 0x0F, 0x3F, 0x7F, 0x7F, 0x00, 0x01, 0x00, 0x10, 0x10, 0x0E, 0x0E, 0x00, 0x0E, 0x21, 
            0x1F, 0x07, 0x8E, 0x80, 0x0E, 0x7F, 0xEF, 0x1F, 0x01, 0x9F, 0x80, 0x6F, 0x80, 0x11, 0xBF, 0x0F, 
            0x7F, 0x7F, 0x8E, 0x8E, 0x3F, 0x80, 0x0E, 0x8F, 0x6F, 0xFF, 0x7F, 0x90, 0x10, 0x10, 0x90, 0x7F, 
            0x7F, 0x01, 0xBF, 0x2F, 0x01, 0x01, 0x8F, 0x0E, 0x7F, 0x0E, 0x7F, 0x0F, 0x0E, 0x0E, 0x01, 0x7F, 
            0x01, 0x10, 0x00, 0x21, 0x00, 0x2F, 0x01, 0x06, 0x1F, 0x06, 0xEF, 0x1F, 0x7F, 0x7F, 0x0F, 0x0E, 
            0x6F, 0x00, 0x0F, 0x7F, 0x21, 0x00, 0x1E, 0x06, 0x7F, 0x2F, 0x10, 0x0E, 0x0E, 0x00, 0x8F, 0x00, 
            0x2F, 0x0E, 0x0F, 0x7F, 0x0F, 0x6F, 0x80, 0x0E, 0x21, 0x0E, 0xFF, 0x1F, 0x06, 0x6F, 0xEF, 0x6F, 
            0x7F, 0x10, 0x7F, 0x1F, 0x0E, 0x0F, 0xFF, 0x8E, 0x21, 0xAF, 0x1F, 0x01, 0x1E, 0x6F, 0x01, 0x0F, 
            0x0E, 0x2F, 0x8E, 0x21, 0x00, 0x0F, 0x00, 0x0F, 0x6F, 0x2F, 0x2F, 0x0F, 0x00, 0x0F, 0x0F, 0x6F, 
            0x2F, 0x0F, 0x0E, 0x7F, 0x01, 0x6F, 0x0E, 0x7F, 0x6F, 0x0E, 0x7F, 0x01, 0x1F, 0x7F, 0x8E, 0x80, 
            0x01, 0x01, 0x8F, 0x7F, 0x7F, 0x01, 0x81, 0x0E, 0x0F, 0x07, 0x6F, 0x7F, 0x7F, 0x01, 0xFF, 0x8F, 
            0x1F, 0xBF, 0x3F, 0x81, 0x1E, 0x7F, 0x10, 0x6F, 0x6F, 0x2F, 0xEF, 0x0E, 0x7F, 0x7F, 0x21, 0x01, 
            0x3F, 0x0E, 0x7F, 0x01, 0x1F, 0x0E, 0x2F, 0x21, 0x21, 0x01, 0x01, 0x7F, 0x7F, 0x0F, 0xFF, 0x7F, 
            0x0E, 0x5F, 0x1F, 0x10, 0x00, 0x06, 0x80, 0x00, 0x1F, 0x21, 0x7F, 0x7F, 0x7F, 0x10, 0x21, 0x1E, 
            0x10, 0x10, 0x0F, 0x3F, 0x1F, 0x7F, 0x7F, 0x7F, 0x3F, 0x7F, 0x0F, 0x90, 0x0E, 0x0E, 0x10, 0x0E, 
            0x6F, 0x00, 0x8E, 0x01, 0x1F, 0x4F, 0x81, 0x7F, 0x00, 0x1E, 0x80, 0x1F, 0x00, 0x00, 0x0F, 0x1F, 
            0x0E, 0x0F, 0x0E, 0x8E, 0x00, 0x8E, 0x0F, 0x8E, 0x3F, 0x6F, 0x3F, 0x0E, 0x0F, 0x21, 0x86, 0x3F, 
            0x01, 0x0E, 0xA1, 0x0F, 0x6F, 0x00, 0xEF, 0x01, 0x3F, 0x0E, 0x21, 0x2F, 0x0E, 0x0F, 0xFF, 0x7F, 
            0x10, 0x6F, 0x7F, 0x90, 0x1F, 0x0E, 0x61, 0x2F, 0x6F, 0x10, 0x90, 0x81, 0x0E, 0x1F, 0xFF, 0x17, 
            0x0F, 0x81, 0x90, 0x0F, 0x9F, 0x0E, 0x00, 0x01, 0x10, 0x01, 0x9F, 0xFF, 0x00, 0xFF, 0x5F, 0x9E, 
            0x0E, 0x1F, 0x7F, 0x0E, 0x0F, 0x00, 0x81, 0x10, 0x1F, 0x7F, 0x4F, 0x0F, 0x10, 0x7F, 0x0E, 0x7F, 
            0x0F, 0x1F, 0x06, 0x0F, 0x1F, 0x7F, 0x71, 0x0F, 0x1F, 0x5F, 0x11, 0x3F, 0x7F, 0x7F, 0x5F, 0x7F, 
            0x1F, 0x7F, 0x81, 0xFF, 0x1F, 0x06, 0x8E, 0x5F, 0x7F, 0xFF, 0x8F, 0x6F, 0x4F, 0x7F, 0x1F, 0x41, 
            0x0E, 0x10, 0x8E, 0x8E, 0x10, 0xA1, 0x81, 0x90, 0x90, 0x00, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 
            0x01, 0x7F, 0xFF, 0x3F, 0x7F, 0x5F, 0x8E, 0x7F, 0x10, 0x2F, 0x01, 0x7F, 0x3F, 0x00, 0xFF, 0xF7, 
            0x1F, 0xFF, 0x81, 0xBF, 0x17, 0x7F, 0x7F, 0x6F, 0x0F, 0x80, 0x8F, 0xBF, 0x0E, 0x07, 0x80, 0x1F, 
            0x06, 0x7F, 0x8E, 0x2F, 0x3F, 0x0E, 0x0F, 0x11, 0x00, 0x3F, 0xAF, 0xBF, 0x00, 0xC1, 0xCF, 0x80, 
            0x0F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x3F, 0x17, 0x1F, 0x7F, 0x01, 0x01, 0x1E, 0x0E, 0x31, 0x10, 
            0x00, 0x10, 0x00, 0x00, 0x10, 0x10, 0x2F, 0x7F, 0x21, 0x0E, 0x10, 0x1F, 0x3F, 0x01, 0x0F, 0x01, 
            0x0E, 0x01, 0x0F, 0x21, 0x0E, 0x0E, 0x0E, 0x1E, 0x7F, 0x0E, 0x0E, 0x7F, 0x1F, 0x1F, 0x01, 0x7F, 
            0x01, 0x7F, 0x2F, 0x1F, 0x7F, 0x7F, 0x7F, 0x0F, 0x7F, 0x7F, 0x1F, 0x0E, 0x1F, 0x01, 0x80, 0x8E, 
            0x7F, 0x0E, 0x8F, 0x1F, 0x0F, 0x2F, 0x90, 0x5F, 0x8E, 0xFF, 0x6F, 0x0E, 0x0E, 0x0F, 0x8F, 0xFF, 
            0x3F, 0xDF, 0xFF, 0x87, 0x1F, 0x01, 0x0F, 0x01, 0x21, 0x80, 0x90, 0x90, 0x10, 0x00, 0x90, 0x10, 
            0x10, 0x10, 0xA1, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x10, 0x90, 0x90, 0x90, 
            0x01, 0x00, 0x00, 0x6F, 0x0E, 0x2F, 0x8E, 0x7F, 0x1F, 0x00, 0x0E, 0x7F, 0x3F, 0x1F, 0x0E, 0x2F, 
            0x01, 0x0E, 0x01, 0x0E, 0x01, 0x31, 0x1F, 0x2F, 0x06, 0x7F, 0xEF, 0x0E, 0x0E, 0x0F, 0x7F, 0x0E, 
            0x0E, 0x00, 0x1F, 0x3F, 0x0F, 0x0F, 0x6F, 0x0E, 0x5F, 0x0F, 0x2F, 0x1F, 0x11, 0x0F, 0x7F, 0x21, 
            0x3F, 0x00, 0x6F, 0x0E, 0x0E, 0x0E, 0x0F, 0x0F, 0x3F, 0x5F, 0x7F, 0x7F, 0x01, 0x01, 0x86, 0x86, 
            0x1E, 0x31, 0x90, 0x10, 0x10, 0x21, 0x90, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x7F, 0x01, 0x0F, 
            0x1F, 0x7F, 0x0E, 0x86, 0x0E, 0x7F, 0x0E, 0x3F, 0x0E, 0x4F, 0xFF, 0x0E, 0x7F, 0x00, 0x2F, 0x1F, 
            0x01, 0x0F, 0x80, 0x61, 0x1F, 0x0E, 0x7F, 0x06, 0x3F, 0x0E, 0x00, 0x0E, 0x7F, 0x1F, 0x0E, 0x3F, 
            0x3F, 0x1E, 0x06, 0x0E, 0x0E, 0x1E, 0x80, 0x0E, 0x0F, 0x7F, 0x17, 0x67, 0x0E, 0x00, 0x0E, 0x6F, 
            0x7F, 0x7F, 0x7F, 0x0E, 0x6F, 0x0E, 0x2F, 0x21, 0x1F, 0x5F, 0x0E, 0x2F, 0x0F, 0x7F, 0x06, 0x0E, 
            0x00, 0x1F, 0x80, 0x7F, 0x7F, 0x00, 0x9F, 0x2F, 0x00, 0x0E, 0x4F, 0x3F, 0x7F, 0x7F, 0x7F, 0x1F, 
            0x0E, 0x17, 0x0E, 0x2F, 0x00, 0x01, 0x00, 0x00, 0x10, 0x01, 0x00, 0x06, 0x10, 0x10, 0x00, 0x00, 
            0x21, 0x00, 0x21, 0x0E, 0x10, 0x10, 0x10, 0x10, 0x10, 0x6F, 0x21, 0x0E, 0x6F, 0x0E, 0x00, 0x0F, 
            0x0E, 0x7F, 0x0E, 0x0E, 0x0E, 0x2F, 0x0E, 0x0E, 0x7F, 0x0F, 0x8E, 0x8E, 0x3F, 0x7F, 0x80, 0x0E, 
            0x0E, 0x00, 0x8E, 0x0E, 0x0E, 0x0F, 0x3F, 0x0F, 0x21, 0xEF, 0x81, 0x0E, 0x0E, 0x0F, 0x2F, 0x81, 
            0x06, 0x1F, 0x0F, 0x00, 0x06, 0x00, 0x0F, 0x7F, 0x01, 0x1F, 0x1F, 0x7F, 0x11, 0x0E, 0x0E, 0x0E, 
            0x0F, 0x7F, 0x21, 0x7F, 0x01, 0x0F, 0x6F, 0x0F, 0x06, 0x3F, 0x01, 0x7F, 0x7F, 0x3F, 0x2F, 0x6F, 
            0x0F, 0x31, 0x80, 0x1F, 0x0F, 0x7F, 0x1F, 0x2F, 0x1F, 0x2F, 0x7F, 0x00, 0x01, 0x0F, 0x06, 0x00, 
            0x00, 0x01, 0x3F, 0x00, 0x21, 0x01, 0x0E, 0x10, 0x10, 0x10, 0x00, 0x31, 0x10, 0x0E, 0x96, 0x0E, 
            0x6F, 0x06, 0x81, 0x01, 0x6F, 0x01, 0x80, 0x1E, 0x0E, 0x7F, 0x0E, 0x0E, 0x7F, 0x0E, 0x8F, 0xAF, 
            0x0E, 0x7F, 0x0E, 0x80, 0x0E, 0x00, 0x01, 0x0E, 0x0F, 0x0F, 0x0F, 0x0E, 0x0E, 0x21, 0x0E, 0x1F, 
            0x0E, 0x0F, 0x7F, 0x00, 0x3F, 0x00, 0x0E, 0x00, 0x7F, 0x3F, 0x1F, 0x6F, 0x6F, 0x11, 0x01, 0x7F, 
            0x4F, 0x3F, 0x4F, 0x00, 0x7F, 0x0E, 0xEF, 0x8F, 0x01, 0x7F, 0xBF, 0x6F, 0x0F, 0xFF, 0x8E, 0x0F, 
            0x3F, 0x7F, 0x0E, 0x1F, 0x06, 0x00, 0xBF, 0x86, 0x00, 0xA1, 0x61, 0x06, 0x21, 0x10, 0x00, 0x10, 
            0x10, 0x00, 0x80, 0x10, 0x00, 0x10, 0x90, 0x0F, 0x0E, 0x0E, 0x8F, 0x0E, 0x7F, 0x00, 0x6F, 0x0F, 
            0x1F, 0x0F, 0x8F, 0x8E, 0x6F, 0x00, 0x0E, 0x0F, 0x00, 0x0E, 0x0F, 0x0F, 0x7F, 0x0F, 0x8E, 0x3F, 
            0x2F, 0x00, 0x8F, 0x01, 0x00, 0x00, 0xAF, 0x0F, 0x0E, 0x0E, 0x0E, 0x0E, 0x1F, 0x0E, 0x8F, 0x8E, 
            0x01, 0xFF, 0x0E, 0xFF, 0x6F, 0x2F, 0x0E, 0x7F, 0x6F, 0x0E, 0xFF, 0x80, 0x0F, 0x0E, 0xCF, 0x3F, 
            0x0E, 0x7F, 0x81, 0x4F, 0x0F, 0x4F, 0x2F, 0x0F, 0x0E, 0x0E, 0x8F, 0x9F, 0x7F, 0xEF, 0x8E, 0xC7, 
            0x8E, 0x0E, 0x01, 0x01, 0x81, 0xA1, 0x81, 0x90, 0x10, 0x10, 0x90, 0x1E, 0x10, 0x7F, 0x8E, 0x41, 
            0x2F, 0x6F, 0x00, 0x7F, 0x0E, 0x0E, 0x8F, 0x8E, 0x3F, 0xFF, 0xBF, 0xA1, 0x00, 0x0E, 0x6F, 0x01, 
            0x01, 0x07, 0x81, 0xEF, 0x0E, 0x2F, 0x9F, 0x2F, 0x2F, 0x7F, 0x21, 0x0E, 0x01, 0x0F, 0x1F, 0x01, 
            0x0F, 0x0F, 0x0F, 0x6F, 0x1F, 0x8E, 0x0E, 0x8E, 0x7F, 0x6F, 0x6F, 0x2F, 0x0E, 0x6F, 0x0E, 0x01, 
            0x00, 0x0E, 0x10, 0x00, 0x10, 0x0E, 0x06, 0x2F, 0x0E, 0x17, 0x1F, 0x1F, 0x0E, 0x01, 0x3F, 0x01, 
            0x2F, 0x0F, 0x0E, 0x0E, 0x7F, 0x1F, 0x0F, 0x7F, 0x2F, 0x0F, 0xAF, 0x8F, 0x0F, 0x0E, 0x8E, 0x5F, 
            0x2F, 0x6F, 0x80, 0x01, 0x6F, 0x0E, 0x0E, 0x21, 0x07, 0x2F, 0x8E, 0xFF, 0x3F, 0x8F, 0x8E, 0x8F, 
            0x0F, 0x0E, 0x0F, 0x0E, 0x3F, 0x81, 0xEF, 0x8F, 0x7F, 0x6F, 0xFF, 0x21, 0x21, 0x01, 0xB1, 0x7F, 
            0x7F, 0x0E, 0x00, 0x0E, 0x0F, 0x0E, 0x8F, 0x9F, 0x0F, 0xBF, 0x6F, 0x80, 0x01, 0x0F, 0x0F, 0x0F, 
            0x0E, 0x71, 0x8F, 0x8E, 0x0E, 0x0E, 0x8E, 0x0F, 0x0E, 0x7F, 0x81, 0x0F, 0x00, 0x6F, 0x4F, 0x7F, 
            0x3F, 0x0E, 0x8E, 0x90, 0x00, 0xFF, 0x80, 0x8E, 0x0F, 0x0E, 0x0F, 0x00, 0x7F, 0x6F, 0x8E, 0x0F, 
            0x01, 0x0E, 0x86, 0x0E, 0x7F, 0x6F, 0x07, 0x00, 0x0F, 0x7F, 0x6F, 0x0E, 0x0E, 0x0F, 0x00, 0x8E, 
            0x0E, 0x8E, 0x0E, 0x01, 0x00, 0x00, 0x0F, 0x6F, 0x0E, 0x2F, 0xEF, 0xFF, 0x01, 0x0E, 0xAF, 0x5F, 
            0x0F, 0x0E, 0xE1, 0x00, 0x0F, 0x1F, 0x6F, 0x6F, 0x00, 0x0E, 0xEF, 0xAF, 0x0E, 0x81, 0x01, 0x80, 
            0x01, 0x1F, 0x01, 0x21, 0x00, 0x00, 0xF1, 0x6F, 0x0F, 0x0F, 0xEF, 0x0F, 0x0E, 0x00, 0x0F, 0x0E, 
            0x0E, 0x0F, 0x2F, 0x0E, 0x0E, 0x0E, 0x0E, 0x1F, 0x00, 0x1F, 0x0E, 0x80, 0x0E, 0x6F, 0x0E, 0x21, 
            0x21, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0F, 0x00, 0x1F, 0x0E, 0x0F, 0x0E, 0x1F, 0x01, 0x0E, 0x6F, 
            0x0F, 0x01, 0x01, 0x0E, 0x00, 0x00, 0x8F, 0x0E, 0x0E, 0x00, 0x00, 0x0E, 0x0E, 0x3F, 0xFF, 0x10, 
            0x7F, 0x10, 0x8F, 0x0F, 0x7F, 0x6F, 0x8F, 0x01, 0x01, 0x0F, 0x31, 0x4E, 0x6F, 0x01, 0xFF, 0xEF, 
            0x01, 0x8F, 0x1F, 0x8F, 0x0F, 0x0E, 0x00, 0x6F, 0x0E, 0x0F, 0x8E, 0x3F, 0x11, 0x0E, 0x9E, 0x0F, 
            0x0E, 0x3F, 0x80, 0x0E, 0x0E, 0x01, 0x01, 0x0F, 0x00, 0x7F, 0x1F, 0xBF, 0x16, 0x90, 0x10, 0x90, 
            0x10, 0x10, 0x10, 0x00, 0x00, 0x0E, 0x8E, 0x8E, 0x01, 0x00, 0xFF, 0x3F, 0x7F, 0x1F, 0x8E, 0x00, 
            0x0E, 0x0F, 0x00, 0x7F, 0x0E, 0x3F, 0x8F, 0x8E, 0x07, 0x8E, 0x0F, 0x80, 0x1F, 0x1E, 0x0E, 0x00, 
            0x00, 0x0F, 0x8E, 0x7F, 0x7F, 0x0E, 0x01, 0x1F, 0x1F, 0x7F, 0x3F, 0x21, 0x10, 0x01, 0x00, 0x3F, 
            0x3F, 0x1F, 0x00, 0x0F, 0x1F, 0x1F, 0xBF, 0x0F, 0x0F, 0x0F, 0x1E, 0x0E, 0x0E, 0x0E, 0x8E, 0x0F, 
            0x1F, 0x0F, 0x0E, 0x0F, 0x3F, 0x0F, 0x3F, 0x7F, 0x0E, 0x3F, 0x10, 0x7F, 0x0E, 0x21, 0x31, 0x10, 
            0x1E, 0x1F, 0x01, 0x0E, 0x7F, 0x01, 0x0E, 0x0F, 0x1E, 0x7F, 0x8F, 0xFF, 0x0F, 0x0E, 0xB7, 0x7F, 
            0x01, 0x00, 0x8E, 0x0F, 0x7F, 0x01, 0x0E, 0x01, 0x0E, 0x3F, 0x8E, 0x0E, 0x2F, 0xFF, 0x31, 0xA1, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x0E, 0x8E, 0x8E, 0x3F, 0x3F, 0x8F, 0x2F, 0x01, 0x01, 0xFF, 0x00, 
            0x7F, 0x0E, 0x7F, 0x7F, 0x0F, 0x0F, 0xAF, 0x9F, 0x7F, 0xBF, 0x3F, 0x8F, 0x0E, 0x3F, 0x0F, 0x01, 
            0x01, 0x2F, 0xBF, 0x9F, 0x01, 0x10, 0x80, 0x10, 0x7F, 0x0F, 0xFF, 0x0E, 0x7F, 0x00, 0x1F, 0x01, 
            0x0E, 0x1F, 0x9E, 0x9F, 0x0E, 0xBF, 0x0E, 0x8E, 0x8F, 0x3F, 0x0E, 0x80, 0x8E, 0x0E, 0x1F, 0x01, 
            0x0E, 0x1F, 0x0F, 0x0E, 0x3F, 0x0E, 0x00, 0x3F, 0x0E, 0x01, 0x0F, 0x3F, 0x0F, 0x01, 0x1F, 0x3F, 
            0x0F, 0x0E, 0x0E, 0x0F, 0x00, 0x01, 0x0E, 0x10, 0x10, 0x31, 0x80, 0xA1, 0x3F, 0x0F, 0xFF, 0x0E, 
            0x0E, 0x0E, 0x8E, 0x3F, 0x0E, 0x6F, 0x0E, 0x7F, 0x0F, 0x3F, 0x8E, 0x8E, 0x0E, 0x8F, 0x0E, 0xEF, 
            0x3F, 0x0F, 0x0E, 0x3F, 0x1F, 0x1F, 0xEF, 0x86, 0x6F, 0x06, 0xAF, 0x0E, 0x8E, 0x8E, 0xBF, 0x8F, 
            0x90, 0x01, 0x01, 0x0E, 0x7F, 0x06, 0xFF, 0x8E, 0x6F, 0x8E, 0x0E, 0x9F, 0x90, 0x10, 0x10, 0xA1, 
            0x00, 0x8E, 0xFF, 0x9F, 0x0F, 0x1F, 0x81, 0x0F, 0x1F, 0x0F, 0x80, 0x11, 0x01, 0x7F, 0x00, 0x0E, 
            0x0F, 0x0E, 0x8E, 0x9F, 0x0E, 0x81, 0x8E, 0x9F, 0x8F, 0x01, 0x0E, 0x0E, 0x0E, 0x0E, 0x8F, 0xFF, 
            0x01, 0x00, 0x6F, 0x0F, 0x06, 0x00, 0x80, 0x01, 0x10, 0x00, 0x0E, 0x3F, 0x1F, 0x1F, 0x0F, 0x3F, 
            0x0E, 0x01, 0x06, 0xFF, 0x0E, 0x1F, 0x0E, 0x0E, 0x0E, 0x0E, 0x7F, 0x2F, 0x0E, 0x3F, 0x0F, 0x0F, 
            0x3F, 0x01, 0x0E, 0x0E, 0x6F, 0x0E, 0x3F, 0x21, 0x0F, 0x0F, 0x1F, 0x3F, 0x0E, 0x00, 0x01, 0x00, 
            0x21, 0x1E, 0x3F, 0x1E, 0x0E, 0x01, 0x81, 0x8E, 0x0E, 0x0F, 0x81, 0x16, 0x0F, 0x06, 0x8E, 0x0F, 
            0xFF, 0x0E, 0x0F, 0x0F, 0x0F, 0x1E, 0x8E, 0x8E, 0x0E, 0x80, 0x1F, 0x8F, 0xCE, 0x0F, 0x10, 0x6F, 
            0x00, 0x81, 0xEF, 0x8F, 0x0E, 0x00, 0xAF, 0x2F, 0x01, 0x7F, 0x8F, 0x6F, 0x0E, 0x00, 0x7F, 0x0F, 
            0x0E, 0x0F, 0x86, 0x9F, 0x00, 0xEF, 0x8F, 0xA1, 0x0E, 0x0F, 0x17, 0x0F, 0x0E, 0x2F, 0xA1, 0xA1, 
            0x0F, 0x2F, 0x81, 0x1F, 0x0F, 0x3F, 0xBF, 0x0E, 0x01, 0x0F, 0x01, 0x1F, 0x0F, 0x0E, 0x01, 0x4F, 
            0x0E, 0xBF, 0xFF, 0xEF, 0x0E, 0x8E, 0x0E, 0x21, 0x0F, 0x0E, 0x8E, 0x0E, 0x0E, 0x01, 0x0F, 0x0E, 
            0x0F, 0x6F, 0x1F, 0x0E, 0x00, 0x0F, 0x6F, 0x0E, 0x0E, 0x3F, 0x01, 0x6F, 0x1F, 0x0E, 0x0E, 0x0E, 
            0x7F, 0x0F, 0x21, 0x0F, 0x3F, 0x11, 0xE1, 0x80, 0x0F, 0x81, 0x8F, 0x0F, 0x7F, 0x7F, 0x8F, 0x00, 
            0x80, 0x01, 0x6F, 0x6F, 0x10, 0x0F, 0x8F, 0xFF, 0x01, 0xFF, 0x0F, 0xEF, 0x01, 0x6F, 0x21, 0x01, 
            0x01, 0x7F, 0xFF, 0x7F, 0x11, 0x10, 0x80, 0x0E, 0x7F, 0x1F, 0x0E, 0x7F, 0x10, 0x0E, 0x10, 0x0E, 
            0x7F, 0x0F, 0x7F, 0x00, 0x0E, 0x2F, 0x8E, 0xFF, 0x00, 0x0F, 0x00, 0x01, 0x06, 0x3F, 0x7F, 0x7F, 
            0x0F, 0x7F, 0x7F, 0x0E, 0x10, 0x10, 0x00, 0x00, 0x7F, 0x0E, 0x00, 0x7F, 0x0E, 0x7F, 0x06, 0x00, 
            0x00, 0x0E, 0x7F, 0x06, 0x00, 0x0E, 0x7F, 0x6F, 0x0F, 0x2F, 0x8F, 0x0F, 0x11, 0x00, 0xEF, 0x0F, 
            0x01, 0x0E, 0xBF, 0x00, 0x3F, 0x00, 0x00, 0x0E, 0x0F, 0x00, 0x0F, 0xFF, 0x0E, 0x90, 0x21, 0x80, 
            0x21, 0x3F, 0x0E, 0x00, 0x61, 0x21, 0x8E, 0x3F, 0x0E, 0x0E, 0x8F, 0x00, 0x1E, 0x0F, 0x8F, 0x21, 
            0x0F, 0x7F, 0x7F, 0x21, 0x21, 0x31, 0x10, 0x80, 0x00, 0x96, 0x0F, 0x0F, 0x0F, 0x0F, 0x1F, 0x65, 
            0x06, 0x01, 0x9F, 0x81, 0xFF, 0x7F, 0x8E, 0x0E, 0x3F, 0xFF, 0x8E, 0x8F, 0xEF, 0x6F, 0x0F, 0x7F, 
            0x00, 0x6F, 0x90, 0x90, 0x10, 0x90, 0x00, 0x8F, 0x7F, 0x00, 0x00, 0x86, 0x06, 0x80, 0x8E, 0x0F, 
            0x1F, 0x1F, 0xAF, 0x7F, 0x7F, 0x61, 0x8E, 0x0F, 0x7F, 0x3F, 0x0E, 0x0E, 0x3F, 0x6F, 0x3F, 0x80, 
            0x00, 0x8F, 0x3F, 0x00, 0x21, 0x0E, 0x0F, 0x6F, 0x00, 0x0E, 0xBF, 0x0E, 0x2F, 0x01, 0x0F, 0x01, 
            0x7F, 0x0E, 0x7F, 0x7F, 0x0E, 0x0F, 0x0F, 0x0E, 0x4F, 0x5F, 0x1F, 0x10, 0x21, 0x21, 0x2F, 0x0E, 
            0x01, 0x1E, 0x0E, 0x0E, 0x00, 0x7F, 0x8F, 0x0F, 0x01, 0x0E, 0xFF, 0x3F, 0x06, 0x0E, 0x8E, 0x06, 
            0x1F, 0x0F, 0x0F, 0x0E, 0x10, 0x0F, 0x7F, 0x0E, 0x6F, 0x01, 0x01, 0x6F, 0x0E, 0x0E, 0x0F, 0x00, 
            0x01, 0x7F, 0x8F, 0x21, 0x7F, 0x01, 0x8F, 0x00, 0x00, 0x07, 0x80, 0x0E, 0x2F, 0x21, 0x00, 0x0F, 
            0x0E, 0x06, 0x8F, 0x0E, 0x00, 0x80, 0x2F, 0x0F, 0x0E, 0x06, 0x0F, 0x0F, 0x0F, 0x0E, 0xBF, 0x7F, 
            0x2F, 0x0F, 0x8F, 0x0F, 0x6F, 0x00, 0xBF, 0x00, 0x00, 0x0E, 0x6F, 0x0F, 0x0E, 0x2F, 0x8F, 0x2F, 
            0x01, 0x8E, 0x6F, 0x0F, 0x2F, 0x01, 0x0E, 0x21, 0x00, 0x0E, 0x0F, 0x17, 0x0F, 0x01, 0x2F, 0x7F, 
            0x00, 0x7F, 0x0E, 0x7F, 0x00, 0x0E, 0x7F, 0x61, 0x2F, 0x01, 0x21, 0x6F, 0x00, 0x01, 0x00, 0x6F, 
            0x01, 0x00, 0x00, 0x6F, 0x01, 0x0F, 0xEF, 0xA7, 0x2F, 0x00, 0xA1, 0x0F, 0x0F, 0x8E, 0x86, 0x0F, 
            0x81, 0x6F, 0x0E, 0x0E, 0x0E, 0x01, 0xAF, 0xA1, 0x00, 0xE1, 0xEF, 0x80, 0x2F, 0x8E, 0x0E, 0xAF, 
            0x01, 0x00, 0xEF, 0xA1, 0x81, 0x81, 0x8F, 0x0F, 0x00, 0x86, 0xEF, 0x80, 0x81, 0xEF, 0x06, 0x2F, 
            0x6F, 0x10, 0x90, 0x80, 0x10, 0x90, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 0x00, 0x90, 0x90, 
            0x10, 0x10, 0xFF, 0x0E, 0x0F, 0x0F, 0x9F, 0x01, 0x01, 0x0E, 0x3F, 0x0E, 0x3F, 0x2F, 0x90, 0x8E, 
            0x0E, 0x8E, 0x8E, 0xFF, 0x0E, 0x1F, 0x3F, 0xA1, 0x0E, 0x0F, 0x80, 0x9F, 0x0E, 0x0F, 0x80, 0x10, 
            0x0E, 0x0F, 0x0E, 0x1F, 0x6F, 0x00, 0x0F, 0x0E, 0x6F, 0x01, 0x0E, 0x8E, 0x0E, 0x01, 0x0E, 0x0F, 
            0x7F, 0x00, 0x6F, 0x6F, 0x0F, 0x01, 0x00, 0x1F, 0x6F, 0x00, 0x6F, 0x00, 0x6F, 0x00, 0x6F, 0x0F, 
            0x2F, 0x0F, 0x2F, 0x6F, 0x01, 0x0F, 0x6F, 0x6F, 0x6F, 0x00, 0x00, 0x6F, 0x00, 0x6F, 0x0F, 0x6F, 
            0x01, 0x00, 0x8F, 0xEF, 0x01, 0x6F, 0x81, 0x0E, 0x00, 0x80, 0xEF, 0x01, 0x8F, 0x6F, 0x00, 0x01, 
            0x0E, 0x6F, 0x80, 0xA1, 0x6F, 0x81, 0x6F, 0xA1, 0x80, 0x0E, 0x6F, 0x6F, 0x2F, 0x01, 0x9F, 0x8E, 
            0x0F, 0x2F, 0x80, 0x01, 0x0E, 0x8F, 0xAF, 0x01, 0x3F, 0x01, 0x0E, 0x00, 0x06, 0x0F, 0x8F, 0x8E, 
            0x6F, 0xAF, 0xAF, 0x81, 0x6F, 0x6F, 0x0F, 0x0F, 0x0F, 0x0E, 0x80, 0xEF, 0x01, 0x01, 0xEF, 0x00, 
            0x6F, 0x0F, 0xEF, 0x6F, 0x00, 0x01, 0x6F, 0x01, 0x0E, 0x6F, 0x00, 0xAF, 0x2F, 0xEF, 0xEF, 0x80, 
            0x6F, 0x6F, 0x6F, 0xEF, 0x0E, 0x4F, 0x8F, 0x0F, 0x00, 0x7F, 0x87, 0x0F, 0x2F, 0x00, 0x01, 0x0F, 
            0x01, 0x06, 0x2F, 0x0F, 0x2F, 0x6F, 0x0F, 0x6F, 0x00, 0x31, 0x10, 0x0F, 0x6F, 0x6F, 0x00, 0x0F, 
            0x01, 0x2F, 0x8F, 0xFF, 0x80, 0x6F, 0x8F, 0x01, 0x6F, 0x0F, 0xAF, 0x00, 0x00, 0x00, 0x6F, 0x01, 
            0x6F, 0x6F, 0x80, 0xEF, 0x6F, 0xEF, 0x6F, 0xA7, 0x6F, 0x01, 0x4E, 0x0E, 0x21, 0x21, 0x81, 0x01, 
            0x6F, 0x00, 0xEF, 0x01, 0x00, 0x00, 0x0F, 0x01, 0x0E, 0x6F, 0x0E, 0x0F, 0x6F, 0x0E, 0x6F, 0x6F, 
            0x0E, 0x00, 0xEF, 0x0F, 0x6F, 0x0E, 0x0F, 0x6F, 0x0F, 0x0F, 0x81, 0x6F, 0x21, 0x6F, 0x00, 0x6F, 
            0x00, 0x0F, 0x6F, 0x0F, 0x0F, 0x0F, 0x6F, 0x0F, 0x0E, 0x06, 0x21, 0x6F, 0x6F, 0x0F, 0xAF, 0x2F, 
            0x2F, 0x6F, 0x2F, 0x00, 0x2F, 0x00, 0xEF, 0xEF, 0x06, 0x00, 0xC0, 0x6F, 0x01, 0x6F, 0xEF, 0x06, 
            0x0F, 0x6F, 0x0E, 0x2F, 0x0F, 0x0F, 0xEF, 0xEF, 0x6F, 0x00, 0x6F, 0x0E, 0x0F, 0x00, 0x6F, 0x0F, 
            0x6F, 0x6F, 0xEF, 0x00, 0x6F, 0x0F, 0x8F, 0x3F, 0x0E, 0x00, 0x0F, 0x0F, 0x21, 0x0E, 0x6F, 0x0E, 
            0x6F, 0x0F, 0x0E, 0x0E, 0x6F, 0x0E, 0x6F, 0x6F, 0x0E, 0x6F, 0x6F, 0x6F, 0x0F, 0x6F, 0x8F, 0x81, 
            0x6F, 0x21, 0x8E, 0x0E, 0x0E, 0x8E, 0xBF, 0x8E, 0xEF, 0x00, 0x0E, 0x6F, 0x6F, 0x01, 0x86, 0x8E, 
            0x0F, 0x8E, 0x00, 0xEF, 0x00, 0x0E, 0x06, 0x8F, 0x80, 0x6F, 0x86, 0x06, 0x0F, 0x0E, 0x2F, 0x00, 
            0x0E, 0x6F, 0x87, 0x01, 0x01, 0x0F, 0x0F, 0x0E, 0x0F, 0x6F, 0x0E, 0x0E, 0x21, 0x01, 0xEF, 0x6F, 
            0x0E, 0x0E, 0x0F, 0x2F, 0x0F, 0x0E, 0xAF, 0x0E, 0x6F, 0x0F, 0x6F, 0x21, 0x6F, 0x0E, 0x01, 0x2F, 
            0x0E, 0x00, 0x00, 0x0F, 0x00, 0x0E, 0x7F, 0x0E, 0x0E, 0x00, 0x0E, 0x2F, 0x7F, 0x07, 0x00, 0x6F, 
            0x6F, 0x21, 0xA1, 0x80, 0x6F, 0x00, 0x81, 0x6F, 0x0F, 0x0F, 0x8E, 0x0E, 0x01, 0x6F, 0x0F, 0x2F, 
            0x6F, 0x01, 0x0E, 0x21, 0x0F, 0x0F, 0x0E, 0x80, 0x06, 0x00, 0x6F, 0x0F, 0x2F, 0x01, 0xAF, 0x01, 
            0x61, 0x00, 0xEF, 0x6F, 0x0F, 0x0F, 0xEF, 0x00, 0x0E, 0x2F, 0x6F, 0x00, 0x0F, 0x00, 0x8F, 0x0F, 
            0x10, 0x90, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x00, 0x10, 0x90, 0x10, 
            0x10, 0x00, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x80, 0x10, 0x90, 0x10, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x90, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x80, 0x10, 0x90, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x90, 0x10, 0x90, 0x10, 0x90, 0x90, 0x90, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 0x90, 0x00, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 
            0x10, 0x90, 0x90, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x90, 0x10, 0x10, 0x80, 0x10, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x10, 0x90, 0x90, 0x90, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0xFF, 0x01, 0x0F, 0x21, 0x0E, 0x0E, 0x0F, 0x0F, 0x6F, 
            0x00, 0x7F, 0x06, 0x0E, 0x00, 0x0F, 0xFF, 0x1F, 0x6F, 0x1F, 0x6F, 0x0F, 0x7F, 0x0E, 0x2F, 0x0F, 
            0x2F, 0x01, 0x00, 0x01, 0x0F, 0x7F, 0x06, 0x0F, 0x01, 0x01, 0x7F, 0x01, 0x0E, 0x0F, 0x00, 0x0E, 
            0x00, 0x7F, 0x3F, 0x01, 0x0F, 0x0E, 0x0F, 0x00, 0x0F, 0x07, 0x7F, 0x7F, 0x2F, 0x0F, 0x80, 0x8E, 
            0x0E, 0x0E, 0x8E, 0x1F, 0x00, 0x8F, 0x8F, 0x0E, 0x3F, 0x7F, 0x7F, 0x0F, 0x21, 0x0E, 0x81, 0x00, 
            0x0E, 0x8E, 0xFF, 0x80, 0x0E, 0x3F, 0x0F, 0x01, 0x00, 0x3F, 0xFF, 0x01, 0x7F, 0x2F, 0x21, 0x0E, 
            0x07, 0x0E, 0x0F, 0x0F, 0x3F, 0x0E, 0x00, 0x0E, 0x3F, 0x0F, 0x0E, 0x01, 0x0F, 0x6F, 0x6F, 0x80, 
            0x6F, 0x6F, 0x6F, 0x0E, 0x0F, 0x06, 0x80, 0xEF, 0x6F, 0x6F, 0x6F, 0x6F, 0x6F, 0x21, 0x21, 0x6F, 
            0x6F, 0x00, 0x2F, 0x6F, 0x6F, 0x0E, 0xEF, 0xEF, 0x6F, 0x8F, 0x8F, 0xEF, 0x6F, 0x6F, 0x0E, 0x6F, 
            0x6F, 0x6F, 0x6F, 0x6F, 0x6F, 0x0F, 0x01, 0x6F, 0x6F, 0x01, 0x6F, 0x01, 0x0E, 0x21, 0x21, 0x07, 
            0x00, 0x6F, 0x0E, 0x6F, 0x00, 0x0F, 0x01, 0x0E, 0x00, 0x0F, 0x6F, 0x21, 0x6F, 0x0F, 0xEF, 0x8E, 
            0x6F, 0x0F, 0xEF, 0x6F, 0x6F, 0x0E, 0xEF, 0x0E, 0x0F, 0x00, 0x6F, 0x00, 0x0F, 0x6F, 0x8E, 0x81, 
            0x2F, 0x06, 0x01, 0x8E, 0x6F, 0x0F, 0x00, 0x01, 0x0F, 0x0E, 0x2F, 0x6F, 0x6F, 0x6F, 0x0F, 0x01, 
            0x06, 0x01, 0x06, 0x01, 0x6F, 0x6F, 0x0E, 0x6F, 0x6F, 0x0F, 0x6F, 0x21, 0x00, 0x6F, 0x00, 0x2F, 
            0x2F, 0x00, 0x21, 0x6F, 0x2F, 0x0E, 0x6F, 0x0F, 0x00, 0x0E, 0x00, 0x0F, 0x00, 0x10, 0x10, 0x10, 
            0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x10, 
            0x10, 0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 
            0x10, 0x10, 0x10, 0x90, 0x7F, 0x01, 0x7F, 0x3F, 0x0E, 0x0F, 0x80, 0xFF, 0x0F, 0x3F, 0x8F, 0x06, 
            0x7F, 0x21, 0x8E, 0x7F, 0x7F, 0x10, 0x3F, 0x7F, 0x0E, 0x06, 0x80, 0x0E, 0x00, 0x86, 0x00, 0x81, 
            0x0E, 0x3F, 0x01, 0x00, 0x1F, 0x7F, 0x00, 0x00, 0x01, 0x06, 0x7F, 0x10, 0x0E, 0x0F, 0x0E, 0x0E, 
            0x07, 0x1F, 0x0E, 0x0E, 0x16, 0x0F, 0x0E, 0x00, 0x00, 0x6F, 0x00, 0x0E, 0x0E, 0x00, 0x01, 0x1F, 
            0x0F, 0x0E, 0x00, 0x7F, 0x00, 0x0E, 0x01, 0x0E, 0x6F, 0x00, 0x0E, 0x0F, 0x0F, 0x0E, 0x0E, 0x0F, 
            0x00, 0x10, 0x0F, 0x7F, 0x1E, 0x1F, 0x06, 0x0F, 0x10, 0x0F, 0x7F, 0x01, 0x1F, 0x00, 0x7F, 0x1E, 
            0x01, 0x0E, 0x2F, 0x10, 0x10, 0x0F, 0x5F, 0x0F, 0x0F, 0x00, 0x01, 0x7F, 0x7F, 0x0E, 0x1F, 0x7F, 
            0x0E, 0x1F, 0x00, 0x0E, 0x01, 0x0F, 0x3F, 0x1F, 0x0E, 0x0F, 0xBF, 0x7F, 0x0E, 0x7F, 0x10, 0x3F, 
            0x0E, 0x6F, 0x0E, 0x7F, 0x1F, 0x01, 0x0E, 0x0F, 0x7F, 0x0E, 0x3F, 0x3F, 0x1F, 0x00, 0x0E, 0xFF, 
            0x0F, 0x0F, 0x0F, 0x7F, 0x0F, 0x31, 0x80, 0x10, 0x10, 0x10, 0x9E, 0x10, 0x2F, 0x1F, 0x8E, 0x2F, 
            0x0E, 0x01, 0x0E, 0x0E, 0x07, 0x1F, 0x8F, 0x8E, 0x21, 0x3F, 0x7F, 0x01, 0x10, 0x0E, 0x00, 0x7F, 
            0x6F, 0x0F, 0x0E, 0x0E, 0x1F, 0x0E, 0x0F, 0x06, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x3F, 0x7F, 0x3F, 
            0x0F, 0x0F, 0x1F, 0x1F, 0x0E, 0x0E, 0x0F, 0x0F, 0x00, 0x1E, 0x71, 0x0E, 0x10, 0x00, 0x9E, 0x90, 
            0x6F, 0x1F, 0x80, 0x0F, 0x2F, 0x7F, 0x8F, 0x01, 0x0F, 0x1F, 0x10, 0x00, 0x0E, 0x1F, 0x8E, 0x86, 
            0x1F, 0x9F, 0x7F, 0x80, 0x7F, 0x0E, 0x00, 0x7F, 0x7F, 0x7F, 0xFF, 0x9F, 0x2F, 0x8F, 0x8F, 0x01, 
            0x2F, 0x90, 0x90, 0x8E, 0x8F, 0x2F, 0x0E, 0x0E, 0x00, 0x7F, 0x8E, 0x8F, 0x0F, 0x8F, 0x8F, 0xEF, 
            0x6F, 0x0F, 0x0F, 0xAF, 0x2F, 0x0E, 0x9E, 0x8E, 0x00, 0x1F, 0x8E, 0x00, 0x1F, 0x1F, 0x80, 0x1F, 
            0x11, 0x01, 0x3F, 0x0E, 0x7F, 0x00, 0xFF, 0x80, 0x0E, 0x81, 0x8F, 0x8F, 0x3F, 0x00, 0x00, 0x10, 
            0x00, 0x7F, 0x8F, 0x81, 0x2F, 0x3F, 0x8E, 0x7F, 0x0E, 0x0F, 0xEF, 0x2F, 0x06, 0x6F, 0x0E, 0x10, 
            0x1F, 0x2F, 0x8F, 0xAF, 0x3F, 0x8F, 0x0E, 0x8E, 0x0E, 0x21, 0x0F, 0x01, 0x10, 0x0F, 0x9F, 0x0E, 
            0x0E, 0x2F, 0x81, 0x0E, 0x0E, 0x0E, 0x8E, 0x01, 0x0E, 0x0F, 0x6F, 0x7F, 0x7F, 0x01, 0x81, 0x31, 
            0x21, 0x21, 0x1E, 0x81, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x9F, 0x81, 0x80, 0x8E, 0xA1, 0x00, 
            0x4F, 0x81, 0x8E, 0x01, 0x81, 0xA1, 0x21, 0x01, 0x21, 0x0E, 0xEF, 0xEF, 0x6F, 0xEF, 0x90, 0xAF, 
            0x00, 0x0F, 0x6F, 0x0E, 0x8F, 0x00, 0x80, 0x86, 0x0E, 0x00, 0x80, 0x2F, 0x0F, 0x0E, 0x8E, 0x6F, 
            0x0E, 0x06, 0x21, 0x21, 0x00, 0x0E, 0x8E, 0x8E, 0x0F, 0x8E, 0x80, 0xEF, 0x0E, 0x01, 0x0F, 0x0E, 
            0x01, 0x0F, 0x80, 0x80, 0x0E, 0x0E, 0x8E, 0x0F, 0x6F, 0x0F, 0x8F, 0x6F, 0x2F, 0x21, 0x6F, 0x0E, 
            0x0F, 0x0E, 0x8F, 0xEF, 0x00, 0xEF, 0x8E, 0x8F, 0x0F, 0x6F, 0x2F, 0x0E, 0x21, 0x0E, 0x86, 0x01, 
            0x0F, 0x0E, 0xEF, 0x6F, 0x6F, 0x6F, 0x8F, 0x0F, 0x0F, 0x21, 0x0E, 0x6F, 0x6F, 0x6F, 0x4F, 0x6F, 
            0x0F, 0x01, 0x00, 0x8F, 0x0E, 0x0F, 0x6F, 0x41, 0x01, 0x6F, 0xEF, 0x80, 0x8F, 0x01, 0x8F, 0x6F, 
            0x00, 0x00, 0x8E, 0x6F, 0x8F, 0x0F, 0x0F, 0x0F, 0x6F, 0x6F, 0x81, 0xEF, 0x00, 0x80, 0x6F, 0xEF, 
            0x0F, 0x01, 0x2F, 0x8F, 0x6F, 0x6F, 0xAF, 0x8F, 0x0E, 0x6F, 0x80, 0x06, 0x21, 0x6F, 0xEF, 0x0F, 
            0x0E, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0E, 0x2F, 0x0F, 0x0E, 0x0E, 0x8E, 0x21, 0x4F, 0x0F, 0x6F, 
            0x0E, 0x2F, 0xEF, 0x2F, 0x2F, 0x01, 0x90, 0x10, 0x10, 0x10, 0x80, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x90, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x10, 
            0x00, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x90, 0x90, 
            0x10, 0x90, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0xFF, 0x7F, 0x01, 0xB1, 0x3F, 
            0x01, 0x01, 0x80, 0x7F, 0x01, 0x01, 0x4F, 0x21, 0x7F, 0x10, 0x80, 0x90, 0x00, 0xEF, 0x2F, 0xEF, 
            0x7F, 0x7F, 0x6F, 0x00, 0x0E, 0x0B, 0x91, 0xA1, 0x01, 0x10, 0xA1, 0x21, 0x21, 0x90, 0x91, 0x2F, 
            0x01, 0x11, 0x7F, 0x00, 0x3F, 0x7F, 0x8F, 0x90, 0x10, 0x8E, 0x01, 0x8F, 0x00, 0x8F, 0x7F, 0x80, 
            0x91, 0x7F, 0x8F, 0x1F, 0x7F, 0x1F, 0x0E, 0x06, 0x10, 0x10, 0x21, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x0F, 0x0E, 0x3F, 0x0F, 0x1F, 0x3F, 0xFF, 0x00, 0x1F, 0x31, 0x7F, 0x7F, 0x00, 0x7F, 0x9F, 0xA1, 
            0x7F, 0x01, 0xC0, 0x10, 0x2F, 0x0F, 0x87, 0x7F, 0x3F, 0x71, 0x2F, 0x0F, 0x0E, 0x7F, 0x8F, 0x0E, 
            0x7F, 0x7F, 0x5F, 0xFF, 0x1F, 0x7F, 0x7F, 0x00, 0x41, 0x10, 0x90, 0xFF, 0x0F, 0x7F, 0xA1, 0x7F, 
            0x7F, 0x7F, 0x90, 0x21, 0x7F, 0x6F, 0x3F, 0x7F, 0x01, 0x6F, 0xFF, 0x9F, 0x0E, 0xFF, 0x7F, 0xFF, 
            0x7F, 0x7F, 0x7F, 0x6F, 0x0E, 0x00, 0x90, 0xA1, 0x01, 0x01, 0x01, 0x00, 0x01, 0x1F, 0xFF, 0x1F, 
            0x00, 0x6F, 0x6F, 0x01, 0x0F, 0x7F, 0xBF, 0x01, 0x7F, 0xA1, 0x00, 0x90, 0x7F, 0x0E, 0x7F, 0x0F, 
            0x00, 0x7F, 0xFF, 0x81, 0x1F, 0x21, 0x80, 0x7F, 0x00, 0x27, 0xEF, 0xEF, 0x01, 0x7F, 0x6F, 0x3F, 
            0x7F, 0x7F, 0xBF, 0xFF, 0x6F, 0xEF, 0x21, 0x90, 0x3F, 0x6F, 0x00, 0x1F, 0x6F, 0x4F, 0x6F, 0x00, 
            0x6F, 0x61, 0x1F, 0x7F, 0x00, 0x31, 0x01, 0x27, 0x3F, 0x6F, 0x00, 0x0E, 0x01, 0x7F, 0x7F, 0x6F, 
            0x0F, 0x00, 0x6F, 0x0E, 0x1F, 0x7F, 0x2F, 0x6F, 0x6F, 0x0F, 0xEF, 0x8E, 0x6F, 0x7F, 0x8E, 0x7F, 
            0x7F, 0xEF, 0xFF, 0x7F, 0x6F, 0x0F, 0x0E, 0x6F, 0x5F, 0xA1, 0xEF, 0x9F, 0x01, 0x8E, 0x01, 0xEF, 
            0x0F, 0x7F, 0x00, 0x10, 0x8E, 0x5F, 0x80, 0x9F, 0x0F, 0xDF, 0x80, 0x1F, 0x00, 0x10, 0x8E, 0x0E, 
            0x0F, 0x1F, 0x5F, 0x7F, 0x00, 0x0F, 0xFF, 0x8E, 0x61, 0x00, 0xFF, 0x80, 0x10, 0x00, 0x10, 0x7F, 
            0x5F, 0xFF, 0x8F, 0x9F, 0x1F, 0x7F, 0x9F, 0x00, 0x7F, 0x10, 0x90, 0x10, 0x00, 0x0F, 0x1F, 0x0E, 
            0x00, 0x7F, 0x80, 0x8F, 0x11, 0x1F, 0x80, 0x9F, 0x0F, 0x00, 0x7F, 0x06, 0x01, 0x00, 0x3F, 0x10, 
            0x10, 0x10, 0x01, 0x10, 0x0E, 0x0F, 0x0E, 0x1F, 0x0F, 0x0F, 0x0E, 0x3F, 0x1F, 0x1F, 0x61, 0x0F, 
            0x0E, 0x7F, 0x3F, 0x8E, 0x2F, 0x0F, 0x10, 0x10, 0x7F, 0x06, 0x0E, 0x1F, 0x06, 0x7F, 0x00, 0x1F, 
            0x0F, 0x0E, 0x0E, 0x0E, 0x1F, 0x6F, 0x01, 0x21, 0x10, 0x0E, 0x0E, 0x0E, 0x0E, 0x7F, 0x1F, 0x0F, 
            0x0F, 0x0E, 0x7F, 0x0E, 0x1F, 0x0F, 0x8F, 0x8E, 0x0F, 0x0E, 0x80, 0x0E, 0x00, 0x0F, 0x8E, 0x0E, 
            0x8F, 0x0E, 0x6F, 0x00, 0x0F, 0x61, 0x8F, 0x8F, 0x0F, 0x7F, 0x8F, 0x8E, 0x0F, 0x0E, 0x1F, 0x0E, 
            0x0F, 0x0E, 0x9F, 0x1F, 0x0E, 0x01, 0x8E, 0x6F, 0x0E, 0x0F, 0x8E, 0x0E, 0x0E, 0x6F, 0x0F, 0x1F, 
            0x2F, 0x1F, 0x6F, 0x0E, 0x0F, 0x0E, 0x0F, 0x01, 0x0E, 0x1F, 0x0F, 0x0E, 0x01, 0x01, 0x0E, 0x0E, 
            0x0E, 0x0E, 0x00, 0x1F, 0x0E, 0x0E, 0x1E, 0x0F, 0x0F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x1F, 
            0x1E, 0x00, 0x7F, 0x0F, 0x21, 0x06, 0x2F, 0x1F, 0x2F, 0x01, 0x1F, 0x01, 0x00, 0x10, 0x90, 0x0E, 
            0x0E, 0x1F, 0x7F, 0x3F, 0x1F, 0x3F, 0x00, 0x01, 0x0E, 0x7F, 0x7F, 0x00, 0x7F, 0x01, 0x1E, 0x1E, 
            0x10, 0x10, 0x3F, 0x2F, 0x1F, 0x7F, 0x90, 0xFF, 0x7F, 0x1F, 0x8F, 0x01, 0x00, 0x90, 0x90, 0x10, 
            0x8E, 0x0F, 0x27, 0x0F, 0x06, 0x1F, 0x8F, 0xFF, 0x00, 0x7F, 0x0F, 0xFF, 0x10, 0x7F, 0x01, 0x06, 
            0x3F, 0x1F, 0xFF, 0x8F, 0x01, 0x01, 0x80, 0x21, 0x01, 0x0E, 0x1E, 0x1E, 0x6F, 0x0E, 0x0F, 0x0E, 
            0x06, 0x0F, 0x3F, 0x1E, 0x21, 0x0E, 0x81, 0x0E, 0x0F, 0x00, 0x3F, 0x6F, 0x4F, 0x1F, 0xDF, 0x7F, 
            0x0F, 0x4F, 0x0F, 0x0F, 0x7F, 0x7F, 0x01, 0x0E, 0x21, 0x0E, 0x3F, 0x01, 0x0F, 0x0F, 0x0E, 0x01, 
            0x6F, 0x2F, 0x8F, 0x0F, 0x00, 0x0F, 0x2F, 0x7F, 0x21, 0x7F, 0x90, 0x6F, 0x7F, 0x7F, 0xFF, 0x7F, 
            0x6F, 0x7F, 0x80, 0x07, 0x0F, 0x0E, 0x21, 0x6F, 0x6F, 0x0F, 0xA7, 0xA1, 0x7F, 0x6F, 0x21, 0x21, 
            0x00, 0x21, 0x0E, 0x6F, 0x2F, 0x21, 0xAF, 0x6F, 0x0E, 0x0F, 0x4F, 0x0E, 0x0F, 0x2F, 0x2F, 0x01, 
            0x01, 0x0F, 0x00, 0x0E, 0x0E, 0x6F, 0x21, 0x0F, 0x0E, 0x0F, 0x00, 0x01, 0x21, 0x0F, 0x00, 0x2F, 
            0x0F, 0x0F, 0x8E, 0x8E, 0x0F, 0x0F, 0xAF, 0x0F, 0x00, 0x0F, 0xAF, 0x00, 0x0F, 0x6F, 0x21, 0x0E, 
            0x6F, 0x6F, 0xCF, 0x81, 0x2F, 0x2F, 0x0F, 0xCE, 0x01, 0x06, 0x0E, 0x00, 0x0F, 0x01, 0xEF, 0x00, 
            0x00, 0x00, 0x00, 0x0F, 0x0F, 0x0F, 0x0F, 0x06, 0x01, 0x21, 0x00, 0x00, 0x21, 0x0E, 0x07, 0x01, 
            0x0F, 0x0E, 0x8E, 0x0F, 0x6F, 0x01, 0x0F, 0x2F, 0x0F, 0x2F, 0xEF, 0x00, 0x21, 0x0E, 0x0F, 0x6F, 
            0x6F, 0x00, 0x01, 0x01, 0x21, 0x6F, 0x0F, 0x01, 0x2F, 0x6F, 0x0F, 0x2F, 0x0E, 0x0F, 0x2F, 0x0F, 
            0x2F, 0x2F, 0x0E, 0x06, 0x0E, 0x00, 0x8E, 0x6F, 0x01, 0x21, 0xAF, 0x6F, 0x01, 0x01, 0x6F, 0x00, 
            0x0F, 0x0F, 0x61, 0x0E, 0x6F, 0x4F, 0x6F, 0x2F, 0x01, 0x01, 0x00, 0x00, 0x0E, 0x0E, 0x01, 0x0F, 
            0x0F, 0x21, 0x06, 0x00, 0x10, 0x01, 0x0F, 0x21, 0x2F, 0x0F, 0x0E, 0x0F, 0x0F, 0x0F, 0x2F, 0x0F, 
            0x6F, 0x00, 0x0E, 0x6F, 0x00, 0x6F, 0x0E, 0x8E, 0x2F, 0x01, 0x07, 0x0E, 0x0E, 0x0F, 0x9F, 0x80, 
            0x00, 0x6F, 0x81, 0x6F, 0x0E, 0x2F, 0xEF, 0x0F, 0xEF, 0x0F, 0x2F, 0x2F, 0x6F, 0x8E, 0x80, 0x81, 
            0x0F, 0x0E, 0x06, 0x0E, 0x0F, 0x0F, 0x0E, 0x2F, 0x0F, 0x06, 0x8E, 0x0F, 0x0E, 0x21, 0x90, 0x01, 
            0x00, 0x00, 0x8F, 0x4E, 0x06, 0x06, 0x4F, 0x6F, 0x01, 0x65, 0x81, 0x0F, 0x0F, 0x01, 0x00, 0x01, 
            0x0F, 0x0F, 0x0F, 0x0F, 0x0E, 0x0F, 0x8F, 0x8F, 0x1F, 0x0E, 0x8E, 0x0F, 0x4F, 0x07, 0x8F, 0x6F, 
            0x2F, 0x0E, 0x6F, 0x2F, 0x00, 0x2F, 0x80, 0xAF, 0x0F, 0xAF, 0x00, 0x80, 0x01, 0x0F, 0x06, 0x0F, 
            0x2F, 0x0F, 0x80, 0x81, 0x6F, 0x01, 0x86, 0xAF, 0x8F, 0x2F, 0xEF, 0x80, 0x81, 0x21, 0x0F, 0x0E, 
            0x0F, 0x8E, 0xA1, 0xA1, 0x00, 0x80, 0x80, 0x8E, 0x6F, 0x2F, 0x2F, 0xAF, 0xEF, 0x00, 0x8E, 0xAF, 
            0x0F, 0x0E, 0xAF, 0x00, 0x4E, 0x00, 0xAF, 0x00, 0x0F, 0x0E, 0x00, 0x0E, 0x06, 0x0F, 0x8E, 0xAF, 
            0x6F, 0x81, 0x8E, 0x86, 0x0F, 0x0F, 0x0E, 0x00, 0x6F, 0x2F, 0xEF, 0x8E, 0x01, 0x2F, 0xEF, 0x4F, 
            0x6F, 0x6F, 0xEF, 0x2F, 0x4F, 0xAF, 0x6F, 0x0F, 0x6F, 0x0F, 0x8F, 0xEF, 0x21, 0x8E, 0x6F, 0xEF, 
            0x00, 0x00, 0x21, 0x8F, 0x0F, 0xAF, 0x8F, 0x01, 0x0E, 0x06, 0xA1, 0x21, 0x0E, 0x00, 0x90, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x21, 0x00, 0x0E, 0x8F, 0x61, 0x0F, 0x4E, 0x6F, 0x0E, 0x6F, 0x0E, 0x0E, 
            0x0F, 0x0F, 0x8F, 0x80, 0x6F, 0x0F, 0xAF, 0xAF, 0x0F, 0x8E, 0x8F, 0xEF, 0xAF, 0x00, 0x0F, 0x00, 
            0x2F, 0x06, 0x80, 0x86, 0x8F, 0x8F, 0x80, 0x8F, 0x80, 0x01, 0x9F, 0x00, 0xAF, 0x0F, 0xE1, 0x81, 
            0x2F, 0x0F, 0x80, 0x01, 0x01, 0x6F, 0x8F, 0x06, 0x00, 0x0E, 0x0F, 0x06, 0x6F, 0x01, 0xEF, 0x81, 
            0x0F, 0x81, 0x0F, 0x8E, 0x00, 0x00, 0x00, 0x00, 0x0F, 0x0E, 0x8F, 0x81, 0xEF, 0x0F, 0x8E, 0x1F, 
            0x00, 0x0E, 0x8F, 0x00, 0xEF, 0x8F, 0x6F, 0x2F, 0x0F, 0x0E, 0xEF, 0x8F, 0x80, 0x8E, 0x8F, 0x8E, 
            0x00, 0x0F, 0x0E, 0xCF, 0xAF, 0x8E, 0x8F, 0x6F, 0x0F, 0x0F, 0x8E, 0x01, 0x6F, 0x61, 0xAF, 0x0E, 
            0x6F, 0x0E, 0x06, 0x0E, 0x07, 0x0F, 0x81, 0x8E, 0x00, 0xA1, 0x81, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x0E, 0x01, 0x8E, 0xEF, 0x0F, 0x01, 0xA1, 0x0F, 0x2F, 0x01, 0x9F, 0x81, 0x8F, 0x0F, 0x8E, 0x0E, 
            0x2F, 0xEF, 0x8E, 0x81, 0x0E, 0x8E, 0x3F, 0xAF, 0x2F, 0x8F, 0x0E, 0x00, 0x0F, 0x2F, 0xEF, 0x8F, 
            0x00, 0x6F, 0x8F, 0x0E, 0x6F, 0x21, 0x8F, 0x0F, 0x2F, 0x00, 0x00, 0x00, 0x0E, 0x00, 0x8E, 0x80, 
            0x00, 0x80, 0x80, 0x80, 0x00, 0x0E, 0x4E, 0x0F, 0x2F, 0x0F, 0x8F, 0x8F, 0x06, 0x0F, 0x8E, 0x0E, 
            0x0E, 0x01, 0x01, 0x2F, 0x41, 0x00, 0x0E, 0x2F, 0x01, 0x0F, 0x86, 0x0E, 0x0E, 0x8F, 0x0F, 0x8F, 
            0x2F, 0x0E, 0x06, 0x2F, 0x2F, 0x0F, 0x8F, 0x81, 0x6F, 0x2F, 0xA1, 0x00, 0x00, 0x00, 0x80, 0x01, 
            0x0E, 0x21, 0x00, 0x00, 0x00, 0x40, 0x80, 0x80, 0x0E, 0x8E, 0x06, 0xEF, 0x0E, 0x00, 0x0F, 0x01, 
            0x06, 0x2F, 0x8F, 0x8F, 0x01, 0x6F, 0xAF, 0x0E, 0x6F, 0x06, 0xAF, 0x0F, 0x0E, 0x0F, 0x10, 0x0F, 
            0x0E, 0x21, 0x80, 0x81, 0x6F, 0xAF, 0x0E, 0x8E, 0x2F, 0x0E, 0x0F, 0x0E, 0x0F, 0x00, 0x80, 0x8E, 
            0x0E, 0x6F, 0xEF, 0x21, 0x00, 0x0F, 0x80, 0x8E, 0x80, 0x01, 0x21, 0x61, 0x01, 0x06, 0xA7, 0x81, 
            0x2F, 0x81, 0x2F, 0x8F, 0x0E, 0x00, 0x00, 0x00, 0x0F, 0x0E, 0xAF, 0x9F, 0x00, 0x0F, 0x8F, 0x0F, 
            0x2F, 0x0F, 0x8F, 0x06, 0x0E, 0x2F, 0x0E, 0x2F, 0x0E, 0x00, 0x80, 0x80, 0x2F, 0x6F, 0xAF, 0xEF, 
            0x0E, 0x2F, 0x0E, 0x0F, 0x01, 0x10, 0x90, 0x90, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 
            0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x80, 0x90, 0x00, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x90, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x90, 0x10, 0x90, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x80, 
            0x10, 0x90, 0x10, 0x80, 0x00, 0x90, 0x10, 0x10, 0x10, 0x10, 0x80, 0x90, 0x10, 0x10, 0x90, 0x10, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x00, 0x10, 0x90, 0x10, 0x90, 0x90, 0x10, 0x90, 0x10, 0x90, 
            0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x10, 0x10, 0x10, 0x90, 0x00, 0x10, 0x10, 0x90, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x90, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x90, 0x80, 0x10, 0x10, 0x10, 
            0x00, 0x90, 0x90, 0x90, 0x10, 0x90, 0x90, 0x80, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 0x90, 0x90, 
            0x10, 0x10, 0x90, 0x00, 0x90, 0x90, 0x90, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x00, 0x90, 0x90, 
            0x10, 0x90, 0x90, 0x80, 0x90, 0x10, 0x10, 0x10, 0x00, 0x10, 0x90, 0x80, 0x10, 0x10, 0x90, 0x10, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x80, 0x90, 0x10, 0x90, 0x90, 0x90, 
            0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x10, 0x10, 0x90, 0x10, 0x90, 0x10, 0x80, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x00, 0x00, 0x90, 0x6F, 0x01, 0x01, 0x0E, 0x8E, 0x0E, 0x0E, 0x01, 0x10, 
            0x6F, 0x00, 0xAF, 0x6F, 0x01, 0x00, 0x8E, 0x21, 0x0F, 0x6F, 0xA1, 0x6F, 0x0F, 0x0F, 0x0F, 0x6F, 
            0x06, 0x6F, 0x6F, 0x6F, 0x6F, 0x00, 0x21, 0x00, 0x6F, 0x21, 0x00, 0x06, 0x0E, 0x01, 0x8F, 0x8F, 
            0x21, 0x0F, 0xA1, 0x6F, 0x6F, 0x6F, 0x81, 0x21, 0xEF, 0x0F, 0x00, 0x0F, 0x0F, 0x6F, 0x8E, 0x80, 
            0x06, 0xCE, 0x21, 0x80, 0x81, 0x0E, 0x0F, 0x0E, 0x00, 0x2F, 0x8F, 0xEF, 0x6F, 0x0F, 0xAF, 0x0F, 
            0x0E, 0x00, 0x80, 0x2F, 0x00, 0x0E, 0x0F, 0x6F, 0x0F, 0x0F, 0xEF, 0x8F, 0x2F, 0xAF, 0x00, 0x80, 
            0x0F, 0x0F, 0x0F, 0x0F, 0x6F, 0x6F, 0xEF, 0x00, 0x21, 0x01, 0x8F, 0x0E, 0x6F, 0x01, 0x8F, 0x0F, 
            0x0F, 0x6F, 0x6F, 0x0E, 0x0F, 0x2F, 0x81, 0x80, 0x10, 0x10, 0x90, 0x90, 0x00, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 
            0x10, 0x10, 0x90, 0x00, 0x10, 0x10, 0x90, 0x10, 0x80, 0x10, 0x10, 0x00, 0x7F, 0x11, 0x8F, 0x90, 
            0x0E, 0xFF, 0x1F, 0x86, 0x8F, 0x81, 0x01, 0x00, 0x2F, 0x80, 0xFF, 0x80, 0x0F, 0x06, 0x7F, 0x2F, 
            0x0E, 0x1F, 0xFF, 0x11, 0x10, 0x10, 0x10, 0x01, 0x01, 0x0F, 0x0F, 0xFF, 0x1F, 0x9F, 0x00, 0xFF, 
            0x7F, 0x01, 0x7F, 0x0E, 0x7F, 0x10, 0x90, 0x10, 0x10, 0x10, 0x0E, 0x7F, 0x7F, 0x7F, 0x0E, 0x2F, 
            0x7F, 0x0E, 0x01, 0x0E, 0x1F, 0x10, 0x01, 0x01, 0x0F, 0x00, 0x80, 0xFF, 0x67, 0x6F, 0x6F, 0x7F, 
            0x00, 0x1F, 0xFF, 0x6F, 0x7F, 0x21, 0xA1, 0x10, 0x10, 0x10, 0xFF, 0x0E, 0x3F, 0x0E, 0x01, 0x06, 
            0x6F, 0x0E, 0xBF, 0xEF, 0x1F, 0xFF, 0x7F, 0xFF, 0x6F, 0x00, 0x21, 0x01, 0x0F, 0x6F, 0x8F, 0x81, 
            0x0E, 0x01, 0x81, 0x0F, 0x0F, 0x7F, 0x7F, 0x0E, 0x3F, 0x0F, 0x6F, 0x7F, 0x00, 0x7F, 0x6F, 0x31, 
            0x10, 0x0F, 0x0E, 0x8E, 0x7F, 0x6F, 0x01, 0x3F, 0x7F, 0x7F, 0x81, 0xEF, 0x7F, 0x21, 0x8E, 0x01, 
            0x21, 0x0E, 0x8E, 0xE1, 0x8F, 0x01, 0x01, 0x7F, 0x6F, 0x0F, 0xEF, 0x80, 0x0E, 0x80, 0x0F, 0xA1, 
            0x3F, 0x6F, 0x21, 0x1F, 0x2F, 0x00, 0xBB, 0x61, 0x2F, 0x3F, 0x01, 0x6F, 0x3F, 0x10, 0x10, 0x0E, 
            0x7F, 0x7F, 0x0E, 0x0E, 0x7F, 0x7F, 0x7F, 0x7F, 0x0E, 0x7F, 0x8F, 0x6F, 0x7F, 0x7F, 0x7F, 0x10, 
            0x00, 0x21, 0x9F, 0x0E, 0x0E, 0x7F, 0x6F, 0x0E, 0x0F, 0x6F, 0x0F, 0x6F, 0x6F, 0x0F, 0x0F, 0x0F, 
            0x10, 0x0E, 0x6F, 0x6F, 0x00, 0x0E, 0x00, 0x00, 0x7F, 0x5F, 0x7F, 0x21, 0x00, 0x00, 0x80, 0xDF, 
            0x6F, 0x0F, 0xEF, 0x10, 0x00, 0x06, 0xFF, 0x7F, 0x0E, 0x7F, 0x0F, 0x6F, 0x00, 0x0E, 0x90, 0x8E, 
            0x7F, 0x90, 0x0F, 0x01, 0x3F, 0x0F, 0x7F, 0x7F, 0x3F, 0x1F, 0xA1, 0x06, 0x00, 0x3F, 0xBF, 0x3F, 
            0x06, 0x6F, 0x86, 0x7F, 0x00, 0x00, 0x7F, 0x00, 0x06, 0x6F, 0x81, 0x01, 0x7F, 0x07, 0x7F, 0x0E, 
            0x06, 0x01, 0x0F, 0x0F, 0x2F, 0x0E, 0x8E, 0xEF, 0x0F, 0x0E, 0xBF, 0x0E, 0x0E, 0x10, 0x8E, 0x0E, 
            0x7F, 0x01, 0x7F, 0x21, 0x00, 0x0E, 0x80, 0x80, 0x3F, 0xFF, 0x0E, 0xA1, 0x00, 0x6F, 0x3F, 0x0E, 
            0x00, 0x01, 0x6F, 0x2F, 0x6F, 0x00, 0x2F, 0x0E, 0x6F, 0x2F, 0x01, 0x0E, 0x00, 0x21, 0x01, 0x21, 
            0x00, 0x61, 0x3F, 0x10, 0x00, 0x21, 0x7F, 0x01, 0x07, 0x31, 0x0F, 0x1F, 0x6F, 0x00, 0xFF, 0x80, 
            0x3F, 0x7F, 0xFF, 0x01, 0x21, 0x90, 0xAF, 0x01, 0xAF, 0x7F, 0x0F, 0x21, 0x0E, 0x61, 0x8F, 0x80, 
            0x0E, 0xA1, 0x0E, 0x9F, 0xFF, 0x00, 0x1F, 0xCF, 0x8F, 0x21, 0xEF, 0x8F, 0x3F, 0x0F, 0x8E, 0x0F, 
            0x8F, 0x2F, 0x8E, 0x0E, 0x2F, 0xBF, 0x21, 0x00, 0x0E, 0x01, 0x8E, 0xFF, 0x00, 0xFF, 0x8E, 0xEF, 
            0x21, 0x10, 0x10, 0x00, 0x1E, 0x01, 0x81, 0x8F, 0x3F, 0x0F, 0x8F, 0x00, 0x2F, 0x06, 0x8F, 0x01, 
            0x7F, 0x0E, 0x07, 0x3F, 0x0E, 0x06, 0xAF, 0x80, 0x6F, 0x80, 0x8E, 0xFF, 0x0F, 0x7F, 0x0F, 0x10, 
            0x00, 0x01, 0x91, 0x2F, 0x31, 0x01, 0x8F, 0x0E, 0x01, 0x0F, 0x01, 0x0E, 0x01, 0x00, 0x01, 0x0F, 
            0x00, 0x0F, 0x00, 0x2F, 0x0F, 0x0F, 0x6F, 0x8F, 0x21, 0x0F, 0x01, 0x6F, 0x0F, 0x0F, 0x0F, 0x0E, 
            0x0F, 0x01, 0x00, 0x6F, 0x0F, 0x0E, 0x06, 0x0F, 0x0F, 0x0F, 0x00, 0x01, 0x6F, 0x0F, 0x0F, 0x0E, 
            0x01, 0x0F, 0x00, 0x0E, 0x01, 0x0E, 0x10, 0x10, 0x00, 0x10, 0x90, 0x90, 0x10, 0x3F, 0xA1, 0x01, 
            0x0E, 0x0F, 0xA1, 0x7F, 0x01, 0x31, 0x7F, 0x01, 0x07, 0x0E, 0x8F, 0xEF, 0x00, 0x8E, 0x8E, 0xEF, 
            0x4E, 0x6F, 0x6F, 0x6F, 0x0E, 0x6F, 0xEF, 0x0F, 0x6F, 0x00, 0x4F, 0x00, 0x6F, 0x0F, 0x0F, 0x2F, 
            0x6F, 0x6F, 0x6F, 0x6F, 0x01, 0x00, 0x0F, 0x6F, 0x6F, 0x01, 0x21, 0x8E, 0x0F, 0x0E, 0x0F, 0x00, 
            0x0F, 0x2F, 0x80, 0x01, 0x2F, 0x01, 0x0F, 0x0E, 0x0E, 0x0E, 0x00, 0x0F, 0x21, 0x6F, 0x01, 0x0F, 
            0x4F, 0x00, 0x0F, 0x01, 0x21, 0x0F, 0x80, 0x2F, 0x6F, 0x4E, 0x00, 0x6F, 0x21, 0x21, 0x01, 0x00, 
            0x00, 0x0E, 0x00, 0x0E, 0x0F, 0x0E, 0x6F, 0x01, 0x0E, 0x0E, 0x0F, 0x21, 0x6F, 0x6F, 0x6F, 0x2F, 
            0x0E, 0x0F, 0x0F, 0x0F, 0x61, 0x21, 0x01, 0x2F, 0x6F, 0x0F, 0xC1, 0xAF, 0x0E, 0x0E, 0xEF, 0x0E, 
            0x0E, 0x00, 0x8F, 0x0F, 0x0E, 0x6F, 0x01, 0x6F, 0x00, 0x0E, 0x8F, 0xEF, 0x01, 0x80, 0x00, 0xEF, 
            0x2F, 0xAF, 0x0E, 0x2F, 0x2F, 0x10, 0x90, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x00, 0x10, 0x10, 0x00, 0x10, 
            0x10, 0x10, 0x80, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x90, 0x10, 0x6F, 0x06, 0x21, 0x01, 0x0E, 0x0F, 0x8F, 0x6F, 
            0x01, 0x6F, 0x8E, 0x0E, 0x01, 0x00, 0xAF, 0x01, 0x0F, 0x00, 0x0F, 0x0F, 0x0E, 0x0E, 0x8E, 0x8F, 
            0x0E, 0x0E, 0x0F, 0x21, 0x6F, 0x01, 0x2F, 0x40, 0x01, 0x0E, 0x00, 0x0E, 0x0F, 0x00, 0x10, 0x00, 
            0x00, 0x10, 0x10, 0x10, 0x00, 0x10, 0x00, 0x00, 0x10, 0x10, 0x10, 0x6F, 0x61, 0x00, 0x10, 0xFF, 
            0x00, 0x41, 0x6F, 0x0F, 0x00, 0x0F, 0x81, 0x9F, 0x10, 0x2F, 0x8F, 0x2F, 0x00, 0x6F, 0xE1, 0x6F, 
            0x00, 0x00, 0x2F, 0x01, 0x6F, 0x00, 0x8F, 0x80, 0x00, 0x0E, 0x0E, 0x80, 0x6F, 0x6F, 0x6F, 0x00, 
            0x0E, 0x00, 0x8F, 0x6F, 0x00, 0x2F, 0x00, 0x0F, 0x0F, 0x6F, 0x6F, 0x00, 0x6F, 0x10, 0x00, 0x00, 
            0x7F, 0x0F, 0x2F, 0x6F, 0x2F, 0x0F, 0x8F, 0x8F, 0x6F, 0x00, 0x0F, 0x0F, 0x01, 0x21, 0x6F, 0x0F, 
            0x61, 0x2F, 0x01, 0x00, 0x2F, 0x0E, 0x01, 0x2F, 0x6F, 0x06, 0x0E, 0x0F, 0x2F, 0x0E, 0x3F, 0x0E, 
            0x0E, 0x0F, 0x0F, 0x0F, 0x01, 0x0F, 0x00, 0x00, 0x00, 0x01, 0x8F, 0x01, 0x0F, 0x2F, 0x2F, 0x0E, 
            0x01, 0x0F, 0x21, 0x0E, 0x00, 0x6F, 0x01, 0x0F, 0x0F, 0x6F, 0x00, 0x6F, 0x6F, 0x41, 0x0F, 0x00, 
            0x2F, 0x6F, 0x6F, 0x06, 0x5F, 0x3F, 0x8E, 0x6F, 0x0F, 0x0F, 0x00, 0x0F, 0x0F, 0x00, 0x0F, 0x0F, 
            0x01, 0x06, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x00, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00, 0x10, 0x00, 0x10, 0x10, 0x00, 0x10, 
            0x10, 0x10, 0x80, 0x00, 0x10, 0x90, 0x10, 0x90, 0x10, 0x00, 0x10, 0x10, 0x00, 0x10, 0x00, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7F, 0x3F, 0x3F, 0x7F, 0x00, 0x01, 0x0E, 0x4F, 0x27, 0x01, 
            0x00, 0x0E, 0x00, 0x0E, 0x00, 0x7F, 0x0F, 0x0E, 0x7F, 0x00, 0x80, 0x86, 0x6F, 0x6F, 0xEF, 0x0E, 
            0x0F, 0x4F, 0x8F, 0x6F, 0x6F, 0x0F, 0x00, 0x00, 0x00, 0x4F, 0x8F, 0x8E, 0x21, 0x0F, 0x00, 0x81, 
            0x80, 0x6F, 0x0E, 0x0F, 0x21, 0xA1, 0xA1, 0x8E, 0x21, 0x0F, 0x80, 0x0E, 0x8E, 0x0E, 0x8E, 0x0E, 
            0x6F, 0x6F, 0x6F, 0x0F, 0x0F, 0x6F, 0x8E, 0x8E, 0x2F, 0xCF, 0x81, 0xAF, 0x8F, 0x6F, 0x01, 0x6F, 
            0x00, 0x00, 0xAF, 0x8E, 0x0E, 0x0E, 0x80, 0x07, 0x00, 0x0E, 0x8F, 0x01, 0x0F, 0x6F, 0x2F, 0x00, 
            0x0F, 0x6F, 0xA1, 0x80, 0x0F, 0x81, 0x8E, 0x8E, 0x2F, 0x0F, 0x0E, 0x2F, 0x0E, 0x0F, 0x8F, 0x6F, 
            0x00, 0x6F, 0x8F, 0x0F, 0x8F, 0x2F, 0x8E, 0x0E, 0x40, 0x0E, 0x0E, 0x0F, 0x01, 0x0E, 0xEF, 0x6F, 
            0x01, 0x0F, 0x21, 0xA1, 0x06, 0x0E, 0x0F, 0x00, 0x00, 0x6F, 0x0E, 0x0E, 0x0E, 0x0E, 0x0F, 0x00, 
            0x0F, 0x00, 0x0F, 0x0F, 0x0F, 0x0E, 0x00, 0x0F, 0x21, 0x0E, 0x0E, 0x6F, 0x0E, 0x0F, 0x0F, 0x00, 
            0x6F, 0x0E, 0x0E, 0x01, 0x0E, 0x0F, 0xCF, 0xEF, 0x0F, 0x0E, 0x8E, 0x00, 0x00, 0x0E, 0xAF, 0x0E, 
            0x6F, 0x0F, 0x2F, 0x6F, 0x0F, 0x6F, 0x8E, 0x81, 0x0E, 0x8E, 0x8F, 0x81, 0x0F, 0x6F, 0x2F, 0x0E, 
            0x06, 0x01, 0x8E, 0x8E, 0x0F, 0x6F, 0x8F, 0x6F, 0x00, 0x0E, 0xEF, 0x6F, 0x00, 0x01, 0x0E, 0x6F, 
            0x0E, 0x00, 0xEF, 0x80, 0x2F, 0xEF, 0x0E, 0x8E, 0x0E, 0x6F, 0x6F, 0x2F, 0x10, 0x10, 0x90, 0x10, 
            0x10, 0x10, 0x80, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x90, 0x10, 0x10, 0x00, 0x80, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x00, 0x00, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x80, 0x90, 0x7F, 0x00, 0x81, 0x06, 0x01, 0x2F, 0x81, 0x0F, 
            0x3F, 0x1E, 0x01, 0x06, 0x0F, 0x01, 0x91, 0x9E, 0x7F, 0x8F, 0x10, 0x87, 0x3F, 0x00, 0x0F, 0x0F, 
            0x3F, 0x1F, 0x9E, 0x01, 0x21, 0x10, 0x8F, 0x0E, 0x01, 0x00, 0x8E, 0x10, 0x10, 0x0E, 0x01, 0x2F, 
            0x0F, 0x3F, 0x0F, 0x7F, 0x6F, 0x0F, 0x0F, 0x81, 0x7F, 0x21, 0x00, 0x01, 0x2F, 0x00, 0x21, 0x3F, 
            0x01, 0x1E, 0x21, 0x2F, 0x01, 0x41, 0x3F, 0x0F, 0x00, 0x01, 0x21, 0x3F, 0x0F, 0x3F, 0x6F, 0x3F, 
            0x00, 0x27, 0x0F, 0x0F, 0x21, 0x01, 0x0E, 0x27, 0x00, 0x1F, 0x8F, 0x3F, 0x0E, 0x0F, 0x8E, 0x01, 
            0x01, 0x0F, 0x81, 0x1F, 0x0F, 0x0F, 0x2F, 0x00, 0x1F, 0x01, 0x8E, 0x8F, 0x0F, 0x8F, 0x0F, 0x90, 
            0x0F, 0x0E, 0x0F, 0x10, 0x00, 0x0E, 0x8F, 0x0E, 0x2F, 0x0F, 0x6F, 0x01, 0x01, 0x00, 0x0E, 0x3F, 
            0x0F, 0x00, 0x2F, 0x3F, 0x0E, 0x2F, 0x81, 0x6F, 0x2F, 0x2F, 0x61, 0x0F, 0x00, 0x01, 0x2F, 0x3F, 
            0x00, 0x6F, 0xBF, 0x8F, 0x01, 0x0F, 0x80, 0x0F, 0x00, 0x01, 0x8F, 0x3F, 0x7F, 0x01, 0x0E, 0x0E, 
            0x00, 0x7F, 0xFF, 0xFF, 0x7F, 0xFF, 0x0E, 0x90, 0x1F, 0x10, 0x0E, 0x0F, 0x0F, 0x3F, 0xAF, 0x7F, 
            0x00, 0x3F, 0x0E, 0x00, 0x7F, 0x07, 0x0F, 0x01, 0x2F, 0x06, 0x6F, 0x07, 0x00, 0x00, 0x01, 0x06, 
            0x0E, 0x0F, 0x80, 0x01, 0x0E, 0x01, 0x0F, 0x0E, 0x0F, 0x00, 0x81, 0x01, 0x07, 0x01, 0x81, 0x6F, 
            0x06, 0x0E, 0x21, 0x01, 0x2F, 0x2F, 0x0E, 0x0F, 0x01, 0x01, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x01, 
            0x0F, 0x00, 0x0E, 0x21, 0x01, 0x0F, 0x8F, 0x0F, 0x0E, 0x01, 0x81, 0x01, 0x00, 0x01, 0xE1, 0x21, 
            0x0F, 0x6F, 0x2F, 0x2F, 0x01, 0x0E, 0xA1, 0xA1, 0x00, 0x80, 0x0F, 0x8E, 0x00, 0x01, 0x0F, 0x61, 
            0x27, 0x0E, 0x86, 0x00, 0x0E, 0x0E, 0x8F, 0x01, 0x2F, 0x01, 0xAF, 0x6F, 0x01, 0x2F, 0x6F, 0x0E, 
            0x01, 0x21, 0x81, 0x00, 0x21, 0x0E, 0x06, 0x81, 0x0F, 0x2F, 0x00, 0x21, 0x01, 0x0E, 0x81, 0x8F, 
            0x2F, 0x0F, 0x80, 0x0E, 0x06, 0x00, 0xAF, 0x0F, 0x0F, 0x6F, 0x2F, 0x00, 0x00, 0x00, 0x81, 0xA1, 
            0x01, 0xA1, 0x21, 0x8E, 0x2F, 0x0E, 0x6F, 0x0F, 0x01, 0x06, 0x0E, 0x2F, 0x06, 0x01, 0x0E, 0x01, 
            0x0E, 0x27, 0x2F, 0x21, 0x6F, 0x07, 0x0E, 0x0F, 0x6F, 0x00, 0x0F, 0x0F, 0x0E, 0x00, 0x00, 0x01, 
            0x2F, 0x21, 0x21, 0x01, 0x00, 0x21, 0x80, 0x8F, 0x0F, 0x01, 0x81, 0x00, 0x00, 0x8F, 0x80, 0x81, 
            0x21, 0x00, 0x01, 0x00, 0x21, 0x0E, 0xA7, 0x8E, 0x2F, 0xAF, 0x21, 0x8F, 0x21, 0x6F, 0x0E, 0x01, 
            0x61, 0x01, 0xEF, 0xAF, 0x27, 0x61, 0x81, 0x0E, 0x01, 0x01, 0x81, 0x21, 0x0F, 0x0E, 0x01, 0x00, 
            0x00, 0x2F, 0x81, 0x8F, 0x27, 0xEF, 0x01, 0x81, 0x0F, 0x0F, 0x01, 0x0E, 0x0E, 0x2F, 0xA1, 0xA1, 
            0x21, 0x0F, 0xEF, 0x0E, 0x0E, 0x01, 0x87, 0x0F, 0x00, 0x2F, 0x21, 0x6F, 0x0E, 0x07, 0xAF, 0x8E, 
            0x0E, 0x8F, 0x80, 0x81, 0x0F, 0x01, 0x2F, 0x61, 0x2F, 0x4F, 0x8F, 0x8E, 0x06, 0x0E, 0x0F, 0x01, 
            0x0F, 0x00, 0x27, 0x01, 0x0F, 0x0E, 0x0F, 0x6F, 0x0F, 0x0E, 0x21, 0x01, 0x01, 0x01, 0x0F, 0x8F, 
            0x2F, 0x00, 0x0E, 0x0F, 0x00, 0x00, 0x0E, 0x2F, 0x0F, 0x01, 0x01, 0x01, 0x00, 0x0E, 0x0F, 0x00, 
            0x01, 0x0E, 0x01, 0x0E, 0x0E, 0x07, 0x21, 0x0F, 0x2F, 0x0E, 0x0E, 0x01, 0x10, 0x00, 0x00, 0x10, 
            0x00, 0x10, 0x90, 0x80, 0x00, 0x10, 0x90, 0x10, 0x10, 0x80, 0x80, 0x10, 0x00, 0x10, 0x10, 0x00, 
            0x10, 0x10, 0x90, 0x80, 0x10, 0x90, 0x80, 0x80, 0x00, 0x00, 0x10, 0x10, 0x10, 0x00, 0x90, 0x90, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x00, 0x10, 0x00, 0x10, 0x90, 0x80, 
            0x10, 0x90, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x90, 0x10, 0x10, 0x10, 0x80, 0x00, 
            0x00, 0x00, 0x80, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00, 0x90, 0x90, 0x10, 0x90, 0x90, 0x90, 
            0x10, 0x00, 0x00, 0x10, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 
            0x00, 0x00, 0x10, 0x00, 0x00, 0x6F, 0x0F, 0x6F, 0x00, 0x6F, 0x0E, 0x21, 0x21, 0x0E, 0x00, 0x00, 
            0x21, 0x0E, 0x8F, 0xEF, 0x6F, 0x0E, 0xEF, 0x0F, 0x00, 0x0F, 0x8E, 0x0E, 0x0E, 0x0E, 0x00, 0x0E, 
            0x0E, 0x00, 0x81, 0xAF, 0x0E, 0x8E, 0x2F, 0xAF, 0x2F, 0x6F, 0x00, 0x01, 0x00, 0x00, 0xA1, 0x80, 
            0x0E, 0x01, 0xAF, 0x00, 0x0E, 0x2F, 0x80, 0x0F, 0x0F, 0x06, 0x00, 0x6F, 0x01, 0x0F, 0x8F, 0x2F, 
            0x0E, 0x00, 0x0E, 0xAF, 0x00, 0x0E, 0x6F, 0x00, 0x6F, 0x0E, 0xA1, 0x21, 0x21, 0x06, 0x0E, 0x0F, 
            0x0F, 0x0E, 0x01, 0x0F, 0x01, 0x00, 0x0E, 0x0E, 0x0E, 0x00, 0x01, 0x6F, 0x01, 0x0F, 0x2F, 0xAF, 
            0x0E, 0x2F, 0x0F, 0x0F, 0x21, 0x0F, 0xA1, 0x01, 0x21, 0x00, 0x0F, 0x0E, 0x0E, 0x00, 0x80, 0x0E, 
            0x21, 0x6F, 0x0E, 0x0F, 0x0F, 0x00, 0x0E, 0x0E, 0x0E, 0x2F, 0x06, 0x0E, 0x2F, 0x6F, 0x21, 0x0F, 
            0x6F, 0x6F, 0x81, 0x01, 0x21, 0x01, 0x00, 0x0E, 0x06, 0x0F, 0x0F, 0x2F, 0x6F, 0x00, 0x00, 0x2F, 
            0x2F, 0x0E, 0x6F, 0x0E, 0x0E, 0x0E, 0x01, 0x0F, 0x0E, 0x0E, 0x21, 0x0F, 0x01, 0x0E, 0x81, 0x8E, 
            0x0E, 0x0E, 0x8E, 0x01, 0x0F, 0x0E, 0x8F, 0x27, 0x0E, 0x2F, 0x0F, 0x0E, 0x0E, 0x01, 0x81, 0xA1, 
            0x0E, 0x80, 0x0F, 0x81, 0x0E, 0x01, 0x0F, 0x0F, 0x0F, 0x0E, 0xAF, 0x0E, 0x0E, 0x0E, 0x8E, 0x0E, 
            0x0E, 0x0F, 0x8E, 0x0E, 0x2F, 0x00, 0x06, 0x0E, 0x0E, 0x6F, 0x0F, 0x21, 0x0F, 0x0E, 0x01, 0xEF, 
            0x00, 0x0F, 0x27, 0x06, 0x6F, 0x0F, 0x8E, 0x0E, 0x27, 0x0F, 0x2F, 0x2F, 0x0F, 0x00, 0x0E, 0x0F, 
            0x00, 0x2F, 0x2F, 0x0F, 0x61, 0x0E, 0x21, 0x0F, 0x0E, 0x01, 0x0F, 0x8E, 0x0E, 0x0E, 0x0E, 0x2F, 
            0x0E, 0x0E, 0x8E, 0xAF, 0x01, 0x0F, 0x8F, 0x4F, 0x0E, 0x2F, 0x8F, 0x0E, 0x0E, 0x0E, 0x0F, 0x0F, 
            0x01, 0x0E, 0x8E, 0x8E, 0x0E, 0x8F, 0x2F, 0x81, 0x06, 0x0F, 0x00, 0x0F, 0x0E, 0x2F, 0x8F, 0x2F, 
            0x0F, 0x00, 0xEF, 0x07, 0x0F, 0x0E, 0x8E, 0x0E, 0x6F, 0x6F, 0x6F, 0x0E, 0x00, 0x2F, 0x8F, 0x0E, 
            0x0E, 0x0E, 0x0F, 0x0E, 0x0E, 0x0E, 0x0E, 0x0F, 0x00, 0x0E, 0x81, 0x8E, 0x00, 0x0E, 0x81, 0x0E, 
            0x0F, 0x0F, 0x8F, 0x0E, 0x0E, 0x0F, 0x01, 0x0E, 0x00, 0x2F, 0xEF, 0xAF, 0x01, 0x0F, 0x6F, 0x90, 
            0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 
            0x00, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00, 0x10, 0x10, 0x10, 
            0x00, 0x10, 0x90, 0x90, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 0x10, 0x00, 0x10, 0x10, 
            0x00, 0x10, 0x80, 0x80, 0x00, 0x90, 0x00, 0x90, 0x10, 0x00, 0x10, 0x10, 0x10, 0x00, 0x90, 0x80, 
            0x00, 0x00, 0x80, 0x10, 0x10, 0x00, 0x90, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x80, 0x80, 
            0x10, 0x90, 0x80, 0x90, 0x00, 0x6F, 0x00, 0x00, 0x21, 0x6F, 0x8F, 0x81, 0x0F, 0x6F, 0x90, 0x7F, 
            0x0F, 0x21, 0x9F, 0x0F, 0x01, 0x01, 0x0E, 0x1F, 0x3F, 0x0E, 0x8E, 0xBF, 0x2F, 0x86, 0x8F, 0x81, 
            0x00, 0x2F, 0x7F, 0x7F, 0x0E, 0x21, 0x81, 0x6F, 0x01, 0x0E, 0x0E, 0x0F, 0x0E, 0x7F, 0x01, 0x7F, 
            0x0E, 0x0E, 0x00, 0x00, 0x0F, 0x6F, 0x31, 0x0E, 0x01, 0x2F, 0x21, 0x80, 0x01, 0x2F, 0x0F, 0x01, 
            0x0F, 0x00, 0x00, 0x01, 0x5F, 0x4F, 0x0E, 0x0E, 0x31, 0x21, 0x21, 0x7F, 0x2F, 0x10, 0x7F, 0x21, 
            0x0E, 0x00, 0x0E, 0x4E, 0x31, 0x00, 0x01, 0x00, 0x0F, 0x10, 0x00, 0x01, 0x2F, 0x7F, 0xFF, 0xBF, 
            0x2F, 0x7F, 0xA1, 0x0E, 0x7F, 0x0F, 0x8E, 0x00, 0xFF, 0x21, 0x0E, 0x7F, 0x7F, 0x3F, 0xEF, 0x9F, 
            0x3F, 0x80, 0x90, 0x80, 0x0F, 0x3F, 0x06, 0x1F, 0x6F, 0x10, 0x90, 0x8E, 0x01, 0x0F, 0x8F, 0x3F, 
            0x0F, 0x01, 0x8F, 0x0E, 0x6F, 0x0F, 0x2F, 0x2F, 0x01, 0x3F, 0x86, 0xBF, 0x3F, 0xAF, 0x10, 0x8F, 
            0x0E, 0x0E, 0x01, 0x01, 0x00, 0x00, 0x8E, 0x6F, 0x61, 0x0F, 0x8E, 0x10, 0x00, 0x10, 0x7F, 0x0F, 
            0x1F, 0x01, 0x0F, 0x7F, 0x01, 0x2F, 0x0F, 0x11, 0x0E, 0x1F, 0x8F, 0x0F, 0x0E, 0x00, 0x06, 0x01, 
            0x7F, 0x21, 0x9F, 0x0E, 0x0E, 0x0E, 0x81, 0x00, 0x0E, 0x0E, 0x0F, 0x0F, 0x3F, 0x0E, 0x0E, 0x1F, 
            0x0E, 0x0F, 0x0F, 0x0E, 0x0F, 0x0E, 0x0E, 0x1F, 0x0E, 0x11, 0x01, 0x7F, 0x01, 0x1F, 0xBF, 0x81, 
            0x0E, 0x0F, 0x8E, 0x0F, 0x11, 0x01, 0x87, 0x0F, 0x06, 0x0E, 0x6F, 0x6F, 0x0E, 0x0E, 0xEF, 0xAF, 
            0x10, 0x90, 0x6F, 0x81, 0x2F, 0x0F, 0x0F, 0x0F, 0x8F, 0x06, 0x81, 0x0E, 0x0E, 0x0F, 0x8F, 0x6F, 
            0x2F, 0x6F, 0x21, 0x2F, 0x0E, 0x0E, 0x2F, 0x6F, 0x01, 0x01, 0x6F, 0x0E, 0x6F, 0x01, 0x0F, 0x0F, 
            0x0F, 0x0F, 0x2F, 0x01, 0x0E, 0x0F, 0xAF, 0x6F, 0x0E, 0x0E, 0x0F, 0x0E, 0x00, 0x01, 0x0E, 0x10, 
            0x10, 0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x6F, 0x00, 0x01, 
            0x4F, 0x07, 0x8F, 0x00, 0x0F, 0x6F, 0x81, 0x01, 0x06, 0x10, 0x10, 0x10, 0x6F, 0x61, 0x41, 0x50, 
            0x7F, 0x71, 0x7F, 0x71, 0x7E, 0xFF, 0x7B, 0xF9, 0x7E, 0x78, 0x7C, 0x7D, 0x7F, 0x7A, 0xFA, 0x7F, 
            0x7B, 0x7E, 0x7A, 0x7E, 0x7E, 0x7B, 0x7A, 0x7D, 0x7D, 0x7F, 0x7F, 0x7E, 0x7C, 0x7E, 0x7A, 0x78, 
            0x78, 0x78, 0x78, 0x7A, 0x7A, 0x79, 0x7C, 0x7C, 0x7D, 0x7C, 0xFC, 0xFE, 0x7C, 0x78, 0xFD, 0x78, 
            0x7E, 0x7D, 0xFA, 0x7E, 0x7C, 0x7A, 0x7E, 0x7A, 0x7A, 0x78, 0xFC, 0xFC, 0x7C, 0xFE, 0x7E, 0xF8, 
            0x7E, 0x7E, 0x7C, 0x7C, 0x7C, 0x7C, 0xFC, 0x7E, 0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7A, 0x78, 0x78, 
            0x7A, 0x7C, 0x7E, 0x7E, 0x7A, 0x7C, 0x7C, 0x7E, 0x7A, 0x78, 0xFC, 0x7E, 0x7C, 0x7E, 0x7A, 0x7C, 
            0x7C, 0x7E, 0xFE, 0x7E, 0x7E, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x78, 0x78, 0x78, 0x7A, 0x78, 
            0x78, 0x78, 0x78, 0x7C, 0x7C, 0x7C, 0x78, 0x7C, 0x7E, 0x78, 0x7C, 0x7C, 0x7C, 0x78, 0xFC, 0xF8, 
            0x78, 0x78, 0xFC, 0x7E, 0x7E, 0x7C, 0xF8, 0x7E, 0x7E, 0x7A, 0x7C, 0x7A, 0x78, 0x7E, 0xF8, 0xFC, 
            0x7E, 0x7E, 0x7A, 0xFE, 0x7C, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0xFE, 0x7C, 0x78, 0x78, 0xFE, 0x7C, 
            0x7A, 0x7C, 0xFC, 0x7E, 0x7E, 0x78, 0x7E, 0x7A, 0x7C, 0x7E, 0xFA, 0x7A, 0x78, 0x7E, 0x7C, 0xFC, 
            0x7A, 0x7E, 0x7A, 0x7E, 0x7C, 0x7C, 0xFC, 0xF8, 0x7E, 0x7C, 0xFC, 0x7E, 0x7C, 0xFC, 0xFE, 0x7E, 
            0xFC, 0x7C, 0x7C, 0x7C, 0x7C, 0x7E, 0xFE, 0xFE, 0x7E, 0xFE, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 
            0x7E, 0x7C, 0xFA, 0x7C, 0x78, 0x7C, 0xFC, 0x7C, 0x78, 0x7A, 0xFE, 0x7E, 0x7A, 0x7E, 0x7E, 0x7E, 
            0x78, 0x7E, 0xFA, 0xFA, 0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7E, 0x7A, 0x7C, 0x7C, 0x7E, 0xFA, 0xF8, 
            0x7C, 0x78, 0xF8, 0x78, 0x78, 0x7E, 0xFF, 0x7E, 0x7F, 0x7E, 0x7E, 0x7E, 0x7F, 0x78, 0xFC, 0xF8, 
            0x7A, 0xF8, 0x7A, 0xFC, 0x7E, 0x7E, 0x7A, 0x7E, 0x78, 0x7F, 0xFA, 0xFC, 0xF8, 0x7E, 0xFE, 0x7A, 
            0x7E, 0x7C, 0xFA, 0x7A, 0xFF, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0xFE, 0xFC, 0x7C, 0xFE, 0xFE, 0xFC, 
            0x7C, 0x7C, 0x7E, 0xFC, 0x78, 0x7C, 0xFC, 0xFE, 0x7C, 0x7C, 0xFC, 0x7C, 0x78, 0x78, 0xFC, 0x7C, 
            0x7C, 0x7E, 0x7E, 0x7C, 0x7A, 0x7E, 0xFC, 0xFC, 0x7E, 0xFE, 0xFC, 0xFE, 0x78, 0x7A, 0x78, 0x7E, 
            0x7E, 0x7E, 0xFE, 0xFE, 0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7C, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x7C, 
            0x78, 0x7E, 0x7A, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7B, 0x7F, 0x7B, 0x7F, 0x7E, 0x7E, 0x7E, 
            0x7E, 0x7E, 0x7E, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7D, 0x79, 0x7F, 0x7C, 0x7E, 0x7E, 
            0x7E, 0x7E, 0x7C, 0x7E, 0x7E, 0x7E, 0x7C, 0x7C, 0x7E, 0x7E, 0xFE, 0xFE, 0x7E, 0x7A, 0xFE, 0x7E, 
            0x7E, 0x7E, 0xFA, 0x7B, 0x7F, 0x7F, 0x7F, 0x7E, 0x7A, 0x7F, 0xFF, 0xFF, 0x7F, 0xFE, 0xFE, 0xFC, 
            0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7E, 0xFE, 0xFA, 0x7C, 0x7E, 0xFA, 0x7E, 0x7E, 0x7E, 0xFC, 0x7A, 
            0x7E, 0x7A, 0x7A, 0x7A, 0x78, 0x7C, 0xFC, 0xFC, 0x7C, 0xFC, 0x7C, 0xFD, 0x79, 0x78, 0x7C, 0x7C, 
            0x7C, 0x78, 0xFC, 0x7A, 0x7E, 0x7C, 0xF8, 0x7A, 0x7E, 0x7C, 0xFE, 0x7E, 0x7A, 0x7C, 0x7E, 0x7E, 
            0x7E, 0x78, 0xF8, 0xF8, 0x7A, 0x7A, 0xFE, 0xFE, 0x7E, 0x7E, 0x7E, 0x7C, 0x7E, 0x7C, 0xFA, 0x7E, 
            0x7E, 0x7C, 0x7A, 0x7E, 0x7A, 0x7E, 0xFE, 0x7C, 0x7E, 0x7E, 0x79, 0x7A, 0x7A, 0x7A, 0x7A, 0xFE, 
            0x7E, 0xFE, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7C, 0x7C, 0x7C, 0xF8, 0xFC, 0x78, 0x7C, 0xFC, 0x78, 
            0x7C, 0x7E, 0xFC, 0x7E, 0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7E, 0xFE, 0xFE, 0x7A, 0xFE, 0x78, 0xFE, 
            0x7C, 0x78, 0x7C, 0x7E, 0x7C, 0x7A, 0xFC, 0x7C, 0x7C, 0x7C, 0x7A, 0x7E, 0x7E, 0x7D, 0x7A, 0x7C, 
            0x7A, 0x7F, 0x7F, 0x7F, 0x7E, 0x7C, 0x7F, 0x7E, 0x7B, 0x7E, 0x7E, 0xFE, 0x7B, 0x7F, 0x7F, 0x7E, 
            0x7E, 0x7B, 0x7C, 0x7E, 0x7A, 0x7A, 0x7A, 0x7B, 0x7F, 0x7C, 0x7E, 0x7A, 0x7E, 0x7E, 0x7E, 0x7F, 
            0x7E, 0x7E, 0x7E, 0x7C, 0x7E, 0x7A, 0x7A, 0x7C, 0x7A, 0x7C, 0x7E, 0x7C, 0x7A, 0x7C, 0xFC, 0x7C, 
            0x78, 0x7A, 0xF8, 0x7C, 0x78, 0x7F, 0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 
            0x7E, 0x7A, 0x7E, 0x7F, 0x7B, 0x7C, 0x7E, 0x7C, 0x7E, 0x79, 0xFF, 0x7C, 0x7E, 0x7C, 0xF8, 0x7E, 
            0x7E, 0x7A, 0xFE, 0x7A, 0x7E, 0x7A, 0x7A, 0x7A, 0x7A, 0x7E, 0x7E, 0xFC, 0x7C, 0xFE, 0x7A, 0x7A, 
            0x7A, 0x78, 0x7A, 0x7E, 0x7E, 0x7A, 0xFC, 0xFE, 0x7E, 0x7E, 0xFE, 0x7A, 0x7E, 0xFC, 0xFC, 0x7E, 
            0xFC, 0x78, 0x7A, 0x7C, 0x7C, 0x7E, 0xFC, 0xFA, 0x7B, 0xF8, 0x7A, 0xFC, 0x7F, 0x78, 0x7C, 0x7C, 
            0x7E, 0x7E, 0xFE, 0x7E, 0x7C, 0x7C, 0x7E, 0x7E, 0x7D, 0x7E, 0x7E, 0x7A, 0x7C, 0x7E, 0x7A, 0x7A, 
            0x7E, 0x7E, 0x7A, 0x7A, 0x7C, 0x78, 0x7C, 0xFE, 0x7A, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x78, 0x78, 
            0x78, 0x7A, 0x7F, 0x7E, 0x7A, 0x7E, 0x7E, 0x7A, 0x7E, 0x7A, 0x7A, 0x7A, 0x7E, 0x7A, 0x7C, 0x7E, 
            0x7E, 0x7C, 0x7E, 0x7E, 0x7E, 0x7C, 0x7E, 0x7A, 0x7C, 0x7E, 0xFA, 0x7E, 0x7E, 0x7E, 0xFE, 0x7A, 
            0x7E, 0x7C, 0xFC, 0x7E, 0x7E, 0x78, 0x7A, 0x7C, 0x7A, 0x7E, 0xFE, 0x7E, 0x7E, 0xFC, 0x7C, 0x7E, 
            0x7E, 0x7C, 0x7E, 0x7C, 0x78, 0x7A, 0xFC, 0x7C, 0x7E, 0x7E, 0xFE, 0x7E, 0x7E, 0x7E, 0xFE, 0x7E, 
            0x7C, 0x7E, 0x7E, 0x7A, 0x7A, 0x7A, 0xFA, 0x78, 0x7E, 0xFC, 0x7E, 0xFE, 0x7C, 0x7E, 0x7A, 0x78, 
            0x7A, 0x7C, 0xFE, 0x7A, 0x7A, 0x7A, 0xFC, 0x7E, 0x7E, 0x7E, 0xFE, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 
            0x7E, 0x7E, 0xFC, 0xFC, 0x7C, 0xFE, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7C, 0x7A, 0x7E, 0x7C, 0x7E, 
            0x78, 0x7A, 0x78, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x78, 0x7A, 0x7E, 0x7A, 0x7C, 0x7C, 0x7E, 0x7E, 
            0x7C, 0x78, 0x7A, 0x7A, 0x7C, 0x7E, 0x7E, 0x7E, 0x7A, 0x7A, 0xFA, 0xF8, 0x7A, 0x7E, 0xFE, 0x7E, 
            0x7E, 0x7E, 0xFE, 0x78, 0x7E, 0x7E, 0x7E, 0x7C, 0x7E, 0x7A, 0xFE, 0xFA, 0x7E, 0xFA, 0x7E, 0xFC, 
            0x7A, 0x7E, 0x7E, 0x7C, 0x7C, 0x7C, 0xFC, 0xFA, 0x7C, 0x7E, 0xFE, 0x7C, 0x7E, 0x78, 0xFE, 0x7E, 
            0x7E, 0x7E, 0x7A, 0xFE, 0x7E, 0x7C, 0xFE, 0xFE, 0x7E, 0xFC, 0x7E, 0xFC, 0x7E, 0x7E, 0x7E, 0x7C, 
            0x7E, 0x7E, 0xFA, 0xFE, 0x7C, 0x7C, 0xFA, 0x78, 0x78, 0x7C, 0xFE, 0x7E, 0x7E, 0x7A, 0x7C, 0x7C, 
            0x78, 0x7A, 0xFC, 0xF8, 0x78, 0xFA, 0xF8, 0xF8, 0x78, 0x78, 0x78, 0x78, 0x7E, 0x7C, 0xFA, 0x7E, 
            0x7C, 0x7C, 0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7C, 0x7E, 0x7E, 0x7C, 0x7C, 
            0x7C, 0x7C, 0x7C, 0xFA, 0x7C, 0x7A, 0x7C, 0x7E, 0x78, 0x7E, 0x7E, 0x7C, 0x78, 0x7C, 0x78, 0x7A, 
            0x7C, 0x7A, 0x7C, 0x7C, 0x78, 0x7A, 0x7E, 0x7C, 0x78, 0x78, 0x78, 0x7E, 0x7E, 0x7C, 0x7A, 0x7C, 
            0x7C, 0x7E, 0x7E, 0x7A, 0x7C, 0x78, 0xFC, 0xF8, 0x7C, 0x7E, 0xFA, 0x7A, 0x78, 0x7A, 0xFC, 0x7E, 
            0xFE, 0x7A, 0x7E, 0x7E, 0x7C, 0x7E, 0xF8, 0xFC, 0x7C, 0xFA, 0x7C, 0xFC, 0x7E, 0x7E, 0x7C, 0x7C, 
            0x7C, 0x7E, 0xFA, 0xFE, 0x7C, 0x7E, 0xFA, 0x7C, 0x7C, 0x7C, 0xFC, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 
            0x7C, 0x78, 0xFC, 0xF8, 0x7C, 0xFE, 0x7E, 0xFC, 0x7E, 0x7C, 0x7E, 0x7E, 0x78, 0x7E, 0xFA, 0xFA, 
            0x78, 0x78, 0xFA, 0x78, 0x78, 0x7A, 0xFE, 0x78, 0x7E, 0x7C, 0x7C, 0x7A, 0x7A, 0x7E, 0xFE, 0xFA, 
            0x7C, 0xFA, 0xFC, 0xFE, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x78, 0xFC, 0x7E, 0x7E, 0x7C, 0xFC, 0x7C, 
            0x7C, 0x7A, 0xFE, 0x7E, 0x7C, 0x7C, 0x7A, 0x7C, 0x7C, 0x7E, 0x7E, 0xF8, 0x7C, 0x7E, 0x7E, 0x7A, 
            0x78, 0x78, 0x7C, 0x7A, 0x7C, 0x7A, 0xFA, 0xFA, 0x7A, 0x78, 0xFA, 0x7E, 0x7E, 0x7E, 0xFC, 0x7E, 
            0x7E, 0x7C, 0x7C, 0xFC, 0x7C, 0x7C, 0xFE, 0xF8, 0x7A, 0xFE, 0x78, 0xFA, 0x7E, 0x7C, 0x7A, 0xFE, 
            0x7E, 0x7E, 0xFE, 0xFF, 0x7D, 0x7D, 0xFD, 0x7F, 0x7D, 0x7F, 0xFF, 0x7F, 0x7D, 0x78, 0x7C, 0x7E, 
            0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0xF8, 0x7E, 0xFB, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 0xFF, 0xFF, 
            0x7F, 0x7E, 0xFE, 0x7A, 0x7E, 0x78, 0x7C, 0x7E, 0x7C, 0x7E, 0x7A, 0x7C, 0x7E, 0x7E, 0x7C, 0x7E, 
            0x7E, 0xFE, 0x7E, 0xFC, 0x7A, 0x7E, 0x7E, 0x7A, 0x7A, 0x7A, 0xFC, 0xFC, 0x7E, 0x7E, 0xFE, 0x7E, 
            0x7E, 0x7E, 0xFE, 0x78, 0x78, 0x7C, 0x7E, 0x7C, 0x7E, 0x78, 0x78, 0xFE, 0x7A, 0xFE, 0x7A, 0xF8, 
            0x7A, 0x78, 0x78, 0x7C, 0x7E, 0x7A, 0xFC, 0x78, 0x7E, 0x7C, 0xFC, 0x7A, 0x78, 0x78, 0xFE, 0x7E, 
            0x7E, 0x7E, 0x7A, 0x7E, 0x7C, 0x7C, 0x7C, 0xFC, 0x7C, 0xFC, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x78, 
            0x7A, 0x7A, 0xFC, 0xFE, 0x7C, 0x7E, 0xFE, 0x7E, 0x7E, 0x7C, 0xF8, 0x78, 0x7A, 0x78, 0x7A, 0xF8, 
            0x78, 0x78, 0xFA, 0x78, 0x78, 0xF8, 0x78, 0xFA, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0xFA, 0x78, 
            0x7A, 0x78, 0xFA, 0x7A, 0x7A, 0x78, 0xF8, 0x7A, 0x7A, 0x7A, 0x78, 0x7A, 0x78, 0x78, 0xFA, 0x78, 
            0x7A, 0x78, 0x78, 0xF9, 0x7A, 0x7B, 0x7A, 0x78, 0x7A, 0x7A, 0xFA, 0xF8, 0x7A, 0x7A, 0xF8, 0x7B, 
            0x7A, 0x7A, 0xF8, 0x7B, 0x7A, 0x78, 0x78, 0x79, 0x78, 0x78, 0x7A, 0x7B, 0x78, 0x7A, 0x7A, 0xFB, 
            0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0xFA, 0xFB, 0x78, 0x78, 0xFA, 0x7A, 0x7A, 0x78, 0xFA, 0x7B, 
            0x78, 0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0xFA, 0xFB, 0x7A, 0xF8, 0x78, 0xF8, 0x78, 0x78, 0x78, 0x7B, 
            0x78, 0x78, 0xFA, 0xF8, 0x7A, 0x7A, 0xFA, 0x78, 0x78, 0x7A, 0xF8, 0x78, 0x7A, 0x78, 0x7A, 0x78, 
            0x78, 0x7A, 0xFA, 0x7B, 0x7A, 0xFA, 0x7A, 0xF8, 0x7A, 0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0xFA, 0xFA, 
            0x78, 0x7B, 0xFA, 0x78, 0xF9, 0xF8, 0xFA, 0xFA, 0x7A, 0x7A, 0x78, 0x78, 0x7A, 0x79, 0xF8, 0xFA, 
            0x79, 0xFB, 0x79, 0xFB, 0x79, 0x7B, 0x7B, 0xF8, 0x7A, 0x78, 0xF8, 0x7A, 0x78, 0x78, 0xF8, 0x78, 
            0x78, 0x7A, 0xF8, 0x7A, 0x7B, 0x7B, 0x7B, 0x7B, 0x7B, 0x79, 0xFB, 0xF9, 0x7B, 0x7B, 0x7B, 0xF9, 
            0x7B, 0x7B, 0x79, 0x78, 0x7A, 0x78, 0xF8, 0xFB, 0x79, 0x7A, 0xF8, 0x78, 0x78, 0x78, 0xF8, 0x7A, 
            0x78, 0x78, 0x78, 0x7B, 0x7B, 0x78, 0xF8, 0xFA, 0x78, 0xFA, 0x78, 0xFA, 0xFA, 0xFA, 0x7A, 0x7A, 
            0x78, 0xFA, 0xFA, 0x7A, 0x7A, 0xF9, 0xF9, 0x78, 0x78, 0x7A, 0x7A, 0xF8, 0x7A, 0xF8, 0xF8, 0xF9, 
            0x79, 0x7A, 0x7A, 0x78, 0x78, 0x78, 0x78, 0x7A, 0x78, 0x7B, 0x79, 0xF8, 0x78, 0x7C, 0xFB, 0x78, 
            0x7E, 0x7D, 0x7D, 0x7C, 0xF8, 0x78, 0x7C, 0xF8, 0x7E, 0x7C, 0x78, 0x78, 0x7C, 0x7A, 0x7E, 0x7E, 
            0x7A, 0x7C, 0x7E, 0x7C, 0x7C, 0x7D, 0x7D, 0x7D, 0x7D, 0x7E, 0x7C, 0x7A, 0x7A, 0x78, 0x7C, 0x78, 
            0x7A, 0x7C, 0x7C, 0x78, 0x7E, 0x7C, 0x7A, 0x7E, 0x7A, 0x7E, 0x7E, 0x7E, 0x78, 0x7A, 0x7E, 0x7E, 
            0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7E, 0x7B, 0x79, 0x7A, 0x7C, 0x7C, 0x7E, 0x78, 0x7C, 
            0x7E, 0x7E, 0x7C, 0x7C, 0x7C, 0x7C, 0x7E, 0x78, 0x7A, 0x7E, 0x78, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 
            0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7A, 0x7E, 0x7E, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x7A, 0x7A, 0x78, 
            0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7E, 0x7C, 0x7C, 0x7C, 0x7E, 0x7E, 0x7E, 0x7A, 0x7A, 0x7A, 0x7A, 
            0x7A, 0x7A, 0x7A, 0x7B, 0x78, 0x7B, 0x7A, 0x78, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x78, 
            0x7A, 0x7A, 0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0x78, 0x7A, 
            0x78, 0x78, 0x78, 0x7A, 0x7A, 0x78, 0x78, 0x78, 0x78, 0x7A, 0x78, 0x78, 0x78, 0x78, 0x78, 0x78, 
            0x78, 0x78, 0x7C, 0x78, 0x7C, 0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7A, 0x7A, 0x7C, 0x78, 
            0x7E, 0x7A, 0x7C, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x78, 0x7E, 
            0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7E, 0x7A, 0x7C, 0x7E, 0x7C, 0x7A, 0xFA, 0x7A, 0xF8, 0x78, 
            0x7E, 0x7E, 0xFE, 0x7E, 0x7C, 0x7E, 0x7E, 0x7E, 0x78, 0x7C, 0x7E, 0x7A, 0x78, 0x7A, 0x7C, 0xFA, 
            0x78, 0x7C, 0x7C, 0x7C, 0x7C, 0x7E, 0x7A, 0x7E, 0x7C, 0x7C, 0xFE, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
            0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7E, 0x7C, 0x7E, 0x7C, 0x7C, 0x7E, 0x7E, 0x7E, 0x7C, 0x7A, 
            0x7A, 0x7A, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7A, 0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7C, 
            0x7C, 0x7C, 0x78, 0x7E, 0x7E, 0x78, 0x78, 0x7C, 0x7C, 0x7C, 0x7C, 0x7E, 0x78, 0x7C, 0x7C, 0x7C, 
            0x78, 0x7C, 0x7C, 0x7A, 0x7C, 0x7A, 0x7A, 0x7E, 0x7C, 0xFA, 0xFE, 0xFA, 0xFA, 0xFE, 0xFE, 0xFE, 
            0xFE, 0xF8, 0xFE, 0x7A, 0x7C, 0x78, 0x7E, 0x7E, 0x7A, 0xFC, 0xFC, 0xF8, 0xFA, 0xFA, 0xFC, 0xFE, 
            0xFA, 0xFE, 0xFE, 0x78, 0x7A, 0x7C, 0x7E, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 
            0x7E, 0x7E, 0x7C, 0x7E, 0x7E, 0x7A, 0x7A, 0x7A, 0x7C, 0xFE, 0xFC, 0xFC, 0xFE, 0x7C, 0x7C, 0xFE, 
            0xFC, 0xFC, 0xFC, 0x7C, 0x7C, 0x7C, 0x78, 0x7C, 0x7C, 0x7C, 0x78, 0x7E, 0x7A, 0x78, 0x7C, 0x7E, 
            0x7D, 0x7B, 0x7B, 0x7B, 0x7B, 0x7F, 0x7F, 0x7F, 0x7F, 0x7B, 0x7B, 0x7F, 0x7F, 0x7F, 0x7F, 0x7D, 
            0x7F, 0xFF, 0xF9, 0x7F, 0x79, 0x7F, 0x7F, 0x79, 0x7F, 0x7B, 0x7F, 0x7F, 0x7D, 0x7F, 0x7D, 0x7F, 
            0x7B, 0x7D, 0x7B, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7D, 0x7D, 0x7B, 0x7B, 0x7F, 0x79, 0x7D, 0x7D, 
            0x79, 0x7B, 0x7D, 0x7F, 0x7F, 0x7D, 0x7F, 0x7B, 0x7F, 0x7F, 0x7F, 0x7F, 0x7D, 0x7D, 0x7D, 0x7F, 
            0xFD, 0x7D, 0x7D, 0x79, 0x7D, 0x7D, 0x7D, 0x79, 0x7D, 0x79, 0x7B, 0x7F, 0x7B, 0x79, 0x7D, 0x7D, 
            0x7F, 0x7B, 0x7B, 0x7F, 0x7F, 0x7F, 0x7B, 0x7B, 0x7F, 0x7F, 0x7D, 0x79, 0x7F, 0x7F, 0x7F, 0x7F, 
            0x7B, 0x7F, 0x7B, 0x7B, 0x7B, 0x79, 0x7B, 0x7B, 0x7F, 0x7B, 0x7B, 0x79, 0x7D, 0x7B, 0x7F, 0x7B, 
            0x7D, 0x7F, 0x7F, 0xF9, 0x7F, 0x79, 0x7F, 0xFF, 0xFF, 0xFB, 0x7B, 0x7B, 0xFF, 0x7B, 0xFD, 0x7F, 
            0x7F, 0x7A, 0xFE, 0xFA, 0xFC, 0x78, 0x7E, 0xFA, 0x7E, 0x7C, 0x7E, 0x78, 0x7C, 0xFC, 0xFC, 0x78, 
            0x7E, 0x7C, 0x7C, 0x7C, 0x7E, 0xFE, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7A, 0x7E, 0x7A, 
            0x7E, 0x7F, 0x7A, 0x7F, 0x7F, 0x7F, 0x7B, 0x7B, 0x7B, 0x7B, 0x7F, 0xFA, 0x79, 0x7D, 0x7F, 0x7F, 
            0x7F, 0x7B, 0x7F, 0x7D, 0x7F, 0x79, 0x7F, 0x7F, 0x7F, 0xFF, 0xFF, 0x7E, 0x7F, 0x7B, 0x7F, 0xF9, 
            0xFC, 0x7C, 0x7C, 0x7C, 0x7E, 0x78, 0x78, 0x7A, 0x78, 0x7C, 0x78, 0x7A, 0x7E, 0x7C, 0x7A, 0x7E, 
            0x78, 0x7E, 0x7A, 0x7A, 0x7A, 0x7E, 0x7A, 0x7E, 0x79, 0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7A, 0x7C, 
            0x78, 0x7F, 0x7E, 0x7F, 0x7E, 0x7E, 0x7C, 0x7A, 0x7B, 0x7F, 0x7F, 0x7F, 0x7F, 0x7C, 0xFC, 0x7D, 
            0x7B, 0x7A, 0xFE, 0x7C, 0x7E, 0x7D, 0x78, 0x7C, 0x79, 0x7C, 0x79, 0x7C, 0x7B, 0x79, 0xFA, 0x7A, 
            0x78, 0x7A, 0x78, 0x7E, 0x7A, 0x7A, 0x7C, 0x7E, 0x7A, 0x7E, 0x7A, 0x7A, 0x7C, 0x7A, 0x7E, 0x7E, 
            0x7E, 0x7C, 0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0xFC, 0x7C, 0x7E, 0x7E, 0x7C, 0x7E, 0x7E, 0x7A, 
            0x7C, 0x7A, 0x7C, 0x7E, 0x7C, 0x7C, 0x7C, 0x7C, 0x7E, 0x7C, 0x78, 0x78, 0x7E, 0x7E, 0x7E, 0x7A, 
            0x78, 0x7C, 0x7C, 0x7E, 0x7C, 0x7A, 0x7C, 0x7E, 0x7A, 0x7A, 0x7E, 0x7A, 0x7C, 0x7E, 0x7E, 0x7E, 
            0x7A, 0x7E, 0x7A, 0x7E, 0x7A, 0x7E, 0x7E, 0x7A, 0x7E, 0x7A, 0x78, 0x7E, 0x7A, 0x7E, 0x7E, 0x7C, 
            0x78, 0x7C, 0x78, 0x7B, 0x7B, 0x7A, 0x7A, 0x7E, 0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x78, 0x7C, 0x7C, 
            0x7C, 0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x7C, 0x7C, 0x7E, 0x7C, 0x7A, 0x78, 0x7A, 0x7E, 0x7E, 0x7E, 
            0x78, 0x7E, 0x7E, 0x7C, 0x7C, 0x7C, 0x7C, 0x78, 0x7C, 0x7C, 0x7E, 0x7E, 0x7A, 0x7E, 0x78, 0x7E, 
            0x7E, 0x7E, 0x7A, 0x7A, 0x78, 0x7C, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7A, 0x78, 0x78, 0x78, 
            0x78, 0x7A, 0x7A, 0x7B, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 
            0x7A, 0x7A, 0x7A, 0x78, 0x78, 0x78, 0x7A, 0x78, 0x7A, 0x7A, 0x78, 0x78, 0x7A, 0x78, 0x78, 0x7A, 
            0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x78, 0x78, 0x7A, 0x7A, 0x78, 0x7A, 0x78, 0x7C, 0x7E, 
            0x7A, 0x78, 0x7C, 0x7E, 0x7C, 0x7E, 0x78, 0x7C, 0x7C, 0x7C, 0x78, 0x78, 0x7E, 0x78, 0x7E, 0x7A, 
            0x7C, 0x7C, 0x7C, 0x7C, 0x7E, 0x7C, 0x7E, 0x7E, 0x7E, 0x7A, 0x7C, 0x78, 0x7E, 0x7A, 0x7A, 0x7E, 
            0x7A, 0x7E, 0x7E, 0x78, 0x78, 0x78, 0x78, 0x7A, 0x7A, 0x7A, 0x78, 0x7A, 0x78, 0x7A, 0x78, 0x7A, 
            0x7A, 0x7E, 0x78, 0x7A, 0x7A, 0x7E, 0x78, 0x7A, 0x7E, 0x7C, 0x78, 0x7C, 0x78, 0x7E, 0x7A, 0x7E, 
            0x7C, 0x7E, 0x7A, 0x7E, 0x7A, 0x7E, 0x78, 0x78, 0x7E, 0x7A, 0x7C, 0x78, 0x7C, 0x78, 0x78, 0x7C, 
            0x7E, 0x78, 0x7C, 0x7E, 0x7C, 0x7A, 0x7E, 0x78, 0x7C, 0x7C, 0x7A, 0x7C, 0x7A, 0x7C, 0x7E, 0x7E, 
            0x7E, 0x7A, 0x7E, 0x7A, 0x7A, 0x7A, 0x7E, 0x7E, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x78, 
            0x7C, 0x7C, 0x78, 0x78, 0x7C, 0x7C, 0x7A, 0x7E, 0x78, 0x7A, 0x7C, 0x7E, 0x78, 0x7E, 0x7E, 0x7E, 
            0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7A, 0x78, 0x78, 0x78, 
            0x7E, 0x7A, 0x7E, 0x7C, 0x7C, 0x7E, 0x78, 0x7E, 0x7A, 0x7C, 0x78, 0x7C, 0x78, 0x7C, 0x7C, 0x7C, 
            0x7A, 0x7E, 0x7C, 0x78, 0x7E, 0x78, 0x7C, 0x7C, 0x7E, 0x7E, 0x7E, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 
            0x7A, 0x7C, 0x7E, 0x7A, 0x7E, 0x7E, 0x78, 0x7E, 0x7A, 0x7A, 0x7A, 0x78, 0x78, 0x78, 0x78, 0x78, 
            0x78, 0x78, 0x78, 0x7A, 0x78, 0x78, 0x78, 0x78, 0x78, 0x78, 0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 
            0x78, 0x7A, 0x7A, 0x78, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x78, 0x78, 0x7A, 0x7A, 0x78, 0x78, 0x78, 
            0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x78, 0x7A, 0x78, 0x7A, 0x7A, 0x78, 0x7C, 0x7C, 0x7C, 0x7C, 
            0x78, 0x7A, 0x7C, 0x7C, 0x7E, 0x7A, 0x7A, 0x7E, 0x7A, 0xFE, 0xFA, 0xFC, 0xFE, 0xFC, 0xFE, 0xF8, 
            0xF8, 0xFE, 0xFE, 0x7E, 0x7C, 0x7E, 0x7E, 0x7F, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0xFA, 0xFA, 0xFE, 
            0xFE, 0xFE, 0xFA, 0xFC, 0xF8, 0xFA, 0xFA, 0x7E, 0x7C, 0x7E, 0x7A, 0x7A, 0x7A, 0x7E, 0x7A, 0x7E, 
            0x7A, 0x7E, 0x7E, 0x7E, 0x7C, 0x7E, 0x7C, 0x7C, 0x7C, 0x7E, 0x7C, 0x7C, 0x7C, 0x7C, 0x7E, 0x7C, 
            0x7A, 0x7E, 0x7E, 0x7E, 0x78, 0x7E, 0x7A, 0x78, 0x7E, 0x7E, 0x7E, 0x7C, 0x7A, 0x7E, 0x7A, 0x7E, 
            0x7C, 0x7A, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7E, 0x7A, 0x7A, 0x7E, 0x7A, 0x7E, 0x7E, 0x7E, 0x7C, 
            0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x7C, 0x78, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x7E, 0x7A, 0x7E, 
            0x7E, 0x7E, 0x7A, 0x7E, 0x7E, 0x7C, 0x7A, 0x7E, 0x7A, 0x7A, 0x7E, 0x7C, 0x7E, 0x7A, 0x7A, 0x7E, 
            0x7E, 0x7C, 0x7E, 0x7E, 0x7C, 0x78, 0x7C, 0x7A, 0x7E, 0x7E, 0x7E, 0x7C, 0x78, 0x7E, 0x78, 0x7C, 
            0x7E, 0x7C, 0x7E, 0x7C, 0x7E, 0x78, 0x7E, 0x7C, 0x7C, 0x7A, 0x7E, 0x7E, 0x7C, 0x7C, 0x7E, 0x7E, 
            0x7E, 0x7A, 0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0xFC, 0x7E, 0xFE, 0x7E, 0x7A, 0x7E, 0x7E, 
            0x7E, 0x78, 0x7C, 0x7E, 0x7E, 0xFE, 0x7E, 0x78, 0x7E, 0xFE, 0x7C, 0x7C, 0x7C, 0xFE, 0x78, 0x7C, 
            0x7C, 0xFE, 0x7A, 0x7A, 0x7E, 0xFE, 0x7E, 0x78, 0x7E, 0x78, 0x7E, 0x7E, 0x7C, 0xFE, 0x7E, 0x7C, 
            0x7E, 0x7E, 0x78, 0x7A, 0x7A, 0xF8, 0x7A, 0x78, 0x7A, 0x78, 0x78, 0x78, 0x78, 0xF8, 0x78, 0x7A, 
            0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0xF8, 0x7A, 0x7A, 0x78, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x78, 
            0x78, 0x78, 0x78, 0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0xFA, 0x7A, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 
            0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x78, 0x78, 0xFA, 0x7A, 0x7A, 0xF8, 0x78, 0x7A, 0x7A, 0x7E, 0x78, 
            0x7A, 0x7E, 0x78, 0xFC, 0x7A, 0x7C, 0xFC, 0xFC, 0xF8, 0xFE, 0xFC, 0xF8, 0xF8, 0x7C, 0x7E, 0x7E, 
            0x78, 0x7E, 0x7E, 0x7A, 0x7E, 0x7C, 0x7C, 0x7C, 0x7E, 0x78, 0xF8, 0xFA, 0xFC, 0xFC, 0xF8, 0xF8, 
            0xFC, 0xFA, 0xFA, 0xFE, 0xFA, 0xFC, 0xFC, 0xFE, 0xFC, 0x7E, 0x7E, 0x7E, 0x7E, 0xF8, 0xFE, 0x7A, 
            0x78, 0x7A, 0x7C, 0x78, 0x78, 0x7E, 0x78, 0x7E, 0x78, 0xFE, 0xFA, 0x78, 0x7E, 0x7C, 0x7A, 0x7A, 
            0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x7C, 0x7A, 0x7C, 0x7C, 0x7C, 0x78, 0xF8, 0xFE, 0x7E, 0x78, 0x7C, 
            0x7C, 0x7E, 0x7C, 0x7E, 0x7C, 0xFA, 0xF8, 0x7C, 0x78, 0x7E, 0x7E, 0x7C, 0x7E, 0x7A, 0x7E, 0xFA, 
            0xFC, 0x7C, 0x7C, 0x7E, 0xFE, 0x7A, 0x7E, 0xFC, 0xFE, 0x7A, 0x7A, 0x7C, 0x7E, 0x7E, 0x7C, 0x7E, 
            0x7C, 0x7A, 0x78, 0x78, 0x7C, 0x7E, 0x7E, 0x7A, 0x7C, 0x7C, 0x7C, 0xFC, 0xFA, 0xFC, 0xFE, 0x7E, 
            0x7A, 0x7C, 0x7A, 0x7A, 0x7C, 0x7E, 0x7A, 0x7E, 0x7E, 0x7E, 0x78, 0x7E, 0x7A, 0x7E, 0x7A, 0x78, 
            0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7E, 0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x7C, 0x7C, 0x78, 0xFC, 0xFA, 
            0x7C, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x78, 0x7E, 0x7E, 0x78, 0x7C, 0x7E, 0x7E, 0x7A, 0x7E, 0x7C, 
            0x7C, 0x7C, 0x7A, 0x7A, 0x78, 0x7E, 0x7E, 0x7E, 0x7A, 0x7A, 0x7E, 0x7A, 0x7E, 0x7E, 0x7C, 0x78, 
            0x78, 0x78, 0x7C, 0x7A, 0x7A, 0x7C, 0x7E, 0x7E, 0x7A, 0x7A, 0x7E, 0x7E, 0x7C, 0x7C, 0x7A, 0x7A, 
            0x7E, 0x7E, 0x7E, 0x7E, 0x7C, 0x7A, 0x7E, 0x78, 0x7E, 0xF8, 0x7A, 0xFE, 0x7E, 0x7E, 0x7C, 0x7A, 
            0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x78, 0x7E, 0x78, 0x7A, 0x78, 0x7A, 
            0x7E, 0x7E, 0x7A, 0x7A, 0x7C, 0x7A, 0x7E, 0x7C, 0x7E, 0x7A, 0x7E, 0x7E, 0x7C, 0x7A, 0x7E, 0x7A, 
            0x7C, 0x7E, 0x7A, 0x7E, 0x7C, 0x7C, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7E, 0x7E, 0x78, 0x7C, 0x7E, 
            0x7A, 0x7A, 0x7A, 0x7E, 0x7A, 0x7E, 0x7E, 0x7C, 0x7A, 0x7C, 0x7C, 0x78, 0x7C, 0x7C, 0x7E, 0x7E, 
            0x7C, 0x7A, 0x78, 0x78, 0x7A, 0x78, 0x7A, 0x78, 0x7A, 0x7C, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7C, 
            0x7E, 0x7E, 0x7C, 0x7A, 0x7E, 0x7A, 0x7C, 0x7C, 0x7E, 0x78, 0x7E, 0x7C, 0x7C, 0x7C, 0x7E, 0x78, 
            0x7C, 0x7C, 0x7E, 0x7A, 0x78, 0x78, 0x7C, 0x78, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0x7E, 0x7E, 0x78, 
            0x78, 0x7A, 0x78, 0x7E, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x78, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 
            0x78, 0x7E, 0x7A, 0x7C, 0x7C, 0x7A, 0x7A, 0x7A, 0x7E, 0x7C, 0x7E, 0x7A, 0x7A, 0x7E, 0x78, 0x7A, 
            0x7A, 0x7A, 0x7E, 0x7E, 0x7A, 0x7A, 0x78, 0x7E, 0x78, 0x7E, 0x7E, 0x7E, 0x7A, 0x78, 0x7E, 0x7E, 
            0x78, 0x7E, 0x7E, 0x7C, 0x78, 0x7A, 0x78, 0x7C, 0x7C, 0x7C, 0x7C, 0x78, 0x7E, 0x7E, 0x78, 0x7E, 
            0x7A, 0x7E, 0x7E, 0x7C, 0x7E, 0x7C, 0x7C, 0x7C, 0x7A, 0x7A, 0x7C, 0x78, 0x7C, 0x78, 0x7C, 0x7E, 
            0x7E, 0x7E, 0x7C, 0x7E, 0x7E, 0x7A, 0x7C, 0x7A, 0x7E, 0x7A, 0x7E, 0x7C, 0x7E, 0x7C, 0x7C, 0x7E, 
            0x7A, 0x7A, 0x7A, 0x7A, 0x7E, 0x7C, 0x7C, 0x7A, 0x7C, 0x7C, 0x7A, 0x7A, 0x7E, 0x7C, 0x7E, 0x78, 
            0x78, 0x7A, 0x78, 0x7E, 0x7E, 0x7A, 0x7A, 0x7E, 0x7A, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 0x7E, 0x7E, 
            0x7E, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0xFA, 0xFA, 0xF8, 0xFA, 0xF8, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 
            0xFA, 0xFA, 0xFA, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 
            0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xF8, 0xFA, 0xFA, 0xF8, 0xFA, 0xFA, 0xFA, 
            0xFA, 0xFA, 0xFA, 0xFA, 0xF8, 0xF8, 0xF8, 0xF8, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xF8, 
            0xFA, 0xF8, 0xFA, 0xF8, 0xFA, 0xF8, 0xFA, 0xFA, 0xFA, 0xF8, 0xFA, 0xF8, 0xFA, 0xFA, 0xFA, 0xF8, 
            0xFA, 0xFA, 0xF8, 0xF8, 0xFA, 0xFA, 0xF8, 0xF8, 0xFA, 0xFA, 0xF8, 0xFA, 0xFA, 0xF8, 0xFA, 0xFA, 
            0xFA, 0xF8, 0xFA, 0xFA, 0xF8, 0xFA, 0xFA, 0xF8, 0xFA, 0xFA, 0xFA, 0xFE, 0xFE, 0xFE, 0xF8, 0xFE, 
            0xFC, 0xFA, 0xFA, 0xFE, 0xFA, 0xFA, 0xF8, 0xFE, 0xFC, 0xFE, 0xFC, 0xFC, 0xFC, 0xFF, 0xF8, 0xFC, 
            0xFC, 0xFE, 0xFE, 0xFE, 0xFA, 0xFE, 0xFC, 0x78, 0x7A, 0x7C, 0x7E, 0x7C, 0x7F, 0x7E, 0x7C, 0x7C, 
            0x7B, 0x7A, 0x7B, 0x7B, 0x79, 0x7B, 0x7F, 0x79, 0x7F, 0x7B, 0x7F, 0x7F, 0x7B, 0x77, 0x75, 0x77, 
            0x73, 0x77, 0xF3, 0xF5, 0x7F, 0x7D, 0xFD, 0xFB, 0x7F, 0xFF, 0xFB, 0xFD, 0x7C, 0xF8, 0x7E, 0x7C, 
            0x7A, 0x7B, 0xF9, 0xFF, 0xFD, 0x7F, 0x7F, 0xFF, 0xFB, 0x7D, 0x7B, 0x79, 0x7F, 0x7F, 0xFD, 0x7B, 
            0x7B, 0x7D, 0xFB, 0xFD, 0xFE, 0x7C, 0x7E, 0x7C, 0x7E, 0x7E, 0x7A, 0x7C, 0x7A, 0x7A, 0x7A, 0x78, 
            0x7E, 0x7E, 0x7C, 0x7A, 0x7A, 0x7C, 0x78, 0x7E, 0x7E, 0x7C, 0x7E, 0x78, 0x7E, 0x7E, 0xFE, 0x7E, 
            0x7E, 0x7E, 0x7C, 0x7C, 0x78, 0x7C, 0x7C, 0x7E, 0x7A, 0x7A, 0x7B, 0x7A, 0x7A, 0x7E, 0x7C, 0x7D, 
            0x7C, 0x7E, 0x7E, 0x7A, 0x78, 0x7E, 0x7C, 0x7D, 0x7E, 0x7F, 0x7E, 0x7E, 0x78, 0x7E, 0xFE, 0x7C, 
            0x78, 0x7C, 0x79, 0x7C, 0x7A, 0x7E, 0x7E, 0x7E, 0x7C, 0x7B, 0x7E, 0x7E, 0x7E, 0x7C, 0x7C, 0x7E, 
            0x7E, 0x7E, 0x7E, 0x7A, 0x78, 0x7A, 0x7E, 0x7A, 0x7E, 0x72, 0x75, 0x72, 0x74, 0x7E, 0x76, 0x74, 
            0x7D, 0x7C, 0x74, 0x76, 0x7E, 0x7E, 0x76, 0x76, 0x7E, 0x7E, 0x76, 0x73, 0x76, 0x7E, 0x7E, 0x77, 
            0x76, 0x72, 0x75, 0x76, 0x72, 0x76, 0x72, 0x76, 0x7C, 0x74, 0x7E, 0x7E, 0x74, 0x76, 0x74, 0x76, 
            0x74, 0x74, 0x74, 0x79, 0x75, 0x74, 0x72, 0x74, 0x76, 0x77, 0x70, 0x74, 0x73, 0x77, 0x76, 0x72, 
            0x77, 0x77, 0x76, 0x76, 0x77, 0x77, 0x76, 0x76, 0x74, 0x75, 0x73, 0x76, 0x74, 0x74, 0x74, 0x76, 
            0x74, 0x74, 0x74, 0x74, 0x77, 0x74, 0x71, 0x77, 0x74, 0x76, 0x76, 0x74, 0x74, 0x72, 0x74, 0x77, 
            0x70, 0x76, 0x74, 0x74, 0x74, 0x76, 0x76, 0x70, 0x76, 0x74, 0x74, 0x74, 0x76, 0x76, 0x76, 0x76, 
            0x76, 0x76, 0x70, 0x76, 0x76, 0x76, 0x76, 0x74, 0x74, 0x76, 0x76, 0x76, 0x76, 0x76, 0x70, 0x74, 
            0x72, 0x76, 0x72, 0x76, 0x72, 0x76, 0x76, 0x74, 0x74, 0x76, 0x76, 0x76, 0x70, 0x76, 0x72, 0x74, 
            0x74, 0x76, 0x72, 0x74, 0x74, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 
            0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 0x72, 0x72, 0x70, 0x70, 0x70, 0x70, 0x72, 0x72, 
            0x72, 0x72, 0x70, 0x72, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x72, 0x70, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x70, 0x70, 0x70, 
            0x70, 0x70, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 0x72, 0x76, 0x72, 0x72, 0x70, 0x74, 
            0x74, 0x70, 0x74, 0x74, 0x70, 0x74, 0x74, 0x70, 0x74, 0x74, 0x70, 0x7A, 0x76, 0x76, 0x76, 0x76, 
            0x76, 0x76, 0x74, 0x76, 0x76, 0x72, 0x70, 0x76, 0x76, 0x76, 0x74, 0x72, 0x72, 0x76, 0x7A, 0x76, 
            0x76, 0x74, 0x74, 0x76, 0x72, 0x74, 0x74, 0x74, 0x70, 0x78, 0x74, 0x74, 0x70, 0x76, 0x70, 0x76, 
            0x70, 0x72, 0x72, 0x7E, 0x74, 0x72, 0x76, 0x71, 0x70, 0x72, 0x74, 0x76, 0x76, 0x76, 0x72, 0x70, 
            0x72, 0x74, 0x76, 0x70, 0x76, 0x72, 0x76, 0x70, 0x74, 0x74, 0x71, 0x72, 0x72, 0x70, 0x74, 0x70, 
            0x72, 0x72, 0x76, 0x76, 0x76, 0x75, 0x77, 0x76, 0x70, 0x76, 0x74, 0x76, 0x76, 0x70, 0x76, 0x73, 
            0x76, 0x74, 0x76, 0x76, 0x76, 0x76, 0x74, 0x72, 0x7A, 0x7A, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x78, 
            0x78, 0x7C, 0x70, 0x76, 0x74, 0x76, 0x76, 0x72, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7E, 0x7A, 0x7E, 
            0x7C, 0x7E, 0x76, 0x74, 0x72, 0x76, 0x76, 0x76, 0x72, 0x70, 0x72, 0x70, 0x76, 0x74, 0x72, 0x76, 
            0x76, 0x72, 0x70, 0x72, 0x75, 0x77, 0x77, 0x73, 0x7D, 0x7D, 0x7B, 0x7D, 0x75, 0x71, 0x76, 0x74, 
            0x74, 0x74, 0x74, 0x72, 0x75, 0x71, 0x75, 0x73, 0x77, 0x75, 0x77, 0x75, 0x73, 0x73, 0x76, 0x74, 
            0x76, 0x76, 0x74, 0x74, 0x76, 0x74, 0x76, 0x76, 0x74, 0x74, 0x74, 0x74, 0x76, 0x74, 0x76, 0x72, 
            0x70, 0x76, 0x72, 0x76, 0x77, 0x73, 0x77, 0x77, 0x74, 0x76, 0x70, 0x72, 0x76, 0x76, 0x76, 0x76, 
            0x74, 0x74, 0x76, 0x74, 0x74, 0x74, 0x70, 0x72, 0x76, 0x72, 0x7E, 0x74, 0x7E, 0x74, 0x76, 0x76, 
            0x70, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x70, 0x74, 0x74, 0x74, 0x76, 0x76, 0x72, 0x72, 
            0x74, 0x76, 0x76, 0x72, 0x76, 0x76, 0x76, 0x76, 0x74, 0x72, 0x76, 0x74, 0x76, 0x76, 0x74, 0x76, 
            0x74, 0x76, 0x72, 0x70, 0x74, 0x72, 0x71, 0x72, 0x7B, 0x70, 0x78, 0x7A, 0x72, 0x70, 0x72, 0x78, 
            0x7A, 0x70, 0x72, 0x7C, 0x72, 0x70, 0x76, 0x76, 0x74, 0x7A, 0x76, 0x76, 0x70, 0x72, 0x74, 0x70, 
            0x70, 0x70, 0x7E, 0x72, 0x72, 0x7A, 0x7E, 0x7A, 0x7C, 0x70, 0x76, 0x76, 0x72, 0x7A, 0x70, 0x78, 
            0x7E, 0x7C, 0x78, 0x7E, 0x7D, 0x72, 0x7B, 0x77, 0x76, 0x72, 0x70, 0x75, 0x7F, 0x78, 0x72, 0x71, 
            0x74, 0x74, 0x72, 0x72, 0x72, 0x78, 0x70, 0x72, 0x72, 0x72, 0x70, 0x76, 0x70, 0x74, 0x75, 0x72, 
            0x70, 0x73, 0x71, 0x73, 0x71, 0x70, 0x76, 0x70, 0x70, 0x73, 0x7A, 0x73, 0x73, 0x73, 0x73, 0x73, 
            0x73, 0x70, 0x73, 0x70, 0x72, 0x72, 0x72, 0x72, 0x7B, 0x7B, 0x72, 0x70, 0x70, 0x72, 0x7A, 0x78, 
            0x72, 0x71, 0x7A, 0x78, 0x70, 0x70, 0x70, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x70, 0x70, 
            0x70, 0x70, 0x72, 0x72, 0x70, 0x72, 0x73, 0x70, 0x72, 0x72, 0x7A, 0x78, 0x70, 0x70, 0x78, 0x7A, 
            0x72, 0x72, 0x72, 0x72, 0x73, 0x71, 0x70, 0x70, 0x72, 0x72, 0x73, 0x71, 0x72, 0x72, 0x73, 0x71, 
            0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 0x72, 0x7A, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x71, 0x73, 0x72, 0xF2, 0x73, 0x73, 0x70, 0x72, 0x72, 0x70, 
            0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 0x7A, 0x70, 0x72, 0x70, 0x70, 0x72, 0xF2, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 0x70, 0x70, 0x70, 0x71, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 
            0x72, 0x72, 0x70, 0x70, 0x70, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 
            0x72, 0x72, 0x72, 0x73, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 
            0x70, 0x70, 0x70, 0x72, 0x70, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x70, 0x72, 0x70, 0x70, 0x72, 0x72, 0x72, 0x72, 0x7A, 0x72, 0x72, 0x72, 0x76, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 
            0x70, 0x76, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 0x76, 0x74, 0x70, 0x72, 
            0x70, 0x70, 0x70, 0x72, 0x72, 0x70, 0x77, 0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 
            0x72, 0x70, 0x70, 0x70, 0x72, 0x70, 0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 
            0x72, 0x70, 0x72, 0x70, 0x72, 0x70, 0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 
            0x70, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 0x70, 0x72, 0x76, 0x76, 0x76, 0x74, 0x74, 
            0x74, 0x76, 0x74, 0x74, 0x74, 0x72, 0x70, 0x70, 0x72, 0x72, 0x70, 0x76, 0x74, 0x76, 0x76, 0x76, 
            0x76, 0x76, 0x74, 0x76, 0x76, 0x70, 0x70, 0x70, 0x70, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 
            0x70, 0x70, 0x72, 0x70, 0x70, 0x70, 0x70, 0x72, 0x70, 0x72, 0x72, 0x74, 0x76, 0x74, 0x76, 0x72, 
            0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 
            0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x70, 0x70, 0x70, 0x72, 0x72, 0x72, 0x70, 0x70, 0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 
            0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x70, 0x72, 0x72, 0x70, 0x76, 0x72, 0x76, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 0x70, 0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 0x72, 0x70, 0x70, 0x70, 0x72, 0x70, 
            0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x76, 0x72, 0x76, 0x76, 0x70, 
            0x70, 0x70, 0x76, 0x76, 0x70, 0x72, 0x76, 0x70, 0x72, 0x72, 0x72, 0x70, 0x74, 0x76, 0x70, 0x72, 
            0x72, 0x70, 0x72, 0x72, 0x72, 0x74, 0x72, 0x70, 0x76, 0x76, 0x76, 0x74, 0x72, 0x72, 0x70, 0x72, 
            0x72, 0x70, 0x76, 0x74, 0x76, 0x76, 0x74, 0x74, 0x70, 0x74, 0x70, 0x70, 0x70, 0x70, 0x72, 0x76, 
            0x76, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x76, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0x7E, 0x7A, 0x7A, 
            0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x78, 0x78, 0x7A, 0x7A, 0x7A, 0x7A, 0x7C, 0x76, 0x70, 0x72, 0x70, 
            0x70, 0x74, 0x74, 0x70, 0x70, 0x76, 0x76, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 
            0x70, 0x72, 0x72, 0x70, 0x71, 0x73, 0x71, 0x71, 0x73, 0x73, 0x73, 0x71, 0x71, 0x77, 0x77, 0x73, 
            0x73, 0x75, 0x77, 0x73, 0x73, 0x71, 0x71, 0x73, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 
            0x74, 0x70, 0x70, 0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 0x72, 0x70, 0x72, 0x76, 0x72, 0x72, 0x70, 0x70, 0x72, 
            0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x70, 0x70, 0x70, 0x70, 0x72, 0x70, 0x70, 
            0x70, 0x70, 0x70, 0x70, 0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x7A, 0x7A, 0x7A, 0x7A, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0xF2, 0x7A, 0x70, 0x72, 0x78, 0x7A, 0x72, 0x72, 0x78, 0x78, 0x72, 0x72, 0x7A, 
            0x7A, 0x72, 0xF2, 0x78, 0x7A, 0x7A, 0xF2, 0x72, 0x78, 0x72, 0x70, 0x7A, 0x78, 0x7C, 0x70, 0x70, 
            0x78, 0x70, 0x70, 0x78, 0x79, 0x73, 0x73, 0x7B, 0x7A, 0x72, 0x72, 0x7A, 0x7A, 0x72, 0x72, 0x7A, 
            0x7B, 0x72, 0x70, 0x7B, 0x7B, 0x72, 0x72, 0xF9, 0x73, 0x72, 0x7A, 0x73, 0x71, 0x72, 0x72, 0x73, 
            0x73, 0x71, 0x72, 0x7A, 0x73, 0x72, 0x70, 0x73, 0xF3, 0x73, 0x72, 0x72, 0x71, 0x72, 0x70, 0x73, 
            0x73, 0x72, 0x72, 0x71, 0x73, 0x70, 0x70, 0xF1, 0xF3, 0xF0, 0xF0, 0xF1, 0xF3, 0xF2, 0xF2, 0xF1, 
            0xF3, 0xF2, 0x70, 0x73, 0xF2, 0xF0, 0x71, 0x72, 0xF0, 0xF2, 0x70, 0x72, 0xF2, 0xF0, 0x70, 0x71, 
            0x72, 0xF0, 0xF0, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 0x72, 0x72, 0xF2, 0x72, 0xF2, 0xF2, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0xF2, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 0x72, 
            0x72, 0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x7A, 0x78, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 
            0x72, 0x72, 0x70, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x7A, 0x7A, 0x72, 0x70, 0x72, 0x72, 
            0x70, 0x72, 0x70, 0x72, 0x78, 0x7A, 0x70, 0x72, 0x72, 0x70, 0x70, 0x70, 0x70, 0x70, 0x7A, 0x7A, 
            0xF2, 0x72, 0x72, 0x7A, 0x71, 0x73, 0x7A, 0x7A, 0x72, 0x72, 0x70, 0xF2, 0x70, 0xF0, 0xF0, 0xF2, 
            0xF2, 0x70, 0x72, 0x70, 0x72, 0x70, 0x73, 0x73, 0x70, 0x72, 0x70, 0x70, 0x72, 0x72, 0x72, 0x72, 
            0x73, 0x73, 0x72, 0x72, 0x72, 0x72, 0x72, 0x7A, 0x72, 0x72, 0x73, 0x73, 0x70, 0x72, 0x70, 0x73, 
            0x70, 0x72, 0x71, 0x73, 0x72, 0x70, 0x72, 0x72, 0xF2, 0x70, 0x72, 0x72, 0x72, 0x7A, 0x7A, 0x72, 
            0x72, 0x72, 0x72, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x72, 0x70, 0x72, 0x70, 0x70, 
            0x70, 0x70, 0x70, 0x73, 0x72, 0x72, 0x70, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x70, 0x71, 0x73, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x7A, 0x72, 0x78, 0x70, 0x72, 0x72, 0x72, 0x72, 
            0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 0xF2, 
            0x72, 0x70, 0x70, 0x70, 0x72, 0xF2, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x70, 0x72, 0xF2, 
            0x72, 0x70, 0xFA, 0x72, 0x71, 0x7A, 0x71, 0x7A, 0x7A, 0x72, 0x70, 0x7E, 0x7C, 0xFC, 0xFC, 0x76, 
            0x7C, 0x7C, 0xF6, 0x76, 0x7E, 0x7E, 0x76, 0xF6, 0x76, 0x74, 0x74, 0x76, 0x76, 0x7E, 0x7E, 0x72, 
            0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x7A, 0x72, 0x78, 0x78, 0x70, 0x7A, 0x73, 0x72, 
            0x72, 0x73, 0x70, 0x7B, 0x71, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x70, 0xF0, 
            0xF0, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0xF2, 0xF2, 0xF2, 0xF0, 0x72, 0xF2, 0xF0, 0xF2, 0xF2, 
            0xF2, 0xF2, 0xF2, 0xF2, 0xF2, 0xF2, 0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 0xF2, 0xF2, 0xF2, 0xF2, 
            0xF0, 0xF2, 0xF0, 0xF2, 0xF2, 0xF0, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 
            0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0xF0, 0xF2, 0xF2, 0xF2, 
            0xF2, 0xF2, 0xF2, 0xF2, 0xF2, 0xF2, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x70, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 0x70, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x7A, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x76, 0x76, 0x76, 0x76, 0x72, 
            0x72, 0x72, 0x70, 0x70, 0x70, 0x70, 0x70, 0x74, 0x70, 0x70, 0x74, 0x74, 0x70, 0x70, 0xF4, 0x74, 
            0xF2, 0x72, 0x76, 0x76, 0x72, 0x72, 0x76, 0x76, 0x76, 0x72, 0x72, 0x76, 0x72, 0x72, 0x76, 0x76, 
            0x76, 0x72, 0x72, 0x76, 0x72, 0x72, 0x76, 0x76, 0x72, 0x70, 0x74, 0x74, 0x72, 0x72, 0x76, 0x74, 
            0x72, 0x72, 0x76, 0x76, 0x72, 0x72, 0x76, 0x74, 0x72, 0x72, 0x76, 0x72, 0xF0, 0x74, 0xF0, 0xF0, 
            0x70, 0x70, 0x70, 0xF0, 0xF0, 0x70, 0x74, 0xF0, 0x70, 0x70, 0x70, 0xF8, 0x70, 0xFA, 0x72, 0x72, 
            0x72, 0x7A, 0x72, 0x70, 0x72, 0x70, 0xF0, 0x72, 0x72, 0xF2, 0xF0, 0x72, 0xF0, 0x72, 0x7A, 0x72, 
            0x72, 0xF2, 0x72, 0xF2, 0xF0, 0xF2, 0xF2, 0xF2, 0xF2, 0x7A, 0xF2, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0xF2, 0xF2, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0xF0, 0xF2, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x72, 0x70, 0x72, 0x72, 0xF2, 0x70, 
            0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 0x76, 0xF6, 0xF2, 0x70, 0x72, 
            0xF2, 0xF2, 0x72, 0x70, 0x70, 0x72, 0xF2, 0xF2, 0x7A, 0x7A, 0x7A, 0x78, 0x7A, 0x7E, 0x7E, 0x7A, 
            0x78, 0x7A, 0x7A, 0x7A, 0x72, 0x72, 0x72, 0x76, 0x7C, 0x7A, 0x78, 0x7A, 0x7A, 0x7A, 0xFA, 0xF8, 
            0x7A, 0x7E, 0xF6, 0xF2, 0x72, 0x72, 0x76, 0x72, 0x72, 0x76, 0x76, 0x70, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0xF2, 0x72, 0x70, 0x79, 0x79, 0x79, 0x79, 0x71, 0x71, 0x78, 0x78, 
            0x78, 0xF8, 0x72, 0x72, 0x70, 0x72, 0x70, 0x72, 0x72, 0x72, 0x76, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 
            0x76, 0x76, 0x72, 0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 
            0x72, 0x70, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x70, 0x70, 0x72, 0x70, 
            0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x50, 0x50, 0x52, 0x50, 0x50, 0x52, 0x52, 
            0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x50, 0x52, 0x52, 0x52, 0x52, 0x56, 0x52, 0x56, 0x50, 0x52, 
            0x52, 0x52, 0x50, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0xD0, 0x52, 
            0x5A, 0x52, 0x52, 0x52, 0x52, 0x52, 0x50, 0x58, 0x52, 0x5A, 0x52, 0x52, 0x52, 0x58, 0x52, 0x50, 
            0x50, 0x50, 0x5A, 0x52, 0x52, 0x5E, 0x5A, 0x58, 0x5E, 0x50, 0x54, 0x5C, 0x54, 0x58, 0x50, 0x58, 
            0x5E, 0x5C, 0x5A, 0x5A, 0x54, 0x56, 0x5A, 0x52, 0x52, 0x52, 0x50, 0x50, 0x58, 0x5E, 0x5E, 0x5C, 
            0x52, 0x52, 0x52, 0x52, 0x52, 0x5A, 0x52, 0x50, 0x50, 0x50, 0x56, 0x52, 0x52, 0x56, 0x52, 0x50, 
            0x5A, 0x52, 0x52, 0x56, 0x5A, 0x52, 0x50, 0x52, 0x52, 0x52, 0x5A, 0x52, 0x52, 0x52, 0x52, 0x52, 
            0x52, 0x50, 0x52, 0x50, 0x52, 0x50, 0x50, 0x52, 0x58, 0x58, 0x52, 0x50, 0x5A, 0x5A, 0x5A, 0x5A, 
            0x52, 0x52, 0x50, 0x50, 0x52, 0x52, 0x58, 0x5A, 0x52, 0x50, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 
            0x50, 0x52, 0x52, 0x52, 0x52, 0x52, 0x50, 0x52, 0x52, 0x52, 0x52, 0x52, 0x50, 0x50, 0x52, 0x50, 
            0x52, 0x52, 0x52, 0x52, 0x50, 0x52, 0x52, 0x52, 0x52, 0x50, 0x50, 0x50, 0x50, 0x58, 0x50, 0x50, 
            0x52, 0x5A, 0x52, 0x52, 0x52, 0x50, 0x52, 0x52, 0x52, 0x52, 0x52, 0x50, 0x52, 0x58, 0x52, 0x52, 
            0x52, 0x50, 0x52, 0x52, 0x52, 0x52, 0x50, 0x52, 0x50, 0x52, 0x50, 0x52, 0x52, 0x52, 0x52, 0x52, 
            0x52, 0x50, 0x50, 0x50, 0x54, 0x52, 0x50, 0x5A, 0x52, 0x52, 0x50, 0x52, 0x50, 0x52, 0x52, 0x50, 
            0x52, 0x52, 0x52, 0x50, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x50, 0x50, 0x52, 0x52, 0x50, 0x52, 
            0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x50, 0x52, 0x52, 0x50, 0x52, 0x12, 0x10, 0x12, 0x10, 
            0x10, 0x12, 0x12, 0x12, 0x12, 0x12, 0x10, 0x10, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x10, 
            0x12, 0x12, 0x12, 0x10, 0x10, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x10, 0x10, 0x12, 0x12, 
            0x12, 0x12, 0x12, 0x12, 0x12, 0x10, 0x10, 0x12, 0x12, 0x10, 0x18, 0x16, 0x12, 0x16, 0x12, 0x10, 
            0x12, 0x14, 0x12, 0x12, 0x12, 0x10, 0x12, 0x10, 0x12, 0x10, 0x12, 0x12, 0x10, 0x12, 0x12, 0x12, 
            0x10, 0x10, 0x10, 0x10, 0x12, 0x10, 0x12, 0x10, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x10, 
            0x10, 0x12, 0x12, 0x12, 0x10, 0x10, 0x10, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x10, 0x12, 
            0x10, 0x12, 0x10, 0x10, 0x12, 0x10, 0x10, 0x12, 0x12, 0x12, 0x12, 0x12, 0x92, 0x92, 0x92, 0x92, 
            0x92, 0x92, 0x90, 0x90, 0x92, 0x92, 0x92, 0x92, 0x92, 0x92, 0x90, 0x10, 0x12, 0x10, 0x10, 0x10, 
            0x90, 0x92, 0x90, 0x90, 0x90, 0x92, 0x92, 0x92, 0x94, 0x96, 0x96, 0x96, 0x96, 0x94, 0x96, 0x14, 
            0x16, 0x16, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x10, 0x12, 0x12, 0x10, 0x12, 0x12, 0x10, 0x12, 
            0x12, 0x12, 0x10, 0x12, 0x12, 0x12, 0x10, 0x12, 0x92, 0x90, 0x90, 0x92, 0x92, 0x92, 0x92, 0x90, 
            0x92, 0x92, 0x92, 0x90, 0x92, 0x90, 0x92, 0x92, 0x96, 0x96, 0x96, 0x96, 0x90, 0x90, 0x94, 0x94, 
            0x96, 0x96, 0x12, 0x12, 0x12, 0x12, 0x10, 0x12, 0x12, 0x10, 0x10, 0x12, 0x12, 0x10, 0x10, 0x12, 
            0x12, 0x10, 0x12, 0x12, 0x12, 0x10, 0x12, 0x12, 0x12, 0x10, 0x10, 0x10, 0x92, 0x92, 0x90, 0x90, 
            0x92, 0x92, 0x90, 0x90, 0x90, 0x90, 0x92, 0x90, 0x92, 0x92, 0x92, 0x92, 0x90, 0x90, 0x90, 0x90, 
            0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x12, 0x10, 0x10, 0x12, 0x10, 0x10, 0x10, 0x10, 0x10, 0x12, 
            0x10, 0x10, 0x12, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x90, 0x90, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 
            0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 
            0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x80, 0x80, 0x80, 0x84, 0x80, 0x80, 
            0x80, 0x80, 0x84, 0x80, 0x80, 0x80, 0x84, 0x84, 0x84, 0x80, 0x80, 0x84, 0x80, 0x84, 0x80, 0x80, 
            0x80, 0x84, 0x84, 0x84, 0x80, 0x80, 0x84, 0x80, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08, 0x0C, 0x08, 0x08, 0x08, 0x08, 0x08, 
            0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x08, 0x08, 0x0C, 0x0C, 
            0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x88, 0x08, 
            0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x88, 0x88, 0x08, 0x88, 
            0x88, 0x88, 0x88, 0x88, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x80, 0x80, 
            0x00, 0x04, 0x80, 0x80, 0x00, 0x00, 0x00, 0x02, 0x80, 0x82, 0x00, 0x00, 0x80, 0x86, 0x00, 0x00, 
            0x00, 0x00, 0x80, 0x80, 0x80, 0x00, 0x00, 0x80, 0x00, 0x20, 0xA0, 0xA0, 0xA0, 0xA0, 0x20, 0x20, 
            0x20, 0x20, 0x20, 0x20, 0x00, 0x00, 0x00, 0x24, 0x20, 0x10, 0x00, 0x20, 0x20, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x20, 0x20, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x20, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09, 0x08, 0x89, 0x88, 0x08, 0x08, 0x08, 0x08, 
            0x08, 0x08, 0x08, 0x08, 0x0B, 0x0A, 0x0A, 0x0A, 0x8B, 0x0A, 0x8A, 0x0A, 0x0B, 0x0A, 0x08, 0x08, 
            0x09, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 
            0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x88, 0x08, 0x88, 0x08, 
            0x08, 0x08, 0x08, 0x08, 0x0A, 0x0A, 0x0A, 0x0A, 0x08, 0x08, 0x0A, 0x0A, 0x0A, 0x0A, 0x08, 0x08, 
            0x08, 0x08, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x09, 0x09, 0x08, 0x08, 0x89, 0x88, 0x08, 0x89, 
            0x88, 0x88, 0x09, 0x88, 0x88, 0x89, 0x88, 0x08, 0x88, 0x88, 0x88, 0x08, 0x88, 0x08, 0x88, 0x88, 
            0x08, 0x08, 0x88, 0x88, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x80, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x80, 0x88, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 
            0x08, 0x88, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x00, 0x00, 0x00, 0x08, 0x08, 0x08, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09, 0x08, 0x00, 0x00, 0x01, 0x00, 
            0x00, 0x00, 0x20, 0x00, 0x29, 0x08, 0x00, 0x00, 0x21, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0A, 0x08, 
            0x00, 0x00, 0x02, 0x0B, 0x02, 0x00, 0x08, 0x0A, 0x00, 0x02, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 
            0x01, 0x21, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x42, 0x48, 0x48, 0x08, 0x18, 0x50, 0x50, 
            0x12, 0x10, 0x50, 0x50, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x52, 0x10, 0x50, 0x50, 0x10, 0x10, 0x10, 0x18, 0x18, 0x10, 
            0x10, 0x58, 0x10, 0x10, 0x12, 0x12, 0x10, 0x10, 0x10, 0x10, 0x12, 0x12, 0x10, 0x10, 0x10, 0x30, 
            0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x11, 0x10, 0x10, 0x10, 0x19, 0x12, 0x19, 0x18, 0x18, 0x1A, 0x18, 0x38, 
            0x38, 0x38, 0x38, 0x18, 0x18, 0x38, 0x38, 0x38, 0x38, 0x9A, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 
            0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x98, 0x18, 0x18, 0x08, 
            0x18, 0x08, 0x09, 0x0A, 0x09, 0x09, 0x09, 0x08, 0x08, 0x18, 0x09, 0x19, 0x19, 0x18, 0x19, 0x19, 
            0x09, 0x08, 0x09, 0x09, 0x09, 0x08, 0x08, 0x08, 0x09, 0x18, 0x08, 0x18, 0x18, 0x18, 0x18, 0x18, 
            0x18, 0x18, 0x19, 0x08, 0x18, 0x18, 0x18, 0x18, 0x18, 0x08, 0x18, 0x19, 0x19, 0x18, 0x18, 0x18, 
            0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x19, 0x18, 0x18, 0x08, 0x18, 0x18, 0x18, 0x18, 0x18, 0x08, 
            0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 
            0x88, 0x88, 0x08, 0x08, 0x08, 0x08, 0x98, 0x88, 0x88, 0x88, 0x88, 0x80, 0x80, 0x80, 0x80, 0x80, 
            0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x60, 0x00, 0x00, 0x00, 0x20, 0x20, 0x20, 
            0x00, 0x00, 0x20, 0x00, 0x20, 0x00, 0x90, 0x80, 0xA0, 0xA0, 0x20, 0x08, 0x48, 0x28, 0x08, 0x00, 
            0x08, 0x00, 0x00, 0x20, 0x20, 0x08, 0x08, 0x08, 0x00, 0x00, 0x00, 0x08, 0x08, 0x08, 0x00, 0x00, 
            0x08, 0x00, 0x08, 0x00, 0x00, 0x08, 0x08, 0x08, 0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x01, 0x00, 
            0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x08, 0x00, 0x00, 0x20, 
            0x20, 0x00, 0x00, 0x00, 0x04, 0x20, 0x20, 0x00, 0x04, 0x00, 0x08, 0x00, 0x0C, 0x00, 0x00, 0x00, 
            0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 0x40, 0x40, 0x48, 0x40, 0x40, 0x40, 0x4C, 0x40, 0x48, 0x40, 
            0x40, 0x40, 0x08, 0x00, 0x04, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x24, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x20, 0x80, 0x40, 0xC0, 0x44, 0x44, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x84, 0x04, 0x04, 0x84, 0x84, 0x04, 0x84, 0x04, 0x04, 0x04, 0x04, 0x84, 0x04, 0x84, 0xA4, 0x84, 
            0x84, 0x94, 0x04, 0x04, 0x85, 0x04, 0x04, 0x04, 0x05, 0x04, 0x84, 0x84, 0x84, 0x80, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x84, 0x04, 
            0x04, 0x04, 0x84, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x80, 0x04, 0x04, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x44, 0x04, 
            0x00, 0x00, 0x00, 0x04, 0x00, 0x40, 0x04, 0x44, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 
            0x40, 0x00, 0x00, 0x40, 0x40, 0x00, 0x40, 0x00, 0x00, 0x00, 0x04, 0xC4, 0x04, 0x44, 0x40, 0x40, 
            0x40, 0x40, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x40, 0x40, 0x40, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x44, 0x00, 
            0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 
            0x04, 0x04, 0x04, 0x00, 0x05, 0x04, 0x04, 0x00, 0x85, 0x04, 0x04, 0x00, 0x00, 0x00, 0x04, 0x40, 
            0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x40, 0x01, 0x00, 0x00, 0x00, 
            0x01, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x02, 0x10, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x02, 0x00, 0x01, 0x00, 0x02, 0x00, 0x01, 0x01, 0x02, 0x00, 0x00, 0x10, 0x02, 0x00, 0x00, 0x00, 
            0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x02, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 
            0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 
            0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x00, 0x00, 0x00, 0x02, 
            0x00, 0x00, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x80, 0x80, 0x82, 0x82, 0x80, 0x80, 0x80, 0x80, 0xA0, 
            0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0xA0, 0x80, 0x80, 0x80, 0xA0, 
            0x80, 0x80, 0x80, 0xA0, 0x80, 0x80, 0x80, 0xA0, 0x80, 0x80, 0x80, 0xA0, 0x80, 0x80, 0x88, 0x80, 
            0x80, 0x80, 0x80, 0xA8, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0xA0, 0x80, 0x80, 0x88, 0x80, 
            0x80, 0x88, 0x00, 0x20, 0x0A, 0x00, 0x02, 0x02, 0x02, 0x00, 0x00, 0x20, 0x0A, 0x02, 0x02, 0x08, 
            0x0A, 0x02, 0x02, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x02, 0x08, 0x00, 0x00, 0x08, 0x28, 
            0x20, 0x28, 0x20, 0x20, 0x22, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 
            0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x22, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 
            0x20, 0x20, 0x20, 0x20, 0x00, 0x00, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x64, 0x60, 0x60, 0x60, 0x64, 0x40, 
            0x40, 0x40, 0x60, 0x60, 0x60, 0x40, 0xC0, 0xC4, 0xC0, 0xC4, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xE0, 
            0xE0, 0xC0, 0xC4, 0xC0, 0xC0, 0xC0, 0xC4, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 
            0xC0, 0xE0, 0xA0, 0x84, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0xA0, 0xA0, 0x80, 0x80, 0x80, 
            0x80, 0x80, 0x80, 0x80, 0x80, 0xA0, 0xA0, 0x80, 0x80, 0x80, 0xA0, 0x80, 0x80, 0xA0, 0xA0, 0x80, 
            0x84, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 
            0x80, 0xA0, 0x20, 0x20, 0x20, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x20, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0xC0, 0xC0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x60, 
            0x40, 0x60, 0x60, 0x60, 0x40, 0x40, 0x42, 0x60, 0xE0, 0xE0, 0x42, 0x60, 0x60, 0x60, 0x40, 0x60, 
            0x60, 0x60, 0x40, 0x40, 0x40, 0x60, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0xC0, 0xC0, 0x40, 0x60, 
            0x40, 0xC0, 0x00, 0x00, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x80, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x80, 0x80, 0x80, 0x00, 0x40, 0x40, 0x80, 0x90, 
            0x90, 0x50, 0x90, 0x90, 0xD0, 0x50, 0x90, 0x10, 0x90, 0x10, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x10, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x90, 
            0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x90, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x40, 0x40, 0x00, 0x00, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x02, 0x00, 0x00, 0x10, 0x12, 0x10, 0x00, 0x10, 0x10, 0x10, 
            0x00, 0x10, 0x00, 0x12, 0x00, 0x02, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 
            0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x10, 0x00, 0x00, 0x00, 0x02, 
            0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x40, 0x00, 0x40, 0x40, 0x00, 0x08, 0x08, 0x08, 0x08, 0x08, 
            0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x0B, 0x09, 0x08, 0x08, 
            0x09, 0x09, 0x09, 0x08, 0x08, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x20, 0x00, 0x10, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 
            0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 
            0x00, 0x00, 0x00, 0x20, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x80, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 
            0x08, 0x08, 0x00, 0x10, 0x00, 0x08, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x08, 0x00, 0x10, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x80, 0x90, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x48, 0x48, 0x00, 0x00, 0x00, 
            0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x01, 0x00, 0x00, 
            0x00, 0x10, 0x00, 0x00, 0x00, 0x10, 0x10, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 0x02, 
            0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x08, 0x00, 0x28, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x08, 
            0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x40, 0x00, 0x40, 0x00, 0x00, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x48, 0x08, 
            0x08, 0x08, 0x48, 0x08, 0x09, 0x08, 0x49, 0x08, 0x08, 0x08, 0x48, 0x08, 0x08, 0x08, 0x48, 0x08, 
            0x09, 0x08, 0x08, 0x08, 0x01, 0x00, 0x40, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x40, 0x00, 
            0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x08, 0x01, 0x00, 0x00, 0x08, 0x00, 0x00, 0x40, 0x08, 
            0x01, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x08, 0x09, 0x0A, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x81, 0x80, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC8, 0xC8, 0xC8, 0xC8, 0xC8, 0xC8, 0xC8, 0xC8, 0xC8, 
            0xC8, 0xC8, 0xC8, 0xC8, 0xC9, 0xC9, 0xC8, 0xC8, 0xC9, 0xC8, 0xC8, 0xC9, 0xC8, 0xC8, 0xC9, 0x88, 
            0x88, 0x89, 0x88, 0x88, 0x89, 0x88, 0x88, 0x89, 0x8A, 0x88, 0x89, 0x88, 0x88, 0x89, 0x88, 0x88, 
            0x89, 0x88, 0xC8, 0x88, 0x88, 0x88, 0xC8, 0x88, 0x88, 0x88, 0xC8, 0x88, 0x88, 0x88, 0xC8, 0x88, 
            0x88, 0x8A, 0xC8, 0xC8, 0xC9, 0xC8, 0x88, 0x88, 0x89, 0x88, 0x88, 0x98, 0x99, 0x98, 0x98, 0x98, 
            0x99, 0x98, 0xD8, 0xD8, 0x99, 0x99, 0x99, 0x99, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 
            0x98, 0x98, 0x98, 0x98, 0x99, 0x99, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 
            0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 0x98, 
            0x98, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 
            0x18, 0x90, 0x10, 0x10, 0x90, 0x90, 0x10, 0x10, 0x98, 0x10, 0x90, 0x10, 0x98, 0x90, 0x18, 0x98, 
            0x10, 0x98, 0x90, 0x18, 0x18, 0x18, 0x98, 0x90, 0x18, 0x18, 0x10, 0x10, 0x90, 0x18, 0x18, 0x18, 
            0x10, 0x10, 0x10, 0x18, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 
            0x00, 0x00, 0x00, 0x10, 0x20, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x10, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 
            0x24, 0x00, 0x00, 0x00, 0x04, 0x00, 0x40, 0x40, 0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
            0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x80, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
            0x04, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 
            0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x04, 
            0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x04, 0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x02, 0x00, 0x02, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x06, 0x04, 0x04, 0x00, 
            0x02, 0x00, 0x00, 0x04, 0x06, 0x00, 0x00, 0x04, 0x06, 0x04, 0x04, 0x04, 0x06, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x06, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x06, 0x00, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x06, 0x00, 0x04, 0x04, 0x00, 0x04, 0x04, 0x00, 0x06, 0x04, 0x00, 0x00, 
            0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x04, 
            0x06, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x06, 0x06, 0x06, 
            0x06, 0x02, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x06, 0x06, 0x06, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x02, 0x00, 0x00, 0x00, 0x02, 0x04, 0x04, 0x04, 0x02, 0x04, 0x04, 0x04, 0x06, 0x04, 0x04, 0x04, 
            0x02, 0x06, 0x02, 0x06, 0x04, 0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x02, 0x06, 0x06, 
            0x06, 0x06, 0x06, 0x06, 0x06, 0x02, 0x06, 0x06, 0x02, 0x06, 0x06, 0x06, 0x06, 0x06, 0x02, 0x02, 
            0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x02, 0x06, 0x06, 0x06, 0x02, 0x06, 
            0x06, 0x06, 0x02, 0x02, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 
            0x06, 0x06, 0x06, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x06, 0x06, 0x06, 0x02, 0x06, 0x06, 0x26, 0x06, 0x26, 0x06, 0x06, 
            0x02, 0x02, 0x02, 0x06, 0x06, 0x02, 0x02, 0x26, 0x06, 0x06, 0x06, 0x26, 0x06, 0x06, 0x06, 0x26, 
            0x06, 0x04, 0x84, 0xA4, 0x84, 0x84, 0x84, 0xA4, 0x84, 0x8C, 0x8C, 0x8C, 0x8C, 0x88, 0x0C, 0xAC, 
            0x8C, 0x8C, 0x8C, 0x8C, 0x8C, 0x80, 0x8C, 0xAC, 0x88, 0x88, 0x8C, 0x8C, 0x8C, 0x8C, 0x88, 0xA8, 
            0x8C, 0x8E, 0x8C, 0x8C, 0x8C, 0x8E, 0x8C, 0xAC, 0x8A, 0x88, 0x88, 0x8C, 0x8E, 0x88, 0x8A, 0x8C, 
            0x8C, 0x8C, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0xA8, 0xAA, 0xE8, 0xA8, 0xE8, 
            0xAC, 0xAC, 0xA8, 0xAC, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xE8, 0xAC, 0xAC, 0xAC, 0xFC, 0xAC, 0xAD, 
            0xBC, 0x6D, 0xBC, 0xBC, 0xA8, 0xE8, 0xA8, 0xA8, 0x80, 0xC8, 0x98, 0x89, 0x98, 0x98, 0x18, 0x99, 
            0x98, 0x58, 0x18, 0x19, 0x18, 0x10, 0x1C, 0x1D, 0x10, 0x54, 0x14, 0x35, 0x14, 0x14, 0x14, 0x30, 
            0x10, 0x50, 0x10, 0x35, 0x14, 0x10, 0x14, 0x34, 0x14, 0x54, 0x10, 0x31, 0x30, 0x34, 0x34, 0x10, 
            0x14, 0x10, 0x14, 0x15, 0x10, 0x10, 0x10, 0x14, 0x14, 0x10, 0x10, 0x31, 0x34, 0x74, 0x74, 0x70, 
            0x70, 0xF0, 0xF4, 0xF4, 0xF4, 0xF4, 0xF4, 0xF4, 0xFC, 0xFC, 0xF8, 0xFC, 0xEC, 0xFC, 0xE8, 0xE9, 
            0xF9, 0xE8, 0xFC, 0xFD, 0xEC, 0xEC, 0xED, 0xE8, 0xE8, 0xE9, 0xBC, 0xAC, 0xAD, 0xA8, 0xAC, 0xAD, 
            0xAC, 0xAC, 0xA9, 0xA8, 0xA8, 0xA9, 0xA8, 0xA8, 0xA9, 0xA8, 0xA8, 0xA9, 0xA8, 0xE8, 0xA8, 0xA8, 
            0xA8, 0xE8, 0xA8, 0xA8, 0xA8, 0xE8, 0xA8, 0xA8, 0xA8, 0xE8, 0xA8, 0xAC, 0xAC, 0xEC, 0xEC, 0xED, 
            0xEC, 0xAC, 0xAC, 0xAD, 0xA8, 0xA8, 0xA8, 0xA9, 0xA8, 0xA8, 0xA8, 0xA8, 0xB8, 0xF8, 0xF8, 0xF8, 
            0xF9, 0xF9, 0xF9, 0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0xF8, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFD, 
            0xFD, 0xF9, 0xFD, 0xFD, 0xFD, 0xF9, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 
            0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xF9, 0xFD, 0xFD, 
            0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xF9, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xF5, 0xF9, 0xFD, 0xFD, 0xFD, 0xD9, 0xDD, 0xDD, 
            0xDD, 0xDD, 0xD9, 0xF9, 0x59, 0x59, 0x59, 0xD9, 0x59, 0x79, 0x51, 0x59, 0x59, 0x59, 0x59, 0x51, 
            0xD9, 0x59, 0x59, 0x59, 0x5D, 0x55, 0x59, 0x51, 0x59, 0x59, 0x5D, 0x5D, 0x55, 0x55, 0x5D, 0x55, 
            0x5D, 0x55, 0x5D, 0x55, 0x59, 0x51, 0x51, 0x51, 0x51, 0x59, 0x59, 0x59, 0x59, 0x59, 0x09, 0x09, 
            0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x0D, 0x0D, 0x09, 0x0D, 0x05, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 
            0x0D, 0x08, 0x0C, 0x0C, 0x0C, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x5C, 0x0C, 
            0x1C, 0x1C, 0x5C, 0x0C, 0x0C, 0x1C, 0x0C, 0x1C, 0x0C, 0x1C, 0x1C, 0x0C, 0x1C, 0x08, 0x5C, 0x1C, 
            0x0C, 0x0C, 0x0C, 0x1C, 0x1C, 0x4C, 0x0C, 0x0D, 0x09, 0x1C, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x08, 
            0x08, 0x09, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x0C, 0x0C, 0x0C, 0x08, 0x04, 0x04, 
            0x0C, 0x0C, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x00, 0x08, 0x08, 
            0x08, 0x08, 0x0C, 0x08, 0x0C, 0x0C, 0x00, 0x01, 0x08, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 
            0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x18, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 
            0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x0C, 0x0D, 0x00, 0x04, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 
            0x0C, 0x08, 0x0C, 0x0C, 0x0C, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 
            0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x0C, 0x0C, 
            0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 
            0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x00, 0x04, 0x04, 
            0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 
            0x1C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0C, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x08, 0x08, 0x08, 0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x03, 0x02, 0x02, 0x03, 0x02, 0x02, 0x03, 0x03, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x03, 0x02, 0x02, 0x02, 0x02, 0x02, 0x03, 
            0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x0A, 0x0A, 
            0x0A, 0x02, 0x0A, 0x0A, 0x0A, 0x02, 0x0A, 0x02, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 
            0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x02, 0x0A, 0x0A, 0x0A, 
            0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 
            0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 
            0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x01, 0x03, 0x01, 0x01, 0x03, 0x01, 0x01, 0x03, 0x03, 0x01, 0x03, 0x01, 0x21, 0x23, 0x23, 
            0x21, 0x23, 0x22, 0x23, 0xA3, 0xA3, 0xA1, 0xA3, 0xA3, 0x83, 0xA2, 0xAA, 0xA8, 0xAA, 0xAA, 0xAA, 
            0x2A, 0xA9, 0xA9, 0xA9, 0xA9, 0xA9, 0xA9, 0xA3, 0xA9, 0xA9, 0xA9, 0xA9, 0xA9, 0xA9, 0xA9, 0xA9, 
            0xA9, 0xAB, 0xA8, 0xA9, 0xA9, 0xA9, 0xA9, 0xA9, 0xAB, 0xAB, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xAA, 
            0xAA, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xAA, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 
            0xA8, 0xAA, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xAA, 0xAA, 0xA8, 0xA8, 0x88, 0xA8, 0xA8, 0xFA, 
            0xB8, 0xB8, 0xF8, 0x38, 0xB8, 0xF8, 0xB8, 0xB8, 0xB8, 0x98, 0xB0, 0xA8, 0xF8, 0xB8, 0xF8, 0xF8, 
            0x78, 0xF9, 0xFA, 0xF8, 0xF8, 0xF8, 0xF8, 0xF0, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xFB, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xFB, 0xFB, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xFB, 0xF9, 0xA9, 0xF9, 
            0xB9, 0xB9, 0xF9, 0xBB, 0xBB, 0xF9, 0xB9, 0xB9, 0xB9, 0xB9, 0xB9, 0xA9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF8, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xD9, 0xD9, 0xD9, 
            0xD9, 0xD9, 0xD9, 0xD9, 0xD9, 0xD9, 0xD9, 0xD9, 0xD9, 0xD9, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 
            0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 0xD1, 
            0xF1, 0xD1, 0xD1, 0x71, 0x71, 0x51, 0x51, 0x51, 0x51, 0x51, 0xD1, 0xD1, 0x71, 0x71, 0x59, 0x51, 
            0x51, 0x59, 0x59, 0x51, 0xD1, 0x51, 0x51, 0x51, 0x51, 0x51, 0x59, 0x59, 0x51, 0x51, 0x51, 0x51, 
            0x51, 0x59, 0x51, 0x51, 0x51, 0x51, 0x51, 0x51, 0x51, 0x51, 0x51, 0x51, 0x51, 0x51, 0x11, 0x01, 
            0x01, 0x01, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x20, 0x20, 0x00, 0x00, 
            0x00, 0x08, 0x00, 0x00, 0x20, 0x20, 0x20, 0x00, 0x20, 0x20, 0x08, 0x08, 0x00, 0x00, 0x20, 0x30, 
            0x00, 0x00, 0x19, 0x28, 0x48, 0x74, 0x3C, 0x0C, 0x34, 0x04, 0x84, 0x15, 0x1C, 0x6C, 0x54, 0x14, 
            0x14, 0x5C, 0x15, 0x4C, 0x1D, 0x14, 0x18, 0x14, 0x14, 0x05, 0x14, 0x1C, 0x14, 0x14, 0x00, 0x20, 
            0x00, 0x00, 0x00, 0x00, 0x10, 0x04, 0x04, 0x24, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0C, 0x14, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x44, 0x1C, 0x10, 0x24, 0x04, 0x04, 0x04, 0x04, 0x14, 0x14, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x10, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x40, 0x10, 0x10, 0x00, 0x00, 0x21, 0x00, 0x00, 0x10, 0x00, 0x08, 0x00, 
            0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x14, 0x1C, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x14, 
            0x14, 0x14, 0x04, 0x04, 0x04, 0x04, 0x14, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x14, 0x14, 0x14, 
            0x14, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x14, 0x14, 0x14, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x14, 0x14, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x10, 0x14, 0x04, 0x04, 0x06, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x06, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x06, 0x26, 0x06, 0x26, 0x04, 0x26, 0x04, 
            0x24, 0x00, 0x24, 0x04, 0x24, 0x04, 0x24, 0x04, 0x2C, 0x04, 0x0C, 0x04, 0x0C, 0x04, 0x0C, 0x04, 
            0x0C, 0x04, 0x0C, 0x04, 0x0C, 0x04, 0x0C, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x04, 0x08, 0x00, 0x04, 0x04, 0x00, 
            0x04, 0x00, 0x00, 0x04, 0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x08, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x06, 0x06, 0x02, 
            0x02, 0x02, 0x02, 0x00, 0x06, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x04, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x06, 0x06, 0x02, 
            0x02, 0x02, 0x02, 0x00, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00, 0x84, 0x00, 0x00, 0x80, 0x00, 0x00, 0x80, 0x80, 0x00, 
            0x00, 0x08, 0x00, 0x08, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x84, 0x04, 0x00, 0x00, 
            0x00, 0x00, 0x80, 0x08, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 
            0x00, 0x00, 0x02, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 
            0x04, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x06, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x12, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x06, 0x06, 0x12, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x12, 0x12, 0x12, 0x12, 0x02, 0x12, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x22, 0x22, 0x22, 
            0x22, 0x22, 0x22, 0x22, 0xA2, 0xA2, 0xA2, 0xA2, 0xA2, 0xA2, 0xA2, 0xAA, 0xAA, 0xAA, 0x8A, 0xAA, 
            0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0x2A, 0x8A, 0x8A, 0x8A, 0x8A, 0x8A, 0x8A, 0x82, 0x2A, 0x2A, 0x2A, 
            0x2A, 0x2A, 0x2A, 0x2A, 0xA2, 0xA2, 0xA2, 0xA2, 0xA2, 0xA2, 0xA2, 0xAA, 0xAA, 0xAA, 0x8A, 0xAA, 
            0xAA, 0xAA, 0xAA, 0xAA, 0xA8, 0x28, 0x88, 0x88, 0x8A, 0x8A, 0x8A, 0x8A, 0x82, 0x0A, 0x0A, 0x0A, 
            0x0A, 0x0A, 0x0A, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x50, 
            0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x52, 0x52, 0x50, 0x50, 0x50, 0x10, 
            0x40, 0x50, 0x50, 0x50, 0x52, 0x50, 0x50, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x40, 0x50, 
            0x50, 0x50, 0x50, 0x50, 0xD0, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x58, 0x70, 0x70, 0x70, 0x30, 
            0x60, 0x70, 0x70, 0xF0, 0xF0, 0xF0, 0xF0, 0xB0, 0xA0, 0xA0, 0xAA, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 
            0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 
            0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xAA, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 
            0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0x88, 0xA8, 0xE8, 0xB8, 
            0xA8, 0xA8, 0xA8, 0x29, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA8, 0xA0, 0xA8, 0xA8, 0xE8, 0xF8, 0xF8, 
            0xF8, 0xF8, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF1, 0xF1, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF1, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xFD, 
            0xFD, 0xFD, 0xF9, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xB9, 0xE9, 0xF9, 
            0xF9, 0xF9, 0xE8, 0xF9, 0xE9, 0xE9, 0xE9, 0xF9, 0xF8, 0xF8, 0xF8, 0xF8, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xE9, 0xE9, 0xF9, 0xF9, 0xF8, 0xE9, 0xF9, 0xF9, 0xE8, 0xE8, 0xE8, 0xE8, 0xF8, 0xF8, 
            0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xF8, 0xE8, 0xE8, 0xE8, 0xE8, 0xF8, 0xE8, 0xF8, 0xD8, 0xC8, 
            0xD8, 0xD8, 0xD8, 0xD8, 0xC8, 0xC8, 0x58, 0x48, 0x58, 0x58, 0x48, 0x58, 0x58, 0x40, 0x40, 0x50, 
            0x40, 0x40, 0x50, 0x50, 0x40, 0x51, 0x41, 0x51, 0x51, 0x41, 0x41, 0x51, 0x41, 0x51, 0x51, 0x51, 
            0x41, 0x41, 0x71, 0x51, 0x41, 0xE1, 0x71, 0x41, 0x51, 0x51, 0x41, 0x41, 0xD1, 0xD1, 0xF1, 0x71, 
            0x41, 0x41, 0x41, 0x49, 0x49, 0xE1, 0xC1, 0x41, 0x41, 0xD1, 0x41, 0xC1, 0x49, 0x49, 0x41, 0x41, 
            0x49, 0x41, 0x49, 0x51, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x51, 0x51, 
            0x41, 0x41, 0x51, 0x11, 0x11, 0x31, 0x11, 0x01, 0x11, 0x01, 0x01, 0x11, 0x91, 0x11, 0x31, 0x21, 
            0x01, 0x11, 0x11, 0x19, 0x11, 0x11, 0x21, 0x21, 0x20, 0x10, 0x20, 0x30, 0x00, 0x00, 0x10, 0x10, 
            0x70, 0x30, 0x10, 0x40, 0x10, 0x31, 0x51, 0x65, 0x35, 0x45, 0x25, 0x15, 0xD5, 0x55, 0x15, 0x65, 
            0x45, 0x45, 0x45, 0x5D, 0x05, 0x45, 0x55, 0x45, 0x41, 0x55, 0x15, 0x45, 0x45, 0x45, 0x45, 0x05, 
            0x11, 0x31, 0x11, 0x11, 0x11, 0x41, 0x11, 0x15, 0x15, 0x35, 0x15, 0x15, 0x05, 0x15, 0x05, 0x45, 
            0x05, 0x05, 0x05, 0x05, 0x15, 0x15, 0x55, 0x55, 0x11, 0x35, 0x15, 0x05, 0x15, 0x55, 0x55, 0x15, 
            0x01, 0x11, 0x01, 0x11, 0x11, 0x51, 0x01, 0x01, 0x11, 0x11, 0x01, 0x01, 0x01, 0x01, 0x01, 0x41, 
            0x11, 0x11, 0x31, 0x11, 0x11, 0x11, 0x71, 0x51, 0x01, 0x01, 0x01, 0x30, 0x00, 0x50, 0x00, 0x10, 
            0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 
            0x50, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x04, 0x10, 0x30, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 
            0x00, 0x00, 0x10, 0x00, 0x00, 0x40, 0x44, 0x04, 0x04, 0x04, 0x04, 0x04, 0x14, 0x14, 0x04, 0x14, 
            0x54, 0x44, 0x44, 0x14, 0x14, 0x05, 0x15, 0x45, 0x15, 0x05, 0x15, 0x15, 0x15, 0x15, 0x55, 0x55, 
            0x45, 0x55, 0x05, 0x15, 0x05, 0x05, 0x15, 0x15, 0x55, 0x55, 0x45, 0x05, 0x15, 0x15, 0x15, 0x15, 
            0x15, 0x55, 0x55, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 
            0x15, 0x15, 0x15, 0x15, 0x15, 0x45, 0x41, 0x17, 0x05, 0x15, 0x17, 0x15, 0x11, 0x03, 0x13, 0x11, 
            0x15, 0x05, 0x05, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x11, 0x15, 0x37, 0x15, 0x31, 0x11, 
            0x31, 0x11, 0x33, 0x11, 0x31, 0x11, 0x31, 0x11, 0x31, 0x11, 0x21, 0x11, 0x01, 0x11, 0x01, 0x11, 
            0x11, 0x11, 0x11, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x14, 0x14, 
            0x04, 0x04, 0x14, 0x14, 0x04, 0x14, 0x14, 0x14, 0x04, 0x14, 0x14, 0x14, 0x04, 0x04, 0x04, 0x14, 
            0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x04, 0x14, 0x04, 0x14, 0x14, 0x14, 0x04, 0x14, 0x14, 0x14, 
            0x14, 0x14, 0x04, 0x14, 0x14, 0x04, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 
            0x14, 0x14, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x14, 0x04, 0x14, 0x14, 
            0x04, 0x14, 0x04, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x04, 0x14, 0x04, 0x00, 0x00, 0x10, 0x20, 
            0x10, 0x00, 0x00, 0x00, 0x01, 0x11, 0x11, 0x11, 0x01, 0x15, 0x05, 0x15, 0x15, 0x15, 0x15, 0x35, 
            0x05, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0x05, 0x15, 0x15, 
            0x15, 0x14, 0x14, 0x14, 0x04, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x04, 0x14, 
            0x14, 0x14, 0x04, 0x14, 0x06, 0x16, 0x06, 0x16, 0x06, 0x06, 0x06, 0x06, 0x16, 0x06, 0x06, 0x06, 
            0x16, 0x16, 0x06, 0x16, 0x06, 0x14, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x14, 0x14, 0x14, 
            0x04, 0x14, 0x04, 0x14, 0x06, 0x16, 0x06, 0x16, 0x06, 0x06, 0x16, 0x16, 0x16, 0x16, 0x16, 0x12, 
            0x12, 0x12, 0x12, 0x12, 0x13, 0x11, 0x03, 0x03, 0x13, 0x13, 0x13, 0x13, 0x03, 0x11, 0x11, 0x11, 
            0x11, 0x01, 0x11, 0x11, 0x11, 0x31, 0x11, 0x31, 0x01, 0x01, 0x11, 0x01, 0x01, 0x11, 0x01, 0x01, 
            0x10, 0x00, 0x10, 0x11, 0x00, 0x10, 0x00, 0x10, 0x00, 0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x00, 
            0x00, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 
            0x00, 0x00, 0x10, 0x10, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 
            0x10, 0x10, 0x10, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x02, 0x12, 0x12, 0x02, 0x12, 0x12, 0x12, 
            0x02, 0x12, 0x12, 0x02, 0x12, 0x12, 0x52, 0x12, 0x12, 0x12, 0x12, 0x12, 0x02, 0x02, 0x12, 0x52, 
            0x02, 0x12, 0x02, 0x12, 0x12, 0x12, 0x12, 0x52, 0x42, 0x42, 0x52, 0x12, 0x42, 0x12, 0x02, 0x02, 
            0x12, 0x02, 0x02, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x02, 0x12, 0x12, 0x12, 0x12, 
            0x12, 0x12, 0x02, 0x10, 0x02, 0x12, 0x12, 0x02, 0x12, 0x10, 0x02, 0x12, 0x02, 0x12, 0x12, 0x12, 
            0x12, 0x02, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x00, 0x00, 0x00, 
            0x10, 0x10, 0x12, 0x12, 0x10, 0x10, 0x02, 0x12, 0x12, 0x02, 0x12, 0x02, 0x12, 0x02, 0x02, 0x10, 
            0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x00, 0x00, 0x10, 0x10, 0x00, 0x00, 0x10, 0x10, 0x00, 
            0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x02, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x02, 0x32, 
            0x32, 0x32, 0x22, 0x32, 0x32, 0x32, 0xB2, 0xB2, 0xB2, 0xB2, 0xA2, 0xB2, 0xB2, 0xBA, 0xAA, 0xAA, 
            0x8A, 0xAA, 0xBA, 0xAA, 0xBA, 0xBA, 0xAA, 0x3A, 0x8A, 0x9A, 0x9A, 0x9A, 0x9A, 0x9A, 0x92, 0x3A, 
            0x2A, 0x3A, 0x3A, 0x3A, 0x2A, 0x3A, 0xA2, 0xB2, 0xB6, 0xB6, 0xA6, 0xB6, 0xB6, 0xBE, 0xBE, 0xAE, 
            0x8E, 0xBE, 0xAE, 0xAE, 0xAE, 0xBE, 0xBE, 0x3E, 0x9E, 0x8E, 0x8E, 0x8E, 0x9E, 0x8E, 0x86, 0x0E, 
            0x1E, 0x0E, 0x1E, 0x0E, 0x0F, 0x0B, 0x03, 0x13, 0x03, 0x10, 0x10, 0x00, 0x01, 0x11, 0x01, 0x11, 
            0x43, 0x53, 0x53, 0x53, 0x51, 0x51, 0x53, 0x53, 0x53, 0x43, 0x43, 0x43, 0x43, 0x43, 0x53, 0x43, 
            0x53, 0x03, 0x53, 0x43, 0x53, 0x53, 0x43, 0x43, 0x43, 0x03, 0x13, 0x13, 0x13, 0x03, 0x03, 0x23, 
            0x53, 0x53, 0x43, 0x43, 0x43, 0x43, 0xD3, 0x53, 0x43, 0x53, 0x43, 0x43, 0x53, 0x5B, 0x73, 0x63, 
            0x63, 0x23, 0x63, 0x63, 0x73, 0xF3, 0xF3, 0xE3, 0xF7, 0xA7, 0xB7, 0xB7, 0xBF, 0xBF, 0xBF, 0xAE, 
            0xBE, 0xAE, 0xAE, 0xAF, 0xBF, 0xAE, 0xAE, 0xBF, 0xBE, 0xAE, 0xAF, 0xAE, 0xBF, 0xBE, 0xBE, 0xAF, 
            0xBF, 0xBF, 0xAF, 0xAF, 0xAB, 0xAB, 0xAB, 0xAF, 0xAB, 0xAB, 0xAB, 0xAB, 0xA9, 0xA9, 0xA9, 0xA9, 
            0xB9, 0xBB, 0xAB, 0xBB, 0xA9, 0xB8, 0xB9, 0xA9, 0xB8, 0xB8, 0xBA, 0xAA, 0xAA, 0xBA, 0x8A, 0xAB, 
            0xFB, 0xBA, 0xAA, 0xBB, 0x9B, 0x1B, 0x9B, 0x9B, 0xAB, 0xAB, 0xAB, 0xAB, 0xB2, 0xBA, 0xBA, 0xFA, 
            0xFA, 0xFA, 0xFA, 0xDA, 0xFA, 0xFA, 0xDA, 0xCA, 0xDA, 0xDA, 0xCA, 0xCA, 0xCA, 0xEA, 0xEA, 0xC8, 
            0xC8, 0xF8, 0xF8, 0xE9, 0xE9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x59, 0xD9, 0xD9, 0xD9, 0xC9, 0xC9, 
            0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 
            0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0x49, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0x89, 
            0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 
            0xC9, 0xC9, 0xC9, 0x49, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 
            0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC8, 0xC8, 0xC8, 0xC8, 0xC8, 0xC8, 0xC8, 0xC8, 0xC9, 0xC9, 
            0xC9, 0xC8, 0x48, 0xC9, 0x49, 0x49, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 
            0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xE9, 0xC9, 0xC9, 0xC9, 
            0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0x49, 0x49, 0xC9, 0x49, 0x49, 0x49, 0xC9, 0xC1, 
            0x41, 0x49, 0x41, 0x41, 0xC9, 0x49, 0x49, 0xC9, 0xC1, 0x49, 0x49, 0xC9, 0x41, 0x41, 0x49, 0xC9, 
            0x49, 0x49, 0x41, 0x41, 0xC1, 0x49, 0x41, 0x41, 0x41, 0x49, 0x41, 0xC9, 0x41, 0x41, 0xC1, 0x41, 
            0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x40, 0x40, 0x40, 0x40, 0xC0, 0x41, 0x41, 
            0x41, 0x40, 0x40, 0xC0, 0x40, 0x40, 0x40, 0xC1, 0x45, 0xC5, 0x45, 0xC5, 0x45, 0x41, 0xC1, 0xC1, 
            0xC5, 0xC5, 0x85, 0x05, 0x45, 0x85, 0x85, 0x45, 0x45, 0x45, 0x45, 0x85, 0x45, 0xC5, 0x45, 0x05, 
            0x85, 0x45, 0xC5, 0xC5, 0x45, 0x65, 0xA5, 0x25, 0x25, 0x35, 0x35, 0x35, 0x74, 0x34, 0x74, 0x34, 
            0x34, 0x34, 0x34, 0x34, 0x34, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x34, 
            0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x36, 0x34, 0x34, 0x34, 0x14, 0x14, 0x14, 0x14, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x04, 0x00, 0x00, 0x04, 0x00, 
            0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x06, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x04, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 
            0x00, 0x04, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 
            0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x02, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x02, 0x02, 0x02, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x02, 0x02, 0x00, 0x00, 0x0A, 0x0A, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x02, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x02, 0x02, 
            0x03, 0x03, 0x02, 0x02, 0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x00, 0x02, 0x02, 0x02, 0x03, 0x01, 
            0x03, 0x02, 0x02, 0x01, 0x00, 0x00, 0x02, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x80, 0x20, 0xA0, 
            0x80, 0x80, 0x18, 0x98, 0x08, 0x80, 0x88, 0x08, 0x88, 0x80, 0x88, 0x80, 0xA8, 0xA8, 0xA8, 0x88, 
            0x98, 0xB8, 0x98, 0x88, 0x88, 0x98, 0x89, 0x88, 0x88, 0x08, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 
            0x81, 0x08, 0x08, 0x89, 0x89, 0x88, 0x88, 0x09, 0x89, 0x89, 0x89, 0x89, 0x89, 0x89, 0x89, 0x89, 
            0x89, 0x89, 0x89, 0x89, 0xA9, 0x89, 0x89, 0x89, 0x99, 0x89, 0x89, 0x89, 0x89, 0x89, 0xA9, 0x89, 
            0x89, 0xA9, 0xB9, 0x89, 0x89, 0xB9, 0xF9, 0x69, 0x69, 0x79, 0x79, 0x79, 0x39, 0x79, 0x29, 0x79, 
            0x79, 0x71, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 0x79, 
            0x69, 0x69, 0x69, 0x21, 0x61, 0x61, 0xE1, 0xF1, 0xE1, 0xF9, 0xE1, 0xE9, 0xE9, 0xE1, 0xE1, 0xE1, 
            0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 
            0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xF1, 0xF1, 0xF1, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 
            0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 
            0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0x58, 0x58, 0x59, 0x58, 0x58, 0x58, 0x58, 0x5A, 0x58, 0x58, 
            0x58, 0x58, 0x58, 0x58, 0x58, 0x58, 0x58, 0x58, 0x59, 0x59, 0x59, 0x58, 0x58, 0x58, 0x48, 0x49, 
            0x48, 0x40, 0x40, 0x40, 0x40, 0x50, 0x70, 0xD0, 0x58, 0xE8, 0x48, 0x78, 0x70, 0xF0, 0x78, 0x78, 
            0xD8, 0x78, 0x78, 0x78, 0x78, 0x58, 0xF8, 0x78, 0x78, 0x78, 0x58, 0x58, 0xD8, 0x58, 0x58, 0x78, 
            0xD8, 0x78, 0x58, 0x78, 0x58, 0xD8, 0xD8, 0xD8, 0xD8, 0xD8, 0xF8, 0xD8, 0xD8, 0xD8, 0x58, 0xF8, 
            0xD8, 0x98, 0x18, 0x98, 0x18, 0x18, 0x18, 0x98, 0x18, 0x98, 0x98, 0x98, 0xB8, 0x18, 0x98, 0x19, 
            0x98, 0x18, 0xB8, 0x18, 0x18, 0xB8, 0xB8, 0x98, 0xB8, 0x98, 0xB8, 0x98, 0x98, 0xB8, 0xB8, 0xB8, 
            0xB8, 0xB8, 0x98, 0x98, 0xB8, 0xB8, 0x18, 0xD8, 0x98, 0x98, 0xBC, 0x9C, 0xBC, 0x38, 0x9C, 0xBC, 
            0x9C, 0xE8, 0xAC, 0x08, 0x04, 0x14, 0x14, 0x14, 0x1C, 0x0C, 0x0C, 0x0C, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x44, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x44, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x24, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x14, 0x00, 0x08, 0x08, 
            0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 
            0x02, 0x02, 0x00, 0x02, 0x02, 0x02, 0x00, 0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x12, 0x12, 
            0x12, 0x1A, 0x1A, 0x0A, 0x0A, 0x0A, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x12, 0x02, 0x02, 
            0x0A, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x52, 0x02, 0x02, 0x4A, 0x42, 0x02, 0x02, 0x12, 0x02, 
            0x02, 0x0A, 0x12, 0x02, 0x02, 0x0A, 0x02, 0x02, 0x02, 0x42, 0x02, 0x02, 0x12, 0x12, 0x02, 0x4A, 
            0x1A, 0x12, 0x02, 0x0A, 0x0A, 0x12, 0x02, 0x02, 0x0A, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x12, 0x02, 0x02, 0x0A, 0x02, 0x12, 0x02, 0x02, 0x0A, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x10, 0x10, 0x00, 0x08, 0x18, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08, 0x00, 
            0x00, 0x10, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 
            0x01, 0x03, 0x01, 0x02, 0x01, 0x00, 0x01, 0x01, 0x00, 0x03, 0x03, 0x03, 0x01, 0x01, 0x01, 0x01, 
            0x01, 0x03, 0x01, 0x01, 0x01, 0x11, 0x11, 0x11, 0x19, 0x09, 0xA8, 0x29, 0xA1, 0xA1, 0xA1, 0xA1, 
            0xA1, 0x01, 0xA1, 0x01, 0xA1, 0xA1, 0xA1, 0xA1, 0xB0, 0xB1, 0xA1, 0xA9, 0xA9, 0xA1, 0xA1, 0xA1, 
            0xA1, 0xA3, 0xA1, 0xA1, 0xA1, 0xA1, 0xB1, 0xB1, 0x11, 0xA9, 0xA9, 0xB9, 0xA1, 0xA1, 0xA9, 0xA1, 
            0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 
            0xA1, 0xA3, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xB1, 0xA1, 
            0xA1, 0xA9, 0xA1, 0xA1, 0xA1, 0xA1, 0xA1, 0xB1, 0xB1, 0xB1, 0xB9, 0xB9, 0xB9, 0xB9, 0xB9, 0xB9, 
            0xB9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xB9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xB9, 0xB9, 0xB9, 0xB9, 0xB9, 0xB9, 
            0xB9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xF9, 0xFD, 0xFD, 
            0xFD, 0xF9, 0xBD, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 
            0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 
            0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0x58, 0x58, 0x59, 0x49, 0x49, 0x49, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x41, 0x41, 0x41, 0x40, 0x40, 0x40, 0x41, 0x41, 0x41, 0x40, 0x40, 0x41, 0x40, 0x41, 0x40, 
            0x40, 0x41, 0x41, 0x41, 0x41, 0xE0, 0x61, 0xC1, 0xF1, 0xD1, 0x51, 0xD8, 0xC8, 0xD0, 0x49, 0x49, 
            0xE1, 0xE1, 0x50, 0x51, 0xD1, 0xE8, 0xE8, 0xE8, 0xD1, 0xD1, 0xF1, 0xC9, 0xE9, 0xD8, 0x40, 0x50, 
            0xE9, 0xE1, 0xF8, 0xD1, 0x70, 0xD8, 0x61, 0xD9, 0x49, 0x61, 0x29, 0x20, 0xB0, 0x90, 0xB1, 0x99, 
            0x89, 0x19, 0x28, 0x81, 0x89, 0x20, 0x00, 0x00, 0x21, 0x91, 0x01, 0x21, 0x09, 0x20, 0x91, 0x20, 
            0x81, 0xA9, 0x20, 0x29, 0xB0, 0xA0, 0x20, 0xB8, 0x10, 0x30, 0x78, 0x68, 0xF8, 0xF8, 0xB0, 0xF8, 
            0xB8, 0x58, 0x78, 0xD8, 0xD8, 0x78, 0x58, 0x18, 0xE8, 0xF8, 0x58, 0x70, 0xB8, 0x28, 0xD8, 0xF8, 
            0xE0, 0xF8, 0x78, 0x38, 0x58, 0x48, 0x0C, 0x1C, 0x04, 0x44, 0x4C, 0x44, 0x44, 0x44, 0x44, 0x44, 
            0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x40, 0x44, 0x44, 0x04, 0x44, 0x44, 0x44, 0x44, 0x40, 
            0x40, 0x00, 0x40, 0x40, 0x40, 0x40, 0x44, 0x04, 0x04, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 
            0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x40, 0x44, 0x44, 0x04, 0x44, 0x44, 0x44, 0x44, 0x44, 
            0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x04, 0x04, 0x44, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x44, 0x44, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x44, 0x44, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0xA4, 0x04, 0x04, 0x14, 0x04, 0x44, 0x4C, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x44, 0x44, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 
            0x00, 0x02, 0x40, 0x02, 0x02, 0x02, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x04, 0x04, 0x04, 
            0x04, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x04, 
            0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x06, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x02, 0x02, 0x06, 
            0x06, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x04, 0x06, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x42, 0x42, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x00, 0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x02, 0x02, 0x03, 0x02, 0x02, 0x00, 0x02, 0x03, 
            0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x03, 0x02, 0x02, 0x03, 0x02, 0x02, 
            0x03, 0x02, 0x02, 0x02, 0x00, 0x02, 0x02, 0x02, 0xA3, 0x02, 0x02, 0x11, 0x03, 0xA2, 0x0A, 0x02, 
            0x12, 0x00, 0x00, 0x0A, 0x02, 0x02, 0x02, 0x03, 0xA2, 0x02, 0x03, 0xB3, 0x02, 0x03, 0xBB, 0x03, 
            0x03, 0x1B, 0x03, 0x03, 0x09, 0x03, 0xA3, 0x03, 0x03, 0xB3, 0xA3, 0x03, 0x1B, 0x13, 0x03, 0x0B, 
            0x0B, 0x01, 0x01, 0x03, 0x01, 0xA1, 0x01, 0x01, 0xB1, 0xA1, 0x01, 0xB9, 0xB1, 0xA1, 0x39, 0x39, 
            0x31, 0x29, 0x29, 0x29, 0x31, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 
            0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 
            0x21, 0x21, 0x21, 0x21, 0x21, 0x23, 0x21, 0x21, 0x21, 0x21, 0x21, 0xA1, 0x21, 0x21, 0x31, 0x21, 
            0x21, 0x29, 0x21, 0x21, 0x21, 0x25, 0x25, 0x21, 0x21, 0x21, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 
            0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x63, 0xE3, 0xE1, 0x65, 0x61, 0x65, 0x61, 0x69, 0x69, 0x61, 
            0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x60, 0x60, 
            0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x62, 0x60, 0x60, 0x60, 
            0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x40, 0x40, 0x41, 0x41, 
            0x41, 0x41, 0x41, 0x41, 0x40, 0x41, 0x43, 0x41, 0x41, 0x41, 0x41, 0x40, 0x41, 0x41, 0x41, 0x41, 
            0x40, 0x40, 0x41, 0x41, 0x41, 0x41, 0x41, 0x40, 0x41, 0x41, 0x40, 0x40, 0x60, 0x61, 0x61, 0x61, 
            0x61, 0x61, 0x41, 0x61, 0x60, 0x61, 0x61, 0x61, 0x61, 0x41, 0x61, 0x61, 0x61, 0x61, 0x40, 0x41, 
            0x61, 0x61, 0x61, 0x61, 0x61, 0x40, 0x60, 0x61, 0x41, 0x41, 0x01, 0x60, 0x60, 0x61, 0x61, 0x61, 
            0x61, 0x61, 0x61, 0x61, 0x60, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x61, 0x60, 0x61, 
            0x61, 0x61, 0x61, 0x41, 0x61, 0x61, 0x60, 0x61, 0x61, 0x21, 0x21, 0x60, 0x60, 0x61, 0x60, 0x60, 
            0x61, 0x21, 0x60, 0x20, 0x60, 0x20, 0x60, 0x20, 0x20, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 
            0x60, 0x60, 0x60, 0x20, 0x60, 0x60, 0x60, 0x60, 0x60, 0x20, 0x60, 0x20, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x40, 0x00, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08, 0x40, 0x00, 0x00, 0x00, 0x01, 
            0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x20, 0x00, 0x04, 
            0x05, 0x05, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x04, 0x44, 0x00, 0x00, 0x01, 0x01, 0x04, 0x04, 
            0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x44, 0x04, 0x24, 0x24, 
            0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x05, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 
            0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x44, 0x44, 0x44, 0x04, 
            0x01, 0x00, 0x04, 0x44, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x40, 0x40, 0x40, 0x00, 0x00, 
            0x00, 0x00, 0x04, 0x04, 0x44, 0x44, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x64, 0x04, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x04, 0x04, 0x06, 0x06, 0x04, 0x04, 0x04, 0x04, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x22, 0x22, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x82, 0x02, 0x82, 0x12, 0x02, 0x12, 0x0A, 0x02, 0x0A, 0x02, 0x02, 0x02, 0x82, 
            0x02, 0x02, 0x12, 0x82, 0x02, 0x0A, 0x12, 0x82, 0x02, 0x0A, 0x12, 0x82, 0x02, 0x0A, 0x12, 0x82, 
            0x02, 0x0A, 0x12, 0x02, 0x02, 0x0A, 0x02, 0x82, 0x02, 0x02, 0x12, 0x02, 0x02, 0x0A, 0x02, 0x82, 
            0x02, 0x02, 0x12, 0x02, 0x02, 0x0A, 0x02, 0x82, 0x02, 0x02, 0x12, 0x02, 0x02, 0x0A, 0x02, 0x82, 
            0x02, 0x02, 0x12, 0x02, 0x02, 0x0A, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x82, 0x82, 0x80, 0x82, 0x80, 0x80, 0x88, 0x88, 0x88, 0x80, 0x88, 0x80, 0x82, 
            0x88, 0xC8, 0x8A, 0xCA, 0x8A, 0xC2, 0xC0, 0xC8, 0x88, 0x88, 0x88, 0x80, 0x88, 0x80, 0x80, 0x8A, 
            0x08, 0x08, 0x0A, 0x0A, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x42, 0xC2, 0x02, 0x02, 0x02, 0x82, 0x02, 0x0A, 0x02, 0x82, 0x02, 0x4A, 0x42, 0x82, 
            0x02, 0x0A, 0x02, 0x82, 0xC2, 0xCA, 0xC2, 0x42, 0x50, 0x4A, 0x4A, 0x4A, 0x4A, 0x42, 0x42, 0x42, 
            0x42, 0x40, 0x42, 0xC2, 0xC2, 0xC0, 0xD2, 0xD0, 0xD2, 0xDA, 0xD8, 0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 
            0xDA, 0xDA, 0xD8, 0xD8, 0xDA, 0xDA, 0xDA, 0xDA, 0xD8, 0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 0xDB, 0xDB, 
            0xDB, 0xD9, 0xDB, 0xDB, 0xDB, 0xD9, 0xDB, 0xD9, 0xDB, 0xDB, 0xD9, 0xDB, 0xDB, 0xDB, 0xDB, 0xDB, 
            0xDB, 0xDB, 0xD9, 0xD9, 0xDB, 0xD9, 0xD9, 0xDB, 0xDB, 0xD9, 0xD9, 0xD9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF8, 0xF9, 0xF9, 0xF9, 0xF9, 
            0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xEB, 0xEB, 0xE9, 0xE1, 0xE9, 0xE1, 0xE9, 0xE1, 0xE1, 
            0xE1, 0xE1, 0xE9, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE1, 0xE8, 0xE1, 0xE0, 0xE1, 0xE0, 0xE1, 
            0xE1, 0xE1, 0xE9, 0xE1, 0xE1, 0x61, 0x61, 0x71, 0xE1, 0x61, 0x69, 0xF1, 0xE1, 0x63, 0x79, 0x71, 
            0x69, 0xE9, 0x79, 0x79, 0xF1, 0x69, 0x69, 0x79, 0xE9, 0x61, 0x69, 0x79, 0xE1, 0x41, 0x69, 0x58, 
            0x49, 0x49, 0x59, 0x41, 0x41, 0xC8, 0xC1, 0x43, 0xD0, 0x51, 0x61, 0xF8, 0x68, 0xE0, 0x59, 0xC1, 
            0x51, 0xC9, 0xD1, 0x49, 0x55, 0x5F, 0x47, 0xCD, 0xCC, 0x45, 0xF5, 0xF5, 0x45, 0x7D, 0xFD, 0x65, 
            0xED, 0xFD, 0x65, 0xF5, 0x7F, 0xE7, 0xFD, 0x6D, 0xF5, 0x7D, 0x65, 0xFD, 0x6D, 0xE5, 0x7D, 0xE5, 
            0x75, 0xED, 0xF5, 0x6D, 0x75, 0x7D, 0x65, 0xED, 0xED, 0x65, 0xF5, 0xF5, 0x65, 0x7D, 0xFD, 0x65, 
            0xED, 0xFD, 0x65, 0x75, 0x7D, 0x65, 0xED, 0xED, 0xE5, 0xF5, 0xF5, 0xF5, 0x7D, 0x5D, 0x7D, 0x6D, 
            0xED, 0xED, 0xE5, 0xF5, 0xF5, 0xF5, 0x7D, 0x7D, 0x7D, 0x6D, 0x4D, 0x4D, 0xE5, 0xE5, 0xE5, 0xF5, 
            0x75, 0x75, 0x7D, 0x6D, 0x6D, 0x6D, 0x65, 0x65, 0x65, 0x65, 0x65, 0x65, 0x65, 0x45, 0x65, 0x64, 
            0x65, 0x65, 0x65, 0x65, 0x65, 0x65, 0x65, 0x65, 0x65, 0x65, 0x45, 0x45, 0x65, 0xC5, 0xC5, 0x45, 
            0x55, 0xD5, 0xC5, 0x4D, 0x5D, 0xD5, 0xC5, 0xCD, 0xDD, 0x55, 0x55, 0xDD, 0xCD, 0x4D, 0x5D, 0x55, 
            0x45, 0x4D, 0x4D, 0x45, 0xC5, 0xC1, 0x45, 0x51, 0xD1, 0xC1, 0x49, 0x79, 0xF1, 0xC1, 0xC9, 0xD9, 
            0x51, 0x51, 0xDD, 0xC9, 0xCD, 0xDD, 0xD1, 0xD1, 0xD9, 0xD9, 0x59, 0x59, 0x59, 0x69, 0x69, 0x49, 
            0x41, 0x41, 0xC5, 0xC1, 0xC1, 0xD5, 0x55, 0x55, 0x5D, 0x4D, 0x4D, 0x4D, 0x45, 0xC5, 0xC5, 0xC5, 
            0xD5, 0xD5, 0xD5, 0x5D, 0x5D, 0x5D, 0x4D, 0x4D, 0x4D, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x65, 
            0x65, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x65, 0x65, 0x65, 0x64, 0x44, 
            0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 
            0x44, 0x44, 0x44, 0x44, 0x44, 0x64, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 
            0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x64, 
            0x44, 0x44, 0x44, 0x44, 0x44, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x41, 0x41, 0x41, 0x41, 
            0x41, 0x41, 0x41, 0x41, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x41, 0x41, 0x40, 0x40, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x40, 0x40, 0x40, 0x40, 0x41, 
            0x40, 0x42, 0x40, 0x41, 0x41, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x44, 0x44, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x46, 0x46, 0x40, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x44, 0x40, 0x40, 
            0x40, 0x40, 0x44, 0x44, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 
            0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 
            0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 
            0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 
            0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 
            0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 
            0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 
            0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0xC2, 0x42, 0x42, 0x52, 0x42, 
            0x42, 0x4A, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0xC2, 0xC2, 0x42, 0xD2, 
            0x52, 0xC2, 0xDA, 0xCA, 0xD2, 0xDA, 0xD3, 0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 
            0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 0xDA, 0xDB, 0xDB, 0xDB, 0xD9, 0xD8, 0xD9, 0xDB, 0xD9, 0xD9, 
            0xDB, 0xDB, 0xD9, 0xF9, 0xD9, 0xD9, 0xDB, 0xD9, 0xDB, 0xDB, 0xD9, 0x99, 0xD9, 0x9B, 0x99, 0xD9, 
            0xD9, 0x9B, 0x99, 0xD9, 0xD9, 0x99, 0x99, 0xD9, 0x99, 0xF9, 0xBB, 0xBB, 0x9B, 0xDB, 0x9B, 0xBB, 
            0xB9, 0x99, 0x9B, 0xB9, 0xBB, 0xB9, 0x9B, 0xFB, 0x99, 0xB9, 0x99, 0x99, 0x99, 0xBB, 0x9B, 0x9B, 
            0x9B, 0x9B, 0x99, 0x9B, 0x9B, 0x9B, 0x99, 0x99, 0x9B, 0x99, 0x9B, 0x9B, 0x9B, 0x9B, 0x9B, 0x9B, 
            0x99, 0x99, 0x9B, 0x99, 0x9B, 0x99, 0x9B, 0x9B, 0x99, 0x99, 0x99, 0x99, 0x99, 0x9B, 0x9B, 0x9B, 
            0x9B, 0x9B, 0x99, 0x9B, 0x9B, 0x9B, 0x99, 0x99, 0x99, 0x9B, 0x9B, 0x99, 0x99, 0x9B, 0x9B, 0x99, 
            0x99, 0x99, 0x19, 0x99, 0x19, 0x9B, 0x0B, 0x09, 0x09, 0x0B, 0x8B, 0x0B, 0x8B, 0x09, 0x89, 0x0B, 
            0x0B, 0x0B, 0x8B, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x89, 0x0B, 0x8B, 0x08, 0x88, 0x08, 
            0x08, 0x8A, 0x8A, 0x0A, 0x8A, 0x88, 0x88, 0x0A, 0x8A, 0x80, 0x80, 0x88, 0x08, 0x80, 0x80, 0x8A, 
            0x8A, 0x82, 0x80, 0x80, 0x82, 0x8A, 0x80, 0x80, 0x88, 0x80, 0x88, 0x8C, 0x08, 0x89, 0x82, 0x8A, 
            0x89, 0x00, 0x00, 0x02, 0x03, 0x0B, 0x83, 0x80, 0x89, 0x09, 0x08, 0x81, 0x08, 0x01, 0x00, 0x0A, 
            0x0A, 0x0C, 0x0C, 0x0D, 0x08, 0x88, 0x08, 0x8C, 0x85, 0x8A, 0x8A, 0x0C, 0x0C, 0x04, 0x89, 0x0E, 
            0x02, 0x8F, 0x0A, 0x0A, 0x0A, 0x04, 0x81, 0x88, 0x88, 0x08, 0x09, 0x82, 0x0A, 0x00, 0x00, 0x08, 
            0x09, 0x09, 0x09, 0x09, 0x08, 0x88, 0x0A, 0x8B, 0x80, 0x89, 0x89, 0x08, 0x09, 0x00, 0x89, 0x08, 
            0x00, 0x09, 0x88, 0x88, 0x08, 0x00, 0x81, 0x80, 0x80, 0x80, 0x01, 0x00, 0x00, 0x00, 0x80, 0x80, 
            0x09, 0x09, 0x09, 0x09, 0x81, 0x80, 0x00, 0x01, 0x80, 0x81, 0x81, 0x80, 0x00, 0x01, 0x01, 0x00, 
            0x00, 0x01, 0x01, 0x01, 0x89, 0x88, 0x08, 0x08, 0x08, 0x09, 0x01, 0x01, 0x89, 0x88, 0x08, 0x09, 
            0x81, 0x80, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x89, 0xC8, 0x00, 0x00, 0x08, 0xA8, 0x80, 0x01, 
            0x01, 0x00, 0x00, 0x80, 0x80, 0x80, 0x80, 0x81, 0x81, 0x20, 0x00, 0x01, 0x01, 0x00, 0x80, 0x80, 
            0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x01, 0x01, 0x81, 0x81, 0x01, 0x01, 0x80, 0x80, 0x01, 0x01, 
            0x01, 0x01, 0x81, 0x81, 0x09, 0x09, 0x08, 0x08, 0x00, 0x01, 0x01, 0x00, 0x80, 0x81, 0x01, 0x00, 
            0x00, 0x01, 0x01, 0x01, 0x89, 0x89, 0x01, 0x01, 0x01, 0x00, 0x00, 0x80, 0x80, 0x81, 0x81, 0x80, 
            0x80, 0x01, 0x01, 0x08, 0x08, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 
            0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x02, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x02, 0x00, 0x00, 0x02, 0x02, 0x42, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x20, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x00, 0x40, 0x00, 0x00, 0x40, 0x40, 0x60, 0x40, 0x40, 0x40, 0x60, 0x60, 0x60, 0x40, 0x00, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x42, 0x42, 
            0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x42, 0x42, 0x42, 0xC2, 0x42, 0x42, 0x52, 0x42, 0x42, 0x52, 
            0x52, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x42, 0xC2, 0x42, 0xC6, 0xD2, 0x52, 
            0x4A, 0x4A, 0x4A, 0x4A, 0x4B, 0x4B, 0x4B, 0x4A, 0x4B, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 
            0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x4B, 0x43, 0x43, 0x43, 0x42, 0x4A, 0x4A, 0x4A, 0x4B, 0x4B, 
            0x4B, 0x4B, 0x4B, 0x4B, 0x4B, 0x43, 0x4B, 0x4F, 0x4A, 0x4B, 0x43, 0x4B, 0x4B, 0x4A, 0x4B, 0x4B, 
            0x4B, 0x0B, 0x0B, 0x0B, 0x0B, 0x0B, 0x0B, 0x0B, 0x0B, 0x03, 0x0B, 0x0B, 0x0A, 0x0B, 0x43, 0x47, 
            0x43, 0x42, 0x42, 0x46, 0x46, 0x46, 0x46, 0x42, 0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 
            0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x42, 0x42, 0x42, 0x42, 0x46, 0x46, 
            0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x46, 0x42, 0x46, 0x46, 0x46, 0x44, 0x40, 0x44, 0x44, 
            0x46, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x46, 0x46, 0x42, 0x46, 0x46, 0x46, 
            0x46, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 
            0x42, 0x42, 0x42, 0x42, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 
            0x00, 0x00, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x08, 0x00, 0x08, 0x09, 0x09, 0x09, 0x01, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 
            0x01, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 
            0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 
            0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 
            0x01, 0x09, 0x01, 0x09, 0x09, 0x09, 0x0D, 0x09, 0x0D, 0x0D, 0x0D, 0x0D, 0x09, 0x0D, 0x0D, 0x0D, 
            0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x05, 0x0D, 
            0x05, 0x05, 0x05, 0x05, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0C, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x00, 0x04, 0x00, 0x00, 0x04, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 
            0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
            0x02, 0x02, 0x0A, 0x0A, 0x28, 0x29, 0x09, 0x29, 0x2B, 0x2B, 0x2B, 0x2B, 0x2B, 0x6B, 0x6B, 0x6B, 
            0x6B, 0x6B, 0x69, 0x6B, 0x6B, 0x6B, 0x6B, 0x69, 0x7B, 0x5B, 0x5B, 0x53, 0x53, 0x52, 0x52, 0x5A, 
            0x5A, 0x5A, 0x5A, 0x5A, 0x5A, 0x5A, 0x50, 0x52, 0x5A, 0x52, 0x52, 0x50, 0x50, 0x50, 0x50, 0x50, 
            0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x58, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 
            0x50, 0x50, 0x50, 0x50, 0x50, 0x54, 0x54, 0x50, 0x54, 0x50, 0x54, 0x54, 0x50, 0x50, 0x54, 0x54, 
            0xD0, 0xD0, 0xD4, 0xD4, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 
            0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 
            0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 
            0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 
            0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0x90, 0x90, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 
            0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 
            0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0xC0, 0xC0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 0xD0, 
            0xD2, 0xD0, 0xD2, 0xD2, 0xD2, 0xD2, 0xD0, 0xD2, 0xD2, 0xD2, 0xD2, 0xD2, 0xD2, 0xD0, 0xD2, 0xD2, 
            0xD2, 0xD2, 0x52, 0x52, 0x52, 0x72, 0x72, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x72, 0x7A, 0x7A, 
            0x7A, 0x7A, 0x7A, 0x7A, 0x72, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
            0x7A, 0x7A, 0xFA, 0x7A, 0xFA, 0xFA, 0xFA, 0x7A, 0xFA, 0xFA, 0x78, 0xFA, 0x78, 0xF8, 0xFA, 0xF8, 
            0xFA, 0x7A, 0xFA, 0xFA, 0xFA, 0xFA, 0x7A, 0xFA, 0xFA, 0xFA, 0x7A, 0xFA, 0x7A, 0xFA, 0xFA, 0xFA, 
            0xFA, 0x7A, 0x7A, 0x7A, 0x72, 0x7A, 0xF2, 0xBA, 0xBA, 0x2A, 0x2A, 0xA8, 0x2A, 0x28, 0x28, 0x28, 
            0x28, 0x28, 0x28, 0x28, 0x28, 0xA8, 0x28, 0xA8, 0xA8, 0xA8, 0xA8, 0x28, 0x28, 0x28, 0xA8, 0x28, 
            0x28, 0xA8, 0x20, 0x08, 0x00, 0x00, 0x80, 0x80, 0x80, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x80, 0xC0, 0x80, 0x80, 0x00, 0x00, 0x00, 0x80, 0x00, 
            0x00, 0x00, 0x80, 0x80, 0x00, 0x00, 0x80, 0x80, 0x84, 0x80, 0x00, 0x00, 0x04, 0x00, 0x80, 0x80, 
            0x00, 0x00, 0x00, 0x00, 0x80, 0x80, 0x00, 0x00, 0x80, 0x80, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x00, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x80, 0x00, 0x00, 
            0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x80, 0x00, 0x00, 0x00, 0x80, 0x80, 0x00, 
            0x00, 0x04, 0x04, 0x84, 0x84, 0x80, 0x80, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x82, 
            0x02, 0x00, 0x02, 0x00, 0x02, 0x02, 0x02, 0x03, 0x83, 0xC3, 0x43, 0x53, 0xD3, 0xD3, 0x53, 0x53, 
            0x53, 0x53, 0xD3, 0xD3, 0x73, 0x73, 0x5B, 0x7B, 0x5B, 0x73, 0x7B, 0x7B, 0xFB, 0xFB, 0x7B, 0x7B, 
            0x7B, 0x7B, 0x7B, 0x6B, 0xEB, 0xEB, 0x6F, 0x6F, 0x6F, 0x6F, 0x6F, 0xEF, 0xEF, 0xEF, 0xEB, 0xEB, 
            0xEB, 0xEB, 0x6B, 0x6B, 0xEB, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 
            0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 
            0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 
            0x6B, 0x6B, 0x6B, 0x6F, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0xEB, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 
            0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 
            0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 
            0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x6B, 0x4B, 0x4B, 0x4B, 0x4B, 0x4B, 0x4B, 
            0x4B, 0x4B, 0x4B, 0x4B, 0x4B, 0x4B, 0x4B, 0x4B, 0x4B, 0x49, 0x4A, 0x49, 0x4B, 0x49, 0x0A, 0x4B, 
            0x0B, 0x4A, 0x0B, 0x4A, 0x4B, 0x4B, 0x4B, 0x4B, 0x4A, 0x49, 0x6B, 0x6B, 0x6B, 0x68, 0x2B, 0x6B, 
            0x6B, 0x68, 0x29, 0x68, 0x6B, 0x6B, 0x2B, 0x2B, 0x2A, 0x6A, 0x6A, 0x6A, 0x6A, 0x6B, 0x6B, 0x6B, 
            0x6A, 0x6A, 0x6B, 0x6A, 0x6A, 0x6A, 0x6A, 0x6A, 0x6A, 0x6A, 0x6A, 0x6A, 0x6B, 0x6A, 0x6B, 0x6B, 
            0x6B, 0x6B, 0x6A, 0x6A, 0x6A, 0x6B, 0x6A, 0x6A, 0x6B, 0x6A, 0x4A, 0x6A, 0x62, 0x6B, 0x6B, 0x6B, 
            0x6A, 0x6A, 0x6B, 0x6A, 0x6A, 0x6A, 0x6A, 0x4A, 0x4A, 0x62, 0x62, 0x6A, 0x6B, 0x6A, 0x6B, 0x6B, 
            0x6B, 0x6B, 0x6A, 0x6A, 0x6A, 0x69, 0x68, 0x68, 0x68, 0x69, 0x09, 0x28, 0x20, 0x29, 0x29, 0x29, 
            0x29, 0x28, 0x28, 0x28, 0x28, 0x29, 0x2D, 0x08, 0x08, 0x20, 0x00, 0x09, 0x01, 0x80, 0x00, 0x01, 
            0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x01, 0x21, 0x20, 0x08, 0x08, 0x00, 0x00, 0x00, 0x01, 
            0x01, 0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x04, 0x04, 0x25, 0x25, 0x0D, 0x0D, 0x05, 0x85, 
            0x84, 0x84, 0x84, 0x84, 0x84, 0x85, 0x85, 0x86, 0x86, 0x84, 0x84, 0x84, 0xC4, 0xC4, 0x84, 0x85, 
            0x85, 0x84, 0x84, 0x85, 0x85, 0x86, 0x86, 0x84, 0x84, 0x81, 0xC1, 0xC0, 0x80, 0x80, 0x80, 0x80, 
            0x80, 0x00, 0x00, 0x01, 0x01, 0x10, 0x10, 0x10, 0x20, 0x20, 0x28, 0xA9, 0xA9, 0x28, 0x28, 0x28, 
            0x08, 0x08, 0x4B, 0x0B, 0x0B, 0x09, 0x09, 0x41, 0x41, 0x41, 0x01, 0x01, 0x02, 0x02, 0x00, 0x01, 
            0x09, 0x0A, 0x4A, 0x48, 0x69, 0x6B, 0x6B, 0x68, 0x69, 0x69, 0x69, 0x6A, 0x69, 0x69, 0x69, 0x6E, 
            0x68, 0x68, 0x69, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x69, 0x68, 0x68, 0x68, 
            0x68, 0x68, 0x68, 0x68, 0x68, 0x69, 0x69, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 
            0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 
            0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 
            0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x6C, 0x68, 0x68, 0x68, 
            0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 
            0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x6C, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 
            0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 
            0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x69, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 
            0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0x68, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 
            0xE8, 0x68, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 
            0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xEC, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 
            0xE8, 0xE8, 0xE8, 0xE8, 0xEC, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 
            0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xE8, 0xF8, 0xF8, 0xF8, 0xF8, 0xF8, 0xFC, 
            0xFC, 0xFC, 0xFC, 0xF8, 0xFC, 0xF8, 0xFC, 0xFC, 0xF8, 0xF8, 0xF8, 0xFC, 0xF8, 0xF8, 0xF8, 0xF8, 
            0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0xF8, 0xFC, 0xFC, 0xF8, 0xF8, 0x7C, 0x7E, 0x7E, 0xFE, 0xFD, 0x71, 
            0x71, 0x7B, 0xFF, 0xFF, 0xF5, 0x7B, 0xFF, 0xFF, 0xF5, 0x7B, 0xFF, 0xFF, 0xF5, 0x71, 0x71, 0x7B, 
            0xF5, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x7B, 0xF5, 0x71, 0x71, 0x71, 0x71, 0x71, 
            0x71, 0x71, 0x7B, 0xFF, 0xF5, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 
            0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x61, 
            0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 
            0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x61, 0x61, 0x61, 0x71, 0x71, 0x71, 
            0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 
            0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x61, 0x61, 0x61, 0x61, 0x71, 0x60, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0xC2, 0x4A, 0xC4, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x41, 
            0x60, 0x40, 0x00, 0x40, 0x60, 0x40, 0x50, 0x40, 0x40, 0x40, 0x41, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x41, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x40, 0x40, 0x40, 0x51, 0x60, 0x60, 0x41, 0x60, 0x41, 0x60, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 
            0x41, 0x60, 0x40, 0x40, 0x40, 0x51, 0x60, 0x40, 0x40, 0x40, 0x40, 0x51, 0x60, 0x40, 0x50, 0x41, 
            0x61, 0x61, 0x60, 0x71, 0x60, 0x60, 0x41, 0x61, 0x61, 0x61, 0x61, 0x60, 0x61, 0x61, 0x61, 0x60, 
            0x41, 0x60, 0x61, 0x61, 0x60, 0x71, 0x61, 0x40, 0x41, 0x61, 0x61, 0x71, 0x71, 0x71, 0x71, 0x71, 
            0x75, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x7B, 0xFF, 0xFF, 0xFF, 0xFF, 
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
            0xFE, 0xBF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xFF, 0xFF, 0xFF, 
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xBE, 0xBE, 0xBF, 0xFE, 0xFF, 0xFF, 0xFF, 
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFD, 0x7F, 0x7D, 0x7F, 0x7F, 0x7F, 
            0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7E, 0x7E, 0x3E, 0x7E, 0x3F, 0x3E, 0x5E, 0x0E, 0x0E, 
            0x1E, 0x0E, 0x1E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0C, 0x0E, 0x1C, 0x1C, 0x1E, 0x1E, 0x0E, 0x0E, 0x0E, 
            0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x2E, 0x2E, 0x2E, 0x2E, 0x2E, 0x3E, 0x2E, 0x0E, 0x1E, 
            0x3F, 0x2E, 0x0E, 0x0E, 0x0E, 0x4E, 0x0E, 0x1E, 0x2E, 0x0E, 0x0E, 0x0E, 0x1E, 0x04, 0x1A, 0x0C, 
            0x0C, 0x1C, 0x2E, 0x1C, 0x0C, 0x0C, 0x04, 0x18, 0x2E, 0x0E, 0x0C, 0x04, 0x08, 0x0C, 0x0C, 0x0C, 
            0x0D, 0x0D, 0x0C, 0x44, 0x4C, 0x0D, 0x04, 0x08, 0x0C, 0x0C, 0x0C, 0x04, 0x00, 0x00, 0x00, 0x00, 
            0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x40, 0x00, 0x01, 0x00, 0x00, 0x00, 0x10, 0x20, 0x10, 0x20, 
            0x01, 0x01, 0x00, 0x5A, 0x74, 0x31, 0x30, 0x30, 0x20, 0x00, 0x10, 0x30, 0xA0, 0x80, 0x80, 0x80, 
            0x10, 0xA0, 0x81, 0x80, 0x80, 0x90, 0xE0, 0xD0, 0xF1, 0xF1, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 
            0xF0, 0xF0, 0xF0, 0xBA, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFE, 0xFE, 0xEE, 0xFF, 0xFE, 0xEF, 0xEF, 
            0xEF, 0xEF, 0xEF, 0xEF, 0xFF, 0xFF, 0xFF, 0xEF, 0xFF, 0xEF, 0xFF, 0xFF, 0xEF, 0xEF, 0xEF, 0xFF, 
            0xFF, 0xEF, 0xEF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xEE, 0xEE, 0xFE, 0xEE, 0xEE, 0xEF, 0xFF, 0xEF, 
            0xEF, 0xEF, 0xEF, 0xEF, 0xEF, 0xEF, 0xEF, 0xEF, 0xFF, 0xEF, 0xEF, 0xEF, 0xEF, 0xFF, 0xFF, 0xFF, 
            0xEE, 0xEE, 0xEE, 0xEE, 0xFE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEF, 0xEF, 0xFF, 
            0xFF, 0x7F, 0x6F, 0xEF, 0xAF, 0xAF, 0xAF, 0xFF, 0xEE, 0xFE, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 
            0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFE, 0xFF, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFE, 0xFE, 0xEE, 0xEE, 
            0xEE, 0xEE, 0xEE, 0xEF, 0xEE, 0xEF, 0xEE, 0xEE, 0xCE, 0xCE, 0xCE, 0xCE, 0xDE, 0xDE, 0xCF, 0xCE, 
            0xDC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCD, 0xCC, 0xCC, 0xCC, 0xDC, 0xDC, 0xCC, 0xDC, 0xCC, 0xCC, 0xCC, 
            0xCC, 0xCD, 0x4C, 0x4C, 0x4C, 0x4C, 0x1C, 0x0C, 0x0C, 0x3C, 0x0C, 0x0C, 0x0F, 0x0E, 0x0E, 0x0E, 
            0x1E, 0x1E, 0x3E, 0x1E, 0x36, 0x16, 0x1A, 0x37, 0x12, 0x12, 0x1A, 0x16, 0x12, 0x12, 0x12, 0x12, 
            0x1A, 0x1E, 0x1E, 0x1E, 0x3E, 0x3E, 0x3E, 0x1E, 0x1E, 0x1E, 0x3E, 0x2E, 0x0E, 0x0E, 0x2E, 0x3F, 
            0x1E, 0x0E, 0x0F, 0x0F, 0x3F, 0x0E, 0x1E, 0x0E, 0x2F, 0x0F, 0x0E, 0x1E, 0x0F, 0x0F, 0x0F, 0x1E, 
            0x0E, 0x1E, 0x8E, 0x8E, 0x8E, 0x9E, 0x8B, 0x9E, 0x9E, 0xCE, 0x0E, 0x4E, 0xCE, 0x8E, 0x8E, 0xCE, 
            0xCE, 0xCE, 0x4E, 0x4E, 0x4E, 0x0E, 0x0E, 0x8E, 0x0E, 0x8E, 0xCE, 0x8E, 0x4E, 0xCE, 0x4E, 0x8E, 
            0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 
            0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x8E, 0x4E, 0x0E, 0x4E, 0x8E, 
            0x4E, 0x0E, 0x4E, 0x8E, 0x4E, 0x0E, 0x4E, 0x8E, 0x4E, 0x0E, 0x4E, 0x8E, 0x4E, 0x0E, 0x4E, 0x8E, 
            0x4E, 0x0E, 0x0E, 0x8E, 0x4E, 0x8E, 0x8E, 0x8E, 0xCE, 0xCE, 0xCE, 0xCE, 0xCE, 0xCE, 0x4E, 0xCE, 
            0x4A, 0x8A, 0x4A, 0x88, 0x4A, 0x8A, 0x4A, 0x8A, 0x48, 0x8A, 0x4A, 0x8A, 0x4A, 0x8A, 0x4A, 0x8A, 
            0x4A, 0x8A, 0x4A, 0x8A, 0x48, 0x88, 0x48, 0x88, 0xC8, 0x88, 0x48, 0xCA, 0xCA, 0x8A, 0xEA, 0xEA, 
            0x4A, 0xC2, 0xC2, 0x8A, 0xCA, 0x40, 0x4C, 0xC4, 0x80, 0x80, 0xC1, 0xC1, 0xC1, 0xC9, 0xC9, 0xC9, 
            0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xCD, 0xC1, 0xC1, 0xC9, 0xCD, 0xC9, 0xC9, 0xD1, 0xD1, 0xC9, 
            0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xC1, 0xD1, 0xC1, 0x81, 0x81, 0x81, 0xC9, 0xC9, 
            0xC9, 0xC9, 0xC9, 0xC9, 0x89, 0x89, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xCD, 0x89, 0x89, 0xC9, 0xC9, 
            0xC9, 0xC9, 0xC9, 0x49, 0x09, 0x18, 0x58, 0xC8, 0xC8, 0x89, 0x89, 0x89, 0x89, 0x81, 0x81, 0x81, 
            0x81, 0x89, 0x81, 0x81, 0x81, 0x81, 0x81, 0x81, 0x80, 0x81, 0x81, 0x85, 0x85, 0x85, 0x85, 0x85, 
            0x85, 0x85, 0x85, 0x87, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x10, 
        };

        /// <summary>
        /// Gets an array containing all the known code pages.
        /// </summary>
        internal static CodePage[] CodePages
        {
            get
            {
                return AllKnownCodePages;
            }
        }

        /// <summary>
        /// Gets the character bitmap.
        /// </summary>
        internal static byte[] Bitmap
        {
            get
            {
                return CharacterBitmap;
            }
        }

        /// <summary>
        /// Represents the range of the code page.
        /// </summary>
        internal struct CodePageRange
        {
            /// <summary>
            /// The start of the code page range.
            /// </summary>
            public ushort First;

            /// <summary>
            /// The end of the code page range.
            /// </summary>
            public ushort Last;

            /// <summary>
            /// The offset for the range.
            /// </summary>
            public ushort Offset;

            /// <summary>
            /// The range mask.
            /// </summary>
            public byte Mask;

            /// <summary>
            /// Initializes a new instance of the <see cref="CodePageRange"/> struct.
            /// </summary>
            /// <param name="first">The start of the code page range.</param>
            /// <param name="last">The end of the code page range.</param>
            /// <param name="offset">The code page offset.</param>
            /// <param name="mask">The code page mask.</param>
            public CodePageRange(ushort first, ushort last, ushort offset, byte mask)
            {
                this.First = first;
                this.Last = last;
                this.Offset = offset;
                this.Mask = mask;
            }
        }

        /// <summary>
        /// Represents information about a code page.
        /// </summary>
        internal struct CodePage
        {
            /// <summary>
            /// The ranges for the code page.
            /// </summary>
            public CodePageRange[] Ranges;

            /// <summary>
            /// The type of the code page.
            /// </summary>
            public CodePageKind Kind;

            /// <summary>
            /// The ASCII support for the code page.
            /// </summary>
            public CodePageAsciiSupport AsciiSupport;

            /// <summary>
            /// The Unicode coverage for the code page.
            /// </summary>
            public CodePageUnicodeCoverage UnicodeCoverage;

            /// <summary>
            /// Type flags for the code page.
            /// </summary>
            public CodePageFlags Flags;

            /// <summary>
            /// The detectable code page identifier.
            /// </summary>
            public ushort DetectId;   
            
            /// <summary>
            /// The code page identifier.
            /// </summary>
            public ushort Id;         
            
            /// <summary>
            /// The Windows code page identifier.
            /// </summary>
            public ushort WindowsId;

            /// <summary>
            /// Initializes a new instance of the <see cref="CodePage"/> struct.
            /// </summary>
            /// <param name="id">The code page identifier.</param>
            /// <param name="windowsId">The Windows code page identifier.</param>
            /// <param name="kind">The type of the code page.</param>
            /// <param name="asciiSupport">The ASCII support for the code page.</param>
            /// <param name="unicodeCoverage">The unicode coverage of the code page.</param>
            /// <param name="flags">Any type flags for the code page.</param>
            /// <param name="detectId">The detection identifier.</param>
            /// <param name="ranges">The ranges of the code page.</param>
            public CodePage(
                ushort id, 
                ushort windowsId, 
                CodePageKind kind, 
                CodePageAsciiSupport asciiSupport, 
                CodePageUnicodeCoverage unicodeCoverage, 
                CodePageFlags flags, 
                ushort detectId, 
                CodePageRange[] ranges)
            {
                this.Id = id;
                this.WindowsId = windowsId;
                this.Kind = kind;
                this.AsciiSupport = asciiSupport;
                this.UnicodeCoverage = unicodeCoverage;
                this.Flags = flags;
                this.DetectId = detectId;
                this.Ranges = ranges;
            }
        }
    }
}
